/*******************************************************************************

   ACTOPTIMIZATION.C

   Author: Peter Loan
           Krystyne Blaikie

   Copyright (c) 2010-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains all of the set-up functions for the static
      optimization code, as well as the callback functions called by each of
      the optimization algorithms (LBFGSB and Levenberg-Marquardt).

   Routines:
      set_up_optimization                 : sets up the static optimization of muscle activations
      cleanup_optimization                : frees memory used by the optimization
      optimize_muscle_activations         : called each time frame to optimize activations
      calc_optimized_muscle_force         : calculates the force in a muscle during optimization
      calc_moment_residuals               : calculates residuals between actual and desired moments
      count_optimized_gencoords           : counts how many gencoords to use in optimization
      append_optimized_gencoord_residuals : appenda residuals to a frame of motion data
      lbfgsb_optimize_activations         : uses LBFGSB algorithm to optimize activations
      calc_lbfgsb_gradient                : calculates gradient for LBFGSB algorithm
      calc_lbfgsb_residuals               : calculates residuals for LBFGSB algorithm
      lmdif_optimize_activations          : uses Levenberg-Marquardt algorithm to optimize activations
      calc_lmdif_residuals                : calculates residuals for Levenberg-Marquardt algorithm

*******************************************************************************/

#include "universal.h"
#include "lbfgsb_optimize.h"


/*************** DEFINES (for this file only) *********************************/

typedef struct
{
   double* moment;
   double* state;
   double time;
   int numQs;
   dpBoolean* q;
   int numMuscles;
   dpMuscleStruct** muscle;
   double* activation;
} OptimizationData;

#define K_NEG 5000.0
#define K_POS 5000.0
#define K_ACT 100.0
#define K_TORQUE 100.0
#define K_ITER 100
#define EMG_ENVELOPE 0.2
#define EMG_THRESHOLD 0.05


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/
static OptimizationData* dataOptimize = NULL;

/**************** GLOBAL VARIABLES (used in only a few files) *****************/


/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern dpModelStruct* sdm;
extern dpSimulationParameters params;
extern char buffer[];   /* defined in readtools.c */

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static double calc_optimized_muscle_force(dpMuscleStruct* muscle, double activation);
static void calc_moment_residuals(OptimizationData* si);
static void lbfgsb_optimize_activations(OptimizationData* si);
static double calc_lbfgsb_residuals(double x[], OptimizationData* si);
static void lmdif_optimize_activations(OptimizationData* si);

void calc_lbfgsb_gradient(double x[], double* f, double g[], void* data);
void calc_lmdif_residuals(int numResid, int numActivations, double activation[],
                          double resid[], int* flag2, void* data);


ReturnCode set_up_optimization(char message[])
{
   int i, j;

   dataOptimize = (OptimizationData*)simm_calloc(1, sizeof(OptimizationData));

   // Create an array of pointers to the muscles whose activations will be optimized.
   dataOptimize->muscle = (dpMuscleStruct**)simm_calloc(sdm->num_muscles, sizeof(dpMuscleStruct*));
   for (i=0; i<params.outputOptions.optimizedActivations_Muscles.num_items; i++)
      dataOptimize->muscle[dataOptimize->numMuscles++] = &sdm->muscles[params.outputOptions.optimizedActivations_Muscles.index[i]];

   if (dataOptimize->numMuscles < 1)
   {
      sprintf(message, "there are no muscles to optimize");
      return code_bad;
   }

   // Create an array of booleans for the Qs, indicating whether or not the
   // moment for that Q will be included in the optimization. The Qs that will be
   // included are those that the user specified and that are either unconstrained
   // or prescribed, and that are used in a joint spanned by at least one muscle.
   dataOptimize->q = (dpBoolean*)simm_calloc(sdm->nu, sizeof(dpBoolean));
   for (i=0; i<sdm->nu; i++)
   {
      dataOptimize->q[i] = dpNo;
      if (find_name_in_list(sdm->q[i].name, &params.outputOptions.optimizedActivations_Gencoords) >= 0 &&
         (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ))
      {
         for (j=0; j<dataOptimize->numMuscles; j++)
         {
            if (NOT_EQUAL_WITHIN_TOLERANCE(dataOptimize->muscle[j]->momentarms[i], 0.0, 0.00001))
            {
               dataOptimize->q[i] = dpYes;
               dataOptimize->numQs++;
               break;
            }
         }
      }
   }

   if (dataOptimize->numQs < 1)
   {
      sprintf(message, "there are no Qs to optimize");
      return code_bad;

   }
   dataOptimize->moment = (double*)simm_calloc(sdm->nu, sizeof(double));
   dataOptimize->activation = (double*)simm_calloc(dataOptimize->numMuscles, sizeof(double));

   return code_fine;
}


void cleanup_optimization(void)
{
   if (dataOptimize)
   {
      FREE_IFNOTNULL(dataOptimize->muscle);
      FREE_IFNOTNULL(dataOptimize->q);
      FREE_IFNOTNULL(dataOptimize->moment);
      FREE_IFNOTNULL(dataOptimize->activation);
      FREE_IFNOTNULL(dataOptimize);
   }
}


/* OPTIMIZE_MUSCLE_ACTIVATIONS: this function calculates a set of muscle activations
 * such that the moments generated by those muscles equal the input moments.
 * The input moments are the ones calculated by calc_gencoord_moments() during an
 * inverse dynamics analysis. This function gets called for each time step
 * during the inverse analysis, after the moments and muscle moment arms have
 * already been calculated. It uses a least-squares algorithm (lmdif) or a
 * bound-constrained, non-linear optimization (LBFGSB) to find the optimal set
 * of muscle activations, starting with the solution from the previous time step.
 */
void optimize_muscle_activations(double time, double state[])
{
   int i, j, numResid, *ipvt;
   int info, num_func_calls, ldfjac, mode = 1, nprint = 0, max_iter = K_ITER;
   double *activation, *diag, *qtf, *wa1, *wa2, *wa3, *wa4, *fjac, *resid;
   double ftol = 1e-4, xtol = 1e-4, gtol = 0.0;
   double epsfcn = 0.0, step_factor = 0.2;

   dataOptimize->state = state;
   dataOptimize->time = time;

   // Store the current activations (the ones calculated by this
   // function for the last time frame) in the activation[] array
   // so they will be used as the starting guess for this time frame.
   // If the current activation is 0.0, make it a small positive
   // number so that the array of activations does not start out
   // equal to zero (which causes problems in lmdif_C).
   for (i=0; i<dataOptimize->numMuscles; i++)
   {
      if (EQUAL_WITHIN_ERROR(dataOptimize->muscle[i]->dynamic_activation, 0.0))
         dataOptimize->activation[i] = 0.0001;
      else
         dataOptimize->activation[i] = dataOptimize->muscle[i]->dynamic_activation;
   }

#if 1
   lbfgsb_optimize_activations(dataOptimize);
#else
   lmdif_optimize_activations(dataOptimize);
#endif

   // Store the optimized activations in the muscle structures so
   // the muscle derivative function can use them and so that they
   // are available for the next time frame.
   for (i=0; i<dataOptimize->numMuscles; i++)
      dataOptimize->muscle[i]->dynamic_activation = dataOptimize->activation[i];

   // Calculate the moment residuals (input/ID moments - muscle moments) so you
   // know if the muscles were able to generate the necessary moment.
   calc_moment_residuals(dataOptimize);
}


/* CALC_OPTIMIZED_MUSCLE_FORCE: this function finds the force in a muscle for
 * use in optimize_muscle_activations(). It contains two methods of calculating
 * force. The default one uses activation level and fiber velocity to calculate
 * the force in a Hill-type model. The second method is a simple approximation
 * using only activation and max_isometric_force.
 */
static double calc_optimized_muscle_force(dpMuscleStruct* muscle, double activation)
{
#if 1

   double c_tau, muscle_width, active_force, passive_force, cos_factor;
   double norm_fiber_length, norm_fiber_velocity, velocity_factor;

   // This code uses the current values of muscle_tendon_length, fiber_length, and
   // fiber_velocity as calculated by the last call to the muscle's derivative function
   // (which should always be muscle model 10, the one specifically designed to work
   // with this function). Pennation angle is also taken into account.
   c_tau = get_muscle_param_value(muscle, "timescale");
   norm_fiber_length = muscle->fiber_length / (*muscle->optimal_fiber_length);
   norm_fiber_velocity = muscle->fiber_velocity * c_tau / (*muscle->optimal_fiber_length);

   muscle_width = *muscle->optimal_fiber_length * sin(DEG_TO_RAD*(*muscle->pennation_angle));

   cos_factor = cos(atan(muscle_width / muscle->muscle_tendon_length));

   active_force = interpolate_function(norm_fiber_length, *muscle->active_force_len_func, zeroth, 1.0, 1.0) * activation;
   if (active_force < 0.0)
      active_force = 0.0;

   passive_force = interpolate_function(norm_fiber_length, *muscle->passive_force_len_func, zeroth, 1.0, 1.0);
   if (passive_force < 0.0)
      passive_force = 0.0;

   velocity_factor = calc_force_velocity(muscle, norm_fiber_velocity);

   muscle->force = (active_force * velocity_factor + passive_force) * (*muscle->max_isometric_force) * cos_factor;

#else

   muscle->force = activation * (*muscle->max_isometric_force);

#endif

   // The force is stored in the muscle structure so that it can be copied to the
   // the output motion structure if the user requests muscle forces. But this value
   // is reset to zero in the derivative function for muscle model 10 so that the
   // force never gets applied to the model.
   return muscle->force;
}


/* CALC_MOMENT_RESIDUALS: calculate the difference between the desired moments
 * (e.g., the inverse dynamics moments) and the actual moments that the muscles
 * generated. These moments may not be equal because the muscles included in the
 * optimization may not have sufficient strength.
 */
static void calc_moment_residuals(OptimizationData* si)
{
   int i, j;
   double muscle_force;

   // Zero-out the current moments. This array will hold the total moments that
   // the muscles generate, given final values of the optimized activations for
   // this time frame.
   for (i=0; i<sdm->nu; i++)
      si->moment[i] = 0.0;

   for (i=0; i<si->numMuscles; i++)
   {
      muscle_force = calc_optimized_muscle_force(si->muscle[i], si->muscle[i]->dynamic_activation);

      // Now find the gencoord moments that this muscle force generates and add them to the array.
      for (j=0; j<sdm->nu; j++)
         si->moment[j] += muscle_force * si->muscle[i]->momentarms[j];
   }

   // Calculate the difference between the muscle moments and the desired moments
   // (the inverse dynamics moments). Assume that the latter have already been calculated).
   // Store the result in the OptimizationData structure.
   for (i=0; i<sdm->nu; i++)
   {
      if (si->q[i] == dpYes)
      {
         si->moment[i] = sdm->q[i].moment - si->moment[i];
      }
   }
}


/* Count the number of gencoords that will be included in the static optimization
 * of muscle activations. A gencoord is included if it is unconstrained or prescribed,
 * and is used in a joint that is spanned by one or more muscles being optimized.
 */
int count_optimized_gencoords(void)
{
   int count = 0;

   if (dataOptimize)
   {
      int i;

      for (i=0; i<sdm->nu; i++)
      {
         if (dataOptimize->q[i] == dpYes)
            count++;
      }
   }

   return count;
}


/* Append the residuals for the optimized gencoords to the specified frame of
 * motion data.
 */
void append_optimized_gencoord_residuals(MotionData* md, int* column, int frame)
{
   if (md && dataOptimize)
   {
      int i;

      for (i=0; i<sdm->nu; i++)
      {
         if (dataOptimize->q[i] == dpYes)
         {
            md->motiondata[(*column)++][frame] = dataOptimize->moment[i];
         }
      }
   }
}


// LBFGSB_OPTIMIZE_ACTIVATIONS
// N - problem dimension. N>0
// M - number of corrections in the BFGS scheme of Hessian approximation
//     update. Recommended value: 3<=M<=7. The smaller value causes worse
//     convergence, the bigger will not cause a considerably better  
//     convergence, but will cause a fall in the performance. M<=N.
// X - initial solution approximation. Array whose index ranges from 1 to N.
// EpsG - positive number which defines a precision of search. The subroutine
//        subroutine finishes its work if the condition ||G|| < EpsG is
//        satisfied, where ||.|| means Euclidian norm, G - gradient projection
//        onto a feasible set, X - current approximation.
// EpsG - positive number which defines a precision of search. The subroutine
//        subroutine finishes its work if the condition ||G|| < EpsG is
//        satisfied, where ||.|| means Euclidian norm, G - gradient projection
//        onto a feasible set, X - current approximation.
// EpsG - positive number which defines a precision of search. The subroutine
//        subroutine finishes its work if the condition ||G|| < EpsG is
//        satisfied, where ||.|| means Euclidian norm, G - gradient projection
//        onto a feasible set, X - current approximation.
// EpsX - positive number which defines a precision of search. The subroutine
//        finishes its work if on iteration number k+1 the condition
//        |X(k+1)-X(k)| <= EpsX is satisfied.
// MaxIts - maximum number of iterations. If MaxIts=0, the number of iterations is unlimited.
// NBD - constraint type. Array whose index ranges from 1 to N.
// If NBD(i) is equal to: 0, X(i) has no constraints,
//                        1, X(i) has only lower boundary,
//                        2, X(i) has both lower and upper boundaries,
//                        3, X(i) has only upper boundary,          
// L - lower boundaries of X(i) variables. Array whose index ranges from 1 to N.
// U - upper boundaries of X(i) variables. Array whose index ranges from 1 to N.
// Info - a return code:  -2: unknown internal error,
//                        -1: wrong parameters were specified,
//                         0: interrupted by user,
//                         1: relative function decreasing is less or equal to EpsF,
//                         2: step is less or equal to EpsX,
//                         4: gradient norm is less or equal to EpsG,
//                         5: number of iterations exceeds MaxIts.
static void lbfgsb_optimize_activations(OptimizationData* si)
{
   int i;
   double start_time, end_time;
   int dp_n = si->numMuscles;    
   int dp_m = 5;//4;
   double dp_epsg = 1e-4;//0.1;//1e-6;
   double dp_epsf = 1e-4;//0.1;//1e-6;
   double dp_epsx = 1e-3;//0.1;      
   int dp_maxits = 1000;  
   int *dp_nbd;
   double *dp_l;
   double *dp_u;
   int dp_info;

   // Allocate space for necessary arrays.
   dp_nbd = (int *)simm_malloc(dp_n * sizeof(int));
   dp_l = (double *)simm_malloc(dp_n * sizeof(double));
   dp_u = (double *)simm_malloc(dp_n * sizeof(double));

   // Set the upper and lower bounds for the activations.
   for (i = 0; i < dp_n; i++)
   {
      dp_nbd[i] = 2;
      dp_l[i] = 0.0;
      dp_u[i] = 1.0;
   }

   lbfgsb_optimize(dp_n, dp_m, si->activation, dp_epsg, dp_epsf, dp_epsx, dp_maxits, dp_nbd,
      dp_l, dp_u, &dp_info, calc_lbfgsb_gradient, si);

#if 0
   // error output
   if (dp_info == -2)
      printf("unknown internal error\n");
   else if (dp_info == -1)
      printf("wrong parameters were specified\n");
   else if (dp_info == 0)
      printf("interrupted by user\n");
   else if (dp_info == 1)
      printf("relative function decreasing is less or equal to EpsF\n");
   else if (dp_info == 2)
      printf("step is less or equal to EpsX\n");
   else if (dp_info == 4)
      printf("gradient norm is less or equal to EpsG\n");
   else if (dp_info == 5)
      printf("number of iterations exceeds MaxIts.\n");
#endif

   free(dp_nbd);
   free(dp_l);
   free(dp_u);
}


/* CALC_LBFGSB_GRADIENT: calculate the gradient for the residual functions for muscle optimizations
 * input params:
 * x[] - input array from 1 to n
 * Output parameters:
    F   -   function value at X.
    G   -   function gradient.
*/
void calc_lbfgsb_gradient(double x[], double* f, double g[], void* data)
{
   double temp, h = 1e-2, a, b, c, d;
   int i, j, numResid;
   OptimizationData* si = (OptimizationData*)data;

   h = 0.01;

   //grad = df/dx = 1/12h * [f(x-2h) - 8 * f(x-h) + 8 * f(x+h) - f(x+2h)]
   for (i = 0; i < si->numMuscles; i++)
   {
      temp = x[i];
      x[i] = temp - 2 * h;
      a = calc_lbfgsb_residuals(x, si);
      x[i] = temp - h;
      b = calc_lbfgsb_residuals(x, si);
      x[i] = temp + h;
      c = calc_lbfgsb_residuals(x, si);
      x[i] = temp + 2 * h;
      d = calc_lbfgsb_residuals(x, si);
      g[i] = 1 / (12 * h) * (a - 8* b + 8 * c - d);
      x[i] = temp;  //restore the x value
   }
   *f = calc_lbfgsb_residuals(x, si);
}


/* calc_lbfgsb_residuals: calculate the residuals for the lbgsf gradient function
 * There is one residual for each muscle's activation level, and one for each gencoord
 * moment that is being optimized.
 * Don't include bounds!  Optimizer takes care of those
 */
static double calc_lbfgsb_residuals(double x[], OptimizationData* si)
{
   int i, j, index;
   double muscle_force, exc, min, max, ssq, norm;
   dpMuscleStruct* muscle;
   double* resid;
   
   resid = (double*)simm_calloc((si->numMuscles + si->numQs), sizeof(double));

   // The first set of residuals is for the muscle activations (in x).
   // If the activation is less than 0.0 or greater than 1.0, there
   // is a [stiff] penalty. If the activation is between 0.0 and 1.0,
   // there is a softer penalty to try to drive the activation close
   // to 0.0. If the muscle has an excitation curve specified in the
   // motion file or muscle file, use it in the residual calculations.
   for (i = 0; i < si->numMuscles; i++)
   {
      muscle = si->muscle[i];

      if (muscle->excitation_func)
      {
         // Calculate the current excitation level of the muscle. The
         // excitation curve may go below 0.0 or above 1.0, but we don't
         // want to allow that here because it may result in optimized
         // activations going outside the 0.0 to 1.0 range.
         exc = calc_excitation(muscle, si->time, si->state);
         if (exc < 0.0)
            exc = 0.0;
         else if (exc > 1.0)
            exc = 1.0;
#if 1
         // Assume that amplitude of excitation/EMG curve is dependable.
         // Try to keep activation within EMG_ENVELOPE of that amplitude.
         min = MAX(0.0, exc - EMG_ENVELOPE);
         max = MIN(1.0, exc + EMG_ENVELOPE);

         if (x[i] < min)
            resid[i] = (exc - x[i]) * K_ACT + (min - x[i]) * K_NEG;
         else if (x[i] > max)
            resid[i] = (x[i] - exc) * K_ACT + (x[i] - max) * K_POS;
         else if (x[i] < exc)
            resid[i] = (exc - x[i]) * K_ACT;
         else
            resid[i] = (x[i] - exc) * K_ACT;
#else
         // Assume that amplitude of excitation/EMG curve is not dependable.
         // If excitation is less than EMG_THRESHOLD, force activation to be zero.
         // If excitation is above EMG_THRESHOLD, allow activation to range
         // from 0.0 to 1.0.
         if (exc < EMG_THRESHOLD)
         {
            if (activation[i] < 0.0)
               resid[i] = -activation[i] * K_NEG;
            else
               resid[i] = activation[i] * K_POS;
         }
         else
         {
            if (activation[i] < 0.0)
               resid[i] = -activation[i] * K_NEG;
            else if (activation[i] > 1.0)
               resid[i] = activation[i] * K_ACT + (activation[i] - 1.0) * K_POS;
            else
               resid[i] = activation[i] * K_ACT;
         }
#endif
      }
      else
      {
         resid[i] = x[i] * K_ACT;
      }
   }

   // Zero-out the current moments. This array will hold the total moments that
   // the muscles generate, given the current guess of activation levels.
   for (j = 0; j < sdm->nu; j++)
      si->moment[j] = 0.0;

   for (i = 0; i < si->numMuscles; i++)
   {
      muscle_force = calc_optimized_muscle_force(si->muscle[i], x[i]);

      // Now find the gencoord moments that this muscle force generates and add them to the array.
      for (j = 0; j < sdm->nu; j++)
         si->moment[j] += muscle_force * si->muscle[i]->momentarms[j];
   }

   // The second set of residuals is the difference between the current moments
   // and the desired moments (the inverse dynamics moments). Assume that the
   // latter have already been calculated).
   for (i = 0, index = 0; i < sdm->nu; i++)
   {
      if (si->q[i] == dpYes)
      {
         resid[si->numMuscles + index] = sdm->q[i].moment - si->moment[i];
         resid[si->numMuscles + index] = DABS(resid[si->numMuscles + index]) * K_TORQUE;
         index++;
      }
   }

   // Calculate the norm to return a value.
   ssq = 0.0;
   for (i = 0; i < si->numMuscles + si->numQs; i++)
   {
      ssq += resid[i] * resid[i];
   }
   norm = sqrt(ssq);

   free(resid);

   return norm;
}


static void lmdif_optimize_activations(OptimizationData* si)
{
   int i, j, *ipvt;
   int info, num_func_calls, mode = 1, nprint = 0, max_iter = K_ITER;
   double *activation, *diag, *qtf, *wa1, *wa2, *wa3, *wa4, *fjac, *resid;
   double ftol = 1e-4, xtol = 1e-4, gtol = 0.0;
   double epsfcn = 0.0, step_factor = 0.2;
   int numResid = si->numMuscles + si->numQs;
   int ldfjac = numResid;

   // Allocate space for necessary arrays.
   resid = (double *)simm_calloc(numResid, sizeof(double));
   ipvt = (int *)simm_malloc(si->numMuscles * sizeof(int));
   diag = (double *)simm_malloc(si->numMuscles * sizeof(double));
   fjac = (double *)simm_malloc(si->numMuscles * numResid * sizeof(double));
   qtf = (double *)simm_malloc(si->numMuscles * sizeof(double));
   wa1 = (double *)simm_malloc(si->numMuscles * sizeof(double));
   wa2 = (double *)simm_malloc(si->numMuscles * sizeof(double));
   wa3 = (double *)simm_malloc(si->numMuscles * sizeof(double));
   wa4 = (double *)simm_malloc(numResid * sizeof(double));

   lmdif_C(calc_lmdif_residuals, numResid, si->numMuscles,
      si->activation, resid, ftol, xtol, gtol, max_iter,
      epsfcn, diag, mode, step_factor, nprint, &info, &num_func_calls,
      fjac, ldfjac, ipvt, qtf, wa1, wa2, wa3, wa4, si);

   // Free all the space allocated.
   free(resid);
   free(ipvt);
   free(diag);
   free(fjac);
   free(qtf);
   free(wa1);
   free(wa2);
   free(wa3);
   free(wa4);
}


/* CALC_LMDIF_RESIDS: this function calculates residuals for optimize_muscle_activations().
 * There is one residual for each muscle's activation level, and one for each gencoord
 * moment that is being optimized.
 * was CALC_OPT_ACT_RESIDS
 */
void calc_lmdif_residuals(int numResid, int numActivations, double activation[],
                          double resid[], int* flag2, void* data)
{
   int i, j, index;
   double muscle_force, exc, min, max;
   dpMuscleStruct* muscle;
   OptimizationData* si = (OptimizationData*)data;

   // The first set of residuals is for the muscle activations.
   // If the activation is less than 0.0 or greater than 1.0, there
   // is a [stiff] penalty. If the activation is between 0.0 and 1.0,
   // there is a softer penalty to try to drive the activation close
   // to 0.0. If the muscle has an excitation curve specified in the
   // motion file or muscle file, use it in the residual calculations.
   for (i = 0; i < si->numMuscles; i++)
   {
      muscle = si->muscle[i];

      if (muscle->excitation_func)
      {
         // Calculate the current excitation level of the muscle. The
         // excitation curve may go below 0.0 or above 1.0, but we don't
         // want to allow that here because it may result in optimized
         // activations going outside the 0.0 to 1.0 range.
         exc = calc_excitation(muscle, si->time, si->state);
         if (exc < 0.0)
            exc = 0.0;
         else if (exc > 1.0)
            exc = 1.0;
#if 1
         // Assume that amplitude of excitation/EMG curve is dependable.
         // Try to keep activation within EMG_ENVELOPE of that amplitude.
         min = MAX(0.0, exc - EMG_ENVELOPE);
         max = MIN(1.0, exc + EMG_ENVELOPE);

         if (activation[i] < min)
            resid[i] = (exc - activation[i]) * K_ACT + (min - activation[i]) * K_NEG;
         else if (activation[i] > max)
            resid[i] = (activation[i] - exc) * K_ACT + (activation[i] - max) * K_POS;
         else if (activation[i] < exc)
            resid[i] = (exc - activation[i]) * K_ACT;
         else
            resid[i] = (activation[i] - exc) * K_ACT;
#else
         // Assume that amplitude of excitation/EMG curve is not dependable.
         // If excitation is less than EMG_THRESHOLD, force activation to be zero.
         // If excitation is above EMG_THRESHOLD, allow activation to range
         // from 0.0 to 1.0.
         if (exc < EMG_THRESHOLD)
         {
            if (activation[i] < 0.0)
               resid[i] = -activation[i] * K_NEG;
            else
               resid[i] = activation[i] * K_POS;
         }
         else
         {
            if (activation[i] < 0.0)
               resid[i] = -activation[i] * K_NEG;
            else if (activation[i] > 1.0)
               resid[i] = activation[i] * K_ACT + (activation[i] - 1.0) * K_POS;
            else
               resid[i] = activation[i] * K_ACT;
         }
#endif
      }
      else
      {
         if (activation[i] < 0.0)
            resid[i] = -activation[i] * K_NEG;
         else if (activation[i] > 1.0)
            resid[i] = activation[i] * K_ACT + (activation[i] - 1.0) * K_POS;
         else
            resid[i] = activation[i] * K_ACT;
      }
   }

   // Zero-out the current moments. This array will hold the total moments that
   // the muscles generate, given the current guess of activation levels.
   for (j = 0; j < sdm->nu; j++)
      si->moment[j] = 0.0;

   for (i = 0; i < si->numMuscles; i++)
   {
      muscle_force = calc_optimized_muscle_force(si->muscle[i], activation[i]);

      // Now find the gencoord moments that this muscle force generates and add them to the array.
      for (j = 0; j < sdm->nu; j++)
         si->moment[j] += muscle_force * si->muscle[i]->momentarms[j];
   }

   // The second set of residuals is the difference between the current moments
   // and the desired moments (the inverse dynamics moments). Assume that the
   // latter have already been calculated).
   for (i = 0, index = 0; i < sdm->nu; i++)
   {
      if (si->q[i] == dpYes)
      {
         resid[si->numMuscles + index] = sdm->q[i].moment - si->moment[i];
         resid[si->numMuscles + index] = DABS(resid[si->numMuscles + index]) * K_TORQUE;
         index++;
      }
   }
}
