/*******************************************************************************

   ASSIGNS.C

   Author: Peter Loan

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description:  This file contains 10 routines to store muscle state values
      after the state variables have been integrated for a given time step.
      For example, in a muscle model where activation is a state, you would
      want to store the activation-state-variable value in the muscle structure
      after each time step so that you can print it out or write it to a file
      without having to figure out which element in the state vector it is
      every time you want to do something with it. The routines are named
      muscle_assign_func1() through muscle_assign_func10(), and each one
      stores all of a muscle's states wherever the user wants.


   Routines:
      muscle_assign_func1  : 3 states (act, fiber_len, fiber_vel), no pennation
      muscle_assign_func2  : 3-state model, pennation included
      muscle_assign_func3  : model 2 plus 4th state for muscle power
      muscle_assign_func4  : 2 states (act, fiber_len), non-zero passive force
      muscle_assign_func5  : 1 state (act), infinitely stiff tendon
      muscle_assign_func6  : 4 state
      muscle_assign_func7  : 0 state, ligament model
      muscle_assign_func8  : model 4 plus two states for fatigue
      muscle_assign_func9  : 0 state, infinitely-stiff tendon (for static optimization)
      muscle_assign_func10 : 1 state (fiber_len) (for static optimization)
      muscle_assign_func11 : unused
      muscle_assign_func12 : unused
      muscle_assign_func13 : unused
      muscle_assign_func14 : unused
      muscle_assign_func15 : unused
      muscle_assign_func16 : unused
      muscle_assign_func17 : unused
      muscle_assign_func18 : unused
      muscle_assign_func19 : unused
      muscle_assign_func20 : unused

*******************************************************************************/

#include "universal.h"


int muscle_assign_func1(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 3;
   int activation = 0, fiber_length = 1, fiber_velocity = 2;

   muscle->dynamic_activation = state_euler[activation];
   muscle->fiber_length = state_euler[fiber_length];
   muscle->fiber_velocity = state_euler[fiber_velocity];

   return numStates;
}


int muscle_assign_func2(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 3;
   int activation = 0, fiber_length = 1, fiber_velocity = 2;

   muscle->dynamic_activation = state_euler[activation];
   muscle->fiber_length = state_euler[fiber_length];
   muscle->fiber_velocity = state_euler[fiber_velocity];

   return numStates;
}


int muscle_assign_func3(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 4;
   int activation = 0, fiber_length = 1, fiber_velocity = 2, energy = 3;

   muscle->dynamic_activation = state_euler[activation];
   muscle->fiber_length = state_euler[fiber_length];
   muscle->fiber_velocity = state_euler[fiber_velocity];
   muscle->mechanical_energy = state_euler[energy];

   return numStates;
}


int muscle_assign_func4(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 2;
   int activation = 0, fiber_length = 1;

   muscle->dynamic_activation = state_euler[activation];
   muscle->fiber_length = state_euler[fiber_length];

   return numStates;
}


int muscle_assign_func5(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 1;
   int activation = 0;

   muscle->dynamic_activation = state_euler[activation];

   return numStates;
}


int muscle_assign_func6(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 4;
   int activation = 0, fiber_length = 1, heat_energy = 2, mechanical_energy = 3;

   muscle->dynamic_activation = state_euler[activation];
   muscle->fiber_length = state_euler[fiber_length];
   muscle->heat_energy = state_euler[heat_energy];
   muscle->mechanical_energy = state_euler[mechanical_energy];

   return numStates;
}


int muscle_assign_func7(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

	muscle->dynamic_activation = 0.0;
	muscle->fiber_length = 0.0;

   return numStates;
}


int muscle_assign_func8(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 4;
   int activation = 0, fiber_length = 1, active_motor_units = 2, fatigued_motor_units = 3;

   muscle->dynamic_activation = state_euler[activation];
   muscle->fiber_length = state_euler[fiber_length];

   return numStates;
}


int muscle_assign_func9(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func10(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 1;
   int fiber_length = 0;

   muscle->fiber_length = state_euler[fiber_length];

   return numStates;
}


int muscle_assign_func11(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func12(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func13(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func14(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func15(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func16(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func17(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func18(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func19(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_assign_func20(double time, dpMuscleStruct *muscle, double state_euler[])
{
   int numStates = 0;

   return numStates;
}
