/*******************************************************************************

   BACKGROUNDSIM.C

   Authors: Peter Loan

   Copyright (c) 2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains the main routine for the DLL version of
      a dynamics simulation (for running within SIMM), and several other
      utility functions that are independent of the musculoskeletal model
      in the simulation. This file is generated by SIMM when you choose the
      "project files" option in the "Save Dynamics" dialog box. You will
      probably want to generate it once (not every time you change your
      model) and then customize it to your specific simulation.

*******************************************************************************/
#include "universal.h"

/*************** DEFINES (for this file only) *********************************/


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/


/**************** GLOBAL VARIABLES (used in only a few files) *****************/


/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern char buffer[];
extern dpBoolean verbose;
extern dpSimulationParameters params;
extern MotionData* input_data;
extern dpModelStruct* sdm;
extern SystemInfo si;
extern dpModelInfo* gModelInfo;
extern double t, *y, *dy, *work;
extern dpMessageStruct mess;
extern int (*dataCallbackFunc)(dpSimmModelID, int, void*);


/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/


DLL int dpRunBackgroundSimulation(dpSimulationParameters* _params, dpModelStruct* model)
{
   int rc = 1;

   copyParameters(_params, &params);

   chdir(params.workingDir);

   rc = set_model(model);
   if (rc < 0)
      return rc;

   return run_simulation();
}


/* set_model: this function is called when the "initialize" button is pressed
 * in the Dynamics Tool. It takes a dpModelStruct containing a complete description
 * of the SIMM model, and copies the information to the global variable sdm. It
 * then assembles the model, initializes the state vector and other variables,
 * then opens the appropriate output files.
 */
int set_model(dpModelStruct* model)
{
   int i, status, num_muscle_states;
   double param[1] = {-1.0};
   char *dllparameters_file_name = "dllparams.txt";
   char *contactparameters_file_name = "dllcontactparams.txt";

   // Copy the dpModelStruct to memory that is local to the DLL.
   sdm = (dpModelStruct*)simm_calloc(1, sizeof(dpModelStruct));
   copyModel(model, sdm);

   // Check to make sure the model is consistent with the one built
   // into the DLL (from model.sd).
   if (checkModel(sdm) < 0)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "SIM> Cannot run this simulation DLL with %s.", model->name);
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   // Count the number of muscle states, and quit the simulation if there are fewer than zero (meaning that there was an
   // error counting the states).
   num_muscle_states = count_muscle_states(sdm);
   if (num_muscle_states < 0)
      return -1;

   sdm->neq = sdm->nu + sdm->nq + num_muscle_states;

   y = (double*)simm_malloc(sdm->neq * sizeof(double));
   dy = (double*)simm_malloc(sdm->neq * sizeof(double));
   work = (double*)simm_malloc(6 * sdm->neq * sizeof(double));

   if (params.inputMotionFile)
      set_up_input_data(params.inputMotionFile, &input_data);

   read_dllparameters_file(dllparameters_file_name, sdm);

#if CONTACT_DETECTION
   if (read_contact_parameters_file(contactparameters_file_name) == code_bad)
      return -1;
#endif

   init_dll_model();

   // Check to make sure the output options have been specified properly, and find indices for the names.
   if (checkOutputOptions() < 0)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "SIM> Cannot run this simulation DLL with %s.", model->name);
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   if (set_initial_conditions(&t, y, dy, input_data) != code_fine)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "SIM> Unable to initialize the dynamic simulation with input motion data.");
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   init_motion();

   sdinit();
   check_for_sderror("SDINIT");
   
   // Enable constraint stabilization.
   sdstab(2.0 * params.baumgarte_stabilization, params.baumgarte_stabilization * params.baumgarte_stabilization);

   if (sdm->newInverseSimulation == dpYes)
   {
      sdm->newInverseSimulation = dpNo;
      if (assemble_model(t, y) == code_bad)
      {
         mess.id = dpFinished;
         sprintf(mess.text, "SIM> Unable to initialize the dynamic simulation.");
         if (dataCallbackFunc)
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         return -1;
      }
      sdm->newInverseSimulation = dpYes;
   }
   else
   {
      if (assemble_model(t, y) == code_bad)
      {
         mess.id = dpFinished;
         sprintf(mess.text, "SIM> Unable to initialize the dynamic simulation.");
         if (dataCallbackFunc)
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         return -1;
      }
   }

   sdstate(t, y, &y[sdm->nq]);

   if (init_muscle_states(sdm, y) == code_bad)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "SIM> Unable to initialize the dynamic simulation.");
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   if (params.outputOptions.normalization != dpNoNormalization)
      params.stepSize = (params.endTime - params.startTime) / 99.0;

   check_for_sderror("DLLMAIN");

#if CONTACT_DETECTION
   if (check_for_initial_penetration(y, dy) == code_bad)
      return 0;
#endif

   // Calculate initial accelerations.
   if (params.outputOptions.gencoordMoments.num_items > 0)
   {
      FILE* fp;
      dpBoolean newInverse = sdm->newInverseSimulation; // save a copy of this flag

      // For an inverse simulation, activate prescribed motion and prescribe
      // the Qs, Us, and As according to the input motion. Then write the
      // initial Qs and Us to a file. This is useful if, for example, you
      // later want to use the gencoord moments to drive a forward simulation
      // to reproduce the motion.
      if (newInverse == dpYes)
      {
         sdm->newInverseSimulation = dpNo;
         set_prescribed_motion(sdm, dpPrescribedQ, 1);
      }
      calc_derivatives(t, y, dy, param, &status);

      fp = fopen("initialConditions.txt","w");
		for (i = 0; i < sdm->nu; i++)
         fprintf(fp, "%s %21.16lf %21.16lf\n", sdm->q[i].name, y[i] * sdm->q[i].conversion, dy[i] * sdm->q[i].conversion);
      fclose(fp);

      // Set the newInverseSimulation flag back to what it was.
      sdm->newInverseSimulation = newInverse;
      if (newInverse == dpYes)
         set_prescribed_motion(sdm, dpPrescribedQ, 0);
   }
   else
   {
      calc_derivatives(t, y, dy, param, &status);
   }

   si.initial_system_energy = MINMDOUBLE;

   if (params.outputOptions.systemEnergy == dpYes)
      calc_system_energy(&si);

   if (params.outputOptions.jointReactionForces.num_items > 0 || params.outputOptions.jointReactionTorques.num_items > 0)
      get_reaction_forces();

   if (params.outputOptions.gencoordMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0) ||
      params.outputOptions.gencoordPowers.num_items > 0)
   {
      if (sdm->newInverseSimulation == dpYes)
            calc_initial_gencoord_moments(t, y, dy, input_data);
      else
         calc_gencoord_moments(t, y, dy);
   }

   // Moment arms should be calculated last because the function
   // calls sdstate() which resets all of the accelerations to zero.
   if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
      (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
      params.outputOptions.totalMuscleMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
      calc_muscle_moment_arms(t, y);

   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
   {
      char message[255];

      if (set_up_optimization(message) != code_fine)
      {
         mess.id = dpFinished;
         sprintf(mess.text, "SIM> Could not set up muscle optimization: %s.", message);
         if (dataCallbackFunc)
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         return -1;
      }
      optimize_muscle_activations(t, y);
   }

   if (params.outputMotionFile || dataCallbackFunc)
   {
      if (input_data && input_data->name)
      {
         // If the name of the input motion ends in "_IN", remove that part and replace it with "_OUT".
         int len = strlen(input_data->name);
         strcpy(buffer, input_data->name);
         if (len >= 3 && STRINGS_ARE_EQUAL(&buffer[len - 3], "_IN"))
            strcpy(&buffer[len - 3], "_OUT");
      }
      else
      {
         strcpy(buffer, "sim_motion_OUT");
      }

      create_output_motion(buffer);
      if (params.outputMotionFile)
         open_output_motion_file(params.outputMotionFile);
   }

   if (store_motion_frame(t, y, dy) != 0)
      return -1;

   return 1;
}


/* run_simulation: this function is called when the "run" button is pressed in
 * the Dynamics Tool. It contains the main integration for loop for the simulation.
 */
int run_simulation(void)
{
   int i, err, which, status, num_impacts = 0;
   double step, step_size, next_t;
   double param[1] = {-1.0};
   dpMessageStruct mess;
   double start_time, end_time, elapsed;

   start_time = get_time_now();

   for (t = params.startTime; t < params.endTime - TINY_TIME; )
   {
      step = step_size = params.stepSize;
      next_t = t + step_size;
      if (next_t > params.endTime)
      {
         next_t = params.endTime;
         step = step_size = next_t - t;
      }

      /* To make the simulation run faster, don't let the muscle excitations
       * change during an integration interval. Thus you can compute them here
       * and store them in the muscle structure. If you are spline-fitting
       * your excitation data, you don't really need to do this since the
       * muscle excitations will change smoothly (whereas with step functions
       * they will not), and thus they will not slow down the integration.
       * Note: Starting with version 2.0.1, all muscle derivative functions in
       * derivs.c depend on the muscle excitation being set here. If you remove
       * the code below, you will need to make sure that excitations are set
       * properly in the derivative function that you use.
       */
      for (i = 0; i < sdm->num_muscles; i++)
         sdm->muscles[i].excitation_level = calc_excitation(&sdm->muscles[i], t, y);
      
      while (t < next_t - TINY_TIME)
      {
         calc_derivatives(t, y, dy, param, &status);
         if (status != NO_FLAG)
         {
            print_final_information(t, y, dy, &si);
            goto END_SIMULATION;
         }
         integrate(calc_derivatives, &t, y, dy, param, step_size, &step, sdm->neq, INTEGRATION_TOL, work, &err, &which);

         if (err == INT_NO_ERROR)
         {
            break;
         }
#if CONTACT_DETECTION
         else if (err == INT_USER_FLAG)
         {
            if (which == IMPACT_FLAG)
            {
               handle_impacts(t, y, dy);
               num_impacts++;
               step = step_size = next_t - t;
            }
            else if (which == CONS_VIOL)
            {
               store_motion_frame(t, y, dy);
               sim_message(abort_action, "CONSTRAINT VIOLATION at time %lf.", t);
               goto END_SIMULATION;
            }
            else if (which == BAD_STATE)
            {
               store_motion_frame(t, y, dy);
               sim_message(abort_action, "BAD STATE at time %lf.", t);
               goto END_SIMULATION;
            }
         } 
#endif
         else if (err == INT_OVER_BUMP)
         {
            if (which >= 0 && which < sdm->nu)
               sim_message(recover, "INTEGRATOR: went over a bump (gencoord %s [position]).", sdm->q[which].name);
            else if (which >= sdm->nu && which < sdm->nq + sdm->nu)
               sim_message(recover, "INTEGRATOR: went over a bump (gencoord %s [velocity]).", sdm->q[which - sdm->nq].name);
            else
               sim_message(recover, "INTEGRATOR: went over a bump (state vector element %d).", which);
            step_size = next_t - t;
         } 
         else if (err == INT_CANT_CONTINUE)
         {
            sim_message(abort_action, "SDVINTEG: Cannot continue after time = %lf (gc %d).", t, which);
            store_motion_frame(t, y, dy);
            print_final_information(t, y, dy, &si);
            goto END_SIMULATION;
         } 
         else 
         {
            if (which == CONS_VIOL)
               sim_message(abort_action, "Constraint Violation.");
            print_final_information(t, y, dy, &si);
            goto END_SIMULATION;
         }
      }

      assign_muscle_states(sdm, t, y);

      if (verbose == dpYes)
         print_state(y, dy);

      if (params.outputOptions.systemEnergy == dpYes)
         calc_system_energy(&si);

      if (params.outputOptions.jointReactionForces.num_items > 0 || params.outputOptions.jointReactionTorques.num_items > 0)
         get_reaction_forces();

      if (params.outputOptions.gencoordMoments.num_items > 0 ||
         (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0) ||
         params.outputOptions.gencoordPowers.num_items > 0)
      {
         if (sdm->newInverseSimulation == dpYes)
         {
            if (t < params.endTime - TINY_TIME)
            {
               calc_next_gencoord_moments(t, y, dy, input_data);
               calc_corrective_moments(t, y, dy, sdm, input_data);
            }
            else
            {
               /* You're at the end of the simulation, so you don't need to
                * calculate corrective moments for the next time step. Just get
                * the corrected moments from this time step and store them in
                * the sdm->q[] array so they can be written to the output motion
                * file and used by the static muscle optimization function.
                */
					for (i = 0; i < sdm->nu; i++)
               {
                  if (input_data->q_moments[i])
                     sdm->q[i].moment = input_data->q_moments[i]->y[input_data->q_moments[i]->numpoints-1];
               }
            }
         }
         else
         {
            calc_gencoord_moments(t, y, dy);
         }
      }

      /* Moment arms should be calculated near the end because the function
       * calls sdstate() which resets all of the accelerations to zero.
       */
      if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
         (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
         params.outputOptions.totalMuscleMoments.num_items > 0 ||
         (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
         calc_muscle_moment_arms(t, y);

      if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
         optimize_muscle_activations(t, y);

      store_motion_frame(t, y, dy);
   }

END_SIMULATION:
   if (params.outputMotionFile)
      close_output_motion_file();

   free(work);
   free(y);
   free(dy);

   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
      cleanup_optimization();

   if (gModelInfo)
   {
      FREE_IFNOTNULL(gModelInfo->joint_type);
      FREE_IFNOTNULL(gModelInfo);
   }

   return 1;
}
