/*******************************************************************************

   DERIVS.C

   Authors: Peter Loan
            Chris Raasch (func3)
            Lisa Schutte (func4)
            Felix Zajac (func5)
            Ning Lan (func6)
            Steven Piazza (func7)

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains 10 routines to calculate state derivatives
      for dimensionless muscle-tendon models. The routines are named
      muscle_deriv_func1() through muscle_deriv_func10(), and each one
      calculates all of a muscle's derivatives for a particular muscle-tendon
      model.

   Routines:
      muscle_deriv_func1  : 3 states (act, fiber_len, fiber_vel), no pennation
      muscle_deriv_func2  : 3-state model, pennation included
      muscle_deriv_func3  : model 2 plus 4th state for muscle power
      muscle_deriv_func4  : 2 states (act, fiber_len), non-zero passive force
      muscle_deriv_func5  : 1 state (act), infinitely-stiff tendon
      muscle_deriv_func6  : 4 state
      muscle_deriv_func7  : 0 state, ligament model
      muscle_deriv_func8  : model 4 plus two states for fatigue
      muscle_deriv_func9  : 0 state, infinitely-stiff tendon (for static optimization)
      muscle_deriv_func10 : 1 state (fiber_len) (for static optimization)
      muscle_deriv_func11 : unused
      muscle_deriv_func12 : unused
      muscle_deriv_func13 : unused
      muscle_deriv_func14 : unused
      muscle_deriv_func15 : unused
      muscle_deriv_func16 : unused
      muscle_deriv_func17 : unused
      muscle_deriv_func18 : unused
      muscle_deriv_func19 : unused
      muscle_deriv_func20 : unused

NOTE: state = full state vector must be passed in using Euler angles, not Quaternions
      dstate = only muscle state derivatives are affected (store calculated values) therefore not converted

*******************************************************************************/

#include "universal.h"


/* A 3-state model, with states of activation, muscle fiber length, and muscle
 * fiber velocity. It should not be used for muscles with resting tendon lengths
 * of zero because it will always calculate a muscle force of zero.
 */
int muscle_deriv_func1(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 3;
   int activation = 0, fiber_length = 1, fiber_velocity = 2;
   double exc, tendon_force, fiber_force, muscle_mass;
   double normstate[3], normdstate[3], norm_tendon_length;
   double norm_muscle_tendon_length;
   double c_tau, mass_param, c_act_1, c_act_2;
   static int first_time = 1;
   static int c_tau_index = 0, mass_param_index = 0, c_act_1_index = 0, c_act_2_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      mass_param_index = get_muscle_param_index(muscle, "mass");
      c_act_1_index = get_muscle_param_index(muscle, "activation1");
      c_act_2_index = get_muscle_param_index(muscle, "activation2");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   mass_param = *(muscle->dynamic_params[mass_param_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in main.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];
   normstate[fiber_length] = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));
   normstate[fiber_velocity] = state_euler[muscle_state_start_index + fiber_velocity] * (c_tau / (*(muscle->optimal_fiber_length)));
  
   // Compute normalized muscle state derivatives
   normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   normdstate[fiber_length] = normstate[fiber_velocity];
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle, state_euler) / (*(muscle->optimal_fiber_length));  // pass full state vector
   norm_tendon_length = norm_muscle_tendon_length - normstate[fiber_length];

   tendon_force = calc_tendon_force(muscle, norm_tendon_length);
   fiber_force = calc_fiber_force(muscle, normstate[activation], normstate[fiber_length], normstate[fiber_velocity]);

   muscle_mass = mass_param * (*(muscle->optimal_fiber_length)/c_tau) * (*(muscle->optimal_fiber_length)/c_tau);
   normdstate[fiber_velocity] = (tendon_force - fiber_force)/muscle_mass;

   // Un-normalize the muscle state derivatives
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / c_tau;
   dstate_quat[muscle_state_start_index + fiber_length] = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   dstate_quat[muscle_state_start_index + fiber_velocity] = normdstate[fiber_velocity] * (*(muscle->optimal_fiber_length)) / (c_tau * c_tau);

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero.
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = state_euler[muscle_state_start_index + fiber_velocity];
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}



/* Model 2 is just like model 1 except that the fiber velocity equation accounts
 * for the pennation angle of the fibers. As with model 1, this one should not be
 * used for muscles with resting tendon lengths of zero.
*/

int muscle_deriv_func2(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 3;
   int activation = 0, fiber_length = 1, fiber_velocity = 2;
   double exc, tendon_force, fiber_force, muscle_mass;
   double normstate[3], normdstate[3], norm_tendon_length, ca, ta;
   double norm_muscle_tendon_length, mass_term, vel_term, pennation_angle;
   double c_tau, mass_param, c_act_1, c_act_2;
   static int first_time = 1;
   static int c_tau_index = 0, mass_param_index = 0, c_act_1_index = 0, c_act_2_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      mass_param_index = get_muscle_param_index(muscle, "mass");
      c_act_1_index = get_muscle_param_index(muscle, "activation1");
      c_act_2_index = get_muscle_param_index(muscle, "activation2");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   mass_param = *(muscle->dynamic_params[mass_param_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in dllmain.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];
   normstate[fiber_length] = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));
   normstate[fiber_velocity] = state_euler[muscle_state_start_index + fiber_velocity] * (c_tau / (*(muscle->optimal_fiber_length)));
  
   // Compute normalized muscle state derivatives
   normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   normdstate[fiber_length] = normstate[fiber_velocity];
   pennation_angle = calc_pennation(normstate[fiber_length], 1.0, (*(muscle->pennation_angle)));
   ca = cos(DEG_TO_RAD*pennation_angle);
   ta = tan(DEG_TO_RAD*pennation_angle);
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle, state_euler) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = norm_muscle_tendon_length - normstate[fiber_length]*ca;
   tendon_force = calc_tendon_force(muscle, norm_tendon_length);
   fiber_force = calc_fiber_force(muscle, normstate[activation], normstate[fiber_length], normstate[fiber_velocity]);

   muscle_mass = mass_param * (*(muscle->optimal_fiber_length) / c_tau) * (*(muscle->optimal_fiber_length)/c_tau);
   mass_term = (tendon_force * ca - fiber_force * ca * ca) / muscle_mass;
   vel_term = normstate[fiber_velocity] * normstate[fiber_velocity] * ta * ta / normstate[fiber_length];
   normdstate[fiber_velocity] = mass_term + vel_term;

   // Un-normalize the muscle state derivatives
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / c_tau;
   dstate_quat[muscle_state_start_index + fiber_length] = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   dstate_quat[muscle_state_start_index + fiber_velocity] = normdstate[fiber_velocity] * (*(muscle->optimal_fiber_length)) / (c_tau * c_tau);

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero.
   
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = state_euler[muscle_state_start_index + fiber_velocity];
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}



/* Model 3 is just like model 2 but with a fourth state for muscle energy,
 * which is the integral of the power applied to the body segments. It should
 * not be used for muscles with resting tendon lengths of zero because it
 * will always calculate a muscle force of zero. This model was designed by
 * Chris Raasch, Ph.D.
 */

int muscle_deriv_func3(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 4;
   int activation = 0, fiber_length = 1, fiber_velocity = 2, energy = 3;
   double exc, ca, ta, tendon_force, fiber_force, muscle_mass;
   double normstate[4], normdstate[4], norm_tendon_length;
   double pennation_angle, mass_term, vel_term, norm_muscle_tendon_length;
   double c_tau, mass_param, c_act_1, c_act_2;
   static int first_time = 1;
   static int c_tau_index = 0, mass_param_index = 0, c_act_1_index = 0, c_act_2_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle,  "timescale");
      mass_param_index = get_muscle_param_index(muscle,  "mass");
      c_act_1_index = get_muscle_param_index(muscle,  "activation1");
      c_act_2_index = get_muscle_param_index(muscle,  "activation2");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   mass_param = *(muscle->dynamic_params[mass_param_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in dllmain.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];
   normstate[fiber_length] = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));
   normstate[fiber_velocity] = state_euler[muscle_state_start_index + fiber_velocity] * (c_tau / (*(muscle->optimal_fiber_length)));
   normstate[energy] = state_euler[muscle_state_start_index + energy];
  
   // Compute normalized muscle state derivatives
   normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   normdstate[fiber_length] = normstate[fiber_velocity];
   pennation_angle = calc_pennation(normstate[fiber_length], 1.0, (*(muscle->pennation_angle)));
   ca = cos(DEG_TO_RAD*pennation_angle);
   ta = tan(DEG_TO_RAD*pennation_angle);
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = norm_muscle_tendon_length - normstate[fiber_length]*ca;
   tendon_force = calc_tendon_force(muscle, norm_tendon_length);
   fiber_force = calc_fiber_force(muscle, normstate[activation], normstate[fiber_length], normstate[fiber_velocity]);

   muscle_mass = mass_param * (*(muscle->optimal_fiber_length)/c_tau) * (*(muscle->optimal_fiber_length)/c_tau);
   mass_term = (tendon_force*ca - fiber_force*ca*ca) / muscle_mass;
   vel_term = normstate[fiber_velocity] * normstate[fiber_velocity] * ta * ta / normstate[fiber_length];
   normdstate[fiber_velocity] = mass_term + vel_term;
   calc_muscle_power(muscle);
   normdstate[energy] = muscle->applied_power;

   // Un-normalize the muscle state derivatives
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / get_muscle_param_value(muscle, "timescale");
   dstate_quat[muscle_state_start_index + fiber_length] = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   dstate_quat[muscle_state_start_index + fiber_velocity] = normdstate[fiber_velocity] * (*(muscle->optimal_fiber_length)) / (c_tau * c_tau);
   dstate_quat[muscle_state_start_index + energy] = normdstate[energy];
  
   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero.  
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = state_euler[muscle_state_start_index + fiber_velocity];
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}



/* Model 4 is a 2-state model which was designed to improve performance
 * and accuracy from the 3-state model (model 2). It does not include
 * muscle mass, but uses an inverse force-velocity relation which
 * includes damping. It should not be used for muscles with resting tendon
 * lengths of zero because it will always calculate a muscle force of zero.
 * This model was developed by Lisa Schutte, Ph.D.
 */
int muscle_deriv_func4(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 2;
   int activation = 0, fiber_length = 1;
   double exc, tendon_force, active_force;
   double passive_force, velocity_dependent_force;
   double normstate[2], normdstate[2], norm_tendon_length, ca;
   double norm_muscle_tendon_length, pennation_angle;
   double h, w, new_fiber_length, new_pennation_angle, new_ca;
   double c_tau, c_act_1, c_act_2;
   static int first_time = 1;
   static int c_tau_index = 0, c_act_1_index = 0, c_act_2_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      c_act_1_index = get_muscle_param_index(muscle, "activation1");
      c_act_2_index = get_muscle_param_index(muscle, "activation2");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in dllmain.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];
   normstate[fiber_length] = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));

   // Compute normalized muscle state derivatives
   if (exc >= normstate[activation])
      normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   else
      normdstate[activation] = (exc - normstate[activation]) * c_act_2;

   pennation_angle = calc_pennation(normstate[fiber_length], 1.0, (*(muscle->pennation_angle)));
   ca = cos(DEG_TO_RAD*pennation_angle);
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = norm_muscle_tendon_length - normstate[fiber_length] * ca;
   tendon_force = calc_tendon_force(muscle,  norm_tendon_length);
   passive_force = calc_nonzero_passive_force(muscle,  normstate[fiber_length],0.0);
   active_force = calc_active_force(muscle,  normstate[fiber_length]);

   // If pennation equals 90 degrees, fiber length equals muscle width and fiber
   // velocity goes to zero.  Pennation will stay at 90 until tendon starts to
   // pull, then "stiff tendon" approximation is used to calculate approximate
   // fiber velocity.
   if (EQUAL_WITHIN_ERROR(ca,0.0))
   {
      if (EQUAL_WITHIN_ERROR(tendon_force,0.0))
      {
         normdstate[fiber_length] = 0.0;
      }
      else
      {
         h = muscle->muscle_tendon_length - (*(muscle->resting_tendon_length));
         w = (*(muscle->optimal_fiber_length)) * sin((*(muscle->pennation_angle)) * DEG_TO_RAD);
         new_fiber_length = sqrt(h*h + w*w) / (*(muscle->optimal_fiber_length));
         new_pennation_angle = calc_pennation(new_fiber_length, 1.0, (*(muscle->pennation_angle)));
         new_ca = cos(new_pennation_angle * DEG_TO_RAD);
         normdstate[fiber_length] = calc_muscle_tendon_velocity(muscle) * c_tau / (*(muscle->optimal_fiber_length)) * new_ca;
      }
   }
   else
   {
      velocity_dependent_force = tendon_force / ca - passive_force;
      normdstate[fiber_length] = calc_fiber_velocity(muscle,  normstate[activation], active_force, velocity_dependent_force);
   }

   // Un-normalize the muscle state derivatives
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / c_tau;
   dstate_quat[muscle_state_start_index + fiber_length] = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero. 
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}



/* Model 5 is a 1-state model that was designed to be significantly faster
 * than any other model. It uses an infinitely stiff tendon to remove
 * the fiber_len state, leaving only activation dynamics. This model can be
 * used for muscles which have resting tendon lengths of zero. The model was
 * designed by Felix Zajac, Ph.D., at Stanford University.
 */

int muscle_deriv_func5(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 1;
   int activation = 0;
   double exc, ca, h, w, fiber_force, tendon_force;
   double norm_fiber_length, norm_fiber_velocity;
   double normstate[1], normdstate[1], norm_tendon_length;
   double norm_muscle_tendon_length, pennation_angle;
   double c_tau, c_act_1, c_act_2;
   static int first_time = 1;
   static int c_tau_index = 0, c_act_1_index = 0, c_act_2_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      c_act_1_index = get_muscle_param_index(muscle, "activation1");
      c_act_2_index = get_muscle_param_index(muscle, "activation2");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in dllmain.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];

   // Compute normalized muscle state derivatives
   if (exc - normstate[activation] >= 0.0)
      normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   else
      normdstate[activation] = (exc - normstate[activation]) * c_act_2;

/* Without pennation:
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = (*(muscle->resting_tendon_length)) / (*(muscle->optimal_fiber_length));
   norm_fiber_length = (norm_muscle_tendon_length - norm_tendon_length);
   norm_fiber_velocity = calc_muscle_tendon_velocity(muscle) * c_tau / (*(muscle->optimal_fiber_length));
   tendon_force = calc_fiber_force(muscle, normstate[activation],norm_fiber_length,
                                   norm_fiber_velocity);
*/

   // With pennation:
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler);
   norm_tendon_length = (*(muscle->resting_tendon_length));
   h = norm_muscle_tendon_length - norm_tendon_length;
   w = (*(muscle->optimal_fiber_length)) * sin((*(muscle->pennation_angle))*DEG_TO_RAD);

   if (EQUAL_WITHIN_ERROR(h,0.0) || h < 0.0)
   {
      norm_fiber_length = w / (*(muscle->optimal_fiber_length));
      norm_fiber_velocity = 0.0;
      tendon_force = 0.0;
   }
   else
   {
      norm_fiber_length = sqrt(h*h + w*w) / (*(muscle->optimal_fiber_length));
      pennation_angle = calc_pennation(norm_fiber_length, 1.0, (*(muscle->pennation_angle)));
      ca = cos(DEG_TO_RAD * pennation_angle);
      norm_fiber_velocity = calc_muscle_tendon_velocity(muscle) * get_muscle_param_value(muscle, "timescale") / (*(muscle->optimal_fiber_length)) * ca;

      fiber_force = calc_fiber_force(muscle,  normstate[activation], norm_fiber_length, norm_fiber_velocity);
      tendon_force = fiber_force * ca;
   }

   // Un-normalize the muscle state derivatives
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / c_tau;

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero.
   muscle->fiber_length = norm_fiber_length * (*(muscle->optimal_fiber_length));
   muscle->fiber_velocity = norm_fiber_velocity * (*(muscle->optimal_fiber_length)) / c_tau;
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}


/* Model 6 is a 4-state model that combines model 4's two states
 * with two new energy states. This model does not include
 * muscle mass, but uses an inverse force-velocity relation which
 * includes damping. It also contains a state for storing the heat
 * energy that a muscle generates as well as a state for storing the
 * mechanical energy that a muscle applies to the body segments.
 * This model should not be used for muscles which have resting tendon
 * lengths of zero because it will always calculate a muscle force of
 * zero. The model was designed and implemented by Ning Lan, Ph.D., at
 * the University of Southern California.
 */

int muscle_deriv_func6(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 4;
   int activation = 0, fiber_length = 1, heat_energy = 2,mechanical_energy = 3;
   double exc, tendon_force, active_force, v_max;
   double passive_force, velocity_dependent_force;
   double normstate[4], normdstate[4], norm_tendon_length, ca;
   double norm_muscle_tendon_length, pennation_angle;
   double h, w, new_fiber_length, new_pennation_angle, new_ca;
   double c_tau, c_act_1, c_act_2, c_heat_r, c_heat_c1, c_heat_alpha;
   static int first_time = 1;
   static int c_tau_index = 0, c_act_1_index = 0, c_act_2_index = 0;
   static int c_heat_r_index = 0, c_heat_c1_index = 0, c_heat_alpha_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      c_act_1_index = get_muscle_param_index(muscle, "activation1");
      c_act_2_index = get_muscle_param_index(muscle, "activation2");
      c_heat_r_index = get_muscle_param_index(muscle, "heat_r");
      c_heat_c1_index = get_muscle_param_index(muscle, "heat_c1");
      c_heat_alpha_index = get_muscle_param_index(muscle, "heat_alpha");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);
   c_heat_r = *(muscle->dynamic_params[c_heat_r_index]);
   c_heat_c1 = *(muscle->dynamic_params[c_heat_c1_index]);
   c_heat_alpha = *(muscle->dynamic_params[c_heat_alpha_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in dllmain.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];
   normstate[fiber_length] = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));
   normstate[heat_energy] = state_euler[muscle_state_start_index + heat_energy];
   normstate[mechanical_energy] = state_euler[muscle_state_start_index + mechanical_energy];

   // Compute normalized muscle state derivatives
   if (exc >= normstate[activation])
      normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   else
      normdstate[activation] = (exc - normstate[activation]) * c_act_2;

   pennation_angle = calc_pennation(normstate[fiber_length],1.0, (*(muscle->pennation_angle)));
   ca = cos(DEG_TO_RAD*pennation_angle);
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = norm_muscle_tendon_length - normstate[fiber_length] * ca;
   tendon_force = calc_tendon_force(muscle, norm_tendon_length);
   passive_force = calc_nonzero_passive_force(muscle,  normstate[fiber_length], 0.0);
   active_force = calc_active_force(muscle,  normstate[fiber_length]);

   // If pennation equals 90 degrees, fiber length equals muscle width and fiber
   // velocity goes to zero.  Pennation will stay at 90 until tendon starts to
   // pull, then "stiff tendon" approximation is used to calculate approximate
   // fiber velocity (a kluge, I know)

   if (EQUAL_WITHIN_ERROR(ca,0.0))
   {
      if (EQUAL_WITHIN_ERROR(tendon_force,0.0))
      {
         normdstate[fiber_length] = 0.0;
      }
      else
      {
         h = norm_muscle_tendon_length - (*(muscle->resting_tendon_length));
         w = (*(muscle->optimal_fiber_length)) * sin((*(muscle->pennation_angle))*DEG_TO_RAD);
         new_fiber_length = sqrt(h*h + w*w) / (*(muscle->optimal_fiber_length));
         new_pennation_angle = calc_pennation(new_fiber_length,1.0, (*(muscle->pennation_angle)));
         new_ca = cos(new_pennation_angle);
         normdstate[fiber_length] = calc_muscle_tendon_velocity(muscle) * c_tau / (*(muscle->optimal_fiber_length)) * new_ca;
      }
   }
   else
   {
      velocity_dependent_force = tendon_force / ca - passive_force;
      normdstate[fiber_length] = calc_fiber_velocity(muscle,  normstate[activation], active_force, velocity_dependent_force);
   }

   v_max = *(muscle->max_contraction_vel);

   if (normdstate[fiber_length] <= (0.051 * v_max / c_tau))
   {
      normdstate[heat_energy] =
         ((0.3 * c_heat_c1 + 0.7 * c_heat_c1 * active_force) * muscle->dynamic_activation +
         c_heat_alpha * v_max) * muscle->dynamic_activation * c_heat_r;
   }
   else
   {
      normdstate[heat_energy] =
         ((0.3 * c_heat_c1 + 0.7 * c_heat_c1 * active_force) * muscle->dynamic_activation -
         c_heat_alpha * v_max - 0.035) * muscle->dynamic_activation * c_heat_r;
   }

   calc_muscle_power(muscle);
   normdstate[mechanical_energy] = muscle->applied_power;

   // Un-normalize the muscle state derivatives
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / c_tau;
   dstate_quat[muscle_state_start_index + fiber_length] = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   dstate_quat[muscle_state_start_index + heat_energy] = normdstate[heat_energy];
   dstate_quat[muscle_state_start_index + mechanical_energy] = normdstate[mechanical_energy];

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero.
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}

/* Model 7 is a zero-state model that computes ligament forces.  Ligaments
 * are assumed to be modeled as springs with force-elongation curves given
 * by the tendon force length curve of the "muscle". This model should not
 * be used for muscles with resting tendon lengths of zero. The model was
 * developed by Stephen Piazza, Ph.D., now at Pennsylvania State University.
 */

int muscle_deriv_func7(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 0;
   double ligament_force, ligament_length, ligament_strain;

   // the entire muscle-tendon unit is ligament
   ligament_length = calc_muscle_tendon_length(muscle, state_euler);
   
   // strain = (length - slack_length) / slack_length
   if (*(muscle->resting_tendon_length) < ROUNDOFF_ERROR)
      ligament_strain = 0.0;
   else
      ligament_strain = (ligament_length - (*(muscle->resting_tendon_length))) / (*(muscle->resting_tendon_length));

   // ligament force = 0 if length < slack_length
   //                = Kx^2 if length > slack_length, x = strain
   // tendon_force_length_curve should be y = X^2
   if (ligament_strain < 0.0)
      ligament_force = 0.0;
   else
      ligament_force = interpolate_function(ligament_strain, *muscle->tendon_force_len_func, zeroth, 1.0, 1.0);
   
   // force = strain^2 * K
   muscle->force = ligament_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;
   muscle->fiber_length = 0.0;
   muscle->fiber_velocity = 0.0;

   *muscle_force = muscle->force;
   
   return numStates;
}


/* Model 8 is model 4 plus two states for modeling fatigue and recovery. The fatigue equations are from:
 * Liu, Jing Z., Brown, Robert, Yue, Guang H., "A Dynamical Model of Muscle
 * Activation, Fatigue, and Recovery," Biophysical Journal, Vol. 82, Issue 5,
 * pp. 2344-2359, 2002.
 */
int muscle_deriv_func8(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 4;
   int activation = 0, fiber_length = 1, active_motor_units = 2, fatigued_motor_units = 3;
   double exc, tendon_force, active_force;
   double passive_force, velocity_dependent_force;
   double normstate[4], normdstate[4], norm_tendon_length, ca;
   double norm_muscle_tendon_length, pennation_angle;
   double h, w, new_fiber_length, new_pennation_angle, new_ca;
   double c_tau, c_act_1, c_act_2, fatigue_factor, recovery_factor;
   static int first_time = 1;
   static int c_tau_index = 0, c_act_1_index = 0, c_act_2_index = 0, fatigue_factor_index = 0, recovery_factor_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      c_act_1_index = get_muscle_param_index(muscle, "activation1");
      c_act_2_index = get_muscle_param_index(muscle, "activation2");
      fatigue_factor_index = get_muscle_param_index(muscle, "fatigue_factor");
      recovery_factor_index = get_muscle_param_index(muscle, "recovery_factor");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);
   c_act_1 = *(muscle->dynamic_params[c_act_1_index]);
   c_act_2 = *(muscle->dynamic_params[c_act_2_index]);
   fatigue_factor = *(muscle->dynamic_params[fatigue_factor_index]);
   recovery_factor = *(muscle->dynamic_params[recovery_factor_index]);

   // Assume that the excitation level has been calculated and stored in the muscle
   // structure before the integrator is called. See note in dllmain.c for details.
   exc = muscle->excitation_level;

   // Normalize the muscle states
   normstate[activation] = state_euler[muscle_state_start_index + activation];
   normstate[fiber_length] = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));
   normstate[active_motor_units] = state_euler[muscle_state_start_index + active_motor_units];
   normstate[fatigued_motor_units] = state_euler[muscle_state_start_index + fatigued_motor_units];

   // Compute normalized muscle state derivatives
   if (exc >= normstate[activation])
      normdstate[activation] = (exc - normstate[activation]) * (c_act_1 * exc + c_act_2);
   else
      normdstate[activation] = (exc - normstate[activation]) * c_act_2;

   pennation_angle = calc_pennation(normstate[fiber_length], 1.0, (*(muscle->pennation_angle)));
   ca = cos(DEG_TO_RAD*pennation_angle);
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = norm_muscle_tendon_length - normstate[fiber_length] * ca;
   tendon_force = calc_tendon_force(muscle,  norm_tendon_length);
   passive_force = calc_nonzero_passive_force(muscle,  normstate[fiber_length],0.0);
   active_force = calc_active_force(muscle,  normstate[fiber_length]);

   // If pennation equals 90 degrees, fiber length equals muscle width and fiber
   // velocity goes to zero.  Pennation will stay at 90 until tendon starts to
   // pull, then "stiff tendon" approximation is used to calculate approximate
   // fiber velocity.
   if (EQUAL_WITHIN_ERROR(ca,0.0))
   {
      if (EQUAL_WITHIN_ERROR(tendon_force,0.0))
      {
         normdstate[fiber_length] = 0.0;
      }
      else
      {
         h = muscle->muscle_tendon_length - (*(muscle->resting_tendon_length));
         w = (*(muscle->optimal_fiber_length)) * sin((*(muscle->pennation_angle)) * DEG_TO_RAD);
         new_fiber_length = sqrt(h*h + w*w) / (*(muscle->optimal_fiber_length));
         new_pennation_angle = calc_pennation(new_fiber_length, 1.0, (*(muscle->pennation_angle)));
         new_ca = cos(new_pennation_angle * DEG_TO_RAD);
         normdstate[fiber_length] = calc_muscle_tendon_velocity(muscle) * c_tau / (*(muscle->optimal_fiber_length)) * new_ca;
      }
   }
   else
   {
      velocity_dependent_force = tendon_force / ca - passive_force;
      normdstate[fiber_length] = calc_fiber_velocity(muscle,  normstate[active_motor_units], active_force, velocity_dependent_force);
   }

	normdstate[active_motor_units] = normdstate[activation] - fatigue_factor * normstate[active_motor_units] + recovery_factor * normstate[fatigued_motor_units];
	normdstate[fatigued_motor_units] = fatigue_factor * normstate[active_motor_units] - recovery_factor * normstate[fatigued_motor_units];

   // Un-normalize the muscle state derivatives.
   dstate_quat[muscle_state_start_index + activation] = normdstate[activation] / c_tau;
   dstate_quat[muscle_state_start_index + fiber_length] = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   dstate_quat[muscle_state_start_index + active_motor_units] = normdstate[active_motor_units];
   dstate_quat[muscle_state_start_index + fatigued_motor_units] = normdstate[fatigued_motor_units];

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero. 
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = normdstate[fiber_length] * (*(muscle->optimal_fiber_length)) / c_tau;
   muscle->force = tendon_force * (*(muscle->max_isometric_force));
   if (muscle->force < 0.0)
      muscle->force = 0.0;

   *muscle_force = muscle->force;

   return numStates;
}


int muscle_deriv_func9(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                       double dstate_quat[], double* muscle_force)
{
   int numStates = 0;
   double ca, h, w, fiber_force, tendon_force;
   double norm_fiber_length, norm_fiber_velocity;
   double normstate[1], normdstate[1], norm_tendon_length;
   double norm_muscle_tendon_length, pennation_angle;
   double c_tau, c_act_1, c_act_2;
   static int first_time = 1;
   static int c_tau_index = 0, c_act_1_index = 0, c_act_2_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);

   // With pennation:
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler);
   norm_tendon_length = (*(muscle->resting_tendon_length));
   h = norm_muscle_tendon_length - norm_tendon_length;
   w = (*(muscle->optimal_fiber_length)) * sin((*(muscle->pennation_angle))*DEG_TO_RAD);

   if (EQUAL_WITHIN_ERROR(h,0.0) || h < 0.0)
   {
      norm_fiber_length = w / (*(muscle->optimal_fiber_length));
      norm_fiber_velocity = 0.0;
      tendon_force = 0.0;
   }
   else
   {
      norm_fiber_length = sqrt(h*h + w*w) / (*(muscle->optimal_fiber_length));
      pennation_angle = calc_pennation(norm_fiber_length, 1.0, (*(muscle->pennation_angle)));
      ca = cos(DEG_TO_RAD * pennation_angle);
      norm_fiber_velocity = calc_muscle_tendon_velocity(muscle) * get_muscle_param_value(muscle, "timescale") / (*(muscle->optimal_fiber_length)) * ca;

      fiber_force = calc_fiber_force(muscle, muscle->dynamic_activation, norm_fiber_length, norm_fiber_velocity);
      tendon_force = fiber_force * ca;
   }

   *muscle_force = tendon_force * (*(muscle->max_isometric_force));
   if (*muscle_force < 0.0)
      *muscle_force = 0.0;

   // Store force, length, and velocity in the muscle structure.
   // Don't let muscle force go below zero.
   muscle->force = 0.0;
   muscle->fiber_length = norm_fiber_length * (*(muscle->optimal_fiber_length));
   muscle->fiber_velocity = norm_fiber_velocity * (*(muscle->optimal_fiber_length)) / c_tau;

   return numStates;
}


/* Model 10 is specifically designed to work with optimize_muscle_activations().
 * It is similar to model 4, except activation is not included because that
 * is the variable being optimized. Also, it is important to set
 * muscle->force to 0.0 at the end of this function so that no force is applied
 * to the body segments. This muscle model is for use during inverse dynamics,
 * and applying muscles forces would change the gencoord moments that are
 * calculated and used in the optimization. Thus this muscle model is used
 * only to calculate up-to-date values of muscle_tendon_length, fiber_length,
 * and fiber_velocity. The actual force in the muscle is calculated by a
 * different function, calc_optimized_muscle_force(), using the state
 * variables calculated by this function.
 */

int muscle_deriv_func10(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 1;
   int fiber_length = 0;
   double tendon_force, active_force;
   double passive_force, velocity_dependent_force;
   double norm_fiber_length, norm_fiber_velocity, norm_tendon_length, ca;
   double norm_muscle_tendon_length, pennation_angle;
   double h, w, new_fiber_length, new_pennation_angle, new_ca;
   double c_tau;
   static int first_time = 1;
   static int c_tau_index = 0;

   if (first_time)
   {
      c_tau_index = get_muscle_param_index(muscle, "timescale");
      first_time = 0;
   }

   c_tau = *(muscle->dynamic_params[c_tau_index]);

   // Normalize the muscle state derivative
   norm_fiber_length = state_euler[muscle_state_start_index + fiber_length] / (*(muscle->optimal_fiber_length));
  
   pennation_angle = calc_pennation(norm_fiber_length, 1.0, (*(muscle->pennation_angle)));
   ca = cos(DEG_TO_RAD*pennation_angle);
   norm_muscle_tendon_length = calc_muscle_tendon_length(muscle,  state_euler) / (*(muscle->optimal_fiber_length));
   norm_tendon_length = norm_muscle_tendon_length - norm_fiber_length * ca;
   tendon_force = calc_tendon_force(muscle,  norm_tendon_length);
   passive_force = calc_nonzero_passive_force(muscle,  norm_fiber_length, 0.0);
   active_force = calc_active_force(muscle,  norm_fiber_length);

   // If pennation equals 90 degrees, fiber length equals muscle width and fiber
   // velocity goes to zero.  Pennation will stay at 90 until tendon starts to
   // pull, then "stiff tendon" approximation is used to calculate approximate
   // fiber velocity.

   if (EQUAL_WITHIN_ERROR(ca,0.0))
   {
      if (EQUAL_WITHIN_ERROR(tendon_force,0.0))
      {
         norm_fiber_length = sin((*(muscle->pennation_angle)) * DEG_TO_RAD);
         norm_fiber_velocity = 0.0;
      }
      else
      {
         h = norm_muscle_tendon_length - (*(muscle->resting_tendon_length));
         w = (*(muscle->optimal_fiber_length)) * sin((*(muscle->pennation_angle)) * DEG_TO_RAD);
         new_fiber_length = sqrt(h*h + w*w) / (*(muscle->optimal_fiber_length));
         new_pennation_angle = calc_pennation(new_fiber_length, 1.0, (*(muscle->pennation_angle)));
         new_ca = cos(new_pennation_angle);
         norm_fiber_velocity = calc_muscle_tendon_velocity(muscle) * c_tau / (*(muscle->optimal_fiber_length)) * new_ca;
      }
   }
   else
   {
      velocity_dependent_force = tendon_force / ca - passive_force;
      norm_fiber_velocity = calc_fiber_velocity(muscle,  muscle->dynamic_activation, active_force, velocity_dependent_force);
   }

   // Un-normalize the muscle state derivative
   dstate_quat[muscle_state_start_index + fiber_length] = norm_fiber_velocity * (*(muscle->optimal_fiber_length)) / c_tau;

   *muscle_force = tendon_force * (*(muscle->max_isometric_force));

   // Store force, length, and velocity in the muscle structure  
   muscle->force = 0.0; // Set muscle force to zero so that it won't apply any force to the segments.
   muscle->fiber_length = norm_fiber_length * (*(muscle->optimal_fiber_length));
   muscle->fiber_velocity = dstate_quat[muscle_state_start_index + fiber_length];

   return numStates;
}


int muscle_deriv_func11(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func12(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func13(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func14(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func15(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func16(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func17(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func18(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func19(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}


int muscle_deriv_func20(double time, dpMuscleStruct* muscle, int muscle_state_start_index, double state_euler[],
                        double dstate_quat[], double* muscle_force)
{
   int numStates = 0;

   return numStates;
}
