/*******************************************************************************

   DP.H

   Authors: Peter Loan
            Krystyne Blaikie

   Version 6.0, June 8, 2011

   Copyright (c) 2004-2011 MusculoGraphics, Inc, a division of Motion Analysis Corp.
   All rights reserved.

   This header file defines the data structures used to hold a musculoskeletal
   model, as well as the interface between SIMM the DLLs made by the Dynamics
   Pipeline. If you are using simulation DLLs, do not modify this file,
   as it is also compiled into SIMM. If you are using stand-alone simulation
   executables, you can modify this file as needed.

*******************************************************************************/

#ifndef DP_H
#define DP_H

#ifdef _WINDOWS
    #define DLL __declspec(dllexport)
#else
    #define DLL
#endif

/* The structures at the top of this file are not expected to change from one version
 * of SIMM and Pipeline to another. Thus they do not need to be versioned and put in
 * dp420.h, dp500.h, etc.
 */

typedef void* dpSimmModelID;

typedef enum
{
   dpNo,
   dpYes
} dpBoolean;

typedef enum
{
   dp123, 
   dp132, 
   dp213, 
   dp231, 
   dp312, 
   dp321, 
   dpUndefinedOrder
} dpEulerAngleOrder;

typedef struct
{
   double xyz[3];
} dpCoord3D;

typedef enum
{
   dpPin,
   dpReversePin,
   dpUniversal,
   dpReverseUniversal,
   dpGimbal,
   dpReverseGimbal,
   dpBall,
   dpReverseBall,
   dpSlider,
   dpReverseSlider,
   dpCylindrical,
   dpReverseCylindrical,
   dpPlanar,
   dpReversePlanar,
   dpBushing,
   dpReverseBushing,
   dpBearing,
   dpReverseBearing,
   dpFree,
   dpReverseFree,
   dpWeld,
   dpSkippable,
   dpUnknownJoint
} dpJointType;

typedef enum
{
   dpTranslational,
   dpRotational,
   dpNoDof
} dpDOFType;

typedef struct
{
   double a, b, c, d;
} dpPlaneStruct;

typedef struct
{
   double x1;
   double x2;
   double y1;
   double y2;
   double z1;
   double z2;
} dpBoundingCube;

/* The collision detection structures below may change from version to version,
 * but they are not used in SIMM, so they do not need to be versioned. They are
 * in this file because they are needed by other structures in the file.
 */
typedef enum
{
   dpUnknownContact,
   dpImpact,
   dpRestingContact,
   dpSeparatingContact
} dpCollisionTypes;

typedef struct
{
   int body1;                     /* SDFast index of body1 */
   int body2;                     /* SDFast index of body2 */
   double pt1[3];                 /* contact point location in body1 frame from b1 com */
   double pt2[3];                 /* contact point location in body2 frame from b2 com */
   double norm1[3];               /* contact normal on body1 in body1 frame */
   double norm2[3];               /* contact normal on body2 in body2 frame */
   double outward_norm[3];        /* from body1 to body2 in ground frame */
   double vel1[3];                /* velocity of contact point on body1 wrt ground */
   double vel2[3];                /* velocity of contact point on body2 wrt ground */
   double rel_vel[3];             /* relative velocity v2 - v1 */
   double cont_frc[3];            /* contact force or impulse */
   double acc;                    /* magnitude of normal relative acceleration */
   double coef_rest;              /* coefficient of restitution */
   double mu_dynamic;             /* dynamic friction coefficient */
   double mu_static;              /* static friction coefficient */
   dpCollisionTypes contact_type; /* IMPACT, CONTACT, or SEPARATING */
   char type1[30];                /* FACE, VERTEX, or EDGE */
   char type2[30];                /* FACE, VERTEX, or EDGE */
   double dist;                   /* for ideal contacts along outward normal */
} dpContactInfo;                  /* contact node */

/* end of versionless structures */

typedef enum
{
   dpSetup,
   dpFrame,
   dpMessage
} dpDataType;

typedef enum
{
   dpModelWarning,
   dpModelError,
   dpSimWarning,
   dpSimError,
   dpInitialized,
   dpFinished,
   dpOverwritable,
   dpNormal,
   dpDllLoaded
} dpMessageID;

typedef enum
{
   dpNoNormalization = 0,
   dpNormalizeToOne,
   dpNormalizeToOneHundred
} dpNormalizeOptions;

/* dpEventType, dpEventSegment, dpEventSource, and dpMotionEvent must
 * exactly match the corresponding definitions in sm.h.
 */
typedef enum {
   dpUnknownEventType = 0,
   dpHeelStrikeEvent,
   dpToeOffEvent
} dpEventType;

typedef enum {
   dpUnknownSegmentEvent = 0,
   dpRightFootEvent,
   dpLeftFootEvent,
   dpOtherSegmentEvent
} dpEventSegment;

typedef enum {
   dpUnknownEvent = 0,
   dpUserEvent,
   dpForceEvent,
   dpMarkerEvent
} dpEventSource;

typedef struct {
   char* name;                         /* event name */
   double xCoord;                      /* xCoord of event (percentage if available, else time?) */
   dpEventType type;                   /* type of event (heel or toe) */
   dpEventSegment segment;             /* segment of event */
   dpEventSource source;               /* source of event (marker or force) */
   dpBoolean displayFlag;              /* unused */
} dpMotionEvent;

typedef struct {
   char* name;                        /* name: rightFootPrint, leftFootPrint */
   double scale;                      /* scale factor */
   double translation[3];             /* translatiion from origin of world frame */
   double euler_rotation[3];          /* XYZ Euler rotation from world frame */
} dpFootPrint;                        /* foot print calculated in SIMM from events */

typedef struct {
   char* name;                        /* name: forceplate1, forceplate2, ... */
   double scale[3];                   /* scale factor */
   double translation[3];             /* translatiion from origin of world frame */
   double euler_rotation[3];          /* XYZ Euler rotation from world frame */
} dpForcePlate;                       /* force plate from motion trial */

typedef struct
{
   char* motionName;
   int numElements;
   char** elementNames;
   double startTime;
   double endTime;
   int numFrames;
   int numEvents;
   dpMotionEvent* event;
   float eventColor[3];
   dpBoolean enforce_constraints;      /* enforce constraints while prescribing motion? */
   dpBoolean enforce_loops;            /* not used in simulation, just copied to output motion */
   dpBoolean show_markers;             /* not used in simulation, just copied to output motion */
   dpBoolean show_marker_trails;       /* not used in simulation, just copied to output motion */
   dpBoolean show_forces;              /* not used in simulation, just copied to output motion */
   dpBoolean show_force_trails;        /* not used in simulation, just copied to output motion */
   dpBoolean show_foot_prints;         /* not used in simulation, just copied to output motion */
   dpBoolean show_force_plates;        /* not used in simulation, just copied to output motion */
   dpBoolean wrap;                     /* not used in simulation, just copied to output motion */
   int num_foot_prints;                /* number of foot prints */
   dpFootPrint** foot_print;           /* array of pointers to foot prints */
   int num_force_plates;               /* number of force plates */
   dpForcePlate** force_plate;         /* array of pointers to force plates */
   int x_column;                       /* x_column from input data */
   int time_column;                    /* time column index from input data */
   double xStart;                      /* not used in simulation, just copied to output motion */
   double xEnd;                        /* not used in simulation, just copied to output motion */
   dpBoolean show_gencoord_moments;    /* not used in simulation, just copied to output motion */
   dpBoolean show_other_objects;       /* not used in simulation, just copied to output motion */
} dpDataSetup;

typedef struct
{
   int numElements;
   double* elements;
} dpDataFrame;

typedef struct
{
   dpMessageID id;
   char text[500];
   void* data;
} dpMessageStruct;

typedef struct
{
   int state;
   int num_items;
   int list_size;
   const char** name;
   int* index;
} dpNameList;

typedef struct
{
   dpNameList muscleActivations;
   dpNameList muscleLengths;
   dpNameList muscleFiberLengths;
   dpNameList muscleFiberVelocities;
   dpNameList muscleForces;
   dpBoolean contactForces;
   dpNameList massCenterPositions;
   dpNameList massCenterVelocities;
   dpBoolean systemEnergy;
   dpNameList jointReactionForces;
   dpNameList jointReactionTorques;
   dpNameList gencoordMoments;
   dpBoolean springForces;
   int numSpringForces;
   dpNameList muscleMomentArms_Muscles;
   dpNameList muscleMomentArms_Gencoords;
   dpNameList muscleMoments_Muscles;
   dpNameList muscleMoments_Gencoords;
   dpNameList totalMuscleMoments;
   dpNameList optimizedActivations_Muscles;
   dpNameList optimizedActivations_Gencoords;
   dpNormalizeOptions normalization;
   dpNameList gencoordValues;
   dpBoolean traditionalInverse;
   dpNameList gencoordVelocities;
   dpNameList gencoordAccelerations;
   dpNameList gencoordPowers;
} dpOutputOptions;

typedef enum {
   dpNoMuscles,                  /* don't include any muscles */
   dpAllMuscles,                 /* include all muscles in the model */
   dpVisibleMuscles,             /* include only the muscles that are currently visible */
   dpMusclesWithEMG,             /* include only the muscles for which there is EMG data in the input motion */
   dpMusclesWithExcitation       /* include only muscles that have excitation patterns */
} dpIncludeMuscles;

/* Structure to mimic "smoothing" and "system parameters" panels in Dynamics Tool in SIMM */
typedef struct
{
   int kinematic_cutoff;                     /* value for smoothing of kinematic data */
   dpBoolean kinematic_cutoff_defined;       /* was the value defined in params.txt? */
   int force_cutoff;                         /* value for smoothing of force data */
   dpBoolean force_cutoff_defined;           /* was the value defined in params.txt? */
   int emg_cutoff;                           /* value for smoothing of emg data */
   dpBoolean emg_cutoff_defined;             /* was the value defined in params.txt? */
   dpBoolean include_kinematics;             /* include kinematics in simulation? - TODO: not yet implemented */
   dpBoolean include_forces;                 /* include forces in simulation? - TODO: not yet implemented */
   dpBoolean include_emg;                    /* include emg in simulation? - TODO: not yet implemented */
   dpIncludeMuscles muscles;                 /* set the subset of muscles to include in the simulation */
   int normalize;                            /* TODO not yet implemented, normalize 0, to 1, to 100 */
} dpInputOptions;

typedef struct
{
   double startTime;
   double endTime;
   double stepSize;
   char* workingDir;
   char* inputMotionFile;
   char* outputMotionFile;
   char* boneDir;
   dpOutputOptions outputOptions;
   dpSimmModelID model;
   dpInputOptions inputOptions;
   double baumgarte_stabilization;
} dpSimulationParameters;

/* dpStepFunction through dpLinearFunction must remain
 * 0 through 3 so that they match the enums in dpSplineType
 * from SIMM 4.2.
 */
typedef enum
{
   dpFunctionTypeUndefined = -1,
   dpStepFunction = 0,
   dpNaturalCubicSpline,
   dpGCVSpline,
   dpLinearFunction
} dpFunctionType;

typedef enum
{
   dpJointFile = 0,
   dpMuscleFile
} dpFunctionSource;

typedef enum
{
   dpFunctionUndefined = 0,
   dpFunctionUserDefined,
   dpFunctionSimmDefined
} dpFunctionStatus;

typedef struct
{
   dpFunctionType type;           /* dpStepFunction, dpLinearFunction, dpNaturalCubicSpline, dpGCVSpline */
   int numpoints;                 /* number of points in the function */
   int coefficient_array_size;    /* current size of x,y,b,c,d arrays */
   int usernum;                   /* user-supplied number for this function*/
   double* x;                     /* list of x coordinates */
   double* y;                     /* list of y coordinates */
   double* b;                     /* list of b coefficients for interpolation */
   double* c;                     /* list of c coefficients for interpolation */
   double* d;                     /* list of d coefficients for interpolation */
   dpFunctionStatus status;       /* undefined, defined in file, created in SIMM GUI */
   dpBoolean used;                /* is this structure currently being used? */
   double cutoff_frequency;       /* for smoothing of GCV splines */
   dpFunctionSource source;       /* is function from joint file or muscle file? */
} dpFunction;                     /* a function with XY control points */

typedef struct
{
   void* gencoord;                /* gencoord number */
   double start;                  /* starting value of range-of-motion */
   double end;                    /* ending value of range-of-motion */
} dpPointRange;                   /* muscle point range (wrapping pts) */

typedef struct
{
   int segment;                   /* body segment this point is fixed to */
   dpBoolean selected;            /* whether or not point has been selected */
   double point[3];               /* xyz coordinates of the point */
   double ground_pt[3];           /* coords of point in ground frame */
   dpFunction* function[3];       /* function for moving point (3 coordinates - 3 possible functions) */
   void* gencoord[3];             /* gencoord for moving point (3 coordinates - 3 possible gencoords) */
   dpBoolean isMovingPoint;       /* whether the point has non-constant coordinates */
   dpBoolean isVia;               /* is the point a via point */
   dpPointRange viaRange;         /* array of ranges */
   dpBoolean is_auto_wrap_point;  /* was this point calc-ed by auto wrapper? */
   double wrap_distance;          /* stores distance of wrap over object surface */
   double* wrap_pts;              /* points wrapping over surface */
   int num_wrap_pts;              /* number of points wrapping over surface */
   double undeformed_point[3];    /* 'point' prior to deformation */
   float normal[3];               /* point normal used for muscle surfaces */
} dpMusclePoint;                  /* properties of a muscle point */

typedef struct
{
   int num_orig_points;           /* number of user-defined muscle points */
   int num_points;                /* total number of muscle points */
   dpMusclePoint* mp_orig;        /* list of user-defined muscle points */
   dpMusclePoint** mp;            /* list of muscle points after auto wrapping */
   int mp_orig_array_size;        /* current size of mp_orig[] */
   int mp_array_size;             /* current size of mp[] */
} dpMusclePathStruct;

typedef enum
{
   dpConstraintSphere = 0,
   dpConstraintCylinder,
   dpConstraintEllipsoid,
   dpConstraintPlane,
   dpConstraintNone
} dpConstraintObjectType;

typedef struct
{
   char *name;                    /* name of the constraint point */
   int segment;                   /* segment the point is attached to */
   double offset[3];              /* location of the point from segment's origin */
   double weight;                 /* weight of point */
   int constraint_num;            /* index of the constraint */
} dpConstraintPoint;

typedef struct
{
   char *name;                             /* name of constraint object */
   dpConstraintObjectType constraint_type; /* type of constraint object */
   dpBoolean active;                       /* is this constraint active? */
   int segment;                            /* body segment object is fixed to */
   double co_radius[3];                    /* constraint object radius */
   double height;                          /* constraint object height (cylinder only) */
   int constraint_axis;                    /* which axis to constrain over */
   int constraint_sign;                    /* which side of constraint object to use */
   int numPoints;                          /* number of constraint points */
   int ptIndex;                            /* index of first pt linked to this constraint */
   double from_local_xform[4][4];          /* parent-to-wrapobj transform matrix */
   dpPlaneStruct plane;                    /* plane parameters (plane only) */
   double to_local_xform[4][4];            /* wrapobj-to-parent transform matrix */
   int cp_array_size;                      /* size of array of constraint points */
   dpConstraintPoint *points;              /* constraint points */
} dpConstraintObject;

typedef struct
{
   int start;                     /* input:  wrap starting muscle point index */
   int end;                       /* input:  wrap ending muscle point index */
   double* wrap_pts;              /* output: array of wrapping path points */
   int num_wrap_pts;              /* output: size of 'wrap_pts' array */
   double wrap_path_length;       /* output: distance along curved r1->r2 path */
   double r1[4];                  /* output: wrap tangent point nearest to p1 */
   double r2[4];                  /* output: wrap tangent point nearest to p2 */
} dpWrapParams;

typedef enum
{
   dpWrapSphere,
   dpWrapCylinder,
   dpWrapEllipsoid,
   dpWrapTorus,
   dpWrapNone
} dpWrapObjectType;

typedef struct {
   struct mss* muscle;
   int start_pt;
   int end_pt;
   int wrap_algorithm;
} dpMuscleWrapAssociation;

typedef struct wos
{
   char* name;                       /* name of wrap object */
   dpWrapObjectType wrap_type;       /* type of wrap object */
   dpBoolean active;                 /* is this wrap object active? */
   int wrap_algorithm;               /* default wrap algorithm for wrap object */
   int segment;                      /* body segment object is fixed to */
   double radius[3];                 /* wrap object radius */
   double height;                    /* wrap object height (cylinder only) */
   int wrap_axis;                    /* which axis to wrap over X=0, Y=1, Z=2 */
   int wrap_sign;                    /* which side of wrap axis to use (0 for both) */
   double from_local_xform[4][4];    /* parent-to-wrapobj transform matrix */
   double to_local_xform[4][4];      /* wrapobj-to-parent transform matrix */
   long display_list;                /* only used for cylinder */
   dpBoolean display_list_is_stale;  /* does display list need updating? */
   dpBoolean visible;                /* is the wrap object visible? */
   dpBoolean show_wrap_pts;          /* draw wrap point xyz coordinates? */
   dpCoord3D rotation_axis;          /* local-to-parent transform parameter */
   double rotation_angle;            /* local-to-parent transform parameter */
   dpCoord3D translation;            /* local-to-parent transform parameter */
   dpBoolean xforms_valid;           /* do xform matrices need recalculation? */
   dpCoord3D undeformed_translation; /* translation before applying deform objects */
   int numMuscles;                   /* number of muscles associated with this wrap object */
   dpMuscleWrapAssociation** muscle; /* arrays of associated muscles */
   struct wos* saved;                /* pointer to the saved version of the wrap object */
} dpWrapObject;

typedef struct
{
   dpWrapObject* wrap_object;        /* muscle is auto-wrapped over this object */
   int wrap_algorithm;               /* wrap algorithm for this muscle */
   int startPoint;                   /* wrap only those muscle segments between startPoint */
   int endPoint;                     /* and endPoint */
   double c[3];                      /* previous c point */
   double r1[3];                     /* previous r1 (tangent) point */
   double r2[3];                     /* previous r2 (tangent) point */
   dpMusclePoint mp_wrap[2];         /* the two muscle points created when the muscle wraps */
} dpMuscleWrapStruct;

typedef struct mss
{
   char* name;                          /* name of muscle */
   int index;                           /* index of this muscle in model's array of muscles */
   dpBoolean display;                   /* whether or not to display this muscle */
   dpBoolean selected;                  /* whether or not this muscle is selected */
   dpMusclePathStruct *path;            /* all the muscle path information */
   double* max_isometric_force;         /* maximum isometric force */
   double* pennation_angle;             /* pennation angle of muscle fibers */
   double* optimal_fiber_length;        /* muscle fiber length */
   double* resting_tendon_length;       /* resting length of tendon */
   double* max_contraction_vel;         /* maximum contraction velocity */
   double muscle_tendon_length;         /* musculotendon length */
   double fiber_length;                 /* muscle fiber length */
   double tendon_length;                /* tendon length */
   double muscle_tendon_vel;            /* musculotendon velocity */
   double fiber_velocity;               /* muscle velocity */
   double tendon_velocity;              /* tendon velocity */
   double force;                        /* force in musculotendon unit */
   double applied_power;                /* power this muscle applies to body segments */
   double heat_energy;                  /* heat energy generated by this muscle */
   double mechanical_energy;            /* energy this muscle gives to body segments */
   double energy;                       /* energy this muscle gives to body segments */
   int nummomentarms;                   /* number of moment arms (= # of gencoords) */
   double* momentarms;                  /* list of moment arm values */
   dpFunction** tendon_force_len_func;  /* tendon force-length function */
   dpFunction** active_force_len_func;  /* muscle active force-length function */
   dpFunction** passive_force_len_func; /* muscle passive force-length function */
   dpFunction** force_vel_func;         /* muscle force-velocity function */
   int num_dynamic_params;              /* size of dynamic_params array */
   char** dynamic_param_names;          /* list of dynamic parameter names */
   double** dynamic_params;             /* array of dynamic (muscle model) parameters */
   double dynamic_activation;           /* dynamic value of muscle activation */
   dpFunction** excitation_func;        /* excitation function */
   void** excitation_abscissa;          /* excitation is function of this gencoord (or 'time') */
   double excitation_level;             /* current level of excitation */
   int* muscle_model_index;             /* index for deriv, init, & assign func arrays */
   int* muscle_model_opt_index;         /* index for deriv, init, & assign func arrays (for static optimization) */
   dpBoolean wrap_calced;               /* has wrapping been calculated/updated? */
   int numWrapStructs;                  /* number of wrap objects used for this muscle */
   dpMuscleWrapStruct** wrapStruct;     /* holds information about the wrap objects */
   int numStateParams;                  /* number of initial state values stored in dllparams.txt */
   double* stateParams;                 /* array of initial state values in dllparams.txt */
   int numgroups;                       /* number of groups to which musc belongs*/
   int* group;                          /* list of muscle group numbers */
   double* min_thickness;               /* minimum thickness of muscle line */
   double* max_thickness;               /* maximum thickness of muscle line */
   int* min_material;                   /* material to draw with for activation = 0.0 */
   int* max_material;                   /* material to draw with for activation = 1.0 */
   struct mss* saved;                   /* pointer to the saved version of the muscle */
} dpMuscleStruct;                       /* properties of a musculotendon unit */

typedef struct
{
   int num_vertices;
   int *vertex_index;
   double normal[3];
   double d;
} dpPolygonStruct;

typedef struct
{
   double coord[3];
   double normal[3];
   int polygon_count;
   int* polygons;
} dpVertexStruct;

typedef struct
{
   char* name;
   int body_segment;
   dpBoundingCube bc;
   dpVertexStruct* vertex;
   int num_vertices;
   dpPolygonStruct* polygon;
   int num_polygons;
} dpPolyhedronStruct;

typedef enum
{
   dpUnconstrainedQ,  /* free gencoord */
   dpConstrainedQ,    /* constrained as a function of another gencoord */
   dpPrescribedQ,     /* gencoord with prescribed pos, vel, and acc */
   dpFixedQ           /* fixed gencoord, vel and acc = 0.0 */
} dpQType;

typedef struct
{
   char* name;                               /* gencoord name */
   dpQType type;                             /* gencoord type - free, fixed, prescribed, constrained */
   int joint;                                /* joint that this gencoord belongs to */
   int axis;                                 /* joint axis that this gencoord controls */
   int q_ind;                                /* independent q a constrained/prescribed q is dependent upon */
   int constraint_num;                       /* index of the constraint used to constrain gencoord */
   double initial_value;                     /* initial gencoord value */
   double initial_velocity;                  /* initial gencoord velocity */
   double conversion;                        /* unit conversion (generally 1.0 for translations and 57.29 (deg/rad) for rotations */
   double range_start;                       /* minimum value of the gencoord */
   double range_end;                         /* maximum value of the gencoord */
   dpFunction* restraint_function;           /* restraint function used to calculate moment necessary to keep gencoord within its range */
   dpFunction* min_restraint_function;       /* support for old restraint functions */
   dpFunction* max_restraint_function;       /* support for old restraint functions */
   dpFunction* constraint_function;          /* function used to constrain gencoord motion */
   dpBoolean function_active;                /* is the restraint function active */
   double moment;                            /* calculated gencoord moment */
   double pd_stiffness;                      /* stiffnes for PD controller */
   int state_index;                          /* index into the state vector (DP) for this gencoord */
   dpBoolean is_part_of_quaternion;          /* is the gencoord part of a joint represented by a quaternion */
   int output_index;                         /* not used */
} dpQStruct;

typedef struct
{
   int sd_segment;               /* SD/FAST segment the floor is associated with */
   char* name;                   /* name of spring_floor */
   dpPlaneStruct plane;          /* parameters of infinite floor plane */
   dpPolyhedronStruct* ph;       /* list of polyhedra from bone file (first one is used to define the infinte plane used in contact detection */
   dpBoolean active;             /* is the floor active? */
   dpBoolean visible;            /* not used in Pipeline */
} dpSpringFloor;                 /* Infinite planar surface used in spring-based contact detection to prevent penetration. */

typedef struct
{
   int sd_segment;               /* SD/FAST segment the surface is associated with */
   char* name;                   /* name of spring_surface */
   dpPolyhedronStruct* ph;       /* polyhedron from bone file */
   dpBoolean active;             /* is floor active? */
   dpBoolean visible;            /* not used in Pipeline */
} dpSpringSurface;               /* Surface used in spring-based contact detection to prevent penetration. */

typedef struct
{
   int segment;                 /* SD/FAST segment force matte associated with */
   char* name;                  /* name of force_matte */
   dpPolyhedronStruct* ph;      /* polyhedron from bone file */
} dpForceMatte;                 /* used for applying ground-based forces to other segments */

typedef struct
{
   int sd_segment;              /* SD/FAST segment the spring is attached to */
   dpBoolean is_floor_contact;  /* does the point contact a floor or surface */
   int contact_index;           /* index of corresponding floor or surface object */
   char* contact_name;          /* name of corresponding contact object */
   double point[3];             /* xyz coordinates of the spring point */
   double force[3];             /* spring force vector */
   double friction;             /* coefficient of friction with floor */
   double param_a;              /* spring parameters */
   double param_b;
   double param_c;
   double param_d;
   double param_e;
   double param_f;
} dpSpringStruct;

typedef struct
{
   double point[3];
   double vector[3];
} dpForceStruct;

typedef struct
{
   char* name;                      /* name of the body (segment) */
   double mass;                     /* mass of the body */
   double inertia[3][3];            /* moment of inertia of the body */
   double mass_center[3];           /* mass center of the body wrt SIMM bone file origin */
   double body_to_joint[3];         /* vector from body mass center to joint origin */
   double inboard_to_joint[3];      /* vector from inboard body (parent) mass center to joint origin */
   dpBoolean contactable;           /* is body a contact object (for contact/collision detection) */
   int num_objects;                 /* how many objects are in the geometrical representation of contact object */
   dpPolyhedronStruct** object;     /* the contact objects representing the body */
   double contact_force[3];         /* the contact force on the body */
   double impact_force[3];          /* impact force on the body */
   double impact_point[3];          /* point of impact on the body */
   dpBoolean* contact_joints;       /* list of joints that are involved in contact for the body */
} dpBodyStruct;

typedef struct
{
   char* name;                    /* name of the joint */
   int inboard_body;              /* index of first (inboard) body of the joint */
   int outboard_body;             /* index of the second (outboard) body of the joint */
   int num_dof;                   /* number of DOFs that the joint allows */
   int first_q;                   /* coordinate index of the first axis (into the state array) */
   int quat;                      /* index of the 4th Euler parameter (quaternion) in the q array */
   double axes[6][3];             /* all the joint axes, rotational and translational */
   double force[3];               /* joint reaction force */
   double torque[3];              /* joint reaction torque */
   dpDOFType dof_type[6];         /* the type (translation or rotation) of each DOF */
   dpJointType jnt_type;          /* the type of joint */
   dpBoolean loop_joint;          /* does the joint close a loop? */
   dpBoolean is_quaternion;       /* is a quaternion used to represent the rotations? */
   dpEulerAngleOrder euler_rots;  /* Euler-angle rotation order */
} dpJointStruct;

typedef struct
{
   char* name;                /* name of the model */
   int nq;                    /* number of sdfast qs (includes 4th quaternions */
   int nu;                    /* number of sdfast q derivatives */
   int neq;                   /* total number of states (q + u + muscle states) */
   int num_muscles;           /* number of muscles in the model */
   int num_body_segments;     /* number of body segments in the model (including ground) */
   int num_joints;            /* number of joints in the model */
   int num_closed_loops;      /* number of closed loops in the model */
   int num_constraints;       /* number of constraints (constrained Qs + user constraints in the model) */
   int num_user_constraints;  /* number of constraints defined by the user */
   int num_contacts;          /* number of contacts for contact detection */
   int num_bilat_contacts;    /* number of bilateral contacts for contact detection */
   int contacts_size;         /* size of array for contacts list */
   int bilat_contacts_size;   /* size of array for bilateral contacts list */
   dpContactInfo* contacts;   /* list of contact points */
   dpContactInfo* bilat_contacts;/* list of bilateral contact points */
   int num_springs;              /* number of spring elements for spring-based contact */
   int num_spring_floors;        /* number of floor elements for spring-based contact */
   int num_spring_surfaces;      /* number of surface elements for spring-based contact */
   int spring_array_size;        /* size of array for spring elements */
   int spring_floor_array_size;  /* size of array for spring floors */
   int spring_surface_array_size;/* size of array for spring surfaces */
   dpSpringStruct* spring;       /* spring points for use with spring-based contact */
   dpSpringFloor* spring_floor;  /* Single polygon plane surface (floor) for use with spring-based contact */
   dpSpringSurface* spring_surface; /* Multi-polygon surface for use with spring-based contact */
   int num_force_mattes;         /* number of force mattes for applying ground-based forces */
   int force_matte_array_size;   /* size of array for force mattes */
   dpForceMatte* force_matte;    /* force mattes used for applying ground-based forces */
   dpQStruct* q;                 /* list of Qs in the model */
   dpBodyStruct* body_segment;   /* list of body segments (including ground) in the model */
   dpMuscleStruct* muscles;      /* muscles in the model */
   dpMuscleStruct default_muscle;   /* default muscle definition */
   dpJointStruct* joint;         /* list of joints in the model */
   int num_wrap_objects;         /* number of wrap objects in the model */
   dpWrapObject* wrap_object;    /* list of wrap objects */
   int num_constraint_objects;   /* number of constraint objects in the model */
   dpConstraintObject* constraint_object; /* list of constraint objects */
   int function_array_size;      /* size of array to store functions */
   int num_functions;            /* number of functions (constraint and moving muscle point) in the model */
   dpFunction** function;        /* array of functions used by the model (for constraints and moving muscle points) */
   double gravity[3];            /* gravity vector */
   dpSimmModelID simmModel;      /* ID of the SIMM model that created the pipeline model */
   int enforce_constraints;      /* enforce constraint objects while integrating? */
   int newInverseSimulation;     /* new-style inverse with corrective moments? */
} dpModelStruct;

typedef struct {
   int num_joints;               /* total number of joints (tree and loop) in the DLL */
   dpJointType* joint_type;      /* joint types of all the DLL joints */
   int num_segments;             /* number of segments in the DLL model */
   int num_tree_joints;          /* number of non-loop joints in the DLL model */
   int num_closed_loops;         /* number of closed loops in the DLL model */
   dpBoolean contactEnabled;     /* was CONTACT_DETECTION parameter enabled when compiling DLL? */
} dpModelInfo;

DLL int dpInitSimulation(void* params);
DLL int dpSetModel(void* model);
DLL int dpRunSimulation(int);
DLL int dpPauseSimulation(int);
DLL int dpResetSimulation(int);
DLL int dpRegisterCallback(int (*dataCallback)(dpSimmModelID, int, void*)); // middle argument is dpDataType
DLL int dpGetVersion(void);
DLL void* dpGetModelInfo(void);
DLL int dpRunBackgroundSimulation(void* params, dpModelStruct* model);

typedef int (*dpInitSimulationFunc)(void* params); // void* is cast to dpSimulationParameters*
typedef int (*dpSetModelFunc)(void* model);        // void* is cast to dpModelStruct*
typedef int (*dpRunSimulationFunc)(int);
typedef int (*dpPauseSimulationFunc)(int);
typedef int (*dpResetSimulationFunc)(int);
typedef int (*dpRegisterCallbackFunc)(int (*dataCallback)(dpSimmModelID, int, void*)); // middle argument is dpDataType
typedef int (*dpGetVersionFunc)(void);
typedef void* (*dpGetModelInfoFunc)(void);  // void* is cast to dpModelInfo*
typedef int (*dpRunBackgroundSimulationFunc)(void* params, void* model); // params is cast to dpSimulationParameters*, model to dpModelStruct*

typedef struct
{
   dpInitSimulationFunc dpInitSimulation; 
   dpSetModelFunc dpSetModel;
   dpRunSimulationFunc dpRunSimulation;
   dpPauseSimulationFunc dpPauseSimulation;
   dpResetSimulationFunc dpResetSimulation;
   dpRegisterCallbackFunc dpRegisterCallback;
   dpGetVersionFunc dpGetVersion;
   dpGetModelInfoFunc dpGetModelInfo;
   dpRunBackgroundSimulationFunc dpRunBackgroundSimulation; 
} dpSimFuncs;

#endif /* DP_H */
