/*******************************************************************************

   INITS.C

   Authors: Peter Loan
            Krystyne Blaikie

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains 10 routines to initialize the states
      for dimensionless muscle-tendon models. The routines are named
      muscle_init_func1() through muscle_init_func10(), and each one
      initializes all of a muscle's states for a particular muscle-tendon
      model. The routine should be called at the beginning of your
      program, before entering the integration loop.

   Routines:
      muscle_init_func1  : 3 states (act, fiber_len, fiber_vel), no pennation
      muscle_init_func2  : 3-state model, pennation included
      muscle_init_func3  : model 2 plus 4th state for muscle power
      muscle_init_func4  : 2 states (act, fiber_len), non-zero passive force
      muscle_init_func5  : 1 state (act), infinitely stiff tendon
      muscle_init_func6  : 4 states
      muscle_init_func7  : 0 state, ligament model
      muscle_init_func8  : model 4 plus two states for fatigue
      muscle_init_func9  : 0 state, infinitely-stiff tendon (for static optimization)
      muscle_init_func10 : 1 state (fiber_len) (for static optimization)
      muscle_init_func11 : unused
      muscle_init_func12 : unused
      muscle_init_func13 : unused
      muscle_init_func14 : unused
      muscle_init_func15 : unused
      muscle_init_func16 : unused
      muscle_init_func17 : unused
      muscle_init_func18 : unused
      muscle_init_func19 : unused
      muscle_init_func20 : unused

*******************************************************************************/

#include "universal.h"

extern char buffer[];

int muscle_init_func1(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 3;
   int activation = 0, fiber_length = 1, fiber_velocity = 2;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 1);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 1);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 1);
      return -1;
   }

   if (muscle->force_vel_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): force_velocity_curve is undefined.", muscle->name, 1);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 1);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 1);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 1);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 1);
      return -1;
   }

   // Initialize the muscle states to default values.
   state_euler[muscle_state_start_index + activation] = 0.0;
   state_euler[muscle_state_start_index + fiber_length] = *(muscle->optimal_fiber_length);
   state_euler[muscle_state_start_index + fiber_velocity] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index +i] = muscle->stateParams[i];

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];
   muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   muscle->fiber_velocity = state_euler[muscle_state_start_index + fiber_velocity];

   return numStates;
}


int muscle_init_func2(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 3;
   int activation = 0, fiber_length = 1, fiber_velocity = 2;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->force_vel_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): force_velocity_curve is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 2);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle undefined", muscle->name, 2);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 2);
      return -1;
   }

   // Initialize the muscle states to default values. Fiber_length is calculated later.
   state_euler[muscle_state_start_index +activation] = 0.0;
   state_euler[muscle_state_start_index +fiber_velocity] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index +i] = muscle->stateParams[i];

   // If the initial value of fiber_length was not specified in the parameters
   // file, calculate it assuming static equilibrium. The function
   // calc_muscle_tendon_force() includes pennation in the muscle model.
   if (muscle->numStateParams < 2)
   {
      calc_muscle_tendon_force(muscle, state_euler[muscle_state_start_index + activation], state_euler);
      state_euler[muscle_state_start_index +fiber_length] = muscle->fiber_length;
   }
   else
   {
      muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   }

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];
   muscle->fiber_velocity = state_euler[muscle_state_start_index + fiber_velocity];

   return numStates;
}


int muscle_init_func3(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 4;
   int activation = 0, fiber_length = 1, fiber_velocity = 2, energy = 3;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->force_vel_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): force_velocity_curve is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 3);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 3);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 3);
      return -1;
   }

   // Initialize the muscle states to default values. Fiber_length is calculated later.
   state_euler[muscle_state_start_index +activation] = 0.0;
   state_euler[muscle_state_start_index +fiber_velocity] = 0.0;
   state_euler[muscle_state_start_index +energy] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index +i] = muscle->stateParams[i];

   // If the initial value of fiber_length was not specified in the parameters
   // file, calculate it assuming static equilibrium. The function
   // calc_muscle_tendon_force() includes pennation in the muscle model.
   if (muscle->numStateParams < 2)
   {
      calc_muscle_tendon_force(muscle, state_euler[muscle_state_start_index +activation], state_euler);
      state_euler[muscle_state_start_index + fiber_length] = muscle->fiber_length;
   }
   else
   {
      muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   }

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];
   muscle->fiber_velocity = state_euler[muscle_state_start_index + fiber_velocity];
   muscle->mechanical_energy = state_euler[muscle_state_start_index + energy];

   return numStates;
}


int muscle_init_func4(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 2;
   int activation = 0, fiber_length = 1;
   double muscle_damping;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 4);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 4);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 4);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 4);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 4);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 4);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 4);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 4);
      return -1;
   }

   muscle_damping = get_muscle_param_value(muscle, "damping");
   if (EQUAL_WITHIN_ERROR(muscle_damping, ERROR_DOUBLE))
   {
      sim_message(none, "Muscle %s (model %d): damping is undefined.", muscle->name, 4);
      return -1;
   }
   else if (muscle_damping < ROUNDOFF_ERROR)
   {
      sim_message(none, "Muscle %s (model %d): damping must be greater than 0.0.", muscle->name, 4);
      return -1;
   }

   // Initialize the muscle states to default values. Fiber_length is calculated later.
   state_euler[muscle_state_start_index + activation] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index + i] = muscle->stateParams[i];

   // If the initial value of fiber_length was not specified in the parameters
   // file, calculate it assuming static equilibrium. The function
   // calc_muscle_tendon_force() includes pennation in the muscle model.
   if (muscle->numStateParams < 2)
   {
      calc_muscle_tendon_force(muscle, state_euler[muscle_state_start_index + activation], state_euler);
      state_euler[muscle_state_start_index + fiber_length] = muscle->fiber_length;
   }
   else
   {
      muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   }

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];

   return numStates;
}


int muscle_init_func5(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 1;
   int activation = 0;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->force_vel_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): force_velocity_curve is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 5);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 5);
      return -1;
   }

   // Initialize the muscle states to default values.
   state_euler[muscle_state_start_index + activation] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index + i] = muscle->stateParams[i];

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];

   return numStates;
}


int muscle_init_func6(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 4;
   int activation = 0, fiber_length = 1, heat_energy = 2, mechanical_energy = 3;
   double muscle_damping;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 6);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 6);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 6);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 6);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 6);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 6);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 6);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 6);
      return -1;
   }

   muscle_damping = get_muscle_param_value(muscle, "damping");
   if (EQUAL_WITHIN_ERROR(muscle_damping, ERROR_DOUBLE))
   {
      sim_message(none, "Muscle %s (model %d): damping is undefined.", muscle->name, 6);
      return -1;
   }
   else if (muscle_damping < ROUNDOFF_ERROR)
   {
      sim_message(none, "Muscle %s (model %d): damping must be greater than 0.0.", muscle->name, 6);
      return -1;
   }

   // Initialize the muscle states to default values. Fiber_length is calculated later.
   state_euler[muscle_state_start_index + activation] = 0.0;
   state_euler[muscle_state_start_index + heat_energy] = 0.0;
   state_euler[muscle_state_start_index + mechanical_energy] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index + i] = muscle->stateParams[i];

   // If the initial value of fiber_length was not specified in the parameters
   // file, calculate it assuming static equilibrium. The function
   // calc_muscle_tendon_force() includes pennation in the muscle model.
   if (muscle->numStateParams < 2)
   {
      calc_muscle_tendon_force(muscle, state_euler[muscle_state_start_index + activation], state_euler);
      state_euler[muscle_state_start_index + fiber_length] = muscle->fiber_length;
   }
   else
   {
      muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   }

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];
   muscle->heat_energy = state_euler[muscle_state_start_index + heat_energy];
   muscle->mechanical_energy = state_euler[muscle_state_start_index + mechanical_energy];

   return numStates;
}


int muscle_init_func7(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int numStates = 0;

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 7);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 7);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 7);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->resting_tendon_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): resting_tendon_length = 0.0.", muscle->name, 7);
      return -1;
   }

   return numStates;
}


int muscle_init_func8(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 4;
   int activation = 0, fiber_length = 1, active_motor_units = 2, fatigued_motor_units = 3;
   double muscle_damping, fatigue_factor, recovery_factor;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 8);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 8);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 8);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 8);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 8);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 8);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 8);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 8);
      return -1;
   }

   muscle_damping = get_muscle_param_value(muscle, "damping");
   if (EQUAL_WITHIN_ERROR(muscle_damping, ERROR_DOUBLE))
   {
      sim_message(none, "Muscle %s (model %d): damping is undefined.", muscle->name, 8);
      return -1;
   }
   else if (muscle_damping < ROUNDOFF_ERROR)
   {
      sim_message(none, "Muscle %s (model %d): damping must be greater than 0.0.", muscle->name, 8);
      return -1;
   }

   fatigue_factor = get_muscle_param_value(muscle, "fatigue_factor");
   if (EQUAL_WITHIN_ERROR(fatigue_factor, ERROR_DOUBLE))
   {
      sim_message(none, "Muscle %s (model %d): fatigue_factor is undefined.", muscle->name, 8);
      return -1;
   }
   else if (fatigue_factor < 0.0 || fatigue_factor > 1.0)
   {
      sim_message(none, "Muscle %s (model %d): fatigue_factor must be between 0.0 and 1.0.", muscle->name, 8);
      return -1;
   }

   recovery_factor = get_muscle_param_value(muscle, "recovery_factor");
   if (EQUAL_WITHIN_ERROR(recovery_factor, ERROR_DOUBLE))
   {
      sim_message(none, "Muscle %s (model %d): recovery_factor is undefined.", muscle->name, 8);
      return -1;
   }
   else if (recovery_factor < 0.0 || recovery_factor > 1.0)
   {
      sim_message(none, "Muscle %s (model %d): recovery_factor must be between 0.0 and 1.0.", muscle->name, 8);
      return -1;
   }

   // Initialize the muscle states to default values. Fiber_length is calculated later.
   state_euler[muscle_state_start_index + activation] = 0.0;
   state_euler[muscle_state_start_index + active_motor_units] = 0.0;
   state_euler[muscle_state_start_index + fatigued_motor_units] = 0.0;

   // Now override these defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates).
   for (i = 0; i < MIN(muscle->numStateParams, numStates); i++)
      state_euler[muscle_state_start_index + i] = muscle->stateParams[i];

   // If the initial value of fiber_length was not specified in the parameters
   // file, calculate it assuming static equilibrium. The function
   // calc_muscle_tendon_force() includes pennation in the muscle model.
   if (muscle->numStateParams < 2)
   {
      calc_muscle_tendon_force(muscle, state_euler[muscle_state_start_index + activation], state_euler);
      state_euler[muscle_state_start_index + fiber_length] = muscle->fiber_length;
   }
   else
   {
      muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   }

   // Now store some of the state values in the muscle structure.
   muscle->dynamic_activation = state_euler[muscle_state_start_index + activation];

   return numStates;
}


int muscle_init_func9(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 0;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->force_vel_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): force_velocity_curve is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 5);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 5);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 5);
      return -1;
   }

   return numStates;
}


int muscle_init_func10(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int i, numStates = 1;
   int fiber_length = 0;
   double muscle_damping;

   if (muscle->active_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): active_force_length_curve is undefined.", muscle->name, 10);
      return -1;
   }

   if (muscle->passive_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): passive_force_length_curve is undefined.", muscle->name, 10);
      return -1;
   }

   if (muscle->tendon_force_len_func == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_force_length_curve is undefined.", muscle->name, 10);
      return -1;
   }

   if (muscle->optimal_fiber_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length is undefined.", muscle->name, 10);
      return -1;
   }

   if (muscle->resting_tendon_length == NULL)
   {
      sim_message(none, "Muscle %s (model %d): tendon_slack_length is undefined.", muscle->name, 10);
      return -1;
   }

   if (muscle->max_isometric_force == NULL)
   {
      sim_message(none, "Muscle %s (model %d): max_force is undefined.", muscle->name, 10);
      return -1;
   }

   if (muscle->pennation_angle == NULL)
   {
      sim_message(none, "Muscle %s (model %d): pennation_angle is undefined.", muscle->name, 10);
      return -1;
   }

   if (EQUAL_WITHIN_ERROR(*(muscle->optimal_fiber_length),0.0))
   {
      sim_message(none, "Muscle %s (model %d): optimal_fiber_length = 0.0.", muscle->name, 10);
      return -1;
   }

   muscle_damping = get_muscle_param_value(muscle, "damping");
   if (EQUAL_WITHIN_ERROR(muscle_damping, ERROR_DOUBLE))
   {
      sim_message(none, "Muscle %s (model %d): damping is undefined.", muscle->name, 10);
      return -1;
   }
   else if (muscle_damping < ROUNDOFF_ERROR)
   {
      sim_message(none, "Muscle %s (model %d): damping must be greater than 0.0.", muscle->name, 10);
      return -1;
   }

   // Activation is not a state in this model, but it is needed to calculate fiber length
   // and fiber velocity, so its initial value can be specified in the parameters file.
   // Fiber_length is calculated later.
   muscle->dynamic_activation = 0.0;
   muscle->fiber_velocity = 0.0;

   // Now override the defaults with as many initial values as were specified in the
   // parameters file (as long as it's not more than numStates + 1). If activation is
   // specified, it comes *before* the states (to make the format consistent with the
   // other muscle models.
   if (muscle->numStateParams >= 1)
   {
      muscle->dynamic_activation = muscle->stateParams[0];
   }

   if (muscle->numStateParams > 1)
   {
      for (i = 0; i < MIN(muscle->numStateParams - 1, numStates); i++)
         state_euler[muscle_state_start_index + i] = muscle->stateParams[i + 1];

      // Now store fiber length in the muscle structure.
      muscle->fiber_length = state_euler[muscle_state_start_index + fiber_length];
   }
   else
   {
      // If the initial value of fiber_length was not specified in the parameters
      // file, calculate it assuming static equilibrium. The function
      // calc_muscle_tendon_force() includes pennation in the muscle model.
      calc_muscle_tendon_force(muscle, muscle->dynamic_activation, state_euler);  //full state vector passed in 
      state_euler[muscle_state_start_index + fiber_length] = muscle->fiber_length;
   }

   return numStates;
}


int muscle_init_func11(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   int numStates = 0;

   return numStates;
}


int muscle_init_func12(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 12 is undefined.");
   return -1;
}


int muscle_init_func13(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 13 is undefined.");
   return -1;
}


int muscle_init_func14(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 14 is undefined.");
   return -1;
}


int muscle_init_func15(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 15 is undefined.");
   return -1;
}


int muscle_init_func16(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 16 is undefined.");
   return -1;
}


int muscle_init_func17(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 17 is undefined.");
   return -1;
}


int muscle_init_func18(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 18 is undefined.");
   return -1;
}


int muscle_init_func19(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 19 is undefined.");
   return -1;
}


int muscle_init_func20(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[])
{
   sim_message(abort_action, "Muscle model 20 is undefined.");
   return -1;
}
