/*******************************************************************************

   LBFGSB_OPTIMIZE.CPP

   Author: Peter Loan
           Krystyne Blaikie

   Copyright (c) 2009-2010 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains wrapper functions so the LBFGSB routines can 
      be called from the Pipeline C code.

   ****************************************************************************/

/*******************************************************
Input parameters:
    N       -   problem dimension. N>0
    M       -   number of  corrections  in  the  BFGS  scheme  of  Hessian
                approximation  update.  Recommended value:  3<=M<=7.   The
                smaller value causes worse convergence,  the  bigger  will
                not  cause  a  considerably  better  convergence, but will
                cause a fall in the performance. M<=N.
    X       -   initial solution approximation.
                Array whose index ranges from 1 to N.
    EpsG    -   positive number which defines a precision of  search.  The
                subroutine finishes its work if the condition ||G|| < EpsG
                is satisfied, where ||.|| means Euclidian norm, G - gradient
                projection onto a feasible set, X - current approximation.
    EpsF    -   positive number which defines a precision of  search.  The
                subroutine  finishes  its  work if on iteration number k+1
                the condition |F(k+1)-F(k)| <= EpsF*max{|F(k)|, |F(k+1)|, 1}
                is satisfied.
    EpsX    -   positive number which defines a precision of  search.  The
                subroutine  finishes  its  work if on iteration number k+1
                the condition |X(k+1)-X(k)| <= EpsX is satisfied.
    MaxIts  -   maximum number of iterations.
                If MaxIts=0, the number of iterations is unlimited.
    NBD     -   constraint type. If NBD(i) is equal to:
                * 0, X(i) has no constraints,
                * 1, X(i) has only lower boundary,
                * 2, X(i) has both lower and upper boundaries,
                * 3, X(i) has only upper boundary,
                Array whose index ranges from 1 to N.
    L       -   lower boundaries of X(i) variables.
                Array whose index ranges from 1 to N.
    U       -   upper boundaries of X(i) variables.
                Array whose index ranges from 1 to N.

Output parameters:
    X       -   solution approximation.
Array whose index ranges from 1 to N.
    Info    -   a return code:
                    * -2 unknown internal error,
                    * -1 wrong parameters were specified,
                    * 0 interrupted by user,
                    * 1 relative function decreasing is less or equal to EpsF,
                    * 2 step is less or equal to EpsX,
                    * 4 gradient norm is less or equal to EpsG,
                    * 5 number of iterations exceeds MaxIts.

FuncGrad routine description. User-defined.
Input parameters:
    X   -   array whose index ranges from 1 to N.
Output parameters:
    F   -   function value at X.
    G   -   function gradient.
            Array whose index ranges from 1 to N.
The memory for array G has already been allocated in the calling subroutine,
and it isn't necessary to allocate it in the FuncGrad subroutine.

    NEOS, November 1994. (Latest revision June 1996.)
    Optimization Technology Center.
    Argonne National Laboratory and Northwestern University.

    Written by Ciyou Zhu in collaboration with
    R.H. Byrd, P. Lu-Chen and J. Nocedal.
void lbfgsbminimize(const int& n,
     const int& m,
     ap::real_1d_array& x,
     const double& epsg,
     const double& epsf,
     const double& epsx,
     const int& maxits,
     const ap::integer_1d_array& nbd,
     const ap::real_1d_array& l,
     const ap::real_1d_array& u,
     int& info);
*************************************************************************/
#include "lbfgsb.h"
#include "lbfgsb_optimize.h"

static int num_controls = 0;
static void (*calc_gradient_callback)(double [], double* , double [], void*) = NULL;
static double* control = NULL;
static double* gradient = NULL;
static void* gradient_callback_data = NULL;

void funcgrad(const ap::real_1d_array& x, double& f, ap::real_1d_array& g);

/* lbfgsb_optimize: function to optimize by calling LBFGSB optimizer */
void lbfgsb_optimize(int dp_n, int dp_m, double dp_x[], double dp_epsg, double dp_epsf, double dp_epsx,
                     int dp_maxits, int dp_nbd[], double dp_l[], double dp_u[], int *dp_info,
                     void (*fcn)(double [], double* , double [], void*), void* fcn_data)
{
   int i, n;
   ap::integer_1d_array nbd;
   ap::real_1d_array x, l, u;

   num_controls = dp_n;
   calc_gradient_callback = fcn;
   gradient_callback_data = fcn_data;
   control = (double*)malloc(num_controls * sizeof(double));
   gradient = (double*)malloc(num_controls * sizeof(double));

   n = dp_n;

   x.setbounds(1, n);
   nbd.setbounds(1, n);
   l.setbounds(1, n);
   u.setbounds(1, n);

   // Copy the input variables from 0-based double arrays used by the Dynamics
   // Pipeline to 1-based _1d_array types used by LBFGSB.
   for (i = 0; i < n; i++)
   {
      x(i+1) = dp_x[i];
      nbd(i+1) = dp_nbd[i];
      l(i+1) = dp_l[i];
      u(i+1) = dp_u[i];
   }

   lbfgsbminimize(n, dp_m, x, dp_epsg, dp_epsf, dp_epsx, dp_maxits, nbd, l, u, *dp_info);

   // After the function has been minimized, copy the output array back into a
   // 0-based double array that can be used by the Dynamics Pipeline.
   for (i = 0; i < n; i++)
      dp_x[i] = x(i+1);

   free(control);
   free(gradient);
}

/* Funcgrad calculates the value of the function F and gradient G in point X.
 * Used by the lbfgsb optimizer. 
*/
void funcgrad(const ap::real_1d_array& x, double& f, ap::real_1d_array& g)
{
   if (calc_gradient_callback)
   {
      // Copy the 1-based read_1d_array x into a zero-based double array
      // so it can be passed to the callback function defined in the Dynamics Pipeline.
      for (int i=0; i<num_controls; i++)
         control[i] = x(i+1);

      // Call the gradient function.
      calc_gradient_callback(control, &f, gradient, gradient_callback_data);

      // Copy the zero-based double array of gradients into the 1-based real_1d_array so
      // lbfgsb can use them.
      for (int i=0; i<num_controls; i++)
         g(i+1) = gradient[i];
   }
}
