/*******************************************************************************

   MAIN.C

   Authors: Peter Loan
            Krystyne Blaikie

   Version 6.0, June 8, 2011

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains the main routine for a dynamics
      simulation, and several other utility functions that are
      independent of the musculoskeletal model in the simulation.
      This file is generated by SIMM when you choose the
      "project files" option in the "Save Dynamics" dialog box.
      You will probably want to generate it once (not every time you
      change your model) and then customize it to your specific simulation.

*******************************************************************************/
#include "universal.h"

/*************** DEFINES (for this file only) *********************************/

// Tolerances and Integration parameters.
#define BAUMGARTE_STAB 20.0

/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/
SystemInfo si;

/**************** GLOBAL VARIABLES (used in only a few files) *****************/
dpModelStruct* sdm;
MotionData* input_data = NULL;
dpSimulationParameters params;
char current_dir[CHARBUFFER];
FILE* kin_out;
dpBoolean isStandAlone = dpYes;
char sd_error_msg[CHARBUFFER];

/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern char buffer[];
extern dpBoolean verbose;

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static void init_model(void);
static void setup_muscle_output_parameters(void);
static void handle_musclesWithEMG_option(double y[], double dy[], double work[]);

main(void)
{
   int i, j, err, which, status, num_muscle_states;
   double t, step, step_size, param[1] = {-1.0};
   double *y, *dy, *work, next_t;
   double start_time, end_time;
   double ke = 0, pe = 0;
   char *muscle_file = NULL;
   char *input_data_file = NULL, *output_motion_file = NULL, *output_kinetics_file = NULL;
   char *parameters_file_name = "params.txt";
   char e_msg[5000] = "";
   FILE* out;
#if CONTACT_DETECTION
   int num_impacts = 0;
#endif

   sim_message(none, "\nDynamics Pipeline (6.0) Simulation.\n");

#if PIPELINE_TRIAL
   check_password();
#endif

   // Fill in sdm struct with model parameters.
   init_model();

   getcwd(current_dir, CHARBUFFER);
   append_if_necessary(current_dir, DIR_SEP_CHAR);

   // Read in the parameters file.
   // Since muscle file has not yet been read in, muscle output options cannot be fully set yet.
   if (read_parameters_file(parameters_file_name, sdm, &muscle_file, &input_data_file, &output_motion_file, &output_kinetics_file) == code_bad)
      sim_message(recover, "Unable to load input parameters: %s%s not found.\n", current_dir, parameters_file_name);

   // If a muscle file was provided, read in the muscle information.
   if (muscle_file)
   {
      if (read_muscle_file(sdm, muscle_file, params.inputOptions.muscles) == code_bad)
      {
         sim_message(recover, "Error reading muscle file %s.", muscle_file);
         sdm->num_muscles = 0;
         sim_message(none, "Running simulation without a muscle file.");
      }
   }
   else
   {
      sdm->num_muscles = 0;
      sim_message(none, "Running simulation without a muscle file.");
   }

   // Initialize the joint functions and the gencoord restraint functions.
   init_joint_functions();
   init_q_restraint_functions();

#if CONTACT_DETECTION
   // allocate space for contact records in sdm
   init_contact_info();
#endif

   // Count the number of muscle states in the model.
   num_muscle_states = count_muscle_states(sdm);
   sdm->neq = sdm->nu + sdm->nq + num_muscle_states;

   y = (double*)simm_malloc(sdm->neq * sizeof(double));
   dy = (double*)simm_malloc(sdm->neq * sizeof(double));
   work = (double*)simm_malloc(6 * sdm->neq * sizeof(double));

   // Read and store input data.
   if (input_data_file)
   {
      int time_column = -1;
      set_up_input_data(input_data_file, &input_data);

      // No gencoords are left as unconstrained. Use index_pres, which is the
      // index of one of the Qs that is prescribed in the file, to get the
      // time of the first and last frames of data in the file. Divide this
      // time range by the number of frames to get the step size (reporting
      // interval) for the integrator. If the startTime and endTime are already
      // equal to the times of the first and last frames, don't change stepSize
      // because the user may have changed it to re-sample the data.
      time_column = input_data->time_column;
      if (time_column >= 0)
      {
         if (NOT_EQUAL_WITHIN_ERROR(params.startTime, input_data->motiondata[time_column][0]) ||
            NOT_EQUAL_WITHIN_ERROR(params.endTime, input_data->motiondata[time_column][input_data->num_frames - 1]))
         {
            params.startTime = input_data->motiondata[time_column][0];
            params.endTime = input_data->motiondata[time_column][input_data->num_frames - 1];
            params.stepSize = (params.endTime - params.startTime) / (input_data->num_frames - 1);
         }
         sim_message(none, "All unconstrained Qs are prescribed with data from %s.", input_data_file);
         sim_message(none, "   assuming inverse dynamics analysis, changing:");
         sim_message(none, "      start time to %lf", params.startTime);
         sim_message(none, "      end time to   %lf", params.endTime);
         sim_message(none, "      step size to  %lf", params.stepSize);
      }
   
      // For simulations using only muscles "withEMG", determine which of the model's muscles have activation data
      // specified in the input file and use only these muscles in the simulation.
      if (params.inputOptions.muscles == dpMusclesWithEMG)
      {
         handle_musclesWithEMG_option(y, dy, work);
      }
   }

   // Now that muscles have been read in, update the muscle output parameters
   setup_muscle_output_parameters();

   // Make sure each muscle specifies a muscle model, and if static optimization is chosen, set the model accordingly.
   for (i=0; i<sdm->num_muscles; i++)
   {
      if (find_name_in_list(sdm->muscles[i].name, &params.outputOptions.optimizedActivations_Muscles) >= 0)
      {
         // This muscle's activation is being optimized. Set its model to muscle_model_opt_index.
         if (sdm->muscles[i].muscle_model_opt_index == NULL)
         {
            sim_message(exit_program, "muscle_model_opt is undefined for muscle %s.", sdm->muscles[i].name);
         }
         else
         {
            if (sdm->muscles[i].muscle_model_index == NULL || sdm->muscles[i].muscle_model_index == sdm->default_muscle.muscle_model_index)
               sdm->muscles[i].muscle_model_index = (int*)simm_calloc(1, sizeof(int));
            *sdm->muscles[i].muscle_model_index = *sdm->muscles[i].muscle_model_opt_index;
         }
      }
      else
      {
         if (sdm->muscles[i].muscle_model_index == NULL)
         {
            sim_message(exit_program, "muscle_model is undefined for muscle %s.", sdm->muscles[i].name);
         }
      }
   }

   // Set the initial conditions.
   if (set_initial_conditions(&t, y, dy, input_data) != code_fine)
   {
      printf("Could not initialize simulation.\n");
      return;
   }

   init_motion();

   sdinit();
   check_for_sderror("SDINIT");
   
   // enable constraint stabilization
   sdstab(2.0 * params.baumgarte_stabilization, params.baumgarte_stabilization * params.baumgarte_stabilization);

   // Assemble the model
   if (sdm->newInverseSimulation == dpYes)
   {
      sdm->newInverseSimulation = dpNo;
      if (assemble_model(t, y) == code_bad)
         return;
      sdm->newInverseSimulation = dpYes;
   }
   else
   {
      if (assemble_model(t, y) == code_bad)
         return;
   }

   sdstate(t, y, &y[sdm->nq]);

   if (init_muscle_states(sdm, y) == code_bad)
      return;

   sim_message(none, "\nSystem assembled (%d bodies, %d Qs, %d Us, %d muscles, %d muscle states).\n",
      sdm->num_body_segments, sdm->nq, sdm->nu, sdm->num_muscles, num_muscle_states);

   if (verbose == dpYes)
      print_body_info();
   
#if CONTACT_DETECTION
   if (verbose == dpYes)
      show_object_info();
#endif

   check_for_sderror("MAIN");

   start_time = get_time_now();
   if (verbose == dpYes)
   {
      print_state(y, dy);
   }

#if CONTACT_DETECTION
   check_for_initial_penetration(y, dy);
#endif

   /**************************************************************************/
   sim_message(none, "Running simulation from time %lf to %lf (step size = %lf).\n",
               params.startTime, params.endTime, params.stepSize);

   // Calculate initial accelerations.
   if (params.outputOptions.gencoordMoments.num_items > 0)
   {
      FILE* fp;
      double *pos_euler = NULL, *vel_euler = NULL, *acc_euler = NULL;
      dpBoolean newInverse = sdm->newInverseSimulation; // save a copy of this flag

      // For an inverse simulation, activate prescribed motion and prescribe
      // the Qs, Us, and As according to the input motion. Then write the
      // initial Qs and Us to a file. This is useful if, for example, you
      // later want to use the joint torques to drive a forward simulation
      // to reproduce the motion.
      if (newInverse == dpYes)
      {
         sdm->newInverseSimulation = dpNo;
         set_prescribed_motion(sdm, dpPrescribedQ, 1);
      }
      calc_derivatives(t, y, dy, param, &status);

      pos_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
      vel_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
      acc_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
      convert_state_from_quaternion(t, y, &y[sdm->nq], &dy[sdm->nq], pos_euler, vel_euler, acc_euler, NULL);
      fp = fopen("initialConditions.txt","w");
      for (i = 0; i < sdm->nu; i++)
         fprintf(fp, "%s %21.16lf %21.16lf\n", sdm->q[i].name, pos_euler[i] * sdm->q[i].conversion, vel_euler[i] * sdm->q[i].conversion);
      fclose(fp);

      // Set the newInverseSimulation flag back to what it was.
      sdm->newInverseSimulation = newInverse;
      if (newInverse == dpYes)
         set_prescribed_motion(sdm, dpPrescribedQ, 0);
   }
   else
   {
      calc_derivatives(t, y, dy, param, &status);
   }

   si.initial_system_energy = MINMDOUBLE;

   if (params.outputOptions.systemEnergy == dpYes)
      calc_system_energy(&si);

   if (params.outputOptions.jointReactionForces.num_items > 0 || params.outputOptions.jointReactionTorques.num_items > 0)
      get_reaction_forces();

   if (params.outputOptions.gencoordMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0) ||
      params.outputOptions.gencoordPowers.num_items > 0)
   {
      if (sdm->newInverseSimulation == dpYes)
         calc_initial_gencoord_moments(t, y, dy, input_data);
      else
         calc_gencoord_moments(t, y, dy);
   }

   // Moment arms should be calculated last because the function
   // calls sdstate() which resets all of the accelerations to zero.
   if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
      (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
      params.outputOptions.totalMuscleMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
      calc_muscle_moment_arms(t, y);

   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0)
   {
      char message[255];

      if (set_up_optimization(message) != code_fine)
         sim_message(exit_program, "Could not set up muscle optimization: %s.", message);
      optimize_muscle_activations(t, y);
   }

   // If any data is to be output, to a motion or kinetics file, open the files, write out the header data,
   // then write out the first frame of data.
   if (output_motion_file)
   {
      create_output_motion("sd_motion");
      open_output_motion_file(output_motion_file);
      if (store_motion_frame(t, y, dy) != 0)
      {
         sim_message(recover, "%s", sd_error_msg);
      }
   }
   if (output_kinetics_file)
   {
      set_up_kinetics_output(output_kinetics_file);
      write_kin_out_frame(t, y, dy);
   }

   for (t = params.startTime; t < params.endTime - TINY_TIME; )
   {
      step = step_size = params.stepSize;
      next_t = t + step_size;
      if (next_t > params.endTime)
      {
         next_t = params.endTime;
         step = step_size = next_t - t;
      }

      print_simulation_info(t, y, &si);      

      // To make the simulation run faster, don't let the muscle excitations
      // change during an integration interval. Thus you can compute them here
      // and store them in the muscle structure. If you are spline-fitting
      // your excitation data, you don't really need to do this since the
      // muscle excitations will change smoothly (whereas with step functions
      // they will not), and thus they will not slow down the integration.
      // Note: Starting with version 2.0.1, all muscle derivative functions in
      // derivs.c depend on the muscle excitation being set here. If you remove
      // the code below, you will need to make sure that excitations are set
      // properly in the derivative function that you use.
      for (j = 0; j < sdm->num_muscles; j++)
         sdm->muscles[j].excitation_level = calc_excitation(&sdm->muscles[j], t, y);
      
      while (t < next_t - TINY_TIME)
      {
         calc_derivatives(t, y, dy, param, &status);
         integrate(calc_derivatives, &t, y, dy, param, step_size, &step, sdm->neq,
                   INTEGRATION_TOL, work, &err, &which);

         if (err == INT_NO_ERROR)
         {
            break;
         }
#if CONTACT_DETECTION
         else if (err == INT_USER_FLAG)
         {
            if (which == IMPACT_FLAG)
            {
               handle_impacts(t, y, dy);
               num_impacts++;
               step = step_size = next_t - t;
            }
            else if (which == CONS_VIOL)
            {
               store_motion_frame(t, y, dy);
               sim_message(abort_action, "CONSTRAINT VIOLATION at time %lf.", t);
               goto END_SIMULATION;
            }
            else if (which == BAD_STATE)
            {
               store_motion_frame(t, y, dy);
               sim_message(abort_action, "BAD STATE at time %lf.", t);
               goto END_SIMULATION;
            }
         } 
#endif
         else if (err == INT_OVER_BUMP)
         {
            if (which >= 0 && which < sdm->nu)
               sim_message(recover, "INTEGRATOR: went over a bump (gencoord %s [position]).", sdm->q[which].name);
            else if (which >= sdm->nu && which < sdm->nq + sdm->nu)
               sim_message(recover, "INTEGRATOR: went over a bump (gencoord %s [velocity]).", sdm->q[which - sdm->nq].name);
            else
               sim_message(recover, "INTEGRATOR: went over a bump (state vector element %d).", which);
            step_size = next_t - t;
         } 
         else if (err == INT_CANT_CONTINUE)
         {
            sim_message(abort_action, "SDVINTEG: Cannot continue after time = %lf (gc %d).", t, which);
            store_motion_frame(t, y, dy);
            goto END_SIMULATION;
         } 
         else 
         {
            if (which == CONS_VIOL)
               sim_message(abort_action, "Constraint Violation.");
            sim_message(abort_action, "%s", sd_error_msg);
            sdprinterr(stdout);
            goto END_SIMULATION;
         }
      }

      assign_muscle_states(sdm, t, y);

      if (verbose == dpYes)
         print_state(y, dy);

      if (output_kinetics_file)
         write_kin_out_frame(t, y, dy);

      if (params.outputOptions.systemEnergy == dpYes)
         calc_system_energy(&si);

      if (params.outputOptions.jointReactionForces.num_items > 0 || params.outputOptions.jointReactionTorques.num_items > 0)
         get_reaction_forces();

      if (params.outputOptions.gencoordMoments.num_items > 0 ||
         (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0) ||
         params.outputOptions.gencoordPowers.num_items > 0)
      {
         if (sdm->newInverseSimulation == dpYes)
         {
            if (t < params.endTime - TINY_TIME)
            {
               calc_next_gencoord_moments(t, y, dy, input_data);
               calc_corrective_moments(t, y, dy, sdm, input_data);
            }
            else
            {
               // You're at the end of the simulation, so you don't need to
               // calculate corrective torques for the next time step. Just get
               // the corrected torques from this time step and store them in
               // the sdm->q[] array so they can be written to the output motion
               // file and used by the static muscle optimization function.
               
               for (i = 0; i < sdm->nu; i++)
               {
                  if (input_data->q_moments[i])
                     sdm->q[i].moment = input_data->q_moments[i]->y[input_data->q_moments[i]->numpoints-1];
               }
            }
         }
         else
         {
            calc_gencoord_moments(t, y, dy);
         }
      }

      // Moment arms should be calculated near the end because the function
      // calls sdstate() which resets all of the accelerations to zero.
      if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
         (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
         params.outputOptions.totalMuscleMoments.num_items > 0 ||
         (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
         calc_muscle_moment_arms(t, y);

      if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
         optimize_muscle_activations(t, y);

      if (output_motion_file)
      {
         if (store_motion_frame(t, y, dy) != 0)
         {
            sim_message(recover, "%s", sd_error_msg);
         }
      }
   }

END_SIMULATION:
   sim_message(none, "*** SIMULATION ENDED ***");
   if (verbose == dpYes)
   {
      print_final_information(t, y, dy, &si);
   }

   if (output_motion_file)
   {
      close_output_motion_file();
      sim_message(none, "Motion data written to: %s.", output_motion_file);
   }

   if (output_kinetics_file)
   {
      fclose(kin_out);
      sim_message(none, "Kinetics data written to: %s.", output_kinetics_file);
   }

   free(work);
   free(y);
   free(dy);

   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
      cleanup_optimization();

   end_time = get_time_now();
   time_elapsed(start_time, end_time);  // dkb to test new optimization
   if (verbose == dpYes)
   {
      time_elapsed(start_time, end_time);

#if CONTACT_DETECTION
      sim_message(none, "%d impact[s] encountered.", num_impacts); 
#endif
   }

   // free sdm, qs,body_segment
   for (i = 0; i < sdm->nu; i++)
      FREE_IFNOTNULL(sdm->q[i].name);
   FREE_IFNOTNULL(sdm->q);
   for (i = 0; i < sdm->num_body_segments; i++)
      FREE_IFNOTNULL(sdm->body_segment[i].name);
   FREE_IFNOTNULL(sdm->body_segment);
   for (i = 0; i < sdm->num_constraint_objects; i++)
      FREE_IFNOTNULL(sdm->constraint_object[i].name);
   FREE_IFNOTNULL(sdm->constraint_object);
   for (i = 0; i < sdm->num_wrap_objects; i++)
      FREE_IFNOTNULL(sdm->wrap_object[i].name);
   FREE_IFNOTNULL(sdm->wrap_object);
   FREE_IFNOTNULL(sdm->joint);
   for (i = 0; i < sdm->function_array_size; i++)
   {
      FREE_IFNOTNULL(sdm->function[i]->x);
      FREE_IFNOTNULL(sdm->function[i]->y);
      FREE_IFNOTNULL(sdm->function[i]->b);
      FREE_IFNOTNULL(sdm->function[i]->c);
      FREE_IFNOTNULL(sdm->function[i]->d);
   }
   FREE_IFNOTNULL(sdm->function);
   for (i = 0; i < sdm->num_force_mattes; i++)
   {
      for (j = 0; j < sdm->force_matte[i].ph->num_polygons; j++)
      {
         int k;
         for (k = 0; k < sdm->force_matte[i].ph->polygon[k].num_vertices; k++)
            FREE_IFNOTNULL(sdm->force_matte[i].ph->polygon[k].vertex_index);
         FREE_IFNOTNULL(sdm->force_matte[i].ph->vertex);
         FREE_IFNOTNULL(sdm->force_matte[i].ph->polygon);
      }
      FREE_IFNOTNULL(sdm->force_matte[i].name);
      FREE_IFNOTNULL(sdm->force_matte[i].ph);
   }
   FREE_IFNOTNULL(sdm->force_matte);
   for (i = 0; i < sdm->num_spring_floors; i++)
   {
      for (j = 0; j < sdm->spring_floor[i].ph->num_polygons; j++)
      {
         int k;
         for (k = 0; k < sdm->spring_floor[i].ph->polygon[j].num_vertices; k++)
            FREE_IFNOTNULL(sdm->spring_floor[i].ph->polygon[j].vertex_index);
         FREE_IFNOTNULL(sdm->spring_floor[i].ph->vertex);
      }
      FREE_IFNOTNULL(sdm->spring_floor[i].ph->polygon);
      FREE_IFNOTNULL(sdm->spring_floor[i].name);
      FREE_IFNOTNULL(sdm->spring_floor[i].ph);
   }
   FREE_IFNOTNULL(sdm->name);

   FREE_IFNOTNULL(input_data_file);
   FREE_IFNOTNULL(output_motion_file);
   FREE_IFNOTNULL(output_kinetics_file);

   sim_message(none, "\nPress Enter to continue.");
   getc(stdin);

}


static void init_model(void)
{
   int i, info[50];

   params.startTime = 0.0;
   params.endTime = 1.0;
   params.stepSize = 0.01;
   params.baumgarte_stabilization = BAUMGARTE_STAB;
   params.outputOptions.contactForces = dpYes;
   params.outputOptions.traditionalInverse = dpYes;
   params.outputOptions.systemEnergy = dpNo;
   params.outputOptions.numSpringForces = 20;
   params.outputOptions.normalization = dpNoNormalization;
   params.inputOptions.emg_cutoff = 0.0;
   params.inputOptions.emg_cutoff_defined = dpNo;
   params.inputOptions.force_cutoff = 10.0;
   params.inputOptions.force_cutoff_defined = dpNo;
   params.inputOptions.kinematic_cutoff = 25.0;
   params.inputOptions.kinematic_cutoff_defined = dpNo;
   params.inputOptions.muscles = dpAllMuscles;
   params.inputOptions.include_emg = dpYes;
   params.inputOptions.include_forces = dpYes;
   params.inputOptions.include_kinematics = dpYes;

   clear_name_list(&params.outputOptions.gencoordMoments);
   clear_name_list(&params.outputOptions.gencoordPowers);
   clear_name_list(&params.outputOptions.jointReactionForces);
   clear_name_list(&params.outputOptions.jointReactionTorques);
   clear_name_list(&params.outputOptions.massCenterPositions);
   clear_name_list(&params.outputOptions.massCenterVelocities);
   clear_name_list(&params.outputOptions.muscleMomentArms_Muscles);
   clear_name_list(&params.outputOptions.muscleMomentArms_Gencoords);
   clear_name_list(&params.outputOptions.muscleMoments_Muscles);
   clear_name_list(&params.outputOptions.muscleMoments_Gencoords);
   clear_name_list(&params.outputOptions.totalMuscleMoments);
   clear_name_list(&params.outputOptions.optimizedActivations_Muscles);
   clear_name_list(&params.outputOptions.optimizedActivations_Gencoords);
   clear_name_list(&params.outputOptions.muscleActivations);
   clear_name_list(&params.outputOptions.muscleLengths);
   clear_name_list(&params.outputOptions.muscleFiberLengths);
   clear_name_list(&params.outputOptions.muscleFiberVelocities);
   clear_name_list(&params.outputOptions.muscleForces);
   clear_name_list(&params.outputOptions.gencoordValues);
   clear_name_list(&params.outputOptions.gencoordVelocities);
   clear_name_list(&params.outputOptions.gencoordAccelerations);

   sdinfo(info);
   sdm = (dpModelStruct*)simm_calloc(1, sizeof(dpModelStruct));
   sdm->num_body_segments = info[1] + 1;    // include ground
   sdm->nq = info[2] + info[7];
   sdm->nu = info[2];
   sdm->num_closed_loops = info[4];
   sdm->num_joints = info[1] + info[4]; // nbod + nloop
   sdm->num_constraints = info[3];
   sdm->num_user_constraints = info[10];
   sdm->num_springs = 0;
   sdm->num_spring_floors = 0;
   sdm->num_force_mattes = 0;
   sdm->spring_array_size = ARRAY_INCREMENT;
   sdm->spring_floor_array_size = ARRAY_INCREMENT;
   sdm->force_matte_array_size = ARRAY_INCREMENT;
   sdm->enforce_constraints = dpYes;
   sdm->newInverseSimulation = dpNo;
   sdm->num_functions = 0;
   sdm->function_array_size = ARRAY_INCREMENT;

   init_qs();
   init_segments();
   init_joints(dpYes);
   init_quaternion_joints();
   init_joint_names();
   init_wrap_objects();
   init_constraint_objects();
   init_functions(sdm);

#if CONTACT_DETECTION
   makepaths();

   for (i = 0; i < sdm->num_body_segments; i++)
   {
      sdm->body_segment[i].num_objects = 0;
      make_vector(sdm->body_segment[i].impact_force, 0.0, 0.0, 0.0);
      make_vector(sdm->body_segment[i].impact_point, 0.0, 0.0, 0.0);
      make_vector(sdm->body_segment[i].contact_force, 0.0, 0.0, 0.0);
   }
#endif
}


/* SIM_MESSAGE: this function prints an error message depending on the string
 * and error status that is passed in. There is a separate version of
 * this function for the DLL version of a simulation.
 */
int sim_message(ErrorAction action, const char* format, ...)
{
   int n;
   char c;
   va_list ap;
   char buffer3[CHARBUFFER];

   va_start(ap, format);
   n = vsprintf(buffer3, format, ap);
   va_end(ap);

   (void)fprintf(stdout,"%s\n", buffer3);

   if (action == recover)
   {
      (void)fprintf(stdout,"Do you want to continue? (y/n)<y>:  ");
      c = getc(stdin);
      if ((c == 'n') || (c == 'N'))
      {
         (void)fprintf(stdout, "Program terminated.\n");
         getc(stdin);
         exit(0);
      }
      else
      {
         (void)fprintf(stdout, "Continuing...\n");
         fflush(stdin);
      }
   }
   else if (action == abort_action)
   {
      (void)fprintf(stdout, "Aborting...\n");
   }
   else if (action == exit_program)
   {
      (void)fprintf(stdout, "Program terminated.\n");
      getc(stdin);
      exit (0);
   }

   return n;

}

static void setup_muscle_output_parameters(void)
{
   int i, index;

   allocate_name_list(&params.outputOptions.muscleActivations, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.muscleFiberLengths, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.muscleFiberVelocities, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.muscleForces, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.muscleLengths, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.muscleMomentArms_Muscles, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.muscleMoments_Muscles, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.optimizedActivations_Muscles, sdm->num_muscles);
   allocate_name_list(&params.outputOptions.totalMuscleMoments, sdm->num_muscles);
   for (i = 0; i < sdm->num_muscles; i++)
   {
      if (params.outputOptions.muscleActivations.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleActivations);
         params.outputOptions.muscleActivations.index[index] = i;
      }
      if (params.outputOptions.muscleFiberLengths.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleFiberLengths);
         params.outputOptions.muscleFiberLengths.index[index] = i;
      }
      if (params.outputOptions.muscleFiberVelocities.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleFiberVelocities);
         params.outputOptions.muscleFiberVelocities.index[index] = i;
      }
      if (params.outputOptions.muscleForces.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleForces);
         params.outputOptions.muscleForces.index[index] = i;
      }
      if (params.outputOptions.muscleLengths.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleLengths);
         params.outputOptions.muscleLengths.index[index] = i;
      }
      if (params.outputOptions.muscleMomentArms_Muscles.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleMomentArms_Muscles);
         params.outputOptions.muscleMomentArms_Muscles.index[index] = i;
      }
      if (params.outputOptions.muscleMoments_Muscles.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleMoments_Muscles);
         params.outputOptions.muscleMoments_Muscles.index[index] = i;
      }
      if (params.outputOptions.optimizedActivations_Muscles.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.optimizedActivations_Muscles);
         params.outputOptions.optimizedActivations_Muscles.index[index] = i;
      }
      if (params.outputOptions.totalMuscleMoments.state)
      {
         index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.totalMuscleMoments);
         params.outputOptions.totalMuscleMoments.index[index] = i;
      }
   }
}

void handle_musclesWithEMG_option(double y[], double dy[], double work[])
{
   int i, num_muscle_states,muscle_index, num_muscles = 0;
   double cutoffFrequency;
   ReturnCode rc;
   dpFunctionType functionType;
   dpMuscleStruct *newList = (dpMuscleStruct*)simm_malloc(sdm->num_muscles*sizeof(dpMuscleStruct));

   // Determine which muscles have the necessary input data.
   for (i = 0; i < input_data->num_elements; i++)
   {
      muscle_index = name_is_muscle(input_data->elementnames[i], sdm, NULL, &functionType, &cutoffFrequency);
      if (muscle_index >= 0)
         copyMuscle(&sdm->muscles[muscle_index], &sdm->default_muscle, &newList[num_muscles++], &sdm->default_muscle, NULL, NULL);
   }

   // Remove the old muscle list
   FREE_IFNOTNULL(sdm->muscles);
   // Copy the new muscle list into the model structure.
   sdm->num_muscles = num_muscles;
   sdm->muscles = (dpMuscleStruct *)simm_realloc(sdm->muscles, num_muscles * sizeof(dpMuscleStruct), &rc);
   for (i = 0; i < sdm->num_muscles; i++)
      copyMuscle(&newList[i], &sdm->default_muscle, &sdm->muscles[i], &sdm->default_muscle, NULL, NULL);

   // Reset the number of muscle states in the model.
   num_muscle_states = count_muscle_states(sdm);
   sdm->neq = sdm->nu + sdm->nq + num_muscle_states;

   y = (double*)simm_realloc(y, sdm->neq * sizeof(double), &rc);
   dy = (double*)simm_realloc(dy, sdm->neq * sizeof(double), &rc);
   work = (double*)simm_realloc(work, 6 * sdm->neq * sizeof(double), &rc);

   // Make space for the muscle moment arms, if appropriate.
   if ((params.outputOptions.muscleMomentArms_Muscles.state && params.outputOptions.muscleMomentArms_Gencoords.state) ||
      (params.outputOptions.muscleMoments_Muscles.state && params.outputOptions.muscleMoments_Gencoords.state) ||
      params.outputOptions.totalMuscleMoments.state ||
      (params.outputOptions.optimizedActivations_Muscles.state && params.outputOptions.optimizedActivations_Gencoords.state))
   {
      for (i = 0; i < sdm->num_muscles; i++)
      {
         sdm->muscles[i].nummomentarms = sdm->nu;
         sdm->muscles[i].momentarms = (double*)simm_malloc(sdm->muscles[i].nummomentarms * sizeof(double));
      }
   }
}