/*******************************************************************************

   MATHTOOLS.C

   Author: Peter Loan
           Krystyne Blaikie

   Copyright (c) 1992-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains functions that implement various
      mathematical tools and functions. Every one operates exclusively
      on the formal parameter list, and so does not change any external
      variables or other general model or plot structures.

   Functions:
      append_matrix             : multiplies two 4x4 matrices and stores result in first
      calc_function_coefficients: finds coefficients of natural cubic spline
      clear_vector              : sets all vector elements to 0.0
      compute_angle_between_vectors : calcs angle between two vectors
      copy_3x3matrix            : copies a 3x3 matrix
      copy_4x4matrix            : copies a 3x3 matrix
      cross_vectors             : finds cros product of two vectors
      distancesqr_between_vertices : calcs square of distance between two 3D points
      distance_between_vertices : calcs distance between two 3D points
      find_end_of_array         : finds the end of an integer array
      format_double             : determines appropriate printf format for a number
      get_distsqr_point_line    : calcs square of shortest distance from point to plane
      get_point_from_point_line : calcs point on line closest to a 3D point
      integrate                 : variable step integrator used by main() and dllmain()
      interpolate_function      : given an x, interpolates a cubic to get a y value
      intersect_line_plane01    : calcs intersection between line segment and plane
      intersect_line_plane      : calcs intersection between line and plane
      intersect_lines           : calcs intersection of two lines in 3D
      intersect_lines_scaled    : alternate form of intersect_lines()
      make_3x3dircos_matrix     : makes a 3x3 direction cosine matrix from axis/angle
      make_3x3matrix            : makes a 3x3 matrix
      make_3x3_xrot_matrix      : makes a 3x3 direction cosine matrix with X rotation
      make_4x4dircos_matrix     : makes a 4x4 direction cosine matrix from axis/angle
      make_vector               : makes a 1x3 vector
      make_quaternion           : makes a quaternion given an angle and axis
      mult_3x3matrices          : multiplies two 3x3 matrices
      mult_3x3matrix_by_vector  : premultiplies a vector by a 3x3 matrix
      mult_4x4matrix_by_vector  : multiplies a 1x4 vector by a 4x4 matrix
      mult_4x4matrices          : multiplies two 4x4 matrices
      normalize_vector          : normalizes a vector
      point_in_polygon3D        : determines if a point is inside a 3D polygon
      point_in_polygon2D        : determines if a point is inside a 2D polygon
      print_3x3matrix           : prints a 3x3 matrix to a file
      project_point_onto_plane  : projects 3D point onto 3D plane
      quat_to_matrix            : converts a quaternion into a 4x4 transform matrix
      rotate_matrix_by_quat     : appends a quaternion rotation onto a 4x4 matrix
      rotate_matrix_axis_angle  : appends an angle/axis rotation onto a 4x4 matrix
      reset_3x3matrix           : sets a 3x3 matrix to the identity matrix
      transform_pt              : multiplies a point by a transform matrix
      transform_vec             : multiplies a vector by a transform matrix
      translate_matrix          : appends a translation onto a 4x4 matrix
      transpose_3x3matrix       : transposes a 3x3 matrix
      vect_equal_within_tol     : checks if two vectors are equal
      zero_3x3matrix            : checks if any row of a 3x3 matrix is all zeros

      convert_rotation_to_quaternion: convert rotation matrix to quaternion
      extract_euler_angles      : extract euler angles from rotation matrix
      set_rotation_matrix_from_euler_angles : set rotation matrix given euler angles
      set_rotation_matrix_from_quaternion   : set rotation matrix given quaternion

*******************************************************************************/

#include "universal.h"


/*************** DEFINES (for this file only) *********************************/
#define LINE_EPSILON 0.00001
#define SMALLNUM 1e-7

#ifndef M_PI
#define M_PI 3.1415926535897
#define M_PI_2 (M_PI/2.0)
#endif

/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/



/*************** EXTERNED VARIABLES (declared in another file) ****************/



/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static void calc_natural_cubic_coefficients(dpFunction* f);
static void make_quaternion(Quat, const double axis[3], double angle);
static void rotate_matrix_by_quat(double m[][4], const Quat);
static void quat_to_matrix(const Quat q, double m[][4]);
static void setThreeAngleThreeAxesBodyFixedForwardCyclicalRotation
  ( double cosAngle1, double sinAngle1, int axis1, 
    double cosAngle2, double sinAngle2, int axis2,
    double cosAngle3, double sinAngle3, int axis3,
    double R[][3]);

/* APPEND_MATRIX: computes m * b and stores the result in m. */
void append_matrix(DMatrix m, DMatrix b)
{
   /* append matrix 'b' to matrix 'm' */ 
   double ta[4][4];
   
   memcpy(ta, m, 16 * sizeof(double));

   m[0][0] = ta[0][0] * b[0][0] + ta[0][1] * b[1][0] + ta[0][2] * b[2][0] + ta[0][3] * b[3][0];
   m[0][1] = ta[0][0] * b[0][1] + ta[0][1] * b[1][1] + ta[0][2] * b[2][1] + ta[0][3] * b[3][1];
   m[0][2] = ta[0][0] * b[0][2] + ta[0][1] * b[1][2] + ta[0][2] * b[2][2] + ta[0][3] * b[3][2];
   m[0][3] = ta[0][0] * b[0][3] + ta[0][1] * b[1][3] + ta[0][2] * b[2][3] + ta[0][3] * b[3][3];

   m[1][0] = ta[1][0] * b[0][0] + ta[1][1] * b[1][0] + ta[1][2] * b[2][0] + ta[1][3] * b[3][0];
   m[1][1] = ta[1][0] * b[0][1] + ta[1][1] * b[1][1] + ta[1][2] * b[2][1] + ta[1][3] * b[3][1];
   m[1][2] = ta[1][0] * b[0][2] + ta[1][1] * b[1][2] + ta[1][2] * b[2][2] + ta[1][3] * b[3][2];
   m[1][3] = ta[1][0] * b[0][3] + ta[1][1] * b[1][3] + ta[1][2] * b[2][3] + ta[1][3] * b[3][3];

   m[2][0] = ta[2][0] * b[0][0] + ta[2][1] * b[1][0] + ta[2][2] * b[2][0] + ta[2][3] * b[3][0];
   m[2][1] = ta[2][0] * b[0][1] + ta[2][1] * b[1][1] + ta[2][2] * b[2][1] + ta[2][3] * b[3][1];
   m[2][2] = ta[2][0] * b[0][2] + ta[2][1] * b[1][2] + ta[2][2] * b[2][2] + ta[2][3] * b[3][2];
   m[2][3] = ta[2][0] * b[0][3] + ta[2][1] * b[1][3] + ta[2][2] * b[2][3] + ta[2][3] * b[3][3];

   m[3][0] = ta[3][0] * b[0][0] + ta[3][1] * b[1][0] + ta[3][2] * b[2][0] + ta[3][3] * b[3][0];
   m[3][1] = ta[3][0] * b[0][1] + ta[3][1] * b[1][1] + ta[3][2] * b[2][1] + ta[3][3] * b[3][1];
   m[3][2] = ta[3][0] * b[0][2] + ta[3][1] * b[1][2] + ta[3][2] * b[2][2] + ta[3][3] * b[3][2];
   m[3][3] = ta[3][0] * b[0][3] + ta[3][1] * b[1][3] + ta[3][2] * b[2][3] + ta[3][3] * b[3][3];
}

/* CALC_FUNCTION_COEFFICIENTS: this function takes an array of x and y values,
 * and computes the coefficients to interpolate the data. It handles linear,
 * natural cubic, and GCV spline functions (step functions do not need
 * coefficients).
 */
ReturnCode calc_function_coefficients(dpFunction* sf)
{
   if (sf == NULL)
      return code_fine;

   if (sf->status == dpFunctionUndefined)
      return code_fine;

   if (sf->b)
      memset(sf->b, 0.0, sf->coefficient_array_size * sizeof(double));
   if (sf->c)
      memset(sf->c, 0.0, sf->coefficient_array_size * sizeof(double));
   if (sf->d)
      memset(sf->d, 0.0, sf->coefficient_array_size * sizeof(double));

	if (sf->type == dpLinearFunction)
	{
		int i;

		if (sf->numpoints < 2)
		{
			sim_message(abort_action, "Error: linear function must have at least 2 points.\n");
			return code_bad;
		}

		/* For a linear function, the slope of each segment is stored in the b array. */
		for (i = 0; i < sf->numpoints - 1; i++)
			sf->b[i] = (sf->y[i+1] - sf->y[i]) / (sf->x[i+1] - sf->x[i]);
		sf->b[sf->numpoints - 1] = sf->b[sf->numpoints - 2];
	}
   else if (sf->type == dpNaturalCubicSpline)
   {
      calc_natural_cubic_coefficients(sf);
   }
   else if (sf->type == dpGCVSpline)
   {
      int i, ier = 0;
      double val, sampling_rate, *weight_x = NULL, *weight_y = NULL, *work = NULL;
      int mode = 1; // Mode for GCVSPL function chosen to specify smoothing parameter directly
      int order = 5; //5th order for quintic spline
      int half_order = (order + 1) / 2; // GCVSPL works with the half order
      int k = 1; // Only one colum of y data is splined at a time

      sampling_rate = 1.0 / (sf->x[1] - sf->x[0]);

      // Allocate memory as needed
      weight_x = (double*)simm_malloc(sf->numpoints*sizeof(double));
      weight_y = (double*)simm_malloc(1*sizeof(double));
      work = (double*)simm_malloc((6*(sf->numpoints*half_order+1)+sf->numpoints)*sizeof(double));

      // Assign weights
      for (i = 0; i < sf->numpoints; i++)
         weight_x[i] = 1.0;
      weight_y[0] = 1.0;

      // Calculate GCVSPL version of cut-off frequency. If cut-off frequency is <= 0.0,
      // set val = 0.0 so that no smoothing is performed.
      if (sf->cutoff_frequency <= 0.0)
         val = 0.0;
      else
         val = sampling_rate / pow(2.0*M_PI*sf->cutoff_frequency/pow(sqrt(2.0)-1.0,0.5/half_order), 2.0*half_order);

      gcvspl(sf->x, sf->y, &sf->numpoints, weight_x, weight_y, &half_order, 
             &sf->numpoints, &k, &mode, &val, sf->c, &sf->numpoints, work, &ier);

		free(weight_x);
      free(weight_y);
      free(work);

      if (ier > 0)
      {
         if (ier == 1)
            sim_message(abort_action, "GCV spline error: Only %d coordinates specified (%d or more are required).",
                    sf->numpoints, order + 1);
         else if (ier == 2)
            sim_message(abort_action, "GCV spline error: X coordinates do not consistently increase in value.");
         else
            sim_message(abort_action, "Error code returned by gcvspl() = %d.", ier);
         return code_bad;
      }
   }

   return code_fine;
}

/* CLEAR_VECTOR: sets all vector elements to 0.0. */
void clear_vector(double a[], int n)
{
   int i;

   for (i=0; i<n; i++)
      a[i] = 0.0;
}


/* COMPUTE_ANGLE_BETWEEN_VECTORS: calculates angle between two vectors.
 * Result is in the range 0 to pi radians.
 */
double compute_angle_between_vectors(double vector1[], double vector2[])
{
   double normal_vector1[3], normal_vector2[3];

   normalize_vector(vector1,normal_vector1);
   normalize_vector(vector2,normal_vector2);

   return (acos(DOT_VECTORS(normal_vector1,normal_vector2)));
}

/* COPY_3X3MATRIX: copies a 3x3 matrix. */
void copy_3x3matrix(double from[][3], double to[][3])
{
   to[0][0] = from[0][0];
   to[0][1] = from[0][1];
   to[0][2] = from[0][2];
   to[1][0] = from[1][0];
   to[1][1] = from[1][1];
   to[1][2] = from[1][2];
   to[2][0] = from[2][0];
   to[2][1] = from[2][1];
   to[2][2] = from[2][2];
}

/* COPY_4x4MATRIX: copies a 4x4 matrix. */
void copy_4x4matrix(double from[][4], double to[][4])
{
   to[0][0] = from[0][0];
   to[0][1] = from[0][1];
   to[0][2] = from[0][2];
   to[0][3] = from[0][3];
   to[1][0] = from[1][0];
   to[1][1] = from[1][1];
   to[1][2] = from[1][2];
   to[1][3] = from[1][3];
   to[2][0] = from[2][0];
   to[2][1] = from[2][1];
   to[2][2] = from[2][2];
   to[2][3] = from[2][3];
   to[3][0] = from[3][0];
   to[3][1] = from[3][1];
   to[3][2] = from[3][2];
   to[3][3] = from[3][3];
}

void copy_vector(double from[], double to[], int length)
{
   int i;

   for (i = 0; i < length; i++)
      to[i] = from[i];
}

/* CROSS_VECTORS: computes the cross product of two 1x3 vectors */
void cross_vectors(double vector1[], double vector2[], double result[])
{
   result[0] = vector1[1] * vector2[2] - vector1[2] * vector2[1];
   result[1] = vector1[2] * vector2[0] - vector1[0] * vector2[2];
   result[2] = vector1[0] * vector2[1] - vector1[1] * vector2[0];
}

/* DISTANCE_BETWEEN_VERTICES: calculates the distance between two 3D points. */
double distance_between_vertices(double vertex1[], double vertex2[])
{
   double vec[3];

   vec[0] = vertex2[0] - vertex1[0];
   vec[1] = vertex2[1] - vertex1[1];
   vec[2] = vertex2[2] - vertex1[2];

   return (sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]));
}

/* DISTANCESQR_BETWEEN_VERTICES: calculates the square of the distance
 * between two 3D points.
 */
double distancesqr_between_vertices(double vertex1[], double vertex2[])
{
   double vec[3];

   vec[0] = vertex2[0] - vertex1[0];
   vec[1] = vertex2[1] - vertex1[1];
   vec[2] = vertex2[2] - vertex1[2];

   return (vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]);
}

/* FIND_END_OF_ARRAY: this routine finds the end of an array of integers,
 * returning the value of the element at the end, and recording the length.
 * The end of the array should be defined with the END_OF_ARRAY constant.
 */
int find_end_of_array(int list[], int* length)
{
   *length = 0;

   while (list[(*length)++] != END_OF_ARRAY)
     ;

   return (list[(*length)-2]);
}

/* FORMAT_DOUBLE: this routine finds a good format (printf style) for a given
 * number by checking its magnitude. For example, if the number is greater than
 * 50, you don't really need to show any decimal places, but if the number is
 * less than 0.01, you should show several decimal places.
 */
void format_double(double number, char format[])
{
   if (number > 50.0)
      (void)strcpy(format,"%.3lf");
   else if (number > 10.0)
      (void)strcpy(format,"%.4lf");
   else if (number > 1.0)
      (void)strcpy(format,"%.5lf");
   else if (number > 0.1)
      (void)strcpy(format,"%.6lf");
   else if (number > 0.01)
      (void)strcpy(format,"%.7lf");
   else
      (void)strcpy(format,"%.8lf");
}

/* GET_DISTSQR_POINT_LINE: calculates the square of the shortest distance from
 * a point (point) to a line (vl, through pl).
 */
double get_distsqr_point_line(double point[], double pl[], double vl[])
{
   double ptemp[3];

   /* find the closest point on line */
   get_point_from_point_line(point,pl,vl,ptemp);

   return (distancesqr_between_vertices(point,ptemp));
}

/* GET_POINT_FROM_POINT_LINE: calculates the closest 3d point to given 3d line.
 * The line is defined as vector(vec) and a point(pt) and has not been normalized
 * to a unit vector. The value hypo*(cosalpha) of a rt triangle is calculated
 * to get the closest point
 */
void get_point_from_point_line(double point[], double pt[], double vec[],
                               double closest_pt[])
{
   double v1[3], v2[3];
   double t, mag;

   v1[0] = point[0] - pt[0];
   v1[1] = point[1] - pt[1];
   v1[2] = point[2] - pt[2];

   v2[0] = vec[0];
   v2[1] = vec[1];
   v2[2] = vec[2];

   mag = normalize_vector(v1,v1);
   normalize_vector(v2,v2);
   t = DOT_VECTORS(v1,v2) * mag;

   closest_pt[0] = pt[0] + t * v2[0];
   closest_pt[1] = pt[1] + t * v2[1];
   closest_pt[2] = pt[2] + t * v2[2];
}

/* A variable-step integrator, taken from the SD/FAST-generated file sdlib.c.
 * Work should be dimensioned 6*neq.
 */
void integrate(void (*func)(),
    double *time,
    double st[],
    double dst[],
    double param[],
    double dt,
    double *step,
    int neqin,
    double tol,
    double work[],
    int *err,
    int *which)
{
    int i,wh,wh2,xst,xst2,xdst,errs,morework,errf,neq,earlyret;
    double ttime,ntime,xtime,tfin,stp,stp2,nstp,maxerr,minstep;
    int lookingForImpact = NO;

    if (neqin < 0) {
        neq = -neqin;
        earlyret = 1;
    } else {
        neq = neqin;
        earlyret = 0;
    }
    xst = 0;
    xst2 = xst+neq;
    xdst = xst2+neq;
    errs = xdst+neq;
    morework = errs+neq;
    minstep = MIN_INTEGRATION_STEP;
    ttime = *time;
    tfin = ttime+dt;
    stp = *step;
    if (stp < minstep) {
        stp = minstep;
    }
    *err = *which = 0;
    if (dt <= 0.) {
        errf = 0;
        func(ttime,st,dst,param,&errf);
        if (errf != 0) {
            *err = 3;
            *which = errf;
        }
        return;
    }

    while (ttime  <  tfin)
    {
        if (ttime+1.25*stp >= tfin)
        {
            stp = tfin-ttime;
        }
        nstp = stp;
        for (;;)
        {
            lookingForImpact = NO;
            sdrk4m(func,ttime,st,dst,param,nstp,&work[xst],neq,&work[morework],&work[errs],&maxerr,&wh);
            xtime = ttime + nstp;
            if (maxerr <= tol)
            {
                func(xtime,&work[xst],&work[xdst],param,&errf);
                if (errf == 1)
                   lookingForImpact = YES;
                else
                   lookingForImpact = NO;
                if (errf == 0)
                {
                   stp2 = 2.0 * nstp;
                   if ((maxerr * 64.0 <= tol) && (stp2 <= dt))
                   {
                        stp = stp2;
                   }
                   else
                   {
                      stp = nstp;
                   }
                   break;
                }
                else // impact encountered
                {
                   stp2 = nstp * 0.5;
                }
            }
            else // maxerr > tol
            {
               stp2 = nstp * 0.5;
            }
            if (lookingForImpact == YES && stp2 < MIN_IMPACT_STEP)
            {
               *err = 3;
               *which = IMPACT_FLAG;
               *time = xtime;
               param[0] = 1.0;
               func(xtime,&work[xst],&work[xdst],param,&errf);
               param[0] = -1.0;
               for (i = 0; i < neq; i++)
               {
                  st[i] = work[xst+i];
                  dst[i] = work[xdst+i];
               }
               return;
            }
            else if (stp2 < minstep)
            {
                func(xtime,&work[xst],&work[xdst],param,&errf);
                if (errf != 0) {
                    *err = 3;
                    *which = errf;
                    *time = ttime;
                    *step = nstp;
                    func(ttime,st,dst,param,&errf);
                    return;
                }
                if (*err == 0)
                {
                    *err = 1;
                    *which = wh;
                }
                if (xtime+nstp > tfin)
                {
                    stp2 = tfin-xtime;
                }
                else
                {
                    stp2 = nstp;
                }
                ntime = xtime+stp2;
                if (ntime == xtime)
                {
                    stp = nstp;
                    break;
                }
                sdrk4m(func,xtime,&work[xst],&work[xdst],param,stp2,&work[xst2]
                  ,neq,&work[morework],&work[errs],&maxerr,&wh2);
                if (work[errs+wh] <= tol) {
                    func(xtime,&work[xst],&work[xdst],param,&errf);
                    stp = nstp;
                    break;
                }
                *err = 2;
                *which = wh;
                *time = ttime;
                *step = nstp;
                func(ttime,st,dst,param,&errf);
                return;
            } else {
                nstp = stp2;
            }
        }
        ttime = xtime;
        for (i = 0; i < neq; i++) {
            st[i] = work[xst+i];
            dst[i] = work[xdst+i];
        }
        if (stp >= minstep) {
            *step = stp;
        }
        ntime = ttime+stp;
        if (ntime == ttime) {
            ttime = tfin;
        }
        if (earlyret != 0) {
            *time = ttime;
            return;
        }
    }

    *time = tfin;
}

/* INTERPOLATE_FUNCTION: given a spline function and an x-value, this routine
 * finds the corresponding y-value by interpolating the spline. It
 * can return the zeroth, first, or second derivative of the spline
 * at that x-value. Interpolation can be done with one of three methods:
 * dpStepFunction (zero-order hold), dpNaturalCubicSpline (3rd order spline), and
 * dpGCVSpline (5th order B-spline).
 */
double interpolate_function(double abscissa, dpFunction* func, Derivative deriv, double velocity, double acceleration)
{
   int i, j, k, n;
   double dx;

   if (func == NULL)
      sim_message(exit_program, "interpolate_function: attempted to access NULL function.");

   if (func->status == dpFunctionUndefined || func->used == dpNo)
      return 0.0;

   if (func->type == dpStepFunction)
   {
      if (deriv != zeroth)
         return 0.0;

      for (i = func->numpoints - 1; i >= 0; i--)
      {
         if (abscissa >= func->x[i] - TINY_TIME)
            return func->y[i];
      }

      /* If the abscissa is less than x[0], return y[0]. */
      return func->y[0];
   }
	else if (func->type == dpLinearFunction)
	{
		if (abscissa < func->x[0])
		{
         if (deriv == zeroth)
            return func->y[0] - (func->x[0] - abscissa) * func->b[0];
         else if (deriv == first)
            return func->b[0] * velocity;
         else // deriv == second
            return 0.0;
		}
		else
		{
			for (i = func->numpoints - 1; i >= 0; i--)
			{
            if (abscissa >= func->x[i] - ROUNDOFF_ERROR)
            {
               if (deriv == zeroth)
					return func->y[i] + (abscissa - func->x[i]) * func->b[i];
               else if (deriv == first)
                  return func->b[i] * velocity;
               else // deriv == second
                  return 0.0;
            }
			}
		}
	}
   else if (func->type == dpGCVSpline)
   {
      int l = 1;
      int order = 5; //5th order for quintic spline
      int half_order = (order + 1) / 2; // GCVSPL works with the half order
      double work[20]; // size of work array must be >= order

      if (deriv == zeroth)
      {
         return splder((int*)&deriv, &half_order, &func->numpoints, &abscissa, func->x, func->c, &l, work);
      }
      else if (deriv == first)
      {
         // f'(x) = velocity * gcv'(x)
         return velocity * splder((int*)&deriv, &half_order, &func->numpoints, &abscissa, func->x, func->c, &l, work);
      }
      else // deriv == second
      {
         // f''(x) = velocity * velocity * gcv''(x) + acceleration * gcv'(x)
         int first_deriv = 1;
         double part1, part2;
         part1 = velocity * velocity * splder((int*)&deriv, &half_order, &func->numpoints, &abscissa, func->x, func->c, &l, work);
         part2 = acceleration * splder(&first_deriv, &half_order, &func->numpoints, &abscissa, func->x, func->c, &l, work);
         return part1 + part2;
      }
   }

   /* The rest of this function is for type = dpNaturalCubicSpline. */

   n = func->numpoints;

   /* Check if the abscissa is out of range of the function. If it is,
    * then use the slope of the function at the appropriate end point to
    * extrapolate. You do this rather than printing an error because the
    * assumption is that this will only occur in relatively harmless
    * situations (like a motion file that contains an out-of-range gencoord
    * value). The rest of the SIMM code has many checks to clamp a gencoord
    * value within its range of motion, so if you make it to this function
    * and the gencoord is still out of range, deal with it quietly.
    */
   if (abscissa < func->x[0] - TINY_TIME)
   {
      if (deriv == zeroth)
         return func->y[0] + (abscissa - func->x[0])*func->b[0];
      if (deriv == first)
         return func->b[0]*velocity;
      if (deriv == second)
         return func->b[0]*acceleration;
   }
   else if (abscissa > func->x[n-1] + TINY_TIME)
   {
      if (deriv == zeroth)
         return func->y[n-1] + (abscissa - func->x[n-1])*func->b[n-1];
      if (deriv == first)
         return func->b[n-1]*velocity;
      if (deriv == second)
         return func->b[n-2]*acceleration;
   }

   /* Check to see if the abscissa is close to one of the end points
    * (the binary search method doesn't work well if you are at one of the
    * end points.
    */
   if (EQUAL_WITHIN_TOLERANCE(abscissa, func->x[0], TINY_TIME))
   {
      if (deriv == zeroth)
         return func->y[0];
      if (deriv == first)
         return func->b[0]*velocity;
      if (deriv == second)
         return func->b[0]*acceleration + 2.0*func->c[0]*velocity*velocity;
   }
   else if (EQUAL_WITHIN_TOLERANCE(abscissa, func->x[n-1], TINY_TIME))
   {
      if (deriv == zeroth)
         return func->y[n-1];
      if (deriv == first)
         return func->b[n-1]*velocity;
      if (deriv == second)
         return func->b[n-1]*acceleration + 2.0*func->c[n-1]*velocity*velocity;
   }

   if (n < 3)
   {
      /* If there are only 2 function points, then set k to zero
       * (you've already checked to see if the abscissa is out of
       * range or equal to one of the endpoints).
       */
      k = 0;
   }
   else
   {
      /* Do a binary search to find which two points the abscissa is between. */
      i = 0;
      j = n;
      while (1)
      {
         k = (i+j)/2;
         if (abscissa < func->x[k])
            j = k;
         else if (abscissa > func->x[k+1])
            i = k;
         else
            break;
      }
   }

   dx = abscissa - func->x[k];

   if (deriv == zeroth)
      return func->y[k] + dx*(func->b[k] + dx*(func->c[k] + dx*func->d[k]));

   if (deriv == first)
      return (func->b[k] + dx*(2.0*func->c[k] + 3.0*dx*func->d[k]))*velocity;

   if (deriv == second)
      return (func->b[k] + dx*(2.0*func->c[k] + 3.0*dx*func->d[k]))*acceleration +
	      (2.0*func->c[k] + 6.0*dx*func->d[k])*velocity*velocity;

   return ERROR_DOUBLE;
}

/* INTERSECT_LINE_PLANE01: calculates intersection between a line segment
 * (pt1,pt2) and plane (plane1,d). Also returns the intersection point(inter)
 * and where on the line it is(t). Ratio t must lie between 0 and 1 or no
 * intersection is found.
 */
dpBoolean intersect_line_plane01(double pt1[], double pt2[], 
                                double plane[], double d, double inter[], double* t)
{
   double dotprod;
   double vec[3];

   MAKE_3DVECTOR(pt1,pt2,vec);
   dotprod = DOT_VECTORS(vec,plane);

   if (DABS(dotprod) < LINE_EPSILON)
      return dpNo;

   *t = (-d - plane[0]*pt1[0] - plane[1]*pt1[1] - plane[2]*pt1[2])/dotprod;

   if ((*t < -LINE_EPSILON) || (*t > 1.0 + LINE_EPSILON))
      return dpNo;

   inter[0] = pt1[0] + ((*t) * vec[0]);
   inter[1] = pt1[1] + ((*t) * vec[1]);
   inter[2] = pt1[2] + ((*t) * vec[2]);

   return dpYes;
}

/* INTERSECT_LINE_PLANE: calculates intersection between a line (pt1,pt2)
 * and plane (plane1,d). Also returns the intersection point(inter) and
 * where on the line it is(t). Ratio t does not have to be between 0 and 1.
 */
dpBoolean intersect_line_plane(double pt1[], double pt2[], 
                              double plane[], double d, double inter[], double* t)
{
   double dotprod;
   double vec[3];

   MAKE_3DVECTOR(pt1,pt2,vec);
   dotprod = DOT_VECTORS(vec,plane);

   if (DABS(dotprod) < LINE_EPSILON)
      return dpNo;

   *t = (-d - plane[0]*pt1[0] - plane[1]*pt1[1] - plane[2]*pt1[2])/dotprod;

   inter[0] = pt1[0] + ((*t) * vec[0]);
   inter[1] = pt1[1] + ((*t) * vec[1]);
   inter[2] = pt1[2] + ((*t) * vec[2]);

   return dpYes;
}

/* INTERSECT_LINES: calculates the intersection of two lines. Returns where
 * on line1 the intersection is (s) and where on line 2 (t). Scales s and t
 * to be between 0 and 1.
 */
int intersect_lines(double p1[], double p2[], double p3[], double p4[],
                    double p_int1[], double* t, double p_int2[], double* s)
{
   double mag1, mag2, cross_prod[3], denom, vec1[3], vec2[3], mat[3][3];

   vec1[0] = p2[0] - p1[0];
   vec1[1] = p2[1] - p1[1];
   vec1[2] = p2[2] - p1[2];
   mag1 = normalize_vector(vec1,vec1);

   vec2[0] = p4[0] - p3[0];
   vec2[1] = p4[1] - p3[1];
   vec2[2] = p4[2] - p3[2];
   mag2 = normalize_vector(vec2,vec2);

   cross_vectors(vec1,vec2,cross_prod);

   denom = cross_prod[0]*cross_prod[0] + cross_prod[1]*cross_prod[1]
      + cross_prod[2]*cross_prod[2];

   if (EQUAL_WITHIN_ERROR(denom,0.0))
   {
      *s = *t = MINMDOUBLE;
      return (0);
   }

   mat[0][0] = p3[0] - p1[0];
   mat[0][1] = p3[1] - p1[1];
   mat[0][2] = p3[2] - p1[2];
   mat[1][0] = vec1[0];
   mat[1][1] = vec1[1];
   mat[1][2] = vec1[2];
   mat[2][0] = cross_prod[0];
   mat[2][1] = cross_prod[1];
   mat[2][2] = cross_prod[2];

   *s = CALC_DETERMINANT(mat) / denom;

   p_int2[0] = p3[0] + (*s) * (vec2[0]);
   p_int2[1] = p3[1] + (*s) * (vec2[1]);
   p_int2[2] = p3[2] + (*s) * (vec2[2]);

   mat[1][0] = vec2[0];
   mat[1][1] = vec2[1];
   mat[1][2] = vec2[2];

   *t = CALC_DETERMINANT(mat) / denom;

   p_int1[0] = p1[0] + (*t) * (vec1[0]);
   p_int1[1] = p1[1] + (*t) * (vec1[1]);
   p_int1[2] = p1[2] + (*t) * (vec1[2]);

   (*s) /= mag2;
   (*t) /= mag1;

   return (1);
}

/* INTERSECT_LINES_SCALED: same as intersect_lines(), but does not scale s and t
 * to be between 0.0 and 1.0. S ranges from 0.0 to mag2, and T ranges from 0.0 to mag1.
 */
int intersect_lines_scaled(double p1[], double p2[], double p3[], double p4[],
                           double p_int1[], double* t, double* mag1,
                           double p_int2[], double* s, double* mag2)
{
   double cross_prod[3], denom, vec1[3], vec2[3], mat[3][3];

   vec1[0] = p2[0] - p1[0];
   vec1[1] = p2[1] - p1[1];
   vec1[2] = p2[2] - p1[2];
   *mag1 = normalize_vector(vec1,vec1);

   vec2[0] = p4[0] - p3[0];
   vec2[1] = p4[1] - p3[1];
   vec2[2] = p4[2] - p3[2];
   *mag2 = normalize_vector(vec2,vec2);

   cross_vectors(vec1,vec2,cross_prod);

   denom = cross_prod[0]*cross_prod[0] + cross_prod[1]*cross_prod[1]
      + cross_prod[2]*cross_prod[2];

   if (EQUAL_WITHIN_ERROR(denom,0.0))
   {
      *s = *t = MINMDOUBLE;
      return (0);
   }

   mat[0][0] = p3[0] - p1[0];
   mat[0][1] = p3[1] - p1[1];
   mat[0][2] = p3[2] - p1[2];
   mat[1][0] = vec1[0];
   mat[1][1] = vec1[1];
   mat[1][2] = vec1[2];
   mat[2][0] = cross_prod[0];
   mat[2][1] = cross_prod[1];
   mat[2][2] = cross_prod[2];

   *s = CALC_DETERMINANT(mat) / denom;

   p_int2[0] = p3[0] + (*s) * (vec2[0]);
   p_int2[1] = p3[1] + (*s) * (vec2[1]);
   p_int2[2] = p3[2] + (*s) * (vec2[2]);

   mat[1][0] = vec2[0];
   mat[1][1] = vec2[1];
   mat[1][2] = vec2[2];

   *t = CALC_DETERMINANT(mat) / denom;

   p_int1[0] = p1[0] + (*t) * (vec1[0]);
   p_int1[1] = p1[1] + (*t) * (vec1[1]);
   p_int1[2] = p1[2] + (*t) * (vec1[2]);

   return (1);
}

/* MAKE_3X3DIRCOS_MATRIX: makes a 3x3 direction cosine matrix from an
 * angle and an axis.
 */
void make_3x3dircos_matrix(double angle, double axis[], double mat[][3])
{

   double rad_angle, cl, sl, omc;

   reset_3x3matrix(mat);

   rad_angle = angle*DTOR;

   cl = cos(rad_angle);
   sl = sin(rad_angle);
   omc = 1.0 - cl;

   /* the following matrix is taken from Kane's 'Spacecraft Dynamics,' pp 6-7 */

   mat[0][0] = cl + axis[0]*axis[0]*omc;
   mat[1][0] = -axis[2]*sl + axis[0]*axis[1]*omc;
   mat[2][0] = axis[1]*sl + axis[2]*axis[0]*omc;
   mat[0][1] = axis[2]*sl + axis[0]*axis[1]*omc;
   mat[1][1] = cl + axis[1]*axis[1]*omc;
   mat[2][1] = -axis[0]*sl + axis[1]*axis[2]*omc;
   mat[0][2] = -axis[1]*sl + axis[2]*axis[0]*omc;
   mat[1][2] = axis[0]*sl + axis[1]*axis[2]*omc;
   mat[2][2] = cl + axis[2]*axis[2]*omc;
}

/* MAKE_3X3MATRIX: makes a 3x3 matrix, given the 9 elements. */
void make_3x3matrix(double m[3][3], double m11, double m12, double m13, 
	                 double m21, double m22, double m23, double m31, double m32, double m33)
{
	m[0][0] = m11;
	m[0][1] = m12;
	m[0][2] = m13;
	m[1][0] = m21;
	m[1][1] = m22;
	m[1][2] = m23;
	m[2][0] = m31;
	m[2][1] = m32;
	m[2][2] = m33;
}

/* MAKE_3X3_XROT_MATRIX: makes a 3x3 direction cosine matrix with
 * a rotation about the X axis.
 */
void make_3x3_xrot_matrix(double a, double m[][3])
{
   m[0][0] = 1.0;
   m[0][1] = 0.0;
   m[0][2] = 0.0;

   m[1][0] = 0.0;
   m[1][1] = cos(a);
   m[1][2] = sin(a);
   
   m[2][0] = 0.0;
   m[2][1] = -sin(a);
   m[2][2] = cos(a);
}

/* MAKE_4X4DIRCOS_MATRIX: this routine creates a 4x4 direction cosine matrix
 * given an arbitrarily oriented axis and a rotation angle in degrees.
 */
void make_4x4dircos_matrix(double angle, double axis[], double mat[][4])
{
   double rad_angle, cl, sl, omc;

   MAKE_IDENTITY_MATRIX(mat);
   normalize_vector(axis,axis);

   rad_angle = angle*DTOR;

   cl = cos(rad_angle);
   sl = sin(rad_angle);
   omc = 1.0 - cl;

   /* the following matrix is taken from Kane's 'Spacecraft Dynamics,' pp 6-7 */

   mat[0][0] = cl + axis[XX]*axis[XX]*omc;
   mat[1][0] = -axis[ZZ]*sl + axis[XX]*axis[YY]*omc;
   mat[2][0] = axis[YY]*sl + axis[ZZ]*axis[XX]*omc;
   mat[0][1] = axis[ZZ]*sl + axis[XX]*axis[YY]*omc;
   mat[1][1] = cl + axis[YY]*axis[YY]*omc;
   mat[2][1] = -axis[XX]*sl + axis[YY]*axis[ZZ]*omc;
   mat[0][2] = -axis[YY]*sl + axis[ZZ]*axis[XX]*omc;
   mat[1][2] = axis[XX]*sl + axis[YY]*axis[ZZ]*omc;
   mat[2][2] = cl + axis[ZZ]*axis[ZZ]*omc;
}

/* MAKE_VECTOR: stores X,Y,Z values in an array. */
void make_vector(double vec[3], double x, double y, double z)
{
	vec[0] = x;
	vec[1] = y;
	vec[2] = z;
}

/* MULT_3X3MATRICES: multiplies two 3x3 matrices. */
void mult_3x3matrices(double mat1[][3], double mat2[][3], double result[][3])
{
   result[0][0] = mat1[0][0]*mat2[0][0] + mat1[0][1]*mat2[1][0] +
      mat1[0][2]*mat2[2][0];

   result[0][1] = mat1[0][0]*mat2[0][1] + mat1[0][1]*mat2[1][1] +
      mat1[0][2]*mat2[2][1];

   result[0][2] = mat1[0][0]*mat2[0][2] + mat1[0][1]*mat2[1][2] +
      mat1[0][2]*mat2[2][2];


   result[1][0] = mat1[1][0]*mat2[0][0] + mat1[1][1]*mat2[1][0] +
      mat1[1][2]*mat2[2][0];

   result[1][1] = mat1[1][0]*mat2[0][1] + mat1[1][1]*mat2[1][1] +
      mat1[1][2]*mat2[2][1];

   result[1][2] = mat1[1][0]*mat2[0][2] + mat1[1][1]*mat2[1][2] +
      mat1[1][2]*mat2[2][2];


   result[2][0] = mat1[2][0]*mat2[0][0] + mat1[2][1]*mat2[1][0] +
      mat1[2][2]*mat2[2][0];

   result[2][1] = mat1[2][0]*mat2[0][1] + mat1[2][1]*mat2[1][1] +
      mat1[2][2]*mat2[2][1];

   result[2][2] = mat1[2][0]*mat2[0][2] + mat1[2][1]*mat2[1][2] +
      mat1[2][2]*mat2[2][2];
}

/* MULT_3X3MATRIX_BY_VECTOR: premultiplies a 3x3 matrix by
 * a 1x3 vector. That is, it does vector*mat, not mat*vector.
 */
void mult_3x3matrix_by_vector(double mat[][3], double vector[], double result[])
{
   result[0] = vector[0]*mat[0][0] + vector[1]*mat[1][0] +
               vector[2]*mat[2][0];
   result[1] = vector[0]*mat[0][1] + vector[1]*mat[1][1] +
               vector[2]*mat[2][1];
   result[2] = vector[0]*mat[0][2] + vector[1]*mat[1][2] +
               vector[2]*mat[2][2];
}

/* MULT_4X4MATRICES: this routine multiplies the two specified 4x4 matrices
 * in the order: mat1*mat2.
 */
void mult_4x4matrices(double mat1[][4], double mat2[][4], double result[][4])
{
   result[0][0] = mat1[0][0]*mat2[0][0] + mat1[0][1]*mat2[1][0] +
      mat1[0][2]*mat2[2][0] + mat1[0][3]*mat2[3][0];

   result[0][1] = mat1[0][0]*mat2[0][1] + mat1[0][1]*mat2[1][1] +
      mat1[0][2]*mat2[2][1] + mat1[0][3]*mat2[3][1];

   result[0][2] = mat1[0][0]*mat2[0][2] + mat1[0][1]*mat2[1][2] +
      mat1[0][2]*mat2[2][2] + mat1[0][3]*mat2[3][2];

   result[0][3] = mat1[0][0]*mat2[0][3] + mat1[0][1]*mat2[1][3] +
      mat1[0][2]*mat2[2][3] + mat1[0][3]*mat2[3][3];


   result[1][0] = mat1[1][0]*mat2[0][0] + mat1[1][1]*mat2[1][0] +
      mat1[1][2]*mat2[2][0] + mat1[1][3]*mat2[3][0];

   result[1][1] = mat1[1][0]*mat2[0][1] + mat1[1][1]*mat2[1][1] +
      mat1[1][2]*mat2[2][1] + mat1[1][3]*mat2[3][1];

   result[1][2] = mat1[1][0]*mat2[0][2] + mat1[1][1]*mat2[1][2] +
      mat1[1][2]*mat2[2][2] + mat1[1][3]*mat2[3][2];

   result[1][3] = mat1[1][0]*mat2[0][3] + mat1[1][1]*mat2[1][3] +
      mat1[1][2]*mat2[2][3] + mat1[1][3]*mat2[3][3];


   result[2][0] = mat1[2][0]*mat2[0][0] + mat1[2][1]*mat2[1][0] +
      mat1[2][2]*mat2[2][0] + mat1[2][3]*mat2[3][0];

   result[2][1] = mat1[2][0]*mat2[0][1] + mat1[2][1]*mat2[1][1] +
      mat1[2][2]*mat2[2][1] + mat1[2][3]*mat2[3][1];

   result[2][2] = mat1[2][0]*mat2[0][2] + mat1[2][1]*mat2[1][2] +
      mat1[2][2]*mat2[2][2] + mat1[2][3]*mat2[3][2];

   result[2][3] = mat1[2][0]*mat2[0][3] + mat1[2][1]*mat2[1][3] +
      mat1[2][2]*mat2[2][3] + mat1[2][3]*mat2[3][3];


   result[3][0] = mat1[3][0]*mat2[0][0] + mat1[3][1]*mat2[1][0] +
      mat1[3][2]*mat2[2][0] + mat1[3][3]*mat2[3][0];

   result[3][1] = mat1[3][0]*mat2[0][1] + mat1[3][1]*mat2[1][1] +
      mat1[3][2]*mat2[2][1] + mat1[3][3]*mat2[3][1];

   result[3][2] = mat1[3][0]*mat2[0][2] + mat1[3][1]*mat2[1][2] +
      mat1[3][2]*mat2[2][2] + mat1[3][3]*mat2[3][2];

   result[3][3] = mat1[3][0]*mat2[0][3] + mat1[3][1]*mat2[1][3] +
      mat1[3][2]*mat2[2][3] + mat1[3][3]*mat2[3][3];
}

/* MULT_4X4MATRIX_BY_VECTOR: this routine premultiplies a 4x4 matrix by
 * a 1x4 vector. That is, it does vector*mat, not mat*vector.
 */
void mult_4x4matrix_by_vector(double mat[][4], double vector[], double result[])
{
   result[0] = vector[0]*mat[0][0] + vector[1]*mat[1][0] +
               vector[2]*mat[2][0] + vector[3]*mat[3][0];
   result[1] = vector[0]*mat[0][1] + vector[1]*mat[1][1] +
               vector[2]*mat[2][1] + vector[3]*mat[3][1];
   result[2] = vector[0]*mat[0][2] + vector[1]*mat[1][2] +
               vector[2]*mat[2][2] + vector[3]*mat[3][2];
   result[3] = vector[0]*mat[0][3] + vector[1]*mat[1][3] +
               vector[2]*mat[2][3] + vector[3]*mat[3][3];
}

/* NORMALIZE_VECTOR: normalizes a 1x3 vector. */
double normalize_vector(double vector[], double norm_vector[])
{
   double magnitude;

   magnitude = VECTOR_MAGNITUDE(vector);

   if (magnitude < TINY_NUMBER)
   {
      norm_vector[0] = vector[0];
      norm_vector[1] = vector[1];
      norm_vector[2] = vector[2];
   }
   else
   {
      norm_vector[0] = vector[0]/magnitude;
      norm_vector[1] = vector[1]/magnitude;
      norm_vector[2] = vector[2]/magnitude;
   }

   return (magnitude);
}

/* POINT_IN_POLYGON2D: determines if a 2D point is inside a 2D polygon. Uses the
 * "crossings test" taken from:
 * Haines, Eric, "Point in Polygon Strategies," Graphics Gems IV, ed. Paul Heckbert,
 * Academic Press, p. 24-46, 1994.
 * Shoot a test ray along +X axis. The strategy, from MacMartin, is to
 * compare vertex Y values to the testing point's Y and quickly discard
 * edges which are entirely to one side of the test ray.
 *
 * Input 2D polygon _pgon_ with _numverts_ number of vertices and test point
 * _point_, returns 1 if inside, 0 if outside.
 */
int point_in_polygon2D(double point[], double pgon[][2], int numverts)
{
   register int j, yflag0, yflag1, inside_flag, xflag0 ;
   register double ty, tx, *vtx0, *vtx1 ;

   tx = point[XX];
   ty = point[YY];

   vtx0 = pgon[numverts-1];

   /* get test bit for above/below X axis */
   yflag0 = ( vtx0[YY] >= ty );
   vtx1 = pgon[0];

   inside_flag = 0;

   for (j = numverts+1; --j ;)
   {
      yflag1 = ( vtx1[YY] >= ty );
	   /* check if endpoints straddle (are on opposite sides) of X axis
	    * (i.e. the Y's differ); if so, +X ray could intersect this edge.
	    */
      if (yflag0 != yflag1)
      {
         xflag0 = (vtx0[XX] >= tx);
	     /* check if endpoints are on same side of the Y axis (i.e. X's
	      * are the same); if so, it's easy to test if edge hits or misses.
	      */
         if ( xflag0 == ( vtx1[XX] >= tx ) )
         {
      		/* if edge's X values both right of the point, must hit */
            if ( xflag0 )
               inside_flag = !inside_flag;
         }
         else
         {
	 	      /* compute intersection of pgon segment with +X ray, note
		       * if >= point's X; if so, the ray hits it.
		       */
            if ((vtx1[XX] - (vtx1[YY]-ty) * (vtx0[XX]-vtx1[XX])/(vtx0[YY]-vtx1[YY])) >= tx)
            {
               inside_flag = !inside_flag;
            }
         }
      }

      /* move to next pair of vertices, retaining info as possible */
      yflag0 = yflag1;
      vtx0 = vtx1;
      vtx1 += 2;
   }

   return inside_flag;
}


/* POINT_IN_POLYGON3D: determines if a 3D point is inside a 3D polygon. The polygon
 * is the first one in the PolyhedronStruct, and is assumed to be planar.
 */
dpBoolean point_in_polygon3D(double pt[], dpPolyhedronStruct* ph, int polygon_index)
{
   int i, j, index, skip_coord;
   double max_value, polygon[50][2], pt2D[2];

   // if the polygon has more than 50 vertices, abort.
   if (ph->polygon[polygon_index].num_vertices > 50)
      return dpNo;

   // Project the polygon onto one of the major planes by ignoring one of the
   // coordinates. Ignore the one most parallel to the plane's normal so that
   // the area of the projected polygon will be as large as possible.
   skip_coord = XX;
   max_value = ph->polygon[polygon_index].normal[XX];
   if (DABS(ph->polygon[polygon_index].normal[YY]) > max_value)
   {
      skip_coord = YY;
      max_value = ph->polygon[polygon_index].normal[YY];
   }
   if (DABS(ph->polygon[polygon_index].normal[ZZ]) > max_value)
      skip_coord = ZZ;

   // Copy the vertices into a 2D array, skipping the skip_coord
   for (i = 0; i < ph->polygon[polygon_index].num_vertices; i++)
   {
      for (j = 0, index = 0; j < 3; j++)
      {
         if (j != skip_coord)
            polygon[i][index++] = ph->vertex[ph->polygon[polygon_index].vertex_index[i]].coord[j];
      }
   }

   // Copy the 3D point into a 2D point, skipping the skip_coord
   for (i = 0, index = 0; i < 3; i++)
   {
      if (i != skip_coord)
         pt2D[index++] = pt[i];
   }

   if (point_in_polygon2D(pt2D, polygon, ph->polygon[polygon_index].num_vertices))
      return dpYes;

   return dpNo;
}


/* PRINT_3X3MATRIX: prints a 3x3 matrix to a file. */
void print_3x3matrix(double mat[][3], char string[], FILE* stream)
{
   if (string != NULL)
      fprintf(stream,"matrix %s:\n", string);
   else
      fprintf(stream,"matrix:\n");

   fprintf(stream,"%lf %lf %lf\n", mat[0][0], mat[0][1], mat[0][2]);
   fprintf(stream,"%lf %lf %lf\n", mat[1][0], mat[1][1], mat[1][2]);
   fprintf(stream,"%lf %lf %lf\n", mat[2][0], mat[2][1], mat[2][2]);
}


/* PROJECT_POINT_ONTO_PLANE: projects a 3D point onto the surface of a 3D plane and returns the distance to the plane.
 * Negative distance - point is below plane, positive distance - point is above plane. 
 */
double project_point_onto_plane(double pt[], dpPlaneStruct* plane, double projpt[])
{
   double plane_normal[3];
   double distance_to_plane;

   plane_normal[XX] = plane->a;
   plane_normal[YY] = plane->b;
   plane_normal[ZZ] = plane->c;

   normalize_vector(plane_normal,plane_normal);

   // distance_to_plane is the distance from pt to the plane along the
   // plane's normal vector. Thus it is negative when pt is below
   // the plane, and positive when pt is above the plane.
   distance_to_plane = DOT_VECTORS(pt, plane_normal) + plane->d;
 
   projpt[XX] = pt[XX] - distance_to_plane*plane_normal[XX];
   projpt[YY] = pt[YY] - distance_to_plane*plane_normal[YY];
   projpt[ZZ] = pt[ZZ] - distance_to_plane*plane_normal[ZZ];

   return distance_to_plane;
}


/* RESET_3X3MATRIX: sets a 3x3 matrix to the identity matrix. */
void reset_3x3matrix(double matrix[][3])
{
   matrix[0][1] = matrix[0][2] = matrix[1][2] = 0.0;
   matrix[1][0] = matrix[2][0] = matrix[2][1] = 0.0;
   matrix[0][0] = matrix[1][1] = matrix[2][2] = 1.0;
}


/* RESET_4X4MATRIX: sets a 4x4 matrix to the identity matrix. */
void reset_4x4matrix(double matrix[][4])
{
   matrix[0][1] = matrix[0][2] = matrix[1][2] = 0.0;
   matrix[1][0] = matrix[2][0] = matrix[2][1] = 0.0;
   matrix[0][0] = matrix[1][1] = matrix[2][2] = 1.0;
   matrix[3][0] = matrix[3][1] = matrix[3][2] = 0.0;
   matrix[0][3] = matrix[1][3] = matrix[2][3] = matrix[3][3] = 0.0;
}


/* ROTATE_MATRIX_AXIS_ANGLE: appends an angle/axis rotation onto a 4x4 transform matrix. */
void rotate_matrix_axis_angle (double m[][4], const double* axis, double angle)
{
    Quat q;

    make_quaternion(q, axis, angle);
    rotate_matrix_by_quat(m, q);
}


/* TRANSFORM_PT: multiplies a point by a transform matrix. */
void transform_pt(double m[][4], double* pt)
{
   /* apply a matrix transformation to a 3d point */
   
   double tx = pt[XX] * m[0][0] + pt[YY] * m[1][0] + pt[ZZ] * m[2][0] + m[3][0];
   double ty = pt[XX] * m[0][1] + pt[YY] * m[1][1] + pt[ZZ] * m[2][1] + m[3][1];

   pt[ZZ] = pt[XX] * m[0][2] + pt[YY] * m[1][2] + pt[ZZ] * m[2][2] + m[3][2];
   pt[XX] = tx;
   pt[YY] = ty;
}


/* TRANSFORM_VEC: multiplies a vector by a transform matrix. */
void transform_vec(double m[][4], double* vec)
{
   /* apply a matrix transformation to a 3d vector */
   
   double tx = vec[XX] * m[0][0] + vec[YY] * m[1][0] + vec[ZZ] * m[2][0];
   double ty = vec[XX] * m[0][1] + vec[YY] * m[1][1] + vec[ZZ] * m[2][1];

   vec[ZZ] = vec[XX] * m[0][2] + vec[YY] * m[1][2] + vec[ZZ] * m[2][2];
   vec[XX] = tx;
   vec[YY] = ty;
}


/* TRANSLATE_MATRIX: appends a translation onto a 4x4 matrix. */
void translate_matrix(double m[][4], double* delta)
{
   /* append translation 'delta' to matrix 'm' */
   
   m[0][0] += m[0][3] * delta[0];
   m[0][1] += m[0][3] * delta[1];   
   m[0][2] += m[0][3] * delta[2];
   m[1][0] += m[1][3] * delta[0];   
   m[1][1] += m[1][3] * delta[1];   
   m[1][2] += m[1][3] * delta[2];
   m[2][0] += m[2][3] * delta[0];   
   m[2][1] += m[2][3] * delta[1];   
   m[2][2] += m[2][3] * delta[2];
   m[3][0] += m[3][3] * delta[0];   
   m[3][1] += m[3][3] * delta[1];   
   m[3][2] += m[3][3] * delta[2];
}


/* TRANSPOSE_3X3MATRIX: transposes a 3x3 matrix. */
void transpose_3x3matrix(double mat[][3], double mat_transpose[][3])
{
   int i, j;

   for (i=0; i<3; i++)
      for (j=0; j<3; j++)
         mat_transpose[j][i] = mat[i][j];
}


/* TRANSPOSE_4X4MATRIX: transposes a 4x4 matrix. */
void transpose_4x4matrix(double mat[][4], double mat_transpose[][4])
{
   int i, j;

   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
         mat_transpose[j][i] = mat[i][j];
}


/* VECT_EQUAL_WITHIN_TOL: checks if two vectors are equal. */
dpBoolean vect_equal_within_tol(double a[3], double b[3], double tol)
{
	if (fabs(a[0] - b[0]) > tol)
		return dpNo;
	if (fabs(a[1] - b[1]) > tol)
		return dpNo;
	if (fabs(a[2] - b[2]) > tol)
		return dpNo;
	return dpYes;
}


/* ZERO_3X3MATRIX: checks if any row of a 3x3 matrix contains all zeros. */
dpBoolean zero_3x3matrix(double m[][3])
{
   int i, j;
   double big, temp;

	for (i = 0; i < 3; i++)
   {
		big = 0.0;
		for (j = 0; j < 3; j++)
      {
			if ((temp = fabs(m[i][j])) > big)
				big = temp;
      }
		if (EQUAL_WITHIN_ERROR(big, 0.0))
			return dpYes;
		return dpNo;
	}
	return dpNo;
}


/* append a rotation (specified in RADIANS) about local x-axis to matrix 'm' */
void x_rotate_matrix_bodyfixed(double m[][4], double angle)
{
   // append rotation about local x-axis to matrix 'm'
   Quat q;
   
   make_quaternion(q, m[0], angle);
   rotate_matrix_by_quat(m, q);
}


/* append a rotation (specified in RADIANS) about local y-axis to matrix 'm' */
void y_rotate_matrix_bodyfixed(double m[][4], double angle)
{
   Quat q;
   
   make_quaternion(q, m[1], angle);
   rotate_matrix_by_quat(m, q);
}


/* append a rotation (specified in RADIANS) about local z-axis to matrix 'm' */
void z_rotate_matrix_bodyfixed(double m[][4], double angle)
{
   Quat q;
   
   make_quaternion(q, m[2], angle);
   rotate_matrix_by_quat(m, q);
}


void x_rotate_matrix_spacefixed(double m[][4], double radians)
{
   // append rotation about global x-axis to matrix 'm'
   double sinTheta = sin(radians);
   double cosTheta = cos(radians);

   double t = m[0][1];
   m[0][1] = t * cosTheta - m[0][2] * sinTheta;
   m[0][2] = t * sinTheta + m[0][2] * cosTheta;

   t = m[1][1];
   m[1][1] = t * cosTheta - m[1][2] * sinTheta;
   m[1][2] = t * sinTheta + m[1][2] * cosTheta;

   t = m[2][1];
   m[2][1] = t * cosTheta - m[2][2] * sinTheta;
   m[2][2] = t * sinTheta + m[2][2] * cosTheta;

   t = m[3][1];
   m[3][1] = t * cosTheta - m[3][2] * sinTheta;
   m[3][2] = t * sinTheta + m[3][2] * cosTheta;
}


void y_rotate_matrix_spacefixed(double m[][4], double radians)
{
   // append rotation about global y-axis to matrix 'm'
   double sinTheta = sin(radians);
   double cosTheta = cos(radians);

   double t = m[0][0];
   m[0][0] = t * cosTheta + m[0][2] * sinTheta;
   m[0][2] = m[0][2] * cosTheta - t * sinTheta;

   t = m[1][0];
   m[1][0] = t * cosTheta + m[1][2] * sinTheta;
   m[1][2] = m[1][2] * cosTheta - t * sinTheta;

   t = m[2][0];
   m[2][0] = t * cosTheta + m[2][2] * sinTheta;
   m[2][2] = m[2][2] * cosTheta - t * sinTheta;

   t = m[3][0];
   m[3][0] = t * cosTheta + m[3][2] * sinTheta;
   m[3][2] = m[3][2] * cosTheta - t * sinTheta;
}


void z_rotate_matrix_spacefixed(double m[][4], double radians)
{
   // append rotation about global z-axis to matrix 'm'
   double sinTheta = sin(radians);
   double cosTheta = cos(radians);

   double t = m[0][0];
   m[0][0] = t * cosTheta - m[0][1] * sinTheta;
   m[0][1] = t * sinTheta + m[0][1] * cosTheta;

   t = m[1][0];
   m[1][0] = t * cosTheta - m[1][1] * sinTheta;
   m[1][1] = t * sinTheta + m[1][1] * cosTheta;

   t = m[2][0];
   m[2][0] = t * cosTheta - m[2][1] * sinTheta;
   m[2][1] = t * sinTheta + m[2][1] * cosTheta;

   t = m[3][0];
   m[3][0] = t * cosTheta - m[3][1] * sinTheta;
   m[3][1] = t * sinTheta + m[3][1] * cosTheta;
}


/* CALC_NATURAL_CUBIC_COEFFICIENTS: this function computes the coefficients
 * of a natural cubic spline defined by the XY control points. The code
 * is translated from a Fortran version printed in "Computer Methods
 * for Mathematical Computations" by Forsythe, Malcolm, and Moler,
 * pp 77-8. To better handle splines which have two or more control points
 * with the same X values, checks were added to make sure that the code never
 * divides by zero.
 */
static void calc_natural_cubic_coefficients(dpFunction* f)
{
   int nm1, nm2, i, j;
   double t;

   if (f->numpoints < 2 || f->used == dpNo || f->type != dpNaturalCubicSpline)
      return;

   if (f->numpoints == 2)
   {
      t = MAX(TINY_NUMBER,f->x[1]-f->x[0]);
      f->b[0] = f->b[1] = (f->y[1]-f->y[0])/t;
      f->c[0] = f->c[1] = 0.0;
      f->d[0] = f->d[1] = 0.0;
      return;
   }

   nm1 = f->numpoints - 1;
   nm2 = f->numpoints - 2;

   // Set up tridiagonal system:
   // b = diagonal, d = offdiagonal, c = right-hand side
   f->d[0] = MAX(TINY_NUMBER,f->x[1] - f->x[0]);
   f->c[1] = (f->y[1]-f->y[0])/f->d[0];
   for (i=1; i<nm1; i++)
   {
      f->d[i] = MAX(TINY_NUMBER,f->x[i+1] - f->x[i]);
      f->b[i] = 2.0*(f->d[i-1]+f->d[i]);
      f->c[i+1] = (f->y[i+1]-f->y[i])/f->d[i];
      f->c[i] = f->c[i+1] - f->c[i];
   }

   // End conditions. Third derivatives at x[0] and x[n-1]
   // are obtained from divided differences.
   f->b[0] = -f->d[0];
   f->b[nm1] = -f->d[nm2];
   f->c[0] = 0.0;
   f->c[nm1] = 0.0;

   if (f->numpoints > 3)
   {
      double d1, d2, d3, d20, d30, d31;

      d31 = MAX(TINY_NUMBER,f->x[3] - f->x[1]);
      d20 = MAX(TINY_NUMBER,f->x[2] - f->x[0]);
      d1 = MAX(TINY_NUMBER,f->x[nm1]-f->x[f->numpoints-3]);
      d2 = MAX(TINY_NUMBER,f->x[nm2]-f->x[f->numpoints-4]);
      d30 = MAX(TINY_NUMBER,f->x[3] - f->x[0]);
      d3 = MAX(TINY_NUMBER,f->x[nm1]-f->x[f->numpoints-4]);
      f->c[0] = f->c[2]/d31 - f->c[1]/d20;
      f->c[nm1] = f->c[nm2]/d1 - f->c[f->numpoints-3]/d2;
      f->c[0] = f->c[0]*f->d[0]*f->d[0]/d30;
      f->c[nm1] = -f->c[nm1]*f->d[nm2]*f->d[nm2]/d3;
   }

   // Forward elimination
   for (i=1; i<f->numpoints; i++)
   {
      t = f->d[i-1]/f->b[i-1];
      f->b[i] -= t*f->d[i-1];
      f->c[i] -= t*f->c[i-1];
   }

   // Back substitution
   f->c[nm1] /= f->b[nm1];
   for (j=0; j<nm1; j++)
   {
      i = nm2 - j;
      f->c[i] = (f->c[i]-f->d[i]*f->c[i+1])/f->b[i];
   }

   // compute polynomial coefficients
   f->b[nm1] = (f->y[nm1]-f->y[nm2])/f->d[nm2] +
               f->d[nm2]*(f->c[nm2]+2.0*f->c[nm1]);
   for (i=0; i<nm1; i++)
   {
      f->b[i] = (f->y[i+1]-f->y[i])/f->d[i] - f->d[i]*(f->c[i+1]+2.0*f->c[i]);
      f->d[i] = (f->c[i+1]-f->c[i])/f->d[i];
      f->c[i] *= 3.0;
   }
   f->c[nm1] *= 3.0;
   f->d[nm1] = f->d[nm2];
}

/* MAKE_QUATERNION: makes a quaternion given an angle and axis. */
static void make_quaternion(Quat q, const double axis[3], double angle)
{
   // make a quaternion given an axis-angle rotation
   // (from java-based vecmath package)
   double n, halfAngle;

   q[XX] = axis[XX];
   q[YY] = axis[YY];
   q[ZZ] = axis[ZZ];

   n = sqrt(SQR(q[XX]) + SQR(q[YY]) + SQR(q[ZZ]));

   halfAngle = 0.5 * angle;

   if (NOT_EQUAL_WITHIN_ERROR(n,0.0))
   {
      double s = sin(halfAngle) / n;

      q[XX] *= s;  q[YY] *= s;  q[ZZ] *= s;

      q[WW] = cos(halfAngle);
   }
}


/* QUAT_TO_MATRIX: converts a quaternion into an equivalent
 * 4x4 transform matrix.
 */
static void quat_to_matrix (const Quat q, double m[][4])
{
   double Nq = SQR(q[XX]) + SQR(q[YY]) + SQR(q[ZZ]) + SQR(q[WW]);
   double s = (Nq > 0.0) ? (2.0 / Nq) : 0.0;

   double xs = q[XX] * s,   ys = q[YY] * s,   zs = q[ZZ] * s;
   double wx = q[WW] * xs,  wy = q[WW] * ys,  wz = q[WW] * zs;
   double xx = q[XX] * xs,  xy = q[XX] * ys,  xz = q[XX] * zs;
   double yy = q[YY] * ys,  yz = q[YY] * zs,  zz = q[ZZ] * zs;

   m[XX][XX] = 1.0 - (yy + zz);  m[XX][YY] = xy + wz;          m[XX][ZZ] = xz - wy;
   m[YY][XX] = xy - wz;          m[YY][YY] = 1.0 - (xx + zz);  m[YY][ZZ] = yz + wx;
   m[ZZ][XX] = xz + wy;          m[ZZ][YY] = yz - wx;          m[ZZ][ZZ] = 1.0 - (xx + yy);

   m[XX][WW] = m[YY][WW] = m[ZZ][WW] = m[WW][XX] = m[WW][YY] = m[WW][ZZ] = 0.0;
   m[WW][WW] = 1.0;
}


/* ROTATE_MATRIX_BY_QUAT: appends a quaternion rotation onto a 4x4 transform matrix. */
static void rotate_matrix_by_quat(double m[][4], const Quat q)
{
   double n[4][4];

   quat_to_matrix(q, n);

   append_matrix(m, n);    /* append matrix 'n' onto matrix 'm' */
}


/* EXTRACT_XYZ_ROT_SPACEFIXED: extracts XYZ space-fixed rotations
 * from a 4x4 transform matrix.
 */
void extract_xyz_rot_spacefixed(double m[][4], double xyz_rot[3])
{
   xyz_rot[YY] = asin(-m[0][2]);

   if (NOT_EQUAL_WITHIN_ERROR(0.0,cos(xyz_rot[YY])))
   {
      xyz_rot[XX] = atan2(m[1][2], m[2][2]);
      xyz_rot[ZZ] = atan2(m[0][1], m[0][0]);
   }
   else
   {
      xyz_rot[XX] = atan2(m[1][0], m[1][1]);
      xyz_rot[ZZ] = 0.0;
   } 
}


/* SET_ROTATION_MATRIX_FROM_QUATERNION: taken from SIMTK code rotation.cpp, function setRotationFromQuaternion
 * NOTE: the quaternion representation is NOT the same as the SD/FAST representation
 * here quaternion is  q0 = cos(theta/2) q1 = lambda1*sin(theta/2) q2 = lambda2*sin(theta/2) q3 = lambda3*sin(theta/2)
 * SDFAST quaternion is q0 = lambda1*sin(theta/2) q1 = lambda2*sin(theta/2) q2 = lambda3*sin(theta/2) q4 = cos(theta/2) 
 */
void set_rotation_matrix_from_quaternion(double q[], double R[][4])
{
   double q00 = q[0]*q[0], q11 = q[1]*q[1], q22 = q[2]*q[2], q33 = q[3]*q[3];
   double q01 = q[0]*q[1], q02 = q[0]*q[2], q03 = q[0]*q[3];
   double q12 = q[1]*q[2], q13 = q[1]*q[3], q23 = q[2]*q[3];

   R[0][0] = q00+q11-q22-q33;
   R[0][1] = 2*(q12-q03);
   R[0][2] = 2*(q13+q02);
   R[0][3] = 0.0;
   R[1][0] = 2*(q12+q03);
   R[1][1] = q00-q11+q22-q33;
   R[1][2] = 2*(q23-q01);
   R[1][3] = 0.0;
   R[2][0] = 2*(q13-q02);
   R[2][1] = 2*(q23+q01);
   R[2][2] = q00-q11-q22+q33;
   R[2][3] = 0.0;
   R[3][0] = 0.0;
   R[3][1] = 0.0;
   R[3][2] = 0.0;
   R[3][3] = 1.0;
}


/* EXTRACT_EULER_ANGLES: taken from function :convertThreeAxesBodyFixedRotationToThreeAngles
 * from rotation.cpp for SIMTK
 * extracts the rotations in the given order from a rotation matrix
 */
void extract_euler_angles(double R[][4], double xyz_rot[3], dpEulerAngleOrder order)
{
   int i, j, k;
   double plusMinus, minusPlus, Rsum, Eps = 1e-5, theta1, theta2, theta3, spos, cpos, sneg, cneg;
   double theta1PlusMinusTheta3, theta1MinusPlusTheta3;

   //if axes are ReverseCyclical
   if (order == dp132 || order == dp213 || order == dp321)
   {
      plusMinus = -1.0;
      minusPlus = 1.0;
   }
   else
   {
      plusMinus = 1.0;
      minusPlus = -1.0;
   }

   if (order == dp123)
   {
      i = 0; j = 1; k = 2;
   }
   else if (order == dp132)
   {
      i = 0; j = 2; k = 1;
   }
   else if (order == dp213)
   {
      i = 1; j = 0; k = 2;
   }
   else if (order == dp231)
   {
      i = 1; j = 2; k = 0;
   }
   else if (order == dp312)
   {
      i = 2; j = 0; k = 1;
   }
   else if (order == dp321)
   {
      i = 2; j = 1; k = 0;
   }
   else
   {
      xyz_rot[0] = xyz_rot[1] = xyz_rot[2] = 0.0;
      return;
   }

   // Calculate theta2 using lots of information in the rotation matrix
   Rsum   =  sqrt((SQR(R[i][i]) + SQR(R[i][j]) + SQR(R[j][k]) + SQR(R[k][k])) / 2);
   theta2 =  atan2( plusMinus*R[i][k], Rsum ); // Rsum = abs(cos(theta2)) is inherently positive

   // There is a "singularity" when cos(theta2) == 0
   if (Rsum > 4*Eps)
   {
      theta1 =  atan2(minusPlus*R[j][k], R[k][k]);
      theta3 =  atan2(minusPlus*R[i][j], R[i][i]);
   }
   else if (plusMinus*R[i][k] > 0)
   {
      spos = R[j][i] + plusMinus*R[k][j];  // 2*sin(theta1 + plusMinus*theta3)
      cpos = R[j][j] + minusPlus*R[k][i];  // 2*cos(theta1 + plusMinus*theta3)
      theta1PlusMinusTheta3 = atan2(spos, cpos);
      theta1 = theta1PlusMinusTheta3;  // Arbitrary split
      theta3 = 0;                      // Arbitrary split
   }
   else
   {
      sneg = plusMinus * (R[k][j] + minusPlus*R[j][i]);  // 2*sin(theta1 + minusPlus*theta3)
      cneg = R[j][j] + plusMinus * R[k][i];              // 2*cos(theta1 + minusPlus*theta3)
      theta1MinusPlusTheta3 = atan2(sneg, cneg);
      theta1 = theta1MinusPlusTheta3;  // Arbitrary split
      theta3 = 0;                      // Arbitrary split
   }

   // Return values have the following ranges:
   // -pi   <=  theta1  <=  +pi
   // -pi/2 <=  theta2  <=  +pi/2   (Rsum is inherently positive)
   // -pi   <=  theta3  <=  +pi
   xyz_rot[0] = theta1;
   xyz_rot[1] = theta2;
   xyz_rot[2] = theta3;
}


/* setThreeAngleThreeAxesBodyFixedForwardCyclicalRotation: taken from rotation.cpp for SIMTK
 * sets the rotation matrix given precalculated cos and sin values as well as rotation axes
 */
static void setThreeAngleThreeAxesBodyFixedForwardCyclicalRotation
  ( double cosAngle1, double sinAngle1, int axis1, 
    double cosAngle2, double sinAngle2, int axis2,
    double cosAngle3, double sinAngle3, int axis3,
    double R[][3]) 
{
   int i = axis1 - 1;
   int j = axis2 - 1;
   int k = axis3 - 1;

   // Repeated calculations (for efficiency)
   double s1c3 = sinAngle1 * cosAngle3;
   double s3c1 = sinAngle3 * cosAngle1;
   double s1s3 = sinAngle1 * sinAngle3;
   double c1c3 = cosAngle1 * cosAngle3;

   reset_3x3matrix(R);

   // check that indices are valid
   if ((i < 0 || i > 2) || (j < 0 || j > 2) || (k < 0 || k > 2))
   {
      printf("setThreeAngleThreeAxesBodyFixedForwardCyclicalRotation - error\n");
      return;
   }
   //set rotation matrix
   R[i][i] =  cosAngle2 * cosAngle3;
   R[i][j] = -sinAngle3 * cosAngle2;
   R[i][k] =  sinAngle2;
   R[j][i] =  s3c1 + sinAngle2 * s1c3;
   R[j][j] =  c1c3 - sinAngle2 * s1s3;
   R[j][k] = -sinAngle1 * cosAngle2;
   R[k][i] =  s1s3 - sinAngle2 * c1c3;
   R[k][j] =  s1c3 + sinAngle2 * s3c1;
   R[k][k] =  cosAngle1 * cosAngle2;
}


/* SET_ROTATION_MATRIX_FROM_EULER_ANGLES :
 * taken from rotation.cpp for SIMTK, function setRotationFromThreeAnglesThreeAxes
 * Given the angles, axes and rotation order, create a rotation matrix
 */
void set_rotation_matrix_from_euler_angles(double angle1, int axis1, double angle2, int axis2, 
                                         double angle3, int axis3, double R[][3], dpEulerAngleOrder order)
{
   double c1, c2, c3, s1, s2, s3;

   // If axes are reverse cyclical, negate the signs of the angles.
   if (order == dp132 || order == dp213 || order == dp321)
   {
      angle1 = -angle1;
      angle2 = -angle2;
      angle3 = -angle3;
   }

   // Calculate the sines and cosines (some hardware can do this more efficiently as one Taylor series).
   c1 = cos( angle1 );
   s1 = sin( angle1 );
   c2 = cos( angle2 );
   s2 = sin( angle2 );
   c3 = cos( angle3 );
   s3 = sin( angle3 );

   // All calculations are based on a body-fixed rotation sequence.
   setThreeAngleThreeAxesBodyFixedForwardCyclicalRotation(c1,s1,axis1, c2,s2,axis2, c3,s3,axis3, R);
}


/* CONVERT_ROTATION_TO_QUATERNION: taken from rotation.cpp for SIMTK, function convertRotationToQuaternion
 * convert the rotation matrix to quaternion format
 * NOTE: the quaternion representation is NOT the same as the SD/FAST representation
 * here quaternion is  q0 = cos(theta/2) q1 = lambda1*sin(theta/2) q2 = lambda2*sin(theta/2) q3 = lambda3*sin(theta/2)
 * SDFAST quaternion is q0 = lambda1*sin(theta/2) q1 = lambda2*sin(theta/2) q2 = lambda3*sin(theta/2) q4 = cos(theta/2) 
 */
void convert_rotation_to_quaternion(double R[][3], double q[])
{
   double scale, magnitude;
   double trace = R[0][0] + R[1][1] + R[2][2];

   if (trace >= R[0][0] && trace >= R[1][1] && trace >= R[2][2]) // trace is larger than any diagonal
   {
      q[0] = 1 + trace;
      q[1] = R[2][1] - R[1][2];
      q[2] = R[0][2] - R[2][0];
      q[3] = R[1][0] - R[0][1];
   }
   else if (R[0][0] >= R[1][1] && R[0][0] >= R[2][2]) // R(0][0] is largest along the diagonal
   {
      q[0] = R[2][1] - R[1][2];
      q[1] = 1 - (trace - 2.0 * R[0][0]);
      q[2] = R[0][1] + R[1][0];
      q[3] = R[0][2] + R[2][0];
   }
   else if(R[1][1] >= R[2][2]) // R[1][1] is largest along the diagonal
   {
      q[0] = R[0][2] - R[2][0];
      q[1] = R[0][1] + R[1][0];
      q[2] = 1 - (trace - 2.0 * R[1][1]);
      q[3] = R[1][2] + R[2][1];
   }
   else // R[2][2] is largest along the diagonal
   {
      q[0] = R[1][0] - R[0][1];
      q[1] = R[0][2] + R[2][0];
      q[2] = R[1][2] + R[2][1];
      q[3] = 1 - (trace - 2.0 * R[2][2]);
   }

   magnitude = sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
   scale = magnitude;
   if (q[0] < 0)
      scale = -scale;

   q[0] /= scale;
   q[1] /= scale;
   q[2] /= scale;
   q[3] /= scale;
}
