/*******************************************************************************

   OUTPUT.C

   Authors: Krystyne Blaikie
            Peter Loan

   Copyright (c) 2000-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains functions that write out simulation variables
      to data files and the shell window.

   Routines:
   get_reaction_forces        : determine joint reaction forces and torques
   print_simulation_info      : print step info to screen
   print_final_information    : print information at end of simulation
   store_motion_frame         : write information to motion and also to file
   set_up_kinetics_output     : sets up the output kinetics file
   write_kin_out_frame        : writes a frame of data to kinetics file
   sum_forces_on_body         : sums the contact forces on a body
   get_time_now               : get the current time
   show_vector                : display a vector on screen
   show_array:                : display an array length n on screen
   show_matrix:               : display a MxN matrix on screen
   show_array_as_matrix       : display an array as a matrix
   time_elapsed               : calculate the time elapsed since given start time
   print_state                : display model states to screen
   print_body_info            : prints body info to console
   get_time_now               : gets the current time
   get_motion_info            :
   get_input_data            :
   
*******************************************************************************/
#include "universal.h"
#include <time.h>

/*************** DEFINES (for this file only) *********************************/
#define STORE_EXTERNAL_FORCES_IN_GROUND_FRAME 1
#define COMBINE_SPRING_FORCES 0
#define STORE_SPRING_FORCES_IN_CONTACT_FRAME 0

/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/
static MotionData md;
static double *previous_state = NULL;
static dpBoolean* transfer_element = NULL;

/**************** GLOBAL VARIABLES (used in only a few files) *****************/
dpBoolean verbose = dpNo;

/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern dpModelStruct* sdm;
extern dpSimulationParameters params;
extern char buffer[];
extern MotionData* input_data;
extern SystemInfo si;
extern FILE* kin_out;

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static void sum_forces_on_body(int body, double p[], double frc[]);
static double find_time_for_x_value(MotionData* motion, double x_value);


/* determine and store the joint reaction forces and torques */
void get_reaction_forces(void)
{
   int i, j;
   static double *forces = NULL, *torques = NULL;

	if (forces == NULL)
		forces = (double *)simm_malloc(sdm->num_joints * 3 * sizeof(double));
	if (torques == NULL)
		torques = (double *)simm_malloc(sdm->num_joints * 3 * sizeof(double));

	sdreac(forces, torques);
	for (i = 0; i < sdm->num_joints; i++)
   {
		for (j = 0; j < 3; j++)
		{
			sdm->joint[i].force[j] = forces[i * 3 + j];
			sdm->joint[i].torque[j] = torques[i * 3 + j];
		}
   }

   FREE_IFNOTNULL(forces);
   FREE_IFNOTNULL(torques);
   check_for_sderror("GET REACTION FORCES");
} 


/************* I/O routines *****************/

/* print information to screen at each step */
void print_simulation_info(double time, double state_quat[], SystemInfo* si)
{

   if (verbose == dpYes)
   {
	   sim_message(none, "time = %.8lf", time);
#if CONTACT_DETECTION
      sim_message(none, "   #contacts = %d", sdm->num_contacts);
#endif 

	   sim_message(none, "kinetic_energy =   %.10", si->kinetic_energy);
	   sim_message(none, "potential_energy = %.10", si->potential_energy);
	   sim_message(none, "system_energy =    %.10", si->system_energy);
   }
   else
   {
      // cannot use sim_message here because the time display 'scrolls' in place.
      printf("time = %.8lf\r", time);
      fflush(stdout);
   }

}

/* display time, state and system energy when simulation stops */
void print_final_information(double time, double y[], double dy[], SystemInfo* si)
{
   double energy_change;

	sim_message(none, "t_final = %.6lf", time);
	sim_message(none, "final state:");
	print_state(y, dy);
#if 0
	sim_message(none, "initial system energy = %f", si->initial_system_energy);
	sim_message(none, "final system energy   = %f", si->system_energy);
	energy_change = si->system_energy - si->initial_system_energy;
	if (energy_change > 0)
   {
		if (energy_change < 0.01) 
			sim_message(none, "gained energy (delta = %e)", energy_change);
		else
         sim_message(none, "gained energy (delta = %f)", energy_change);
   }
	else if (energy_change < 0)
   {
		if (energy_change > -0.01)
			sim_message(none, "lost energy (delta = %e)", energy_change);
		else
         sim_message(none, "lost energy (delta = %f)", energy_change);
   }
	else
   {
		sim_message(none, "no change in energy");
   }

   sim_message(none, "Simulated from time %lf to %lf.", params.startTime, time);
#endif
}


/* OPEN_OUTPUT_MOTION_FILE: Open the file and write the previously set header information and column names. */
void open_output_motion_file(const char filename[])
{
   int i;

	md.fp = fopen(filename, "w");
   if (md.fp == NULL)
   {
      sim_message(exit_program, "Unable to open output motion file %s.", filename);
      return;
   }

	fprintf(md.fp, "name %s\ndatacolumns %d\n", md.name, md.num_elements);
	fprintf(md.fp, "datarows %d\n", md.num_frames);
	fprintf(md.fp, "range %lf %lf\n", md.range_start, md.range_end);
   if (md.x_column >= 0)
      fprintf(md.fp, "x_column %d\n", md.x_column+1);    // make it 1-based
	fprintf(md.fp, "cursor 1.0 1.0 0.0\n");
   fprintf(md.fp, "keys m_key\n");
   if (md.wrap == dpYes)
      fprintf(md.fp, "wrap\n");
   fprintf(md.fp, "enforce_constraints %s\n", (md.enforce_constraints == dpYes) ? "yes" : "no");
   fprintf(md.fp, "enforce_loops %s\n", (md.enforce_loops == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_markers %s\n", (md.show_markers == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_marker_trails %s\n", (md.show_marker_trails == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_forces %s\n", (md.show_forces == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_force_trails %s\n", (md.show_force_trails == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_foot_prints %s\n", (md.show_foot_prints == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_force_plates %s\n", (md.show_force_plates == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_gencoord_moments %s\n", (md.show_gencoord_moments == dpYes) ? "yes" : "no");
   fprintf(md.fp, "show_other_objects %s\n", (md.show_other_objects == dpYes) ? "yes" : "no");
	fprintf(md.fp, "calc_derivatives no\n");
   for (i = 0; i < md.num_events; i++)
   {
      if (md.event[i].source == dpForceEvent)
         fprintf(md.fp, "force_event %lf %s\n", md.event[i].xCoord, md.event[i].name);
      else if (md.event[i].source == dpMarkerEvent)
         fprintf(md.fp, "marker_event %lf %s\n", md.event[i].xCoord, md.event[i].name);
      else
         fprintf(md.fp, "event %lf %s\n", md.event[i].xCoord, md.event[i].name);
   }
   fprintf(md.fp, "event_color %f %f %f\n", md.event_color[0], md.event_color[1], md.event_color[2]);
   for (i=0; i<md.num_foot_prints; i++)
      fprintf(md.fp, "%s %lf %lf %lf %lf %lf %lf %lf\n", md.foot_print[i]->name, md.foot_print[i]->scale,
         md.foot_print[i]->translation[0], md.foot_print[i]->translation[1], md.foot_print[i]->translation[2], 
         md.foot_print[i]->euler_rotation[0], md.foot_print[i]->euler_rotation[1], md.foot_print[i]->euler_rotation[2]);
   for (i=0; i<md.num_force_plates; i++)
      fprintf(md.fp, "%s %lf %lf %lf %lf %lf %lf %lf %lf %lf\n", md.force_plate[i]->name,
         md.force_plate[i]->scale[0], md.force_plate[i]->scale[1], md.force_plate[i]->scale[2], 
         md.force_plate[i]->translation[0], md.force_plate[i]->translation[1], md.force_plate[i]->translation[2], 
         md.force_plate[i]->euler_rotation[0], md.force_plate[i]->euler_rotation[1], md.force_plate[i]->euler_rotation[2]);
   fprintf(md.fp, "endheader\n\n");

   for (i = 0; i < md.num_elements; i++)
      fprintf(md.fp, "%s\t", md.elementnames[i]);

	fprintf(md.fp, "\n");
	fflush(md.fp);
}


void close_output_motion_file()
{
   if (md.fp)
      fclose(md.fp);
}


/* Open an old-style kinetics data file for storing Qs, Us, and accelerations.
 * This file can be read by the SIMM-generated inverse dynamics
 * code in order to find joint moments.
 */
void set_up_kinetics_output(char filename[])
{
   int i, num_steps, num_gencoords = 0;

   // num_steps should add 1 to account for the first frame of data written
   // at the start of the simulation, and 0.99999 so that any fraction of a
   // time step at the end of the simulation gets counted.
   num_steps = (params.endTime - params.startTime) / params.stepSize + 1.99999;

   remove(filename);  // work around CodeWarrior fopen(,"w") bug 
   kin_out = fopen(filename, "w");
   if (kin_out == NULL)
   {
      sim_message(exit_program, "Unable to open output kinetics file %s.", filename);
      return;
   }

   // Count the number of Qs to output.
   for (i=0; i<sdm->nu; i++)
      if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
         num_gencoords++;

   fprintf(kin_out, "frames %d\n", num_steps);
   fprintf(kin_out, "elements %d\n", num_gencoords * 3 + 1);
   fprintf(kin_out, "otherdata 0\n\n");

   fprintf(kin_out, "time ");
   for (i=0; i<sdm->nu; i++)
      if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
 	      fprintf(kin_out, "%s %s_vel %s_acc\n", sdm->q[i].name, sdm->q[i].name, sdm->q[i].name);
   fprintf(kin_out, "\n");
}


/* Writes a frame of Qs, Us, and accelerations to a data file
 * that can be used for inverse dynamics.
 */
void write_kin_out_frame(double time, double y[], double dy[])
{
   int i;
   double *position_euler, *velocity_euler, *acceleration_euler;

   position_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   velocity_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   acceleration_euler = (double *)simm_malloc(sdm->nu * sizeof(double));

   copy_vector(y, position_euler, sdm->nu);
   copy_vector(dy, velocity_euler, sdm->nu);
   copy_vector(&dy[sdm->nq], acceleration_euler, sdm->nu);

   convert_state_from_quaternion(time, y, &y[sdm->nq], &dy[sdm->nq], position_euler, velocity_euler, acceleration_euler, NULL);

   fprintf(kin_out, "%lf ", time);
   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
         fprintf(kin_out, "%13.8lf %13.8lf %13.8lf\n", position_euler[i]*sdm->q[i].conversion,
                 velocity_euler[i]*sdm->q[i].conversion, acceleration_euler[i]*sdm->q[i].conversion);
   }
   fprintf(kin_out, "\n");
   fflush(kin_out);

   FREE_IFNOTNULL(position_euler);
   FREE_IFNOTNULL(velocity_euler);
   FREE_IFNOTNULL(acceleration_euler);
}


#if CONTACT_DETECTION
/* for each segment, sum all the contact forces acting on it, and average
 * the points of force application.  If forces are due to contact with
 * more than one body, display the force at the body's mass center.
 */
static void sum_forces_on_body(int body, double point[], double force[])
{
   int i,j, num = 0, other_body, check, change;
   double sum[3], force_mag;

   make_vector(force, 0.0, 0.0, 0.0);
   make_vector(point, 0.0, 0.0, 0.0);
   make_vector(sum, 0.0, 0.0, 0.0);
   num = 0;

   for (i = 0; i < sdm->num_contacts; i++)
   {
      force_mag = VECTOR_MAGNITUDE(sdm->contacts[i].cont_frc);
      if (EQUAL_WITHIN_TOLERANCE(force_mag, 0.0, 1e-3))
         continue;
      if (sdm->contacts[i].body1 == body)
      {
         sdvmul(force_mag, sdm->contacts[i].norm1, force);
         sdvadd(force, sum, sum);
         sdvadd(sdm->contacts[i].pt1, point, point);
         num++;
      }
      if (sdm->contacts[i].body2 == body)
      {
         sdvmul(force_mag, sdm->contacts[i].norm2, force);
         sdvadd(force, sum, sum);
         sdvadd(sdm->contacts[i].pt2, point, point);
         num++;
      }
   }
   sdvcopy(sum, force);
   if (num != 0)
      sdvmul(1.0/num, point, point);

   // if the body contacts two or more other bodies, move the 
   // contact point to the body center of mass.
   for (i = 0; i < sdm->num_contacts; i++)
   {
      if (sdm->contacts[i].body1 == body)
      {
         other_body = sdm->contacts[i].body2;
         check = 2;
         break;
      }
      else if (sdm->contacts[i].body2 == body)
      {
         other_body = sdm->contacts[i].body1;
         check = 1;
         break;
      }
   }
   // if the body contacts more than one other body, set the point to the mass center
   change = NO;
   for (j = i + 1; j < sdm->num_contacts; j++)
   {
      if ((check == 1) && (sdm->contacts[j].body1 != other_body))
      {
         change = YES;
         break;
      }
      else if ((check == 2) && (sdm->contacts[j].body2 != other_body))
      {
         change = YES;
         break;
      }
   }

	if (change == YES)
      clear_vector(point, 3);
} 
#endif


void show_vector(char name[], double vect[])
{
   sim_message(none, "%s = %f  %f  %f", name, vect[0], vect[1], vect[2]);
}


void show_array(char name[], double vect[], int start, int end)
{
   int i;

   // Cannot use sim_message() in the first case because the array is printed
   // onto one line. For consistency, it is not used at all in this function.
   if (end <= 3)
   {
      printf("%s = ", name);
      for (i = start; i < end; i++) 
         printf("% f ", vect[i]);
      printf("\n");
   }
   else
   {
      for (i = start; i < end; i++)
         printf("%s[%d] = % f", name, i, vect[i]);
   }
}


void show_matrix(char name[], double *mat[], int m, int n)
{
   int i, j;

   // Cannot use sim_message() here because each row is printed onto one line.
   printf("%s = \n", name);
   for (i = 0; i < m; i++)
   {
      for (j = 0; j < n; j++)
         printf("%f ", mat[i][j]);
      printf("\n");
   }
}

void show_array_as_matrix(char name[], double vec[], int n, int m)
{
   int i, j;

   // Cannot use sim_message() here because each row is printed onto one line.
   for (i = 0; i < (n/m); i++)
   {
      printf("%s%d = ", name, i);
      for (j = 0; j < m; j++)
      {
         printf("%f ", vec[i*m + j]);
      }
      printf("\n");
   }
} 

/* calculate the time elapsed between given start and end times */
void time_elapsed(double t_in, double t_out)
{
   double delta;
   int hr, min, sec;

	if (t_out < t_in) t_out += 24.0;
	delta = t_out - t_in;
	hr = (int) floor(delta);
	delta -= hr;
	delta *= 60;
	min = (int) floor(delta);
	delta -= min;
	delta *= 60;
	sec = (int)(delta);
	sim_message(none, "time elapsed: %d hours %d minutes and %d seconds.", hr, min, sec);
}

/* print the state vectors (quaternion format) and gencoord information to the screen */
void print_state(double y[], double dy[])
{
   int i, pres, axis;
   char buf[1024];

   for (i = 0; i < sdm->nu; i++)
   {
      sprintf(buf, "y[%2d] = %-15.15s= %12.5f %12.5f % 12.5lf", i, sdm->q[i].name, sdm->q[i].conversion * y[i], sdm->q[i].conversion * y[i+ sdm->nq], dy[i + sdm->nq]);
      if (sdm->q[i].type == dpFixedQ)
         strcat(buf, "   FIXED");
      else if (sdm->q[i].type == dpUnconstrainedQ)
         strcat(buf, "   FREE");
      else if (sdm->q[i].type == dpPrescribedQ)
         strcat(buf, "   PRESCRIBED");
      else if (sdm->q[i].type == dpConstrainedQ)
         strcat(buf, "   CONSTRAINED");
      axis = sdm->q[i].axis;
      if (sdm->q[i].is_part_of_quaternion)
         axis = abs(sdm->q[i].axis) - 1;
      sdgetpres(sdm->q[i].joint, axis, &pres);
      if (pres == 0)
         strcat(buf, "  (0)");
      else
         strcat(buf, "  (1)");
      sim_message(none, buf);
   }
}

/* print all the segment information to the screen */
void print_body_info(void)
{
   int i;

   for (i = 1; i < sdm->num_body_segments; i++)
   {
      sim_message(none, "BODY %d - %s", i, sdm->body_segment[i].name);
      sim_message(none, "   mass = %f", sdm->body_segment[i].mass);
      sim_message(none, "   inertia = %f %f %f", sdm->body_segment[i].inertia[0][0],
         sdm->body_segment[i].inertia[1][1], sdm->body_segment[i].inertia[2][2]);
      show_vector("   mc", sdm->body_segment[i].mass_center);
   }
}

/* get the current time, print it to the screen and save it in hours */
double get_time_now(void)
{
   double 		hr, min, sec, time_now;
   struct tm	*ptr;
   time_t		lt;
   char			time_str[100];

	lt = time(NULL);
	ptr = localtime(&lt);
	 
	strftime(time_str, 100, "%a %b %d %H:%M:%S %p", ptr);
	sim_message(none, "%s", time_str);
	hr = ptr->tm_hour; min = ptr->tm_min; sec = ptr->tm_sec;
	time_now = hr + min / 60.0 + sec / 3600.0;

	return time_now;
}


static double find_time_for_x_value(MotionData* motion, double x_value)
{
   int i, index, frame = 0;
   double time_range, x_range, percent;

   if (!motion || motion->x_column < 0 || motion->time_column < 0 || motion->num_frames < 2)
      return x_value;

   if (motion->time_column == motion->x_column)
      return x_value;

   // Find the first frame that is greater than x_value, then figure out whether current value
   // is closer to that frame or the previous frame.
   for (i=0; i<motion->num_frames; i++)
   {
      if (motion->motiondata[motion->x_column][i] + TINY_NUMBER >= x_value)
      {
         if (i == 0)
         {
            frame = 0;
            goto calc_time;
         }
         else
         {
            double leftDiff = x_value - motion->motiondata[motion->x_column][i - 1] - TINY_NUMBER;
            double rightDiff = motion->motiondata[motion->x_column][i] + TINY_NUMBER - x_value;

            if (leftDiff >= 0.0 && rightDiff >= 0.0)
            {
               if (leftDiff < rightDiff)
                  frame = i - 1;
               else
                  frame = i;
               goto calc_time;
            }
         }
      }
   }

calc_time:
   if (frame == motion->num_frames - 1)
      index = frame - 1;
   else
      index = frame;

   x_range = motion->motiondata[motion->x_column][index + 1] - motion->motiondata[motion->x_column][index];
   percent = (x_value - motion->motiondata[motion->x_column][index]) / x_range;
   time_range = motion->motiondata[motion->time_column][index + 1] - motion->motiondata[motion->time_column][index];

   return motion->motiondata[motion->time_column][index] + percent * time_range;
}


/* CREATE_OUTPUT_MOTION: Create output motion structure, fill in header information (event information,
 * footprint information, forceplate information, SIMM display parameters), determine the number of data columns,
 * and set the column names. */
void create_output_motion(const char name[])
{
   int i, j, count, num_steps, num_cols = 0, body;
#if COMBINE_SPRING_FORCES
	dpBoolean *segmentHasSprings;
#endif

   mstrcpy(&md.name, name);

   // num_frames should add 1 to account for the first frame of data written
   // at the start of the simulation, and 0.99999 so that any fraction of a
   // time step at the end of the simulation gets counted.
   md.num_frames = (params.endTime - params.startTime) / params.stepSize + 1.99999;

   md.fp = 0;
   md.range_start = params.startTime;
   md.range_end = params.endTime;
   md.step_size = params.stepSize;
   md.x_column = -1;
   md.time_column = -1;

   if (input_data)
   {
      md.num_events = 0;
      if (input_data->num_events > 0 && input_data->event)
      {
         ReturnCode rc;

         md.event = (dpMotionEvent*)simm_malloc(input_data->num_events * sizeof(dpMotionEvent));

         for (i=0; i<input_data->num_events; i++)
         {
            // Copy only the events that fall within the simulation time range.
            double event_time = find_time_for_x_value(input_data, input_data->event[i].xCoord);

            if (event_time >= md.range_start && event_time <= md.range_end)
            {
               memcpy(&md.event[md.num_events], &input_data->event[i], sizeof(dpMotionEvent));
               mstrcpy(&md.event[md.num_events].name, input_data->event[i].name);
               md.num_events++;
            }
         }
         md.event = (dpMotionEvent*)simm_realloc(md.event, md.num_events * sizeof(dpMotionEvent), &rc);
      }
      else
      {
         md.event = NULL;
      }

      for (i = 0; i < 3; i++)
         md.event_color[i] = input_data->event_color[i];

      md.num_foot_prints = input_data->num_foot_prints;
      if (md.num_foot_prints > 0)
      {
         md.foot_print = (dpFootPrint**)simm_malloc(md.num_foot_prints * sizeof(dpFootPrint*));
         if (md.foot_print)
         {
            for (i=0; i<md.num_foot_prints; i++)
            {
               md.foot_print[i] = (dpFootPrint*)simm_malloc(sizeof(dpFootPrint));
               memcpy(md.foot_print[i], input_data->foot_print[i], sizeof(dpFootPrint));
               mstrcpy(&md.foot_print[i]->name, input_data->foot_print[i]->name);
            }
         }
      }
      else
      {
         md.foot_print = NULL;
      }

      md.num_force_plates = input_data->num_force_plates;
      if (md.num_force_plates > 0)
      {
         md.force_plate = (dpForcePlate**)simm_malloc(md.num_force_plates * sizeof(dpForcePlate*));
         if (md.force_plate)
         {
            for (i=0; i<md.num_force_plates; i++)
            {
               md.force_plate[i] = (dpForcePlate*)simm_malloc(sizeof(dpForcePlate));
               memcpy(md.force_plate[i], input_data->force_plate[i], sizeof(dpForcePlate));
               mstrcpy(&md.force_plate[i]->name, input_data->force_plate[i]->name);
            }
         }
      }
      else
      {
         md.force_plate = NULL;
      }

      md.enforce_constraints = input_data->enforce_constraints;
      md.enforce_loops = input_data->enforce_loops;
      md.show_markers = input_data->show_markers;
      md.show_marker_trails = input_data->show_marker_trails;
      md.show_forces = input_data->show_forces;
      md.show_force_trails = input_data->show_force_trails;
      md.show_foot_prints = input_data->show_foot_prints;
      md.show_force_plates = input_data->show_force_plates;
      md.wrap = input_data->wrap;
   }
   else
   {
      md.event_color[0] = 1.0;
      md.event_color[1] = 0.0;
      md.event_color[2] = 1.0;
   }

   // current_frame indicates the first empty row of data
   md.current_frame = 0;

   // normalization
   if (params.outputOptions.normalization != dpNoNormalization)
      num_cols++;

   // time
	num_cols++;

   num_cols += params.outputOptions.gencoordValues.num_items;
   num_cols += params.outputOptions.gencoordVelocities.num_items;
   num_cols += params.outputOptions.gencoordAccelerations.num_items;

   // The set of muscle activations to output is the boolean addition of the
   // muscle activations and the optimized muscle activations.
   num_cols += params.outputOptions.muscleActivations.num_items;
   for (i=0; i<params.outputOptions.optimizedActivations_Muscles.num_items; i++)
      if (find_name_in_list(params.outputOptions.optimizedActivations_Muscles.name[i], &params.outputOptions.muscleActivations) == -1)
         num_cols++;

   num_cols += params.outputOptions.muscleLengths.num_items;
   num_cols += params.outputOptions.muscleFiberLengths.num_items;
   num_cols += params.outputOptions.muscleFiberVelocities.num_items;
   num_cols += params.outputOptions.muscleForces.num_items;

   // external forces and torques
   // All torques, and all forces that are not specified in the ground frame,
   // are stored directly in the output motion file. All forces that are
   // specified in the ground frame are merged into one force per force matte.
   if (input_data)
   {
      // 9 columns per force matte (3 are for the free torque component)
      num_cols += (9 * sdm->num_force_mattes);

      // 6 columns per force not specified in the ground frame
      for (i = 0; i < input_data->num_forces; i++)
      {
         if (input_data->forces[i]->segment != GROUND)
            num_cols += 6;
      }

      // 3 columns per torque
      num_cols += input_data->num_torques * 3;
   }

#if CONTACT_DETECTION
	// contact forces (px, py, pz, vx, vy, vz) (cr, cg, cb
	if (params.outputOptions.contactForces == dpYes)
   {
		for (i = 0; i < sdm->num_body_segments; i++)
      {
			if (sdm->body_segment[i].contactable == dpYes)
         {
				num_cols += 6 + 3;
         }
		}
   }
#endif

   // spring forces (px, py, pz, vx, vy, vz)
   if (params.outputOptions.springForces == dpYes)
   {
#if COMBINE_SPRING_FORCES
		// All the spring forces on a body segment will be merged into one vector.
		// The params.outputOptions.numSpringForces parameter is ignored.
		segmentHasSprings = (dpBoolean*)simm_calloc(sdm->num_body_segments, sizeof(dpBoolean));
      for (i = 0; i < sdm->num_springs; i++)
			segmentHasSprings[sdm->spring[i].segment + 1] = dpYes;
		for (i = 0; i < sdm->num_body_segments; i++)
		{
			if (segmentHasSprings[i] == dpYes)
				num_cols += 6;
		}
#else
      for (i = 0; i < MIN(params.outputOptions.numSpringForces, sdm->num_springs); i++)
      {
         num_cols += 6;
      }
#endif
   }

	// joint reaction forces (px, py, pz, rx, ry, rz)
   num_cols += params.outputOptions.jointReactionForces.num_items * 6;
	
   // joint reaction torques (x, y, z)
   num_cols += params.outputOptions.jointReactionTorques.num_items * 3;

   num_cols += params.outputOptions.gencoordMoments.num_items;
   num_cols += params.outputOptions.gencoordPowers.num_items;

   // muscle moment arms
   num_cols += params.outputOptions.muscleMomentArms_Muscles.num_items * params.outputOptions.muscleMomentArms_Gencoords.num_items;

   // muscle moments
   num_cols += params.outputOptions.muscleMoments_Muscles.num_items * params.outputOptions.muscleMoments_Gencoords.num_items;

   // total muscle moments (list specifies Qs; all muscles are used)
   num_cols += params.outputOptions.totalMuscleMoments.num_items;

	// mass center positions (x, y, z)
   num_cols += params.outputOptions.massCenterPositions.num_items * 3;

	// total system energy
	if (params.outputOptions.systemEnergy == dpYes)
		num_cols++;

	// output mass center velocities (px, py, pz, vx, vy, vz)
   num_cols += params.outputOptions.massCenterVelocities.num_items * 6;

   // residual moments of gencoords used for optimized muscle activations 
   // The Qs that will be included are those that the user specified and that are either unconstrained
   // or prescribed, and that are used in a joint spanned by at least one muscle.
   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0)
   {
      for (i = 0; i < sdm->nu; i++)
      {
         if (find_name_in_list(sdm->q[i].name, &params.outputOptions.optimizedActivations_Gencoords) >= 0)
         {
            for (j = 0; j < sdm->num_muscles; j++)
            {
               if (NOT_EQUAL_WITHIN_TOLERANCE(sdm->muscles[j].momentarms[i], 0, 0.00001))
               {
                  num_cols++;
                  break;
               }
            }
         }
      }
   }

   // Count how many columns of data should be transferred from the input motion (if there is one)
   // to the output motion.
   if (input_data)
   {
      FREE_IFNOTNULL(transfer_element);
      transfer_element = (dpBoolean*)simm_calloc(input_data->num_elements, sizeof(dpBoolean));
      num_cols += identify_elements_to_transfer(input_data, transfer_element);
   }

   // set the display parameters
   // turn display of forces on if have any of: spring forces, contact forces, joint reaction forces
   if ((params.outputOptions.contactForces == dpYes) || (params.outputOptions.jointReactionForces.num_items > 0) ||
      (params.outputOptions.springForces == dpYes) || sdm->num_force_mattes > 0 || 
      (params.outputOptions.jointReactionTorques.num_items > 0))
      md.show_forces = dpYes;
   if (params.outputOptions.massCenterPositions.num_items > 0)
      md.show_other_objects = dpYes;


   md.num_elements = num_cols;
   md.elementnames = (char**)simm_calloc(md.num_elements, sizeof(char*));
   count = 0;

	// normalization
   if (params.outputOptions.normalization == dpNormalizeToOne)
      mstrcpy(&md.elementnames[count++], "factor");
   else if (params.outputOptions.normalization == dpNormalizeToOneHundred)
   {
      md.x_column = count;
      mstrcpy(&md.elementnames[count++], "percent");
   }

	// step, time
   md.time_column = count;
	mstrcpy(&md.elementnames[count++], "time");

	// gencoords
   for (i=0; i<params.outputOptions.gencoordValues.num_items; i++)
   {
      sprintf(buffer, "%s_gcv", params.outputOptions.gencoordValues.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }
			
	// gencoord velocities
   for (i=0; i<params.outputOptions.gencoordVelocities.num_items; i++)
   {
      sprintf(buffer, "%s_vel_gcv", params.outputOptions.gencoordVelocities.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

	// gencoord accelerations
   for (i=0; i<params.outputOptions.gencoordAccelerations.num_items; i++)
   {
      sprintf(buffer, "%s_acc_gcv", params.outputOptions.gencoordAccelerations.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

   // The set of muscle activations to output is the boolean addition of the
   // muscle activations and the optimized muscle activations.
   for (i=0; i<params.outputOptions.muscleActivations.num_items; i++)
      mstrcpy(&md.elementnames[count++], params.outputOptions.muscleActivations.name[i]);
   for (i=0; i<params.outputOptions.optimizedActivations_Muscles.num_items; i++)
      if (find_name_in_list(params.outputOptions.optimizedActivations_Muscles.name[i], &params.outputOptions.muscleActivations) == -1)
         mstrcpy(&md.elementnames[count++], params.outputOptions.optimizedActivations_Muscles.name[i]);

   // muscle length
   for (i=0; i<params.outputOptions.muscleLengths.num_items; i++)
   {
      sprintf(buffer, "%s_len", params.outputOptions.muscleLengths.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

	// muscle fiber length
   for (i=0; i<params.outputOptions.muscleFiberLengths.num_items; i++)
   {
      sprintf(buffer, "%s_fiberlen", params.outputOptions.muscleFiberLengths.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }
	
   // muscle fiber velocity
   for (i=0; i<params.outputOptions.muscleFiberVelocities.num_items; i++)
   {
      sprintf(buffer, "%s_fibervel", params.outputOptions.muscleFiberVelocities.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

   // muscle force
   for (i=0; i<params.outputOptions.muscleForces.num_items; i++)
   {
      sprintf(buffer, "%s_frc", params.outputOptions.muscleForces.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

   // external forces
   if (input_data)
   {
      const char* function_suffix = "_gcv";

      // Because ground-based forces are merged together using force mattes, you cannot
      // necessarily preserve the function type of individual forces when outputting them.
      // So just use the function type of the X component of the first force vector.
      if (input_data->num_forces > 0)
         function_suffix = get_function_suffix(&input_data->forces[0]->force_vec[XX]);

#if STORE_EXTERNAL_FORCES_IN_GROUND_FRAME
      // Use this code to store external forces in the ground frame.

      // 9 columns per force matte (3 are for the free torque component).
      for (i = 0; i < sdm->num_force_mattes; i++)
      {
         sprintf(buffer, "%s_force_px%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_py%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_pz%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_vx%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_vy%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_vz%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_x%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_y%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_z%s", sdm->body_segment[GROUND+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
      }

      // 6 columns per other external force.
      for (i = 0; i < input_data->num_forces; i++)
      {
         if (input_data->forces[i]->segment != GROUND)
         {
            sprintf(buffer, "%s_force_px%s", sdm->body_segment[GROUND+1].name,
               get_function_suffix(&input_data->forces[i]->appl_point[XX]));
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_force_py%s", sdm->body_segment[GROUND+1].name,
               get_function_suffix(&input_data->forces[i]->appl_point[YY]));
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_force_pz%s", sdm->body_segment[GROUND+1].name,
               get_function_suffix(&input_data->forces[i]->appl_point[ZZ]));
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_force_vx%s", sdm->body_segment[GROUND+1].name,
               get_function_suffix(&input_data->forces[i]->force_vec[XX]));
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_force_vy%s", sdm->body_segment[GROUND+1].name,
               get_function_suffix(&input_data->forces[i]->force_vec[YY]));
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_force_vz%s", sdm->body_segment[GROUND+1].name,
               get_function_suffix(&input_data->forces[i]->force_vec[ZZ]));
            mstrcpy(&md.elementnames[count++], buffer);
         }
      }
#else
      // Use this code to store external forces in the reference frame
      // of the segment to which they are applied.

      // 9 columns per force matte (3 are for the free torque component).
      for (i = 0; i < sdm->num_force_mattes; i++)
      {
         sprintf(buffer, "%s_force_px%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_py%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_pz%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_vx%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_vy%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_force_vz%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_x%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_y%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_z%s", sdm->body_segment[sdm->force_matte[i].segment+1].name, function_suffix);
         mstrcpy(&md.elementnames[count++], buffer);
      }

      // 6 columns per external force.
      for (i = 0; i < input_data->num_forces; i++)
      {
         if (input_data->forces[i]->segment != GROUND)
         {
				sprintf(buffer, "%s_force_px%s", sdm->body_segment[input_data->forces[i]->segment+1].name,
					get_function_suffix(&input_data->forces[i]->appl_point[XX]));
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_force_py%s", sdm->body_segment[input_data->forces[i]->segment+1].name,
					get_function_suffix(&input_data->forces[i]->appl_point[YY]));
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_force_pz%s", sdm->body_segment[input_data->forces[i]->segment+1].name,
					get_function_suffix(&input_data->forces[i]->appl_point[ZZ]));
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_force_vx%s", sdm->body_segment[input_data->forces[i]->segment+1].name,
					get_function_suffix(&input_data->forces[i]->force_vec[XX]));
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_force_vy%s", sdm->body_segment[input_data->forces[i]->segment+1].name,
					get_function_suffix(&input_data->forces[i]->force_vec[YY]));
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_force_vz%s", sdm->body_segment[input_data->forces[i]->segment+1].name,
					get_function_suffix(&input_data->forces[i]->force_vec[ZZ]));
            mstrcpy(&md.elementnames[count++], buffer);
         }
      }
#endif

      // 3 columns per torque (always specified in the frame in which they are applied)
      for (i=0; i<input_data->num_torques; i++)
      {
         sprintf(buffer, "%s_torque_vx%s", sdm->body_segment[input_data->torques[i]->segment+1].name,
				get_function_suffix(&input_data->torques[i]->torque_vec[XX]));
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_vy%s", sdm->body_segment[input_data->torques[i]->segment+1].name,
				get_function_suffix(&input_data->torques[i]->torque_vec[YY]));
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_torque_vz%s", sdm->body_segment[input_data->torques[i]->segment+1].name,
				get_function_suffix(&input_data->torques[i]->torque_vec[ZZ]));
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }

#if CONTACT_DETECTION
	// contact forces
	if (params.outputOptions.contactForces == dpYes)
   {
		for (i = 0; i < sdm->num_body_segments; i++)
      {
			if (sdm->body_segment[i].contactable == dpYes)
         {
            sprintf(buffer, "%s_contact_cr", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_cg", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_cb", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_px", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_py", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_pz", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_vx", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_vy", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
            sprintf(buffer, "%s_contact_vz", sdm->body_segment[i].name);
            mstrcpy(&md.elementnames[count++], buffer);
			}
      }
   }
#endif
					
   // spring forces
   if (params.outputOptions.springForces == dpYes)
   {
#if COMBINE_SPRING_FORCES
  #if STORE_SPRING_FORCES_IN_CONTACT_FRAME
		for (i = 0; i < sdm->num_body_segments; i++)
		{
			if (segmentHasSprings[i] == dpYes)
			{
				// Find first spring which attaches to this segment and use its
				// floor/surface to write a segment name into the column label. This
				// assumes that all springs attaching to one segment use either
				// a single contact surface or multiple surfaces that are fixed to the same segment
				int segmentIndex = GROUND + 1;
				for (j = 0; j < sdm->num_springs; j++)
				{
					if (sdm->spring[j].segment + 1 == i)
					{
                  segmentIndex = get_spring_contact_segment_sd_index(j);
						break;
					}
				}
				sprintf(buffer, "%s_spring_force_px", sdm->body_segment[segmentIndex+1].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_py", sdm->body_segment[segmentIndex+1].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_pz", sdm->body_segment[segmentIndex+1].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_vx", sdm->body_segment[segmentIndex+1].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_vy", sdm->body_segment[segmentIndex+1].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_vz", sdm->body_segment[segmentIndex+1].name);
				mstrcpy(&md.elementnames[count++], buffer);
			}
		}
  #else
		for (i = 0; i < sdm->num_body_segments; i++)
		{
			if (segmentHasSprings[i] == dpYes)
			{
				sprintf(buffer, "%s_spring_force_px", sdm->body_segment[i].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_py", sdm->body_segment[i].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_pz", sdm->body_segment[i].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_vx", sdm->body_segment[i].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_vy", sdm->body_segment[i].name);
				mstrcpy(&md.elementnames[count++], buffer);
				sprintf(buffer, "%s_spring_force_vz", sdm->body_segment[i].name);
				mstrcpy(&md.elementnames[count++], buffer);
			}
		}
  #endif
		FREE_IFNOTNULL(segmentHasSprings);
#else
  #if STORE_SPRING_FORCES_IN_CONTACT_FRAME
      for (i = 0; i < MIN(params.outputOptions.numSpringForces, sdm->num_springs); i++)
      {
         int segmentIndex = get_spring_contact_segment_sd_index(i);
         sprintf(buffer, "%s_spring_force_px", sdm->body_segment[segmentIndex + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_py", sdm->body_segment[segmentIndex + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_pz", sdm->body_segment[segmentIndex + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_vx", sdm->body_segment[segmentIndex + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_vy", sdm->body_segment[segmentIndex + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_vz", sdm->body_segment[segmentIndex + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
      }
  #else
      for (i = 0; i < MIN(params.outputOptions.numSpringForces, sdm->num_springs); i++)
      {
         sprintf(buffer, "%s_spring_force_px", sdm->body_segment[sdm->spring[i].sd_segment + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_py", sdm->body_segment[sdm->spring[i].sd_segment + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_pz", sdm->body_segment[sdm->spring[i].sd_segment + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_vx", sdm->body_segment[sdm->spring[i].sd_segment + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_vy", sdm->body_segment[sdm->spring[i].sd_segment + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_spring_force_vz", sdm->body_segment[sdm->spring[i].sd_segment + 1].name);
         mstrcpy(&md.elementnames[count++], buffer);
      }
  #endif
#endif
   }

   // joint reaction forces
   for (i=0; i<params.outputOptions.jointReactionForces.num_items; i++)
   {
      body = sdm->joint[params.outputOptions.jointReactionForces.index[i]].outboard_body + 1;
      sprintf(buffer, "%s_joint_force_px", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_force_py", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_force_pz", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_force_vx", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_force_vy", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_force_vz", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
   }
	
   // joint reaction torques
   for (i=0; i<params.outputOptions.jointReactionTorques.num_items; i++)
   {
      body = sdm->joint[params.outputOptions.jointReactionTorques.index[i]].outboard_body + 1;
      sprintf(buffer, "%s_joint_torque_vx", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_torque_vy", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
      sprintf(buffer, "%s_joint_torque_vz", sdm->body_segment[body].name);
      mstrcpy(&md.elementnames[count++], buffer);
   }

   // gencoord moments
   if (params.outputOptions.traditionalInverse == dpNo)
   {
      for (i=0; i<params.outputOptions.gencoordMoments.num_items; i++)
      {
         sprintf(buffer, "%s_moment_lin", params.outputOptions.gencoordMoments.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }
   else
   {
      for (i=0; i<params.outputOptions.gencoordMoments.num_items; i++)
      {
         sprintf(buffer, "%s_moment_gcv", params.outputOptions.gencoordMoments.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }

   // gencoord powers
   for (i=0; i<params.outputOptions.gencoordPowers.num_items; i++)
   {
      sprintf(buffer, "%s_power", params.outputOptions.gencoordPowers.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

   // muscle moment arms
   for (i=0; i<params.outputOptions.muscleMomentArms_Muscles.num_items; i++)
   {
      for (j=0; j<params.outputOptions.muscleMomentArms_Gencoords.num_items; j++)
      {
         sprintf(buffer, "%s_ma_%s", params.outputOptions.muscleMomentArms_Muscles.name[i],
            params.outputOptions.muscleMomentArms_Gencoords.name[j]);
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }

   // muscle moments
   for (i=0; i<params.outputOptions.muscleMoments_Muscles.num_items; i++)
   {
      for (j=0; j<params.outputOptions.muscleMoments_Gencoords.num_items; j++)
      {
         sprintf(buffer, "%s_moment_%s", params.outputOptions.muscleMoments_Muscles.name[i],
            params.outputOptions.muscleMoments_Gencoords.name[j]);
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }

   // total muscle moments
   for (i=0; i<params.outputOptions.totalMuscleMoments.num_items; i++)
   {
      sprintf(buffer, "%s_musc_moment", params.outputOptions.totalMuscleMoments.name[i]);
      mstrcpy(&md.elementnames[count++], buffer);
   }

   // mass center positions
   for (i=0; i<params.outputOptions.massCenterPositions.num_items; i++)
   {
      if (STRINGS_ARE_EQUAL(params.outputOptions.massCenterPositions.name[i], "ground"))
      {
         // to hold mass center for the entire system
         mstrcpy(&md.elementnames[count++], "system_mc_px");
         mstrcpy(&md.elementnames[count++], "system_mc_py");
         mstrcpy(&md.elementnames[count++], "system_mc_pz");
      }
      else
      {
         sprintf(buffer, "%s_mc_px", params.outputOptions.massCenterPositions.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mc_py", params.outputOptions.massCenterPositions.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mc_pz", params.outputOptions.massCenterPositions.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }

   // system energy
	if (params.outputOptions.systemEnergy == dpYes)
      mstrcpy(&md.elementnames[count++], "system_energy");

   // mass center velocities
   for (i=0; i<params.outputOptions.massCenterVelocities.num_items; i++)
   {
      if (STRINGS_ARE_EQUAL(params.outputOptions.massCenterVelocities.name[i], "ground"))
      {
         // to hold mass center velocity for the entire system
         mstrcpy(&md.elementnames[count++], "system_mcvel_vector_px");
         mstrcpy(&md.elementnames[count++], "system_mcvel_vector_py");
         mstrcpy(&md.elementnames[count++], "system_mcvel_vector_pz");
         mstrcpy(&md.elementnames[count++], "system_mcvel_vector_vx");
         mstrcpy(&md.elementnames[count++], "system_mcvel_vector_vy");
         mstrcpy(&md.elementnames[count++], "system_mcvel_vector_vz");
      }
      else
      {
         sprintf(buffer, "%s_mcvel_vector_px", params.outputOptions.massCenterVelocities.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mcvel_vector_py", params.outputOptions.massCenterVelocities.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mcvel_vector_pz", params.outputOptions.massCenterVelocities.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mcvel_vector_vx", params.outputOptions.massCenterVelocities.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mcvel_vector_vy", params.outputOptions.massCenterVelocities.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
         sprintf(buffer, "%s_mcvel_vector_vz", params.outputOptions.massCenterVelocities.name[i]);
         mstrcpy(&md.elementnames[count++], buffer);
      }
   }

   // moment residuals for gencoords optimized in static muscle optimization
   // The Qs that will be included are those that the user specified and that are either unconstrained
   // or prescribed, and that are used in a joint spanned by at least one muscle.
   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0)
   {
      for (i = 0; i < sdm->nu; i++)
      {
         if (find_name_in_list(sdm->q[i].name, &params.outputOptions.optimizedActivations_Gencoords) >= 0)
         {
            for (j = 0; j < sdm->num_muscles; j++)
            {
               if (NOT_EQUAL_WITHIN_TOLERANCE(sdm->muscles[j].momentarms[i], 0, 0.00001))
               {
                  sprintf(buffer, "%s_residual", sdm->q[i].name);
                  mstrcpy(&md.elementnames[count++], buffer);
                  break;
               }
            }
         }
      }
   }

   // columns from the input motion that are being transferred to the output motion
   // if there is a "percent" column, store the column index
   if (input_data)
   {
      for (i=0; i<input_data->num_elements; i++)
      {
         if (transfer_element[i] == dpYes)
         {
            if (STRINGS_ARE_EQUAL(input_data->elementnames[i], "percent"))
               md.x_column = count;
            mstrcpy(&md.elementnames[count++], input_data->elementnames[i]);
         }
      }
   }

   // if no "percent" column was found, set x_column = time_column
   if (md.x_column < 0)
      md.x_column = md.time_column;

   // Now make room for the data, which will get filled in later.
   md.motiondata = (double**)simm_calloc(md.num_elements, sizeof(double*));
   for (i = 0; i < md.num_elements; i++)
      md.motiondata[i] = (double*)simm_calloc(md.num_frames, sizeof(double));
}


/* Add a force vector to one already stored in a ForceMatteForce. */
static void add_vector(double pt[], double vec[], double freeTorque[], ForceMatteForce* fmf)
{
   int i;
   double oldMagnitude = fmf->magnitude;
   double vecMagnitude = VECTOR_MAGNITUDE(vec);
   double newMagnitude = oldMagnitude + vecMagnitude;

   for (i=0; i<3; i++)
   {
      // Add in the force.
      fmf->force[i] += vec[i];
      // The new point of application is the weighted sum of the two vectors.
      fmf->pointOfApplication[i] = (fmf->pointOfApplication[i]*oldMagnitude + pt[i]*vecMagnitude) / newMagnitude;
      // Add in the free torque. TODO: 2 free torques may complicate the calculations for force and COP addition.
      fmf->freeTorque[i] += freeTorque[i];
   }
   fmf->magnitude = newMagnitude;
}


/* Store a frame of data in the output motion structure, and write it to a file, if the file has been opened.
 */
int store_motion_frame(double time, double state_quat[], double dstate_quat[])
{
   int i, j, index, frame, imp = NO, col = 0;
   double torque, com[3] = {0.0, 0.0, 0.0}, vel[3];
   double point[3], force[3], free_torque[3], point2[3], force2[3], free_torque2[3], intersectpt[3];
   double *state_euler, *velocity_euler, *acceleration_euler;

   frame = md.current_frame++;

	// normalization
   if (params.outputOptions.normalization == dpNormalizeToOne)
      md.motiondata[col++][frame] = (double)frame * 0.01;
   else if (params.outputOptions.normalization == dpNormalizeToOneHundred)
      md.motiondata[col++][frame] = (double)frame;

	// step, time
   md.motiondata[col++][frame] = time;

   state_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   velocity_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   acceleration_euler = (double *)simm_malloc(sdm->nq * sizeof(double));

   convert_state_from_quaternion(time, state_quat, &state_quat[sdm->nq], &dstate_quat[sdm->nq], state_euler, velocity_euler, acceleration_euler, previous_state);

	// Store the gencoord values in the motion structure
   for (i=0; i<params.outputOptions.gencoordValues.num_items; i++)
   {
      index = params.outputOptions.gencoordValues.index[i];
      md.motiondata[col++][frame] = state_euler[index] * sdm->q[index].conversion;
   }

   if (previous_state == NULL)
      previous_state = (double *)simm_malloc(sdm->nq * sizeof(double));
   for (i = 0; i < sdm->nq; i++)
      previous_state[i] = state_euler[i];

	// gencoord velocities
   for (i=0; i<params.outputOptions.gencoordVelocities.num_items; i++)
   {
      index = params.outputOptions.gencoordVelocities.index[i];
      md.motiondata[col++][frame] = velocity_euler[index] * sdm->q[index].conversion;
   }

	// gencoord accelerations
   for (i=0; i<params.outputOptions.gencoordAccelerations.num_items; i++)
   {
      index = params.outputOptions.gencoordAccelerations.index[i];
      md.motiondata[col++][frame] = acceleration_euler[index] * sdm->q[index].conversion;
   }

   // The set of muscle activations to output is the boolean addition of the
   // muscle activations and the optimized muscle activations.
   for (i=0; i<params.outputOptions.muscleActivations.num_items; i++)
   {
      index = params.outputOptions.muscleActivations.index[i];
      md.motiondata[col++][frame] = sdm->muscles[index].dynamic_activation;
   }
   for (i=0; i<params.outputOptions.optimizedActivations_Muscles.num_items; i++)
   {
      if (find_name_in_list(params.outputOptions.optimizedActivations_Muscles.name[i], &params.outputOptions.muscleActivations) == -1)
      {
         index = params.outputOptions.optimizedActivations_Muscles.index[i];
         md.motiondata[col++][frame] = sdm->muscles[index].dynamic_activation;
      }
   }

   // muscle length
   for (i=0; i<params.outputOptions.muscleLengths.num_items; i++)
   {
      index = params.outputOptions.muscleLengths.index[i];
      md.motiondata[col++][frame] = sdm->muscles[index].muscle_tendon_length;
   }

   // muscle fiber length
   for (i=0; i<params.outputOptions.muscleFiberLengths.num_items; i++)
   {
      index = params.outputOptions.muscleFiberLengths.index[i];
      md.motiondata[col++][frame] = sdm->muscles[index].fiber_length;
   }

   // muscle fiber velocity
   for (i=0; i<params.outputOptions.muscleFiberVelocities.num_items; i++)
   {
      index = params.outputOptions.muscleFiberVelocities.index[i];
      md.motiondata[col++][frame] = sdm->muscles[index].fiber_velocity;
   }

   // muscle force
   for (i=0; i<params.outputOptions.muscleForces.num_items; i++)
   {
      index = params.outputOptions.muscleForces.index[i];
      md.motiondata[col++][frame] = sdm->muscles[index].force;
   }

   // external forces
   if (input_data)
   {
      ForceMatteForce* fmf = (ForceMatteForce*)simm_calloc(sdm->num_force_mattes, sizeof(ForceMatteForce));

      // For the ground-based forces, see which force_matte they intersect (if any), and add them to
      // to that matte's force structure.
      for (i = 0; i < input_data->num_forces; i++)
      {
         if (input_data->forces[i]->segment == GROUND)
         {
            get_applied_force(time, input_data, i, point, force, free_torque);

            for (j = 0; j < sdm->num_force_mattes; j++)
            {
               sdtrans(GROUND, force, sdm->force_matte[j].segment, force2);
               sdtrans(GROUND, free_torque, sdm->force_matte[j].segment, free_torque2);
               sdvcopy(point, point2);
               convert_point(point2, GROUND, sdm->force_matte[j].segment);
               // If the force intersects force_matte j, add it to that force_matte's force structure and
               // skip over the remaining force mattes.
               if (vector_intersects_matte(point2, force2, &sdm->force_matte[j], intersectpt) == dpYes)
               {
                  add_vector(point2, force2, free_torque2, &fmf[j]);
                  break;
               }
            }
         }
      }

      // Store the force_matte forces in the motion structure.
      for (i = 0; i < sdm->num_force_mattes; i++)
      {
#if STORE_EXTERNAL_FORCES_IN_GROUND_FRAME
         sdtrans(sdm->force_matte[i].segment, fmf[i].force, GROUND, fmf[i].force);
         sdtrans(sdm->force_matte[i].segment, fmf[i].freeTorque, GROUND, fmf[i].freeTorque);
         convert_point(fmf[i].pointOfApplication, sdm->force_matte[i].segment, GROUND);
#else
         // Convert the point of application from the SD/FAST frame to the SIMM frame (by adding the mass center location).
         sdvadd(fmf[i].pointOfApplication, sdm->body_segment[sdm->force_matte[i].segment + 1].mass_center, fmf[i].pointOfApplication);
#endif
         md.motiondata[col++][frame] = fmf[i].pointOfApplication[XX];
         md.motiondata[col++][frame] = fmf[i].pointOfApplication[YY];
         md.motiondata[col++][frame] = fmf[i].pointOfApplication[ZZ];
         md.motiondata[col++][frame] = fmf[i].force[XX];
         md.motiondata[col++][frame] = fmf[i].force[YY];
         md.motiondata[col++][frame] = fmf[i].force[ZZ];
         md.motiondata[col++][frame] = fmf[i].freeTorque[XX];
         md.motiondata[col++][frame] = fmf[i].freeTorque[YY];
         md.motiondata[col++][frame] = fmf[i].freeTorque[ZZ];
      }

      // Store the non-ground-based forces in the motion structure.
      for (i = 0; i < input_data->num_forces; i++)
      {
         if (input_data->forces[i]->segment != GROUND)
         {
            double mc[3];
            get_applied_force(time, input_data, i, point, force, free_torque);
#if STORE_EXTERNAL_FORCES_IN_GROUND_FRAME
            sdtrans(input_data->forces[i]->segment, force, GROUND, force);    // convert the force from segment frame to ground frame
            // want to output the point with respect to the segment ORIGIN not its mass center,
            // subtract the mass center to get the point wrt origin in segment frame, then convert to ground frame
            point[XX] -= sdm->body_segment[input_data->forces[i]->segment+1].mass_center[XX];
            point[YY] -= sdm->body_segment[input_data->forces[i]->segment+1].mass_center[YY];
            point[ZZ] -= sdm->body_segment[input_data->forces[i]->segment+1].mass_center[ZZ];
            convert_point(point, input_data->forces[i]->segment, GROUND);
#endif
            md.motiondata[col++][frame] = point[XX];
            md.motiondata[col++][frame] = point[YY];
            md.motiondata[col++][frame] = point[ZZ];
            md.motiondata[col++][frame] = force[XX];
            md.motiondata[col++][frame] = force[YY];
            md.motiondata[col++][frame] = force[ZZ];
         }
      }

      // Store the torques (always done in the frame in which they are applied).
      for (i = 0; i < input_data->num_torques; i++)
      {
         get_applied_torque(time, input_data, i, force);
         md.motiondata[col++][frame] = force[XX];
         md.motiondata[col++][frame] = force[YY];
         md.motiondata[col++][frame] = force[ZZ];
      }

      FREE_IFNOTNULL(fmf);
   }

#if CONTACT_DETECTION
	// contact forces  - show contact forces in blue, impacts in red
	if (params.outputOptions.contactForces == dpYes)
   {
		for (i = 0; i < sdm->num_body_segments; i++)
      {
			if (sdm->body_segment[i].contactable == dpYes)
         {
            sdvcopy(sdm->body_segment[i].impact_force, force);
            sdvcopy(sdm->body_segment[i].impact_point, point);
            if (EQUAL_WITHIN_TOLERANCE(VECTOR_MAGNITUDE(force), 0.0, 1e-4))
            {
               sum_forces_on_body(i - 1, point, force);
               // get point from segment origin
               sdvadd(sdm->body_segment[i].mass_center, point, point);
               md.motiondata[col++][frame] = 0.0;
               md.motiondata[col++][frame] = 0.0;
               md.motiondata[col++][frame] = 1.0;
            }
            else 
            {
               md.motiondata[col++][frame] = 1.0;
               md.motiondata[col++][frame] = 0.0;
               md.motiondata[col++][frame] = 0.0;
            }

            md.motiondata[col++][frame] = point[XX];
            md.motiondata[col++][frame] = point[YY];
            md.motiondata[col++][frame] = point[ZZ];
            md.motiondata[col++][frame] = force[XX];
            md.motiondata[col++][frame] = force[YY];
            md.motiondata[col++][frame] = force[ZZ];
			}
      }
   }
	// reset impact forces in body segment structure
	for (i = 0; i < sdm->num_body_segments; i++)
		make_vector(sdm->body_segment[i].impact_force, 0.0, 0.0, 0.0);
#endif

   // spring forces
   if (params.outputOptions.springForces == dpYes)
   {
#if COMBINE_SPRING_FORCES
		int *segmentSpringCount = (int*)simm_calloc(sdm->num_body_segments, sizeof(int));
		dpForceStruct *segmentSpringBin = (dpForceStruct*)simm_calloc(sdm->num_body_segments, sizeof(dpForceStruct));

		for (i = 0; i < sdm->num_springs; i++)
		{
			for (j = 0; j < 3; j++)
			{
				segmentSpringBin[sdm->spring[i].segment + 1].point[j] += sdm->spring[i].point[j];
				segmentSpringBin[sdm->spring[i].segment + 1].vector[j] += sdm->spring[i].force[j];
			}
			segmentSpringCount[sdm->spring[i].segment + 1]++;
		}

		for (i = 0; i < sdm->num_body_segments; i++)
		{
			if (segmentSpringCount[i] > 0)
			{
				int segmentIndex = GROUND + 1;

				for (j = 0; j < 3; j++)
					segmentSpringBin[i].point[j] /= segmentSpringCount[i];

  #if STORE_SPRING_FORCES_IN_CONTACT_FRAME
				// Find first spring which attaches to this segment and use its
				// floor segment to output the combined spring force in. This
				// assumes that all springs attaching to one segment use either
				// a single floor or multiple floors that are fixed to the same
				// segment.
				for (j = 0; j < sdm->num_springs; j++)
				{
					if (sdm->spring[j].segment + 1 == i)
					{
                  if (sdm->spring[j].floorContact)
                     segmentIndex = sdm->spring_floor[sdm->spring[j].contact_index].segment + 1;
                  else
                     segmentIndex = sdm->spring_surface[sdm->spring[j].contact_index].segment + 1;
						break;
					}
				}

            sdvcopy(segmentSpringBin[i].point, point);
            convert_point(point, i - 1, segmentIndex - 1);
            sdvadd(sdm->body_segment[segmentIndex].mass_center, point, point);
            sdtrans(i - 1, segmentSpringBin[i].vector, segmentIndex - 1, force);
  #else
            sdvadd(sdm->body_segment[i].mass_center, segmentSpringBin[i].point, point);
            sdvcopy(segmentSpringBin[i].vector, force);
  #endif
            md.motiondata[col++][frame] = point[XX];
            md.motiondata[col++][frame] = point[YY];
            md.motiondata[col++][frame] = point[ZZ];
            md.motiondata[col++][frame] = force[XX];
            md.motiondata[col++][frame] = force[YY];
            md.motiondata[col++][frame] = force[ZZ];
			}
		}

		FREE_IFNOTNULL(segmentSpringCount);
		FREE_IFNOTNULL(segmentSpringBin);
#else
  #if STORE_SPRING_FORCES_IN_CONTACT_FRAME
      for (i = 0; i < MIN(params.outputOptions.numSpringForces, sdm->num_springs); i++)
      {
         int segmentIndex = get_spring_contact_segment_sd_index(i);
         if (segmentIndex < -1)     //-1 is ground
            continue;
         sdvcopy(sdm->spring[i].point, point);
         convert_point(point, sdm->spring[i].sd_segment, segmentIndex);
         sdvadd(sdm->body_segment[segmentIndex + 1].mass_center, point, point);
         sdtrans(sdm->spring[i].sd_segment, sdm->spring[i].force, segmentIndex, force);
         md.motiondata[col++][frame] = point[XX];
         md.motiondata[col++][frame] = point[YY];
         md.motiondata[col++][frame] = point[ZZ];
         md.motiondata[col++][frame] = force[XX];
         md.motiondata[col++][frame] = force[YY];
         md.motiondata[col++][frame] = force[ZZ];
      }
  #else
      for (i = 0; i < MIN(params.outputOptions.numSpringForces, sdm->num_springs); i++)
      {
         // Output the point with respect to segment origin in the segment frame
         sdvadd(sdm->body_segment[sdm->spring[i].sd_segment + 1].mass_center, sdm->spring[i].point, point);
         md.motiondata[col++][frame] = point[XX];
         md.motiondata[col++][frame] = point[YY];
         md.motiondata[col++][frame] = point[ZZ];
         md.motiondata[col++][frame] = sdm->spring[i].force[XX];
         md.motiondata[col++][frame] = sdm->spring[i].force[YY];
         md.motiondata[col++][frame] = sdm->spring[i].force[ZZ];
      }
  #endif
#endif
   }

	// joint reaction force - force applied by the inboard body to the 
   // outboard body displayed at the origin of the outboard body
   for (i=0; i<params.outputOptions.jointReactionForces.num_items; i++)
   {
      index = params.outputOptions.jointReactionForces.index[i];
      md.motiondata[col++][frame] = com[XX];
      md.motiondata[col++][frame] = com[YY];
      md.motiondata[col++][frame] = com[ZZ];
      md.motiondata[col++][frame] = sdm->joint[index].force[XX];
      md.motiondata[col++][frame] = sdm->joint[index].force[YY];
      md.motiondata[col++][frame] = sdm->joint[index].force[ZZ];
   }

   // joint reaction torques
   for (i=0; i<params.outputOptions.jointReactionTorques.num_items; i++)
   {
      index = params.outputOptions.jointReactionTorques.index[i];
      md.motiondata[col++][frame] = sdm->joint[index].torque[XX];
      md.motiondata[col++][frame] = sdm->joint[index].torque[YY];
      md.motiondata[col++][frame] = sdm->joint[index].torque[ZZ];
   }

   // gencoord moments
   for (i=0; i<params.outputOptions.gencoordMoments.num_items; i++)
   {
      index = params.outputOptions.gencoordMoments.index[i];
      md.motiondata[col++][frame] = sdm->q[index].moment;
   }

   // gencoord powers
   for (i=0; i<params.outputOptions.gencoordPowers.num_items; i++)
   {
      index = params.outputOptions.gencoordPowers.index[i];
      md.motiondata[col++][frame] = sdm->q[index].moment * velocity_euler[index] * sdm->q[index].conversion;
   }

   // muscle moment arms
   for (i=0; i<params.outputOptions.muscleMomentArms_Muscles.num_items; i++)
   {
      int m_index = params.outputOptions.muscleMomentArms_Muscles.index[i];
      for (j=0; j<params.outputOptions.muscleMomentArms_Gencoords.num_items; j++)
      {
         index = params.outputOptions.muscleMomentArms_Gencoords.index[j];
         md.motiondata[col++][frame] = sdm->muscles[m_index].momentarms[index];
      }
   }

   // muscle moments
   for (i=0; i<params.outputOptions.muscleMoments_Muscles.num_items; i++)
   {
      int m_index = params.outputOptions.muscleMoments_Muscles.index[i];
      for (j=0; j<params.outputOptions.muscleMoments_Gencoords.num_items; j++)
      {
         index = params.outputOptions.muscleMoments_Gencoords.index[j];
         md.motiondata[col++][frame] = sdm->muscles[m_index].momentarms[index] * sdm->muscles[m_index].force;
      }
   }

   // total muscle moments
   for (i=0; i<params.outputOptions.totalMuscleMoments.num_items; i++)
   {
      index = params.outputOptions.totalMuscleMoments.index[i];
      for (j=0, torque=0.0; j<sdm->num_muscles; j++)
         torque += sdm->muscles[j].momentarms[index] * sdm->muscles[j].force;
      md.motiondata[col++][frame] = torque;
   }

	// mass centers positions
   for (i=0; i<params.outputOptions.massCenterPositions.num_items; i++)
   {
      if (STRINGS_ARE_EQUAL(params.outputOptions.massCenterPositions.name[i], "ground"))
      {
         // to hold mass center for the entire system
         double systemMass, systemMassCenter[3], systemInertia[3][3];
         sdsys(&systemMass, systemMassCenter, systemInertia);
         md.motiondata[col++][frame] = systemMassCenter[XX];
         md.motiondata[col++][frame] = systemMassCenter[YY];
         md.motiondata[col++][frame] = systemMassCenter[ZZ];
      }
      else
      {
         // The stored index is the sdm segment index, subtract one to get the SD/FAST segment index.
         index = params.outputOptions.massCenterPositions.index[i] - 1;
         sdpos(index, com, point);
         md.motiondata[col++][frame] = point[XX];
         md.motiondata[col++][frame] = point[YY];
         md.motiondata[col++][frame] = point[ZZ];
      }
   }

	// system energy
	if (params.outputOptions.systemEnergy == dpYes)
   {
      md.motiondata[col++][frame] = si.system_energy;
	}
	
   // mass center velocities
   for (i=0; i<params.outputOptions.massCenterVelocities.num_items; i++)
   {
      if (STRINGS_ARE_EQUAL(params.outputOptions.massCenterVelocities.name[i], "ground"))
      {
         double systemVelocity[3], systemMass, systemMassCenter[3], systemInertia[3][3];
         sdsys(&systemMass, systemMassCenter, systemInertia);
         memset(systemVelocity, 0, 3 * sizeof(double));
         for (j=0; j<sdm->num_body_segments; j++)
         {
            sdvel(j-1, com, vel);      // decrease by one to get the corresponding SD/FAST segment index
            systemVelocity[XX] += vel[XX] * sdm->body_segment[j].mass;
            systemVelocity[YY] += vel[YY] * sdm->body_segment[j].mass;
            systemVelocity[ZZ] += vel[ZZ] * sdm->body_segment[j].mass;
         }
         md.motiondata[col++][frame] = systemMassCenter[XX];
         md.motiondata[col++][frame] = systemMassCenter[YY];
         md.motiondata[col++][frame] = systemMassCenter[ZZ];
         md.motiondata[col++][frame] = systemVelocity[XX] / systemMass;
         md.motiondata[col++][frame] = systemVelocity[YY] / systemMass;
         md.motiondata[col++][frame] = systemVelocity[ZZ] / systemMass;
      }
      else
      {
         // The stored index is the sdm segment index, subtract one to get the SD/FAST segment index.
         index = params.outputOptions.massCenterVelocities.index[i] - 1;
         sdpos(index, com, point);
         md.motiondata[col++][frame] = point[XX];
         md.motiondata[col++][frame] = point[YY];
         md.motiondata[col++][frame] = point[ZZ];
         sdvel(index, com, vel);
         md.motiondata[col++][frame] = vel[XX];
         md.motiondata[col++][frame] = vel[YY];
         md.motiondata[col++][frame] = vel[ZZ];
      }
   }

   // moment residuals for optimized gencoords
   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0)
      append_optimized_gencoord_residuals(&md, &col, frame);

   // If there are any columns from the input motion to transfer to the output, do that now.
   if (input_data && col < md.num_elements)
   {
      int index1 = -1, index2 = -1;
      double percent = 0.0;

      // Find out which two frames of data the current time is between.
      if (time <= input_data->motiondata[input_data->time_column][0])
      {
         index1 = 0;
      }
      else if (time >= input_data->motiondata[input_data->time_column][input_data->num_frames-1])
      {
         index1 = input_data->num_frames-1;
      }
      else
      {
         for (i=1; i<input_data->num_frames; i++)
         {
            if (time <= input_data->motiondata[input_data->time_column][i])
            {
               index1 = i - 1;
               index2 = i;
               percent = (time - input_data->motiondata[input_data->time_column][index1]) /
                  (input_data->motiondata[input_data->time_column][index2] - input_data->motiondata[input_data->time_column][index1]);
               break;
            }
         }
      }

      // If necessary, use linear interpolation to get the appropriate value from the input motion.
      for (i=0; i<input_data->num_elements; i++)
      {
         if (transfer_element[i] == dpYes)
         {
            if (index2 == -1)
               md.motiondata[col++][frame] = input_data->motiondata[i][index1];
            else
               md.motiondata[col++][frame] = input_data->motiondata[i][index1] +
                  percent * (input_data->motiondata[i][index2] - input_data->motiondata[i][index1]);
         }
      }
   }

   // Now write the frame to the file.
   if (md.fp)
   {
      for (i = 0; i < md.num_elements; i++)
         fprintf(md.fp, "%21.16lf\t", md.motiondata[i][frame]);

      fprintf(md.fp, "\n");
      fflush(md.fp);
   }

   FREE_IFNOTNULL(state_euler);
   FREE_IFNOTNULL(velocity_euler);
   FREE_IFNOTNULL(acceleration_euler);

	return (check_for_sderror("Writing to motion file"));
}


/* GET_MOTION_INFO: get the information about the motion created (for SIMM)
 * 
 */
void get_motion_info(dpDataSetup* dataSetup)
{
   int i;
   ReturnCode rc;

   if (md.name)
      mstrcpy(&dataSetup->motionName, md.name);

   dataSetup->numFrames = md.num_frames;
   dataSetup->numElements = md.num_elements;
   dataSetup->elementNames = (char**)simm_malloc(dataSetup->numElements * sizeof(char*));
   dataSetup->time_column = md.time_column;
   dataSetup->x_column = md.x_column;
   if (md.x_column != md.time_column)
   {
      dataSetup->xStart = input_data->motiondata[input_data->x_column][0];
      dataSetup->xEnd = input_data->motiondata[input_data->x_column][input_data->num_frames-1];
   }

   for (i = 0; i < dataSetup->numElements; i++)
      mstrcpy(&dataSetup->elementNames[i], md.elementnames[i]);

   dataSetup->numEvents = 0;
   if (md.num_events > 0)
   {
      dataSetup->event = (dpMotionEvent*)simm_malloc(md.num_events * sizeof(dpMotionEvent));
      if (dataSetup->event)
      {
         for (i = 0; i < md.num_events; i++)
         {
            // Copy only the events that are within the simulation time range.  
            if (md.x_column != md.time_column && md.x_column >= 0)
            {
               // If the x_column ("percent") exists and is not the same as the time column, the motion has been normalized.
               // In this case, the xCoord of the event is not a time value but a percent value, so check that its value falls
               // within the first and last values in the percent column.
               if (md.event[i].xCoord >= input_data->motiondata[input_data->x_column][0] && md.event[i].xCoord <= input_data->motiondata[input_data->x_column][input_data->num_frames-1])
               {
                  memcpy(&dataSetup->event[dataSetup->numEvents], &md.event[i], sizeof(dpMotionEvent));
                  mstrcpy(&dataSetup->event[dataSetup->numEvents].name, md.event[i].name);
                  dataSetup->numEvents++;
               }
            }
            else
            {
               if (md.event[i].xCoord >= md.range_start && md.event[i].xCoord <= md.range_end)
               {
                  memcpy(&dataSetup->event[dataSetup->numEvents], &md.event[i], sizeof(dpMotionEvent));
                  mstrcpy(&dataSetup->event[dataSetup->numEvents].name, md.event[i].name);
                  dataSetup->numEvents++;
               }
            }
         }
         dataSetup->event = (dpMotionEvent*)simm_realloc(dataSetup->event, dataSetup->numEvents * sizeof(dpMotionEvent), &rc);
      }
   }
   else
   {
      dataSetup->event = NULL;
   }

   for (i = 0; i < 3; i++)
      dataSetup->eventColor[i] = md.event_color[i];

   dataSetup->num_foot_prints = md.num_foot_prints;
   if (dataSetup->num_foot_prints > 0)
   {
      dataSetup->foot_print = (dpFootPrint**)simm_malloc(dataSetup->num_foot_prints * sizeof(dpFootPrint*));
      if (dataSetup->foot_print)
      {
         for (i=0; i<dataSetup->num_foot_prints; i++)
         {
            dataSetup->foot_print[i] = (dpFootPrint*)simm_malloc(sizeof(dpFootPrint));
            memcpy(dataSetup->foot_print[i], md.foot_print[i], sizeof(dpFootPrint));
            mstrcpy(&dataSetup->foot_print[i]->name, md.foot_print[i]->name);
         }
      }
   }
   else
   {
      dataSetup->foot_print = NULL;
   }

   dataSetup->num_force_plates = md.num_force_plates;
   if (dataSetup->num_force_plates > 0)
   {
      dataSetup->force_plate = (dpForcePlate**)simm_malloc(dataSetup->num_force_plates * sizeof(dpForcePlate*));
      if (dataSetup->force_plate)
      {
         for (i=0; i<dataSetup->num_force_plates; i++)
         {
            dataSetup->force_plate[i] = (dpForcePlate*)simm_malloc(sizeof(dpForcePlate));
            memcpy(dataSetup->force_plate[i], md.force_plate[i], sizeof(dpForcePlate));
            mstrcpy(&dataSetup->force_plate[i]->name, md.force_plate[i]->name);
         }
      }
   }
   else
   {
      dataSetup->force_plate = NULL;
   }

   dataSetup->enforce_constraints = md.enforce_constraints;
   dataSetup->enforce_loops = md.enforce_loops;
   dataSetup->show_markers = md.show_markers;
   dataSetup->show_marker_trails = md.show_marker_trails;
   dataSetup->show_forces = md.show_forces;
   dataSetup->show_force_trails = md.show_force_trails;
   dataSetup->show_foot_prints = md.show_foot_prints;
   dataSetup->show_force_plates = md.show_force_plates;
   dataSetup->show_gencoord_moments = md.show_gencoord_moments;
   dataSetup->show_other_objects = md.show_other_objects;
   dataSetup->wrap = md.wrap;
}


/* GET_MOTION_DATA: This function fills in a data structure with the last frame of
 * data that was stored in the output motion structure.
 */
void get_motion_data(dpDataFrame* dataFrame)
{
   int i, frame;

   if (md.current_frame > 0)
      frame = md.current_frame - 1;
   else
      frame = 0;

   for (i = 0; i < dataFrame->numElements; i++)
      dataFrame->elements[i] = md.motiondata[i][frame];
}
