/*******************************************************************************

   PIPETOOLS.C

   Authors: Peter Loan
            Krystyne Blaikie

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains routines that implement various
      mathematical functions and other utilities. Every one operates
      exclusively on the formal parameter list, and so does not change any
      external variables or other general model or plot structures.

   Routines:
      mstrcpy                    : copies one string to another, mallocing first
      convert_point              : converts a point from one frame to another
      convert_point2             : alternate version of convert_point
      malloc_function            : mallocs space for a function
      realloc_function           : reallocs a function
      free_function              : frees a function
      make_zero_func             : makes a Y=0.0 2-point function
      init_function              : fills in part of a function
      setup_functions            : sets up functions in MotionData struct
      verify_assembly            : checks gencoord values in assembled model
      vector_intersects_matte    : checks intersection of vector and force matte
      apply_external_forces      : applies external forces and torques to segs
      get_applied_force          : gets current force value from MotionData struct
      get_applied_torque         : gets current torque value from MotionData struct
      set_fixed_gencoords        : prescribed Q,U,A of 'fixed' joints
      set_prescribed_gencoords   : prescribed Q,U,A of 'prescribed' gencoords
      set_prescribed_motion      : turns on/off prescribed motion for certain Qs
      calc_gencoord_moments      : calcs gencoord moments for a time step
      calc_initial_gencoord_moments : calcs gencoord moments for first two steps
      calc_next_moments          : calcs gencoord moments for next time step
      calc_corrective_moments    : calcs corrective moment to track motion
      add_moments_to_motion_data : adds data to MotionData to hold ID moments
      apply_gencoord_restraint_moments : applies moments to restrain joint motion
      get_muscle_param_value     : returns current value of named parameter
      check_for_sderror          : check if any SD/FAST errors have been posted
      init_joints                : initialize joint structure in sdm
      calc_system_energy         : calculates total system energy
      free_motion_data           : frees a MotionData struct

*******************************************************************************/

#include "universal.h"
#include "gcvspl.h"

/*************** DEFINES (for this file only) *********************************/


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/


/**************** GLOBAL VARIABLES (used in only a few files) *****************/


/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern char buffer[];
extern dpModelStruct* sdm;
extern dpSimulationParameters params;
extern dpBoolean isStandAlone;
extern char sd_error_msg[];

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static void set_prescribed_quaternions(double time, double q_quat[], double u_quat[], dpModelStruct* sdm, MotionData* data);
static void convert_quaternions_to_euler(double time, int frame, double input_state[], double previous_state[], double output_state[]);

/* MSTRCPY: this routine is like strcpy(), but it first mallocs space for
 * the copy of the string.
 */
ReturnCode mstrcpy(char* dest_str[], const char original_str[])
{
   if ((*dest_str = (char*)simm_malloc(STRLEN(original_str)*sizeof(char))) == NULL)
      return code_bad;

   (void)strcpy(*dest_str,original_str);
   return code_fine;
}


/* CONVERT_POINT: converts a point from one reference frame to another,
 * overwriting the original point.
 */
void convert_point(double pt[], int frame1, int frame2)
{
   int i;
   double vec[3], pt_g[3], origin_g[3], origin[] = {0.0,0.0,0.0};

   // Convert pt to the ground frame
   if (frame1 != GROUND)
   {
      sdpos(frame1,pt,pt_g);
   }
   else
   {
      for (i=0; i<3; i++)
         pt_g[i] = pt[i];
   }

   // Convert the origin of frame2 to the ground frame 
   if (frame2 != GROUND)
   {
      sdpos(frame2,origin,origin_g);
   }
   else
   {
      for (i=0; i<3; i++)
         origin_g[i] = origin[i];
   }

   // Form a vector from the origin of frame2 to pt,
   // expressed in the ground frame.
   for (i=0; i<3; i++)
      vec[i] = pt_g[i] - origin_g[i];

   // Convert vec from the ground frame to frame2,
   // storing the result in pt.
   if (frame2 != GROUND)
      sdtrans(GROUND,vec,frame2,pt);
   else
   {
      for (i=0; i<3; i++)
         pt[i] = vec[i];
   }
}


/* CONVERT_POINT2: converts a point from frame1 to frame2, storing result
 * in separate array.
 */
void convert_point2(int frame1, double p1[], int frame2, double p2[])
{
   double loc1[3], loc2[3], com[3] = {0.0, 0.0, 0.0}, cm2[3];

   sdpos(frame1, p1, loc1);
   sdpos(frame2, com, cm2);
   sdvsub(loc1, cm2, loc2);
   sdtrans(GROUND, loc2, frame2, p2);
}


/* MALLOC_FUNCTION: this routine mallocs space for a spline function. */
ReturnCode malloc_function(dpFunction* func, int numpoints)
{
   if (func == NULL)
      return code_bad;

   func->coefficient_array_size = 0;
   func->used = dpNo;
   func->status = dpFunctionUndefined;

   if ((func->x = (double*)simm_calloc(numpoints, sizeof(double))) == NULL)
      return code_bad;
   if ((func->y = (double*)simm_calloc(numpoints, sizeof(double))) == NULL)
      return code_bad;
   if ((func->b = (double*)simm_calloc(numpoints, sizeof(double))) == NULL)
      return code_bad;
   if ((func->c = (double*)simm_calloc(numpoints, sizeof(double))) == NULL)
      return code_bad;
   if ((func->d = (double*)simm_calloc(numpoints, sizeof(double))) == NULL)
      return code_bad;

   func->coefficient_array_size = numpoints;
   func->type = dpNaturalCubicSpline;
   func->used = dpYes;

   return code_fine;
}


/* REALLOC_FUNCTION: this routine reallocs the arrays in a spline structure. */
ReturnCode realloc_function(dpFunction* func, int size)
{
   ReturnCode rc1, rc2, rc3, rc4, rc5;

   func->x = (double*)simm_realloc(func->x,size*sizeof(double),&rc1);
   func->y = (double*)simm_realloc(func->y,size*sizeof(double),&rc2);
   func->b = (double*)simm_realloc(func->b,size*sizeof(double),&rc3);
   func->c = (double*)simm_realloc(func->c,size*sizeof(double),&rc4);
   func->d = (double*)simm_realloc(func->d,size*sizeof(double),&rc5);

   if (rc1 == code_bad || rc2 == code_bad || rc3 == code_bad || rc4 == code_bad || rc5 == code_bad)
      return code_bad;

   func->coefficient_array_size = size;

   return code_fine;
}


/* FREE_FUNCTION: frees a spline function. */
void free_function(dpFunction* func)
{
   if (func == NULL)
      return;

   FREE_IFNOTNULL(func->x);
   FREE_IFNOTNULL(func->y);
   FREE_IFNOTNULL(func->b);
   FREE_IFNOTNULL(func->c);
   FREE_IFNOTNULL(func->d);

   func->numpoints = func->coefficient_array_size = 0;
   func->used = dpNo;
   func->status = dpFunctionUndefined;
}


/* MAKE_ZERO_FUNC: makes a 2-point spline function from min_x to max_x
 * with Y values of 0.0.
 */
ReturnCode make_zero_func(dpFunction* sf, double min_x, double max_x)
{
   int num_points = 2;
   sf->type = dpStepFunction;
   sf->cutoff_frequency = 0.0;
   sf->coefficient_array_size = sf->numpoints = 0;
   sf->used = dpNo;
   sf->status = dpFunctionUndefined;
   sf->x = sf->y = sf->b = sf->c = sf->d = NULL;

   if ((sf->x = (double*)simm_malloc(num_points * sizeof(double))) == NULL)
      return code_bad;
   if ((sf->y = (double*)simm_malloc(num_points * sizeof(double))) == NULL)
      return code_bad;

   sf->coefficient_array_size = sf->numpoints = num_points;
   sf->used = dpYes;
   sf->status = dpFunctionSimmDefined;

   sf->x[0] = min_x;
   sf->x[1] = max_x;

   sf->y[0] = sf->y[1] = 0.0;

   return code_fine;
}


/* INIT_FUNCTION: this function mallocs space for the X and Y coordinates
 * of a spline function, and fills in the Y coords. It sets the function type
 * so that later on (in setup_functions()) the X coords can be filled in and
 * the spline coefficients calculated (depending on the type). If type = step_func,
 * then the X and Y coords are used as control points in a step function, not a
 * spline.
 */
ReturnCode init_function(dpFunction* sf, double Ycoords[], int numpoints, dpFunctionType type, double cutoffFrequency)
{
   int i;

   sf->type = type;
   sf->cutoff_frequency = cutoffFrequency;
   sf->coefficient_array_size = sf->numpoints = 0;
   sf->used = dpNo;
   sf->status = dpFunctionUndefined;
   sf->x = sf->y = sf->b = sf->c = sf->d = NULL;

   if ((sf->x = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
      return code_bad;
   if ((sf->y = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
      return code_bad;

   if (sf->type == dpNaturalCubicSpline)
   {
      if ((sf->b = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
         return code_bad;
      if ((sf->c = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
         return code_bad;
      if ((sf->d = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
         return code_bad;
   }
   else if (sf->type == dpGCVSpline)
   {
      if ((sf->c = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
         return code_bad;
   }
   else if (sf->type == dpLinearFunction)
   {
      if ((sf->b = (double*)simm_malloc(numpoints*sizeof(double))) == NULL)
         return code_bad;
   }

   sf->coefficient_array_size = sf->numpoints = numpoints;
   sf->used = dpYes;
   sf->status = dpFunctionSimmDefined;

   for (i = 0; i < numpoints; i++)
      sf->y[i] = Ycoords[i];

   return code_fine;
}


/* SETUP_FUNCTIONS: this function finishes initializing the spline structs
 * in a MotionData struct, after if it has been determined what the X axis is
 * (time, or frame index), and whether each data column should be splined or not.
 */
ReturnCode setup_functions(dpModelStruct* sdm, MotionData* data)
{
   int i, j, k;
   double *time_values;
   ReturnCode rc = code_fine;

   // Make a temporary array to hold the time coordinates for all data columns 
   time_values = (double*)simm_malloc(data->num_frames * sizeof(double));
   if (time_values == NULL)
      sim_message(exit_program, "Unable to allocate enough memory to read motion file.");

   if (data->time_column >= 0)
   {
      for (i = 0; i < data->num_frames; i++)
         time_values[i] = data->motiondata[data->time_column][i];
   }
   else
   {
      // If there is no time data, assume that the frames of data should be spread
      // evenly over the entire simulation period, with the first frame starting at
      // the start of the simulation, and the last frame *starting* at the end of
      // the simulation.
      if (EQUAL_WITHIN_ERROR((params.endTime - params.startTime), 0.0))
      {
         params.startTime = 0.0;
         params.endTime = data->num_frames - 1.0;
      }

      for (i = 0; i < data->num_frames; i++)
         time_values[i] = params.startTime + i * (params.endTime - params.startTime) / (data->num_frames - 1);
   }

   // Now finish setting up the spline structs 
   for (i = 0; i < sdm->nu; i++)
   {
      if (data->q_data[i])
      {
         for (j = 0; j < data->q_data[i]->numpoints; j++)
            data->q_data[i]->x[j] = time_values[j];
         if (calc_function_coefficients(data->q_data[i]) == code_bad)
            goto error;
      }

      if (data->u_data[i])
      {
         for (j = 0; j < data->u_data[i]->numpoints; j++)
            data->u_data[i]->x[j] = time_values[j];
         if (calc_function_coefficients(data->u_data[i]) == code_bad)
            goto error;
      }

      if (data->udot_data[i])
      {
         for (j = 0; j < data->udot_data[i]->numpoints; j++)
            data->udot_data[i]->x[j] = time_values[j];
         if (calc_function_coefficients(data->udot_data[i]) == code_bad)
            goto error;
      }

      if (data->q_moments[i])
      {
         for (j = 0; j < data->q_moments[i]->numpoints; j++)
            data->q_moments[i]->x[j] = time_values[j];
         if (calc_function_coefficients(data->q_moments[i]) == code_bad)
            goto error;
      }
   }

   for (i = 0; i < data->num_musc_excitations; i++)
   {
      for (j = 0; j < data->musc_excitations[i]->numpoints; j++)
         data->musc_excitations[i]->x[j] = time_values[j];
      if (calc_function_coefficients(data->musc_excitations[i]) == code_bad)
         goto error;
   }

   // For forces and torques, some of the individual components may not
   // have been defined in the input file. If a component is not defined,
   // initialize it to a two-point zero function. If it is defined, finish
   // setting it up.
   for (i = 0; i < data->num_forces; i++)
   {
      for (j = 0; j < 3; j++)
      {
         if (data->forces[i]->appl_point[j].status == dpFunctionUndefined)
         {
            rc = make_zero_func(&data->forces[i]->appl_point[j], time_values[0], time_values[data->num_frames]);
         }
         else
         {
            for (k = 0; k < data->forces[i]->appl_point[j].numpoints; k++)
               data->forces[i]->appl_point[j].x[k] = time_values[k];
            rc = calc_function_coefficients(&data->forces[i]->appl_point[j]);
         }
         if (rc == code_bad)
            goto error;
      }
      for (j = 0; j < 3; j++)
      {
         if (data->forces[i]->force_vec[j].status == dpFunctionUndefined)
         {
            rc = make_zero_func(&data->forces[i]->force_vec[j], time_values[0], time_values[data->num_frames]);
         }
         else
         {
            for (k = 0; k < data->forces[i]->force_vec[j].numpoints; k++)
               data->forces[i]->force_vec[j].x[k] = time_values[k];
            rc = calc_function_coefficients(&data->forces[i]->force_vec[j]);
         }
         if (rc == code_bad)
            goto error;
      }
      for (j = 0; j < 3; j++)
      {
         if (data->forces[i]->free_torque[j].status == dpFunctionUndefined)
         {
            rc = make_zero_func(&data->forces[i]->free_torque[j], time_values[0], time_values[data->num_frames]);
         }
         else
         {
            for (k = 0; k < data->forces[i]->free_torque[j].numpoints; k++)
               data->forces[i]->free_torque[j].x[k] = time_values[k];
            rc = calc_function_coefficients(&data->forces[i]->free_torque[j]);
         }
         if (rc == code_bad)
            goto error;
      }
   }

   for (i = 0; i < data->num_torques; i++)
   {
      for (j = 0; j < 3; j++)
      {
         if (data->torques[i]->torque_vec[j].status == dpFunctionUndefined)
         {
            rc = make_zero_func(&data->torques[i]->torque_vec[j], time_values[0], time_values[data->num_frames]);
         }
         else
         {
            for (k = 0; k < data->torques[i]->torque_vec[j].numpoints; k++)
               data->torques[i]->torque_vec[j].x[k] = time_values[k];
            rc = calc_function_coefficients(&data->torques[i]->torque_vec[j]);
         }
         if (rc == code_bad)
            goto error;
      }
   }

   goto cleanup;

error:
   sim_message(recover, "Error setting up spline functions for kinetics data.");
   rc = code_bad;

cleanup:
   free(time_values);

   return rc;
}


/* VERIFY_ASSEMBLY: this routine checks to see if the assembled SD/FAST model has
 * any Qs (gencoords) that violate their allowable ranges of motion. A gencoord
 * is allowed to violate its range of motion if a restraint function is defined
 * for it (the assumption being that the restraint torque will eventually pull
 * the gencoord back into range.
 */
ReturnCode verify_assembly(dpModelStruct* sdm, double state_quat[])
{
   int i;
   dpBoolean valid = dpYes;
   double *state_euler;

   state_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   copy_vector(state_quat, state_euler, sdm->nq);

   // convert the state from quaternions to Euler angles
   convert_state_from_quaternion(0, state_quat, NULL, NULL, state_euler, NULL, NULL, NULL);

   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type != dpUnconstrainedQ)
         continue;
      if ((sdm->q[i].restraint_function == NULL) &&
         ((state_euler[i]*sdm->q[i].conversion < sdm->q[i].range_start && sdm->q[i].min_restraint_function == NULL) ||
         (state_euler[i]*sdm->q[i].conversion > sdm->q[i].range_end && sdm->q[i].max_restraint_function == NULL)))
      {
         sim_message(none, "%s out of range: value = %lf, range = %lf to %lf.",
            sdm->q[i].name, state_euler[i]*sdm->q[i].conversion, sdm->q[i].range_start, sdm->q[i].range_end);
         valid = dpNo;
      }
   }

   FREE_IFNOTNULL(state_euler);

   if (valid == dpNo)
   {
      sim_message(none, "Assembly invalid.");
      return code_bad;
   }

   return code_fine;
}


/* VECTOR_INTERSECTS_MATTE: This function checks for intersection of a vector
 * and a 'force matte,' which is a planar polygon stored in a ForceMatte struct.
 */
dpBoolean vector_intersects_matte(double pt[], double vec[], dpForceMatte* fm, double inter[])
{
   int i;
   double pt2[3], t;

   pt2[0] = pt[0] + vec[0];
   pt2[1] = pt[1] + vec[1];
   pt2[2] = pt[2] + vec[2];

   // Check to see if the vector intersects any of the polygons in the force matte.
   // intersect_line_plane() is used to get the point of intersection between the
   // vector and the plane of the polygon, and then this point is checked to see if
   // it is inside the boundaries of the polygon. Return the first intersection found.
   for (i = 0; i < fm->ph->num_polygons; i++)
   {
      if (intersect_line_plane(pt, pt2, fm->ph->polygon[i].normal, fm->ph->polygon[i].d, inter, &t) == dpYes &&
         point_in_polygon3D(inter, fm->ph, i) == dpYes)
      {
         return dpYes;
      }
   }

   return dpNo;
}


/* APPLY_EXTERNAL_FORCES: This routine applies external, user-defined forces
 * and torques. A typical example of an external force is a ground-reaction.
 * force. The forces and torques are specified in a data file and are loaded
 * into a MotionData structure.  External forces in the struct are expected
 * to be with respect to the body segment origins, not their mass centers.
 * Therefore, the mass center coordinates are subtracted from the points
 * of application when the forces are applied.
 */
void apply_external_forces(double time, double y[], MotionData* data)
{
   int i, j, axis, first, index, haveTorques = 0;
   double appl_point[3], force_vec[3], free_torque[3], torque_vec[3];
   double intersectpt[3], appl_point2[3], force_vec2[3], free_torque2[3];
   double *moments_euler = NULL, *moments_quat = NULL;

   // Apply point forces to the body segments
   for (i = 0; i < data->num_forces; i++)
   {
      get_applied_force(time, data, i, appl_point, force_vec, free_torque);
      sdvsub(appl_point, sdm->body_segment[data->forces[i]->segment + 1].mass_center, appl_point);

      // Forces applied to the ground segment have no effect on the simulation.
      // So assume that all ground forces should be applied to other body segments
      // using the force matte mechanism. The force vector at each time step is
      // checked for intersection with an array of force mattes (polygons defined in
      // PolyhedronStructs). If the force vector intersects a segment's force matte,
      // apply the force to that segment.
      // Forces specified in the ground frame have a free-torque component, which is
      // transformed into the force matte segment and applied. Forces applied to other
      // body segments do not have a free-torque component.
      if (data->forces[i]->segment == GROUND)
      {
         // Intersect the force vector with the force mattes. 
         for (j = 0; j < sdm->num_force_mattes; j++)
         {
            sdtrans(GROUND, force_vec, sdm->force_matte[j].segment, force_vec2);
            sdtrans(GROUND, free_torque, sdm->force_matte[j].segment, free_torque2);
            sdvcopy(appl_point, appl_point2);
            convert_point(appl_point2, GROUND, sdm->force_matte[j].segment);
            if (vector_intersects_matte(appl_point2, force_vec2, &sdm->force_matte[j], intersectpt) == dpYes)
            {
               // Apply the force and free torque, and break so that no other mattes are checked.
               sdpointf(sdm->force_matte[j].segment, appl_point2, force_vec2);
               sdbodyt(sdm->force_matte[j].segment, free_torque2);
               break;
            }
         }
      }
      else
      {
         sdpointf(data->forces[i]->segment, appl_point, force_vec);
      }
   }

   // Apply torques to the body segments.
   for (i = 0; i < data->num_torques; i++)
   {
      get_applied_torque(time, data, i, torque_vec);
      sdbodyt(data->torques[i]->segment, torque_vec);
   }

   // Apply moments/forces to the Qs.
   moments_euler = (double *)simm_calloc(sdm->nu, sizeof(double));
   moments_quat = (double *)simm_calloc(sdm->nu, sizeof(double));

   // Put all moments that need to be applied in an Euler-based array.
   for (i = 0; i < sdm->nu; i++)
   {
      sdm->q[i].moment = 0.0;

      if (data->q_moments[i])
      {
         moments_euler[sdm->q[i].state_index] = interpolate_function(time, data->q_moments[i], zeroth, 1.0, 0.0);
         haveTorques = 1;
      }
   }

   // Convert the moments from Euler to quaternion representation.
   if (haveTorques)
   {
      convert_moment_array_to_quaternion(y, moments_euler, moments_quat);

      // Store the moments in the Q array and apply them
      for (i = 0; i < sdm->nu; i++)
      {
         sdm->q[i].moment = moments_quat[sdm->q[i].state_index];
         axis = sdm->q[i].axis;
         if (sdm->q[i].is_part_of_quaternion)
         {
            axis = abs(axis) - 1;
            if (sdm->joint[sdm->q[i].joint].jnt_type == dpFree)
               axis += 3;
         }
         sdhinget(sdm->q[i].joint, axis, sdm->q[i].moment);
      }
   }

   FREE_IFNOTNULL(moments_euler);
   FREE_IFNOTNULL(moments_quat);

   if (check_for_sderror("Apply_external_force"))
      printf("Apply external forces error\n");
}


/* GET_APPLIED_FORCE: this function gets the current value of force in a MotionData struct.
 * It uses the current simulation time to get the value, as the X coordinate of the force
 * spline functions is time. The free torque component is typically used only for ground-reaction
 * forces.
 */
void get_applied_force(double time, MotionData* data, int frc_index, double appl_point[], double force_vec[], double free_torque[])
{
   appl_point[XX] = interpolate_function(time, &data->forces[frc_index]->appl_point[XX], zeroth, 1.0, 0.0);
   appl_point[YY] = interpolate_function(time, &data->forces[frc_index]->appl_point[YY], zeroth, 1.0, 0.0);
   appl_point[ZZ] = interpolate_function(time, &data->forces[frc_index]->appl_point[ZZ], zeroth, 1.0, 0.0);
   force_vec[XX] = interpolate_function(time, &data->forces[frc_index]->force_vec[XX], zeroth, 1.0, 0.0);
   force_vec[YY] = interpolate_function(time, &data->forces[frc_index]->force_vec[YY], zeroth, 1.0, 0.0);
   force_vec[ZZ] = interpolate_function(time, &data->forces[frc_index]->force_vec[ZZ], zeroth, 1.0, 0.0);
   free_torque[XX] = interpolate_function(time, &data->forces[frc_index]->free_torque[XX], zeroth, 1.0, 0.0);
   free_torque[YY] = interpolate_function(time, &data->forces[frc_index]->free_torque[YY], zeroth, 1.0, 0.0);
   free_torque[ZZ] = interpolate_function(time, &data->forces[frc_index]->free_torque[ZZ], zeroth, 1.0, 0.0);
}


/* GET_APPLIED_TORQUE: this function gets the current value of torque in a MotionData struct.
 * It uses the current simulation time to get the value, as the X coordinate of the torque
 * spline functions is time.
 */
void get_applied_torque(double time, MotionData* data, int trq_index, double torque_vec[])
{
   torque_vec[XX] = interpolate_function(time, &data->torques[trq_index]->torque_vec[XX], zeroth, 1.0, 0.0);
   torque_vec[YY] = interpolate_function(time, &data->torques[trq_index]->torque_vec[YY], zeroth, 1.0, 0.0);
   torque_vec[ZZ] = interpolate_function(time, &data->torques[trq_index]->torque_vec[ZZ], zeroth, 1.0, 0.0);
}


/* SET_FIXED_GENCOORDS: This routine prescribes the position, velocity, and
 * acceleration of the fixed gencoords in the model. Fixed gencoords are
 * used to implement joints in SD/FAST that should be fixed, but cannot be
 * modeled that way. Instead, a fixed joint is modeled as a pin joint whose
 * motion is prescribed to be zero.  Do not call sdprespos for quaternions.
 */
void set_fixed_gencoords(dpModelStruct* sdm)
{
   int i, axis;

   // for all fixed gencoords:
   //   pos = initial value
   //   vel = 0.0
   //   acc = 0.0
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == dpFixedQ)
      {
         axis = sdm->q[i].axis;
         if (sdm->q[i].is_part_of_quaternion)
         {
            axis = abs(sdm->q[i].axis) - 1;
            if (sdm->joint[sdm->q[i].joint].jnt_type == dpFree)
               axis += 3;
         }
         if (!sdm->q[i].is_part_of_quaternion)
            sdprespos(sdm->q[i].joint, axis, sdm->q[i].initial_value / sdm->q[i].conversion);
         sdpresvel(sdm->q[i].joint, axis, 0.0);
         sdpresacc(sdm->q[i].joint, axis, 0.0);
      }
   }
}


/* SET_PRESCRIBED_GENCOORDS: This routine prescribes the position, velocity, and
 * acceleration of the prescribed gencoords in the model, using data from the input motion
 * file. Position data must be explicitly specified (though it can be splined or not) for
 * non-quaternion gencoords.
 * Velocity can either be specified explicitly, or calculated from position (if position
 * data is splined). Acceleration can be specified explicitly or calculated from position
 * or velocity.
 * The input state is in quaternion frame.
 */
void set_prescribed_gencoords(double time, double q_quat[], double u_quat[], dpModelStruct* sdm, MotionData* data)
{
   int i, axis;
   double Q, U, A;

   // set the position, velocity, and acceleration for all non-quaternion prescribed gencoords
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].is_part_of_quaternion)
         continue;
      axis = sdm->q[i].axis;
      if (sdm->q[i].type == dpPrescribedQ)
      {
         if (data && data->q_data[i])
         {
            Q = interpolate_function(time, data->q_data[i], zeroth, 0.0, 0.0) / sdm->q[i].conversion;

            if (data->u_data[i])
               U = interpolate_function(time, data->u_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->q_data[i]->type != dpStepFunction)
               U = interpolate_function(time, data->q_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
            else
               U = 0.0;

            if (data->udot_data[i])
               A = interpolate_function(time, data->udot_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->u_data[i] && data->u_data[i]->type != dpStepFunction)
               A = interpolate_function(time, data->u_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->q_data[i]->type != dpStepFunction)
               A = interpolate_function(time, data->q_data[i], second, 1.0, 0.0) / sdm->q[i].conversion;
            else
               A = 0.0;

            sdprespos(sdm->q[i].joint, axis, Q);
            sdpresvel(sdm->q[i].joint, axis, U);
            sdpresacc(sdm->q[i].joint, axis, A);
         }
         else
         {
            // If the Q is prescribed but there is no data for it in the kinetics
            // file, then prescribe it to remain fixed at its initial value. This
            // is the mechanism used to handle Qs which are locked in SIMM.
            sdprespos(sdm->q[i].joint, axis, sdm->q[i].initial_value / sdm->q[i].conversion);
            sdpresvel(sdm->q[i].joint, axis, 0.0);
            sdpresacc(sdm->q[i].joint, axis, 0.0);
         }
      }
   }

   if (data)
      set_prescribed_quaternions(time, q_quat, u_quat, sdm, data);
            
   check_for_sderror("SDUMOTION");
}

/* SET_PRESCRIBED_QUATERNIONS: This routine prescribes the velocity, and acceleration
 * of the prescribed gencoords that are quaternions, using data from the input motion
 * file. Position data cannot be prescribed for quaternions.
 * The Euler angle derivatives are calculated from the position data, then converted
 * to the necessary angular velocities and angular accelerations.
 */
static void set_prescribed_quaternions(double time, double q[], double u[], dpModelStruct* sdm, MotionData* data)
{
   int i, j, axis, first_index, state_index[3], axes[3];
   dpBoolean valid = dpYes;
   double q_euler[3], u_euler[3], a_euler[3], q_quat[4], u_quat[3], a_quat[3];

   for (i = 0; i < sdm->num_joints; i++)
   {
      if (sdm->joint[i].jnt_type == dpBall || sdm->joint[i].jnt_type == dpFree)
      {
         valid = dpYes;
         first_index = sdm->joint[i].first_q;
         if (sdm->joint[i].jnt_type == dpFree)
            first_index += 3;

         // all THREE gencoords must be prescribed and Q data must exist for them
         for (j = 0; j < 3; j++)
         {
            axes[j] = sdm->q[first_index+j].axis;
            state_index[j] = sdm->q[first_index+j].state_index;
            if (sdm->q[first_index+j].type != dpPrescribedQ)
               valid = dpNo;
            if (data->q_data[state_index[j]] == NULL)
               valid = dpNo;
         }   

         if (!valid)
            continue;


         // determine Euler angles and derivatives
         for (j = 0; j < 3; j++)
         {
            q_euler[j] = interpolate_function(time, data->q_data[state_index[j]], zeroth, 0.0, 0.0) / sdm->q[first_index+j].conversion;

            if (data->u_data[state_index[j]])
               u_euler[j] = interpolate_function(time, data->u_data[state_index[j]], zeroth, 1.0, 0.0) / sdm->q[first_index+j].conversion;
            else if (data->q_data[state_index[j]]->type != dpStepFunction)
               u_euler[j] = interpolate_function(time, data->q_data[state_index[j]], first, 1.0, 0.0) / sdm->q[first_index+j].conversion;
            else
               u_euler[j] = 0.0;

            if (data->udot_data[state_index[j]])
               a_euler[j] = interpolate_function(time, data->udot_data[state_index[j]], zeroth, 1.0, 0.0) / sdm->q[first_index+j].conversion;
            else if (data->u_data[state_index[j]] && data->u_data[state_index[j]]->type != dpStepFunction)
               a_euler[j] = interpolate_function(time, data->u_data[state_index[j]], first, 1.0, 0.0) / sdm->q[first_index+j].conversion;
            else if (data->q_data[state_index[j]]->type != dpStepFunction)
               a_euler[j] = interpolate_function(time, data->q_data[state_index[j]], second, 1.0, 0.0) / sdm->q[first_index+j].conversion;
            else
               a_euler[j] = 0.0;
         }

         //convert from Euler-space to quaternion-space
         convert_angles_to_quaternions(q_euler, u_euler, a_euler, q_quat, u_quat, a_quat, sdm->joint[i].euler_rots, axes);

         // prescribe the velocities and accelerations
         for (j = 0; j < 3; j++)
         {
            // if the joint type is Free (6 dof with quaternion) add 3 to the axis so it specifies 
            // the joint axis to be prescribed.  Value stored is the rotation order.
            if (sdm->joint[i].jnt_type == dpFree)
            {
               sdpresvel(sdm->q[first_index+j].joint, j+3, u_quat[j]);
               sdpresacc(sdm->q[first_index+j].joint, j+3, a_quat[j]);
            }
            else
            {
               sdpresvel(sdm->q[first_index+j].joint, j, u_quat[j]);
               sdpresacc(sdm->q[first_index+j].joint, j, a_quat[j]);  
            }
         }
      }
   }
}


/* SET_PRESCRIBED_MOTION: This routine turns on/off prescribed motion for all 
 * the Qs of the specified type.
 */
void set_prescribed_motion(dpModelStruct* sdm, dpQType type, int value)
{
   int i, pres, axis;

   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == type)
      {
         // Get the current state (prescribed/unprescribed) of the Q, and change it only if necessary
         axis = sdm->q[i].axis;
         // if the joint type is Free (6 dof with quaternion) add 3 to the axis so it specifies 
         // the joint axis to be prescribed.  Value stored is the rotation order
         if (sdm->q[i].is_part_of_quaternion)
         {
            axis = abs(sdm->q[i].axis) - 1;
            if (sdm->joint[sdm->q[i].joint].jnt_type == dpFree)
               axis += 3;
         }

         sdgetpres(sdm->q[i].joint, axis, &pres);
         if (value != pres)
            sdpres(sdm->q[i].joint, axis, value);
         if (check_for_sderror("SET_PRESCRIBED_MOTION") == 19)
            fprintf(stderr,"Unable to change prescribed state of %s\n", sdm->q[i].name);
      }
   }
}


/* CALC_GENCOORD_MOMENTSS: this function calculates the moments on the Qs that result from
 * the user-specified prescribed motion. It first applies all user-defined forces
 * (muscles, restraint moments, spring forces, body forces and torques, etc.) and then
 * calculates the *additional* forces/moments needed to produce the accelerations in the
 * prescribed motion. If you prescribe the motion of the system while applying muscle
 * forces or gencoord moments that exactly produce that motion, this function will
 * calculate and return gencoord moments of zero.
 * The calculated gencoord moments are written to the motion file if the output_gencoord_moments
 * parameter is set to "dpYes" in the parameters file.
 */
void calc_gencoord_moments(double time, double state_quat[], double dstate_quat[])
{
   int i, rank, info[50], *multmap;
   double *mults, *mults2, *moments_quat, *moments_euler, *state_euler;

   // Allocate space 
   mults = (double*)simm_calloc(sdm->num_constraints, sizeof(double));
   mults2 = (double*)simm_malloc(sdm->num_constraints*sizeof(double));
   multmap = (int*)simm_malloc(sdm->num_constraints*sizeof(int));
   moments_quat = (double*)simm_malloc(sdm->nu*sizeof(double));
   moments_euler = (double*)simm_malloc(sdm->nu*sizeof(double));

   // Get the multipliers for the current system state. These multipliers are the
   // torques that were applied to the Qs at the end of the last integration step.
   sdmult(mults2, &rank, multmap);

   // Form a new multiplier array with non-zero multipliers for the constrained
   // Qs and the loop constraints (if there are any closed loops). This is done
   // so that sdfulltrq() will return zero torques for those Qs, and non-zero
   // torques for the unconstrained ones (and the fixed ones, but those are not
   // output).
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].constraint_num != -1)
      {
         // get the index of the multiplier for this Q's constraint (stored in info[1])
         sdcons(sdm->q[i].constraint_num, info);
         
         // copy this multiplier to the new array
         mults[info[1]] = mults2[info[1]];
      }
   }

   // Now copy over the multipliers for the loop constraints.
   if (sdm->num_closed_loops > 0)
   {
      int start = sdm->num_constraints - sdm->num_user_constraints -
                  6 * sdm->num_closed_loops;
      int end = sdm->num_constraints - sdm->num_user_constraints;

      for (i = start; i < end; i++)
         mults[i] = mults2[i];
   }

   // Reset the system to the state for which you want the torques.
   sdstate(time, state_quat, &state_quat[sdm->nq]);

   // Apply all user forces, including external forces, muscle forces, and spring forces.
   // This must be done because these forces are included in the multipliers. If you do
   // not include them, you will get non-zero forces/torques at the constrained Qs.
   sduforce(time, state_quat, &state_quat[sdm->nq]);

   // Get the joint torques.
   sdfulltrq(&dstate_quat[sdm->nq], mults, moments_quat);

   // returned torques are in quaternion frame, convert to Euler for output
   state_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   convert_state_from_quaternion(time, state_quat, NULL, NULL, state_euler, NULL, NULL, NULL);
   convert_moment_array_from_quaternion(state_euler, moments_quat, moments_euler);

   // Store the torques in the SDM structure
   for (i = 0; i < sdm->nu; i++)
      sdm->q[i].moment = moments_euler[i] * DSIGN(sdm->q[i].conversion);

   free(mults);
   free(mults2);
   free(multmap);
   free(moments_quat);
   free(moments_euler);
   free(state_euler);

   check_for_sderror("CALC_GENCOORD_MOMENTS");
}


/* CALC_INITIAL_GENCOORD_MOMENTS: at the start of an inverse simulation, this function
 * calls calc_gencoord_moments() for the first and second time steps and stores the
 * moments in the q_moments[] array in the MotionData structure. This is done so
 * that the simulation knows what moments to apply during the first integration
 * interval, before any corrective moments have been calculated.
 */
void calc_initial_gencoord_moments(double time, double y[], double dy[], MotionData* data)
{
   int i, state_index, status;
   double param[1] = {-1.0};
   double *initialMoments, *savedY, *savedDy;
   double *q_euler, *u_euler, *a_euler, *q_quat, *u_quat, *a_quat;

   savedY = (double *)simm_malloc(sdm->neq * sizeof(double));
   savedDy = (double *)simm_malloc(sdm->neq * sizeof(double));
   q_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   u_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   a_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   q_quat = (double *)simm_malloc(sdm->nq * sizeof(double));
   u_quat = (double *)simm_malloc(sdm->nu * sizeof(double));
   a_quat = (double *)simm_malloc(sdm->nu * sizeof(double));

   initialMoments = (double *)simm_malloc(sdm->nu * sizeof(double));
   for (i = 0; i < sdm->neq; i++)
   {
      savedY[i] = y[i];
      savedDy[i] = dy[i];
   }

   // turn on prescribed motion and turn off NewInverseSimulation to calculate moments
   // at first and second time steps
   set_prescribed_motion(sdm, dpPrescribedQ, 1);
   sdm->newInverseSimulation = dpNo;

   // First calculate the moments for the initial time step and store them
   calc_derivatives(time, y, dy, param, &status);
   calc_gencoord_moments(time, y, dy);
   for (i = 0; i < sdm->nu; i++)
      initialMoments[i] = sdm->q[i].moment;

   // reset the moments so they don't affect second calculation ????
   for (i = 0; i < sdm->nu; i++)
      sdm->q[i].moment = 0.0;

   // Now calculate the moments for the second time step.

   // ?? what happens to gencoords that are NOT prescribed???
   // Set the state to the desired state (from input motion file)
   // determine any prescribed gencoord values
   // do the constrainedQs after the prescribedQs since the independent Q values must be calculated first
   clear_vector(q_euler, sdm->nu);
   clear_vector(u_euler, sdm->nu);
   clear_vector(a_euler, sdm->nu);
   for (i = 0; i < sdm->nu; i++)
   {
      state_index = sdm->q[i].state_index;

      if (sdm->q[i].type == dpPrescribedQ)
      {
         if (data->q_data[i])
         {
            q_euler[state_index] = interpolate_function(time + params.stepSize, data->q_data[i], zeroth, 0.0, 0.0) / sdm->q[i].conversion;

            if (data->u_data[i])
               u_euler[state_index] = interpolate_function(time + params.stepSize, data->u_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->q_data[i]->type != dpStepFunction)
               u_euler[state_index] = interpolate_function(time + params.stepSize, data->q_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
            else
               u_euler[state_index] = 0.0;

            if (data->udot_data[i])
               a_euler[state_index] = interpolate_function(time + params.stepSize, data->udot_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->u_data[i] && data->u_data[i]->type != dpStepFunction)
               a_euler[state_index] = interpolate_function(time + params.stepSize, data->u_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->q_data[i]->type != dpStepFunction)
               a_euler[state_index] = interpolate_function(time + params.stepSize, data->q_data[i], second, 1.0, 0.0) / sdm->q[i].conversion;
            else
               a_euler[state_index] = 0.0;
         }
      }
      else if (sdm->q[i].type == dpFixedQ)
      {
         q_euler[state_index] = sdm->q[i].initial_value  / sdm->q[i].conversion;
         u_euler[state_index] = 0.0;
         a_euler[state_index] = 0.0;
      }
   }
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         int q_ind = sdm->q[i].q_ind;
         double q_ind_value, u_ind_value, a_ind_value;
         state_index = sdm->q[i].state_index;
         
         q_ind_value = q_euler[q_ind] * sdm->q[q_ind].conversion;
         q_euler[state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, zeroth, 0.0, 0.0) / sdm->q[i].conversion;
         
         u_ind_value = u_euler[q_ind] * sdm->q[q_ind].conversion;
         u_euler[state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, first, u_ind_value, 0.0)/sdm->q[i].conversion;
         
         a_ind_value = a_euler[q_ind] * sdm->q[q_ind].conversion;
         a_euler[state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, second, u_ind_value, a_ind_value)/sdm->q[i].conversion;
      }
   }

   // convert euler params to quaternions
   // first, copy over all values
   for (i = 0; i < sdm->nu; i++)
   {
      q_quat[i] = q_euler[i];
      u_quat[i] = u_euler[i];
      a_quat[i] = a_euler[i];
   }
   // how to check that all 3 DOFs of quaternion joints were prescribed
   {
      int j, first_index, state_index[3], axes[3];
      dpBoolean valid;
      double qe[3], qq[4], ue[3], uq[3], ae[3], aq[3];

      for (i = 0; i < sdm->num_joints; i++)
      {
         if (sdm->joint[i].jnt_type == dpBall || sdm->joint[i].jnt_type == dpFree)
         {
            first_index = sdm->joint[i].first_q;
            if (sdm->joint[i].jnt_type == dpFree)
               first_index += 3;

            state_index[0] = sdm->q[first_index].state_index;
            state_index[1] = sdm->q[first_index+1].state_index;
            state_index[2] = sdm->q[first_index+2].state_index;
            for (j = 0; j < 3; j++)
            {
               axes[j] = sdm->q[first_index+j].axis;
               qe[j] = q_euler[state_index[j]];
               ue[j] = u_euler[state_index[j]];
               ae[j] = a_euler[state_index[j]];
            }

            convert_angles_to_quaternions(qe, ue, ae, qq, uq, aq, sdm->joint[i].euler_rots, axes);

            // copy back into state array
            for (j = 0; j < 3; j++)
            {
               q_quat[state_index[j]] = qq[j];
               u_quat[state_index[j]] = uq[j];
               a_quat[state_index[j]] = aq[j];
            }
            q_quat[sdm->joint[i].quat] = qq[3];
         }
      }
   }
   // set the new state
   for (i = 0; i < sdm->nu; i++)
   {
      state_index = sdm->q[i].state_index;
      y[state_index] = q_quat[state_index];
      y[i + sdm->nq] = u_quat[state_index];
      dy[state_index] = u_quat[state_index];
   }
   for (i = sdm->nu; i < sdm->nq; i++)
      y[i] = q_quat[i];

   sdstate(time + params.stepSize, y, &y[sdm->nq]);

   
   calc_derivatives(time + params.stepSize, y, dy, param, &status);
   calc_gencoord_moments(time + params.stepSize, y, dy);

   // Store the moments in the motion data structure.
   for (i = 0; i < sdm->nu; i++)
   {
      if (data->q_moments[i])
      {
         dpFunction* sf = data->q_moments[i];

         sf->y[1] = sdm->q[i].moment;
         sf->y[0] = initialMoments[i];
         sdm->q[i].moment = initialMoments[i];
      }
   }

   // determine the b[0] coefficients
   for (i = 0; i < sdm->nu; i++)
   {
      if (data->q_moments[i])
      {
         dpFunction* sf = data->q_moments[i];
         sf->b[0] = (sf->y[1] - sf->y[0]) / (sf->x[1] - sf->x[0]);
      }
   }

   // reset the original state
   sdstate(time, savedY, savedDy);
   set_prescribed_motion(sdm, dpPrescribedQ, 0);
   sdm->newInverseSimulation = dpYes;

   {
      int err;
      err = check_for_sderror("calc initial moments:");

   }
   //delete malloced structs
   FREE_IFNOTNULL(savedY);
   FREE_IFNOTNULL(savedDy);
   FREE_IFNOTNULL(initialMoments);
   FREE_IFNOTNULL(q_euler);
   FREE_IFNOTNULL(u_euler);
   FREE_IFNOTNULL(a_euler);
   FREE_IFNOTNULL(q_quat);
   FREE_IFNOTNULL(u_quat);
   FREE_IFNOTNULL(a_quat);
}


/* CALC_NEXT_GENCOORD_MOMENTS: similar to calc_gencoord_moments(), but calculates
 * the moments for the next time step after the current one. This is done
 * so that the inverse simulation knows what moments to apply for the
 * next integration period.
 */
void calc_next_gencoord_moments(double time, double y[], double dy[], MotionData* data)
{
   int i, j, status, state_index;
   double *savedState, *savedDy;
   double param[1] = {-1.0};
   double *q_euler, *u_euler, *a_euler, *q_quat, *u_quat, *a_quat;

   q_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   u_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   a_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   q_quat = (double *)simm_malloc(sdm->nq * sizeof(double));
   u_quat = (double *)simm_malloc(sdm->nq * sizeof(double));
   a_quat = (double *)simm_malloc(sdm->nq * sizeof(double));

   // Save the current state vector so you can restore it later.
   savedState = (double*)simm_malloc(sdm->neq*sizeof(double));
   savedDy = (double*)simm_malloc(sdm->neq*sizeof(double));
   memcpy(savedState, y, sdm->neq * sizeof(double));
   memcpy(savedDy, dy, sdm->neq * sizeof(double));

   // Enable prescribed motion for the prescribed gencoords,
   // then prescribe them according to the data from the time
   // step after the current one.
   sdm->newInverseSimulation = dpNo;
   set_prescribed_motion(sdm, dpPrescribedQ, 1);

   clear_vector(q_euler, sdm->nu);
   clear_vector(u_euler, sdm->nu);
   clear_vector(a_euler, sdm->nu);
   for (i = 0; i < sdm->nu; i++)
   {
      state_index = sdm->q[i].state_index;

      if (sdm->q[i].type == dpPrescribedQ)
      {
         if (data->q_data[i])
         {
            q_euler[state_index] = interpolate_function(time + params.stepSize, data->q_data[i], zeroth, 0.0, 0.0) / sdm->q[i].conversion;

            if (data->u_data[i])
               u_euler[state_index] = interpolate_function(time + params.stepSize, data->u_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->q_data[i]->type != dpStepFunction)
               u_euler[state_index] = interpolate_function(time + params.stepSize, data->q_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
            else
               u_euler[state_index] = 0.0;

            if (data->udot_data[i])
               a_euler[state_index] = interpolate_function(time + params.stepSize, data->udot_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->u_data[i] && data->u_data[i]->type != dpStepFunction)
               a_euler[state_index] = interpolate_function(time + params.stepSize, data->u_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
            else if (data->q_data[i]->type != dpStepFunction)
               a_euler[state_index] = interpolate_function(time + params.stepSize, data->q_data[i], second, 1.0, 0.0) / sdm->q[i].conversion;
            else
               a_euler[state_index] = 0.0;
         }
      }
      else if (sdm->q[i].type == dpFixedQ)
      {
         q_euler[state_index] = sdm->q[i].initial_value  / sdm->q[i].conversion;
         u_euler[state_index] = 0.0;
         a_euler[state_index] = 0.0;
      }

   }
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         int q_ind = sdm->q[i].q_ind;
         double q_ind_value, u_ind_value, a_ind_value;
         state_index = sdm->q[i].state_index;
         
         q_ind_value = q_euler[q_ind] * sdm->q[q_ind].conversion;
         q_euler[state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, zeroth, 0.0, 0.0) / sdm->q[i].conversion;
         
         u_ind_value = u_euler[q_ind] * sdm->q[q_ind].conversion;
         u_euler[state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, first, u_ind_value, 0.0)/sdm->q[i].conversion;
         
         a_ind_value = a_euler[q_ind] * sdm->q[q_ind].conversion;
         a_euler[state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, second, u_ind_value, a_ind_value)/sdm->q[i].conversion;

      }
   }

   // convert euler params to quaternions
   // first, copy over all values
   for (i = 0; i < sdm->nu; i++)
   {
      q_quat[i] = q_euler[i];
      u_quat[i] = u_euler[i];
      a_quat[i] = a_euler[i];
   }
   // how to check that all 3 DOFs of quaternion joints were prescribed
   {
      int j, first_index, state_index[3], axes[3];
      dpBoolean valid;
      double qe[3], qq[4], ue[3], uq[3], ae[3], aq[3];

      for (i = 0; i < sdm->num_joints; i++)
      {
         if (sdm->joint[i].jnt_type == dpBall || sdm->joint[i].jnt_type == dpFree)
         {
            first_index = sdm->joint[i].first_q;
            if (sdm->joint[i].jnt_type == dpFree)
               first_index += 3;

            state_index[0] = sdm->q[first_index].state_index;
            state_index[1] = sdm->q[first_index+1].state_index;
            state_index[2] = sdm->q[first_index+2].state_index;
            for (j = 0; j < 3; j++)
            {
               axes[j] = sdm->q[first_index+j].axis;
               qe[j] = q_euler[state_index[j]];
               ue[j] = u_euler[state_index[j]];
               ae[j] = a_euler[state_index[j]];
            }

            convert_angles_to_quaternions(qe, ue, ae, qq, uq, aq, sdm->joint[i].euler_rots, axes);

            // copy back into state array
            for (j = 0; j < 3; j++)
            {
               q_quat[state_index[j]] = qq[j];
               u_quat[state_index[j]] = uq[j];
               a_quat[state_index[j]] = aq[j];
            }
               q_quat[sdm->joint[i].quat] = qq[3];
         }
      }
   }


   // set the new state
   for (i = 0; i < sdm->nu; i++)
   {
      state_index = sdm->q[i].state_index;
      y[state_index] = q_quat[state_index];
      y[i + sdm->nq] = u_quat[state_index];
      dy[state_index] = u_quat[state_index];
   }
   for (i = sdm->nu; i < sdm->nq; i++)
      y[i] = q_quat[i];
   sdstate(time + params.stepSize, y, &y[sdm->nq]);

   calc_derivatives(time + params.stepSize, y, dy, param, &status);

   // Now calculate the moments required to achieve the
   // accelerations at this next time step.
   calc_gencoord_moments(time + params.stepSize, y, dy);

   // Store the moments in the q_moments[] functions,
   // putting them in the appropriate time step. There is
   // no need to adjust the function coefficients (the
   // b[] array for dpLinear) because they will be calculated
   // after the "corrective" moment has been added later.
   for (i = 0; i < sdm->nu; i++)
   {
      if (data->q_moments[i])
      {
         dpFunction* sf = data->q_moments[i];

         for (j = 0; j < sf->numpoints; j++)
         {
            if (EQUAL_WITHIN_TOLERANCE(time + params.stepSize, sf->x[j], 0.0001))
            {
               sf->y[j] = sdm->q[i].moment;
            }
         }
      }
   }

   // Disable prescribed motion and restore the state
   // to the saved one.
   set_prescribed_motion(sdm, dpPrescribedQ, 0);
   sdm->newInverseSimulation = dpYes;
   sdstate(time, savedState, &savedState[sdm->nq]);

   free(savedState);
   free(savedDy);
   FREE_IFNOTNULL(q_euler);
   FREE_IFNOTNULL(u_euler);
   FREE_IFNOTNULL(a_euler);
   FREE_IFNOTNULL(q_quat);
   FREE_IFNOTNULL(u_quat);
   FREE_IFNOTNULL(a_quat);
}


/* CALC_CORRECTIVE_MOMENTS: this function calculates the additional moment needed at each
 * gencoord to keep the model tracking the input motion during an inverse simulation.
 * At each time step, the error between the actual and desired kinematics is calculated
 * and used to determine a corrective moment. This moment is added to the applied moment
 * (stored in the MotionData structure) at the *next* time step, so that during the
 * next integration interval that moment can be applied (usually with a piece-wise linear
 * interpolation).
 */
void calc_corrective_moments(double time, double q_quat[], double u_quat[], dpModelStruct* sdm, MotionData* data)
{
   int i, j;
   double desQ, desU, errQ, errU;
   double k1, k2, correctiveMoment;
   dpFunction* sf;
   double *q_euler, *u_euler;

   q_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   u_euler = (double *)simm_malloc(sdm->nq * sizeof(double));
   convert_state_from_quaternion(time, q_quat, &q_quat[sdm->nq], NULL, q_euler, u_euler, NULL, NULL);

   for (i = 0; i < sdm->nu; i++)
   {
      // Corrective moments are calculated only for Qs which have kinematic data
      // to track as well as a moment function to hold the applied moment.
      if (data && data->q_data[i] && data->q_moments[i] && sdm->q[i].type == dpPrescribedQ)
      {
         // Calculate the desired kinematics: desQ, desU.
         desQ = interpolate_function(time, data->q_data[i], zeroth, 0.0, 0.0) / sdm->q[i].conversion;

         if (data->u_data[i])
            desU = interpolate_function(time, data->u_data[i], zeroth, 1.0, 0.0) / sdm->q[i].conversion;
         else if (data->q_data[i]->type != dpStepFunction)
            desU = interpolate_function(time, data->q_data[i], first, 1.0, 0.0) / sdm->q[i].conversion;
         else
            desU = 0.0;

         // The error is the desired minus the actual kinematics.
         errQ = desQ - q_euler[i];
         errU = desU - u_euler[i];

         k1 = sdm->q[i].pd_stiffness;
         k2 = 2.0 * sqrt(k1);

         // Calculate the corrective moment and add it to the q_moments[]
         // function holding this gencoord's applied moment. The corrective
         // moment is added to the moment value at the next step, not the
         // current one (because that moment has already been applied).
         correctiveMoment = k1 * errQ + k2 * errU;

         sf = data->q_moments[i];

         for (j = 0; j < sf->numpoints; j++)
         {
            if (EQUAL_WITHIN_TOLERANCE(time + params.stepSize, sf->x[j], 0.00001))
            {
               sf->y[j] += correctiveMoment;
               if (j - 1 >= 0)
                  sf->b[j-1] = (sf->y[j] - sf->y[j-1]) / (sf->x[j] - sf->x[j-1]);
               if (j < sf->numpoints - 1)
                  sf->b[j] = (sf->y[j+1] - sf->y[j]) / (sf->x[j+1] - sf->x[j]);

               // Now store the original+corrected moments from the current time
               // step (the one before the one calculated in this function) in the
               // sdm->q[] array so that they can be written to the output motion
               //  file and also used by the static muscle optimization function.
               if (j > 0)
                  sdm->q[i].moment = sf->y[j - 1];
               break;
            }
         }
      }
   }
   FREE_IFNOTNULL(q_euler);
   FREE_IFNOTNULL(u_euler);
}


/* ADD_MOMENTS_TO_MOTION_DATA: this function makes sure there is a
 * function in q_moments[] for each gencoord whose motion is prescribed.
 * During an inverse simulation, the moments for each gencoord are
 * calculated and stored in these functions so they can be applied
 * to the gencoords to make the model track the input motion. If the
 * user changed params.stepSize away from the default value (the
 * time increment in the input gencoord data), then these functions
 * will not have the same number of points as the gencoord functions.
 */
void add_moments_to_motion_data(dpModelStruct* sdm, MotionData* data)
{
   int i, j, numMomentFrames;
   double *initialY;

   // Add 1.99999 to account for the first frame of data, and any fraction of
   // a time step at the end of the simulation.
   numMomentFrames = (params.endTime - params.startTime) / params.stepSize + 1.99999;

   initialY = (double*)simm_calloc(numMomentFrames, sizeof(double));

   for (i = 0; i < sdm->nu; i++)
   {
      if (data->q_moments[i] == NULL && sdm->q[i].type == dpPrescribedQ)
      {
         data->q_moments[i] = (dpFunction*)simm_malloc(sizeof(dpFunction));
         init_function(data->q_moments[i], initialY, numMomentFrames, dpLinearFunction, 0);
         for (j = 0; j < numMomentFrames - 1; j++)
         {
            data->q_moments[i]->x[j] = params.startTime + j * params.stepSize;
            data->q_moments[i]->b[j] = 0.0;
         }
         // The last time frame should be at exactly the end time.
         data->q_moments[i]->x[numMomentFrames - 1] = params.endTime;
         data->q_moments[i]->b[numMomentFrames - 1] = 0.0;
      }
   }

   free(initialY);
}


/* APPLY_GENCOORD_RESTRAINT_MOMENTS: This routine uses the user-defined restraint moment
 * functions to apply moments to the degrees of freedom in order to keep them from
 * going out of the desirable ranges of motion. The new method of specifying restraint
 * moments uses one function (restraint_function) which covers the complete range of motion
 * as well as off both ends. If this function is defined, use it, otherwise use the
 * old method (min_restraint_function and max_restraint_function).
 *   If you want to include damping in the restraint moments, then uncomment the
 * appropriate lines in the function, and set the dampingFactor to the desired value.
 */
void apply_gencoord_restraint_moments(dpModelStruct* sdm, double state_quat[])
{
   int i;
   double state_value, difference, moment, dampingFactor = 1.0;
   double *state_euler, *dstate_euler, *moment_euler, *moment_quat;

   state_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   dstate_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   moment_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   moment_quat = (double *)simm_malloc(sdm->nu * sizeof(double));
   copy_vector(state_quat, state_euler, sdm->nu);

   // convert the input quaternion state to Euler angles
   convert_state_from_quaternion(0, state_quat, &state_quat[sdm->nq], NULL, state_euler, dstate_euler, NULL, NULL);

   // determine the moment to be applied for each Q that is out of range
   for (i=0; i<sdm->nu; i++)
   {
      moment_euler[i] = moment_quat[i] = 0.0;
      if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
      {
         state_value = state_euler[i] * sdm->q[i].conversion;
         moment = 0.0;
         if ((sdm->q[i].restraint_function != NULL) && (sdm->q[i].function_active == dpYes))
         {
            moment = interpolate_function(state_value, sdm->q[i].restraint_function,zeroth, 1.0, 1.0);
            // Because the moment in this case is applied for all gencoord values
            // (though it's usually zero within the range of motion), you only want
            // to include the damping term when the moment is not zero.
            //if (NOT_EQUAL_WITHIN_ERROR(moment, 0.0))
            //moment -= dampingFactor * dstate_euler[i];
         }
         else if (state_value < sdm->q[i].range_start && sdm->q[i].min_restraint_function != NULL)
         {
            difference = sdm->q[i].range_start - state_value;
            moment = interpolate_function(difference, sdm->q[i].min_restraint_function, zeroth, 1.0, 1.0);
            //moment -= dampingFactor * dstate_euler[i];
         }
         else if (state_value > sdm->q[i].range_end && sdm->q[i].max_restraint_function != NULL)
         {
            difference = state_value - sdm->q[i].range_end;
            moment = -interpolate_function(difference, sdm->q[i].max_restraint_function, zeroth, 1.0, 1.0);
            //moment -= dampingFactor * dstate_euler[i];
         }
         moment_euler[i] = moment;
      }
   }

   // convert the moments in Euler parameters to quaternion frame to be applied at the joints
   convert_moment_array_to_quaternion(state_quat, moment_euler, moment_quat);

   // apply the moments
   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
      {
         int axis = sdm->q[i].axis;
         if (sdm->q[i].is_part_of_quaternion)
         {
            axis = abs(sdm->q[i].axis) - 1;
            if (sdm->joint[sdm->q[i].joint].jnt_type == dpFree)
               axis += 3;
         }
         sdhinget(sdm->q[i].joint, axis, moment_quat[i]);
      }
   }

   FREE_IFNOTNULL(state_euler);
   FREE_IFNOTNULL(dstate_euler);
   FREE_IFNOTNULL(moment_euler);
   FREE_IFNOTNULL(moment_quat);
}


/* GET_MUSCLE_PARAM_VALUE: This routine gives the Pipeline programmer access to the
 * muscle parameters. Since the parameters used in the dynamics code are defined
 * at runtime, the programmer who writes the muscle model code in derivs.c cannot
 * just access the parameters by looking into the muscle structure. Instead, one
 * must pass the name of the parameter into this routine, which returns the value
 * for that parameter. For completeness, this routine also gives access to the
 * SIMM/static parameters (e.g. max_isometric_force), although this routine is not
 * required for this access.
 */
double get_muscle_param_value(dpMuscleStruct* muscle, char param_name[])
{
   int i;

   for (i=0; i<muscle->num_dynamic_params; i++)
   {
      if (STRINGS_ARE_EQUAL(param_name,muscle->dynamic_param_names[i]))
      {
         if (muscle->dynamic_params[i] == NULL)
         {
            sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);
            return ERROR_DOUBLE;
         }
         else
         {
            return *muscle->dynamic_params[i];
         }
      }
   }

   if (STRINGS_ARE_EQUAL(param_name,"max_force"))
   {
      if (muscle->max_isometric_force == NULL)
      {
         sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);
         return ERROR_DOUBLE;
      }
      else
      {
         return *muscle->max_isometric_force;
      }
   }

   if (STRINGS_ARE_EQUAL(param_name,"pennation_angle"))
   {
      if (muscle->pennation_angle == NULL)
      {
         sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);
         return ERROR_DOUBLE;
      }
      else
      {
         return *muscle->pennation_angle;
      }
   }

   if (STRINGS_ARE_EQUAL(param_name,"optimal_fiber_length"))
   {
      if (muscle->optimal_fiber_length == NULL)
      {
         sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);
         return ERROR_DOUBLE;
      }
      else
      {
         return *muscle->optimal_fiber_length;
      }
   }

   if (STRINGS_ARE_EQUAL(param_name,"tendon_slack_length"))
   {
      if (muscle->resting_tendon_length == NULL)
      {
         sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);
         return ERROR_DOUBLE;
      }
      else
      {
         return *muscle->resting_tendon_length;
      }
   }

   if (STRINGS_ARE_EQUAL(param_name,"max_contraction_velocity"))
   {
      if (muscle->max_contraction_vel == NULL)
      {
         sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);
         return ERROR_DOUBLE;
      }
      else
      {
         return *muscle->max_contraction_vel;
      }
   }

   sim_message(none, "Error: attempted to access undefined muscle parameter (%s).", param_name);
   return ERROR_DOUBLE;
}


/* GET_MUSCLE_PARAM_INDEX: This function is similar to get_muscle_param_value(),
 * but returns the index of the parameter, so it can be stored for faster access
 * later.
 */
int get_muscle_param_index(dpMuscleStruct* muscle, char param_name[])
{
   int i;

   for (i = 0; i < muscle->num_dynamic_params; i++)
   {
      if (STRINGS_ARE_EQUAL(param_name, muscle->dynamic_param_names[i]))
      {
         return i;
      }
   }

   sim_message(none, "Error: %s was not specified for muscle %s.", param_name, muscle->name);

   return 0; // should probably return -1, but other code may not handle it properly
}


/* Display any posted SD/FAST error messages with the name of the function
 * in which they occurred. Clear the errors. */
int check_for_sderror(char caller[])
{
   int routine, err;

   sderror(&routine, &err);

   switch (err)
   {
      case 0:
         break;
      case 19:
         if (isStandAlone)
            sim_message(none, "%s: trying to set a non-? parameter.", caller);
         break;
      default:
         if (isStandAlone)
         {
        //    sim_message(none, "%s:", caller);
            printf("\n%s:", caller);
            sdprinterr(stdout);
         }
         sprintf(sd_error_msg, "SD/FAST error in %s", caller);
         break;
   }
   sdclearerr();

   return err;
}


/* initialize the joint records.
 */
void init_joints(dpBoolean allocateSpace)
{
   int info[50], slider[6], i, j;
   int axis, num_axes, q, axis1, axis2, axis3;
   double pin[3];
   dpEulerAngleOrder axis_order;

   if (allocateSpace == dpYes)
      sdm->joint = (dpJointStruct*)simm_malloc(sdm->num_joints*sizeof(dpJointStruct));

   for (i = 0; i < sdm->num_joints; i++)
   {
      sdjnt(i, info, slider);

      sdm->joint[i].inboard_body = info[2];     /* get the index of the inboard (first) body of the joint */
      sdm->joint[i].outboard_body = info[3];    /* get the index of the outboard (second) body of the joint */
      sdm->joint[i].num_dof = info[4];          /* number of DOFs that the joint has */
      sdm->joint[i].quat = info[8];             /* if the joint is ball or 6dof, the index of the 4th Euler param/quaternion */
      sdm->joint[i].loop_joint = dpNo;
      sdm->joint[i].is_quaternion = dpNo;
      sdm->joint[i].first_q = info[7];          /* coordinate index of the first axis */

      for (j = 0; j < 6; j++)
      {
         if (slider[j] == 0)
            sdm->joint[i].dof_type[j] = dpRotational;
         else if (slider[j] == 1)
            sdm->joint[i].dof_type[j] = dpTranslational;
         else
            sdm->joint[i].dof_type[j] = dpNoDof;
      }

      switch(info[0])
      {
         case(1):
            sdm->joint[i].jnt_type = dpPin;
            break;
         case(2):
            sdm->joint[i].jnt_type = dpUniversal;
            break;
         case(3):
            sdm->joint[i].jnt_type = dpGimbal;
            break;
         case(4):
            sdm->joint[i].jnt_type = dpBall;
            sdm->joint[i].is_quaternion = dpYes;
            sdm->q[sdm->joint[i].first_q].is_part_of_quaternion = dpYes;
            sdm->q[sdm->joint[i].first_q+1].is_part_of_quaternion = dpYes;
            sdm->q[sdm->joint[i].first_q+2].is_part_of_quaternion = dpYes;
            break;
         case(5):
            sdm->joint[i].jnt_type = dpSlider;
            break;
         case(6):
            sdm->joint[i].jnt_type = dpFree;
            sdm->joint[i].is_quaternion = dpYes;
            // only rotational Qs are quaternions
            for (j = 0; j < 6; j++)
               if (sdm->joint[i].dof_type[j] == dpRotational)
                  sdm->q[sdm->joint[i].first_q+j].is_part_of_quaternion = dpYes;
            break;
         case(7):
            sdm->joint[i].jnt_type = dpCylindrical;
            break;
         case(8):
            sdm->joint[i].jnt_type = dpPlanar;
            break;
         case(9):
            sdm->joint[i].jnt_type = dpWeld;
            sdm->joint[i].loop_joint = dpYes;
            break;
         case(10):
            sdm->joint[i].jnt_type = dpBushing;
            break;
         case(11):
            sdm->joint[i].jnt_type = dpBearing;
            break;
         case(20):
            sdm->joint[i].jnt_type = dpReversePlanar;
            break;
         case(21):
            sdm->joint[i].jnt_type = dpReverseFree;
            break;
         case(22):
            sdm->joint[i].jnt_type = dpReverseBushing;
            break;
         case(23):
            sdm->joint[i].jnt_type = dpReverseBearing;
            break;
         default:
            sdm->joint[i].jnt_type = dpUnknownJoint;
            break;
      }
      
      num_axes = sdm->joint[i].num_dof;
      if (sdm->joint[i].jnt_type == dpBall)
         num_axes = 0;
      for (j = 0; j < num_axes; j++)
      {
         if (j < 3) axis = j;
         else axis = j - 3;
         sdgetpin(i, axis, pin);
         sdvcopy(pin, sdm->joint[i].axes[j]);
      }
      for (j = 0; j < 3; j++)
         sdm->joint[i].force[j] = sdm->joint[i].torque[j] = 0.0;      
      if (check_for_sderror("INIT_JOINT") == 18)
         fprintf(stderr,"Unable to get pin for joint %d (%s -> %s)\n", i, sdm->body_segment[sdm->joint[i].inboard_body+1].name, sdm->body_segment[sdm->joint[i].outboard_body+1].name);
   }
}


/* CALC_SYSTEM_ENERGY: this routine calculates the total energy of the body
 * segments in the system. It does not keep track of the energy that the
 * muscles add/remove from the system, so if any muscles in the model generate
 * force, or if there are non-energy-conserving collisions, system energy will
 * not be conserved.
 */
void calc_system_energy(SystemInfo* si)
{
   sdmom(si->linear_momentum, si->angular_momentum, &si->kinetic_energy);   
   sdgetgrav(si->gravity);
   sdsys(&si->mass, si->mass_center, si->inertia);

   si->potential_energy = -(si->mass_center[1]) * si->gravity[1] * si->mass;
   si->system_energy = si->kinetic_energy + si->potential_energy;

   // If initial_system_energy has not yet been set, set it.
   if (EQUAL_WITHIN_ERROR(si->initial_system_energy,MINMDOUBLE))
      si->initial_system_energy = si->system_energy;
}


/* FREE_MOTION_DATA: frees a MotionData struct. */
void free_motion_data(MotionData* md)
{
   if (md)
   {
      int i;

      FREE_IFNOTNULL(md->name);

      for (i = 0; i < md->num_elements; i++)
      {
         FREE_IFNOTNULL(md->elementnames[i]);
         FREE_IFNOTNULL(md->motiondata[i]);
      }
      FREE_IFNOTNULL(md->elementnames);
      FREE_IFNOTNULL(md->motiondata);

      FREE_IFNOTNULL(md->q_data);
      FREE_IFNOTNULL(md->u_data);
      FREE_IFNOTNULL(md->udot_data);
      FREE_IFNOTNULL(md->q_moments);
      FREE_IFNOTNULL(md->musc_excitations);

      for (i = 0; i < md->num_forces; i++)
         FREE_IFNOTNULL(md->forces[i]);
      FREE_IFNOTNULL(md->forces);

      for (i = 0; i < md->num_torques; i++)
         FREE_IFNOTNULL(md->torques[i]);
      FREE_IFNOTNULL(md->torques);
   }
}


/* CONVERT_QUATERNIONS_TO_EULER: given the input state with possible quaternions,
 * convert the quaternions to Euler angles and store in output state.
 * NOTE: the quaternion representation is NOT the same as the SD/FAST representation
 * here quaternion is  q0 = cos(theta/2) q1 = lambda1*sin(theta/2) q2 = lambda2*sin(theta/2) q3 = lambda3*sin(theta/2)
 * SDFAST quaternion is q0 = lambda1*sin(theta/2) q1 = lambda2*sin(theta/2) q2 = lambda3*sin(theta/2) q4 = cos(theta/2) 
 */
static void convert_quaternions_to_euler(double time, int frame, double input_state[], double previous_state[], double output_state[])
{
   int i, j, first_index, quat_index, have_flip;
   Quat quat;
   double R[4][4], rots[] = {0.0, 0.0, 0.0};

   for (i = 0; i < sdm->num_joints; i++)
   {
      if (sdm->joint[i].is_quaternion)
      {
         first_index = sdm->joint[i].first_q;
         if (sdm->joint[i].jnt_type == dpFree)
            first_index += 3;
         quat_index = sdm->joint[i].quat;
         quat[XX] = input_state[quat_index];
         quat[YY] = input_state[first_index];
         quat[ZZ] = input_state[first_index+1];
         quat[WW] = input_state[first_index+2];

         set_rotation_matrix_from_quaternion(quat, R);

         extract_euler_angles(R, rots, sdm->joint[i].euler_rots);

         // if the axis is "mirrored", eg -X, -Y or -Z, flip its angle
         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               rots[j] = -rots[j];
            output_state[sdm->q[first_index+j].state_index] = rots[j];
         }

         //////// see if r2 is near 90 then check if had a jump
         if (EQUAL_WITHIN_TOLERANCE(rots[1], M_PI/2, 0.1) && previous_state != NULL)
         {
            have_flip = 0;
            for (j = 0; j < 3; j++)
            {
               double a1 = RAD2DEG(output_state[sdm->q[first_index+j].state_index]);
               double a2 = RAD2DEG(previous_state[sdm->q[first_index+j].state_index]);
               double diff = output_state[sdm->q[first_index+j].state_index] - previous_state[sdm->q[first_index+j].state_index];
               if (fabs(diff) > 3.0)
                  have_flip = 1;
            }
            if (have_flip)
               extract_euler_angles(R, rots, sdm->joint[i].euler_rots);
         }

      /*   if (!have_flip && previous_state != NULL)
         {
            for (j = 0; j < 3; j++)
            {
               int index = sdm->q[first_index+j].state_index;
               double a1 = RAD2DEG(output_state[index]);
               double a2 = RAD2DEG(previous_state[index]);
               double diff = output_state[index] - previous_state[index];
               if (fabs(diff) > 3.0)
               {
                  if ((0 < diff && diff <= M_PI))
                     output_state[index] -= M_PI;
                  else if (M_PI < diff && diff <= 2 * M_PI)
                     output_state[index] -= 2 * M_PI;
                  else if (2 * M_PI < diff)
                     output_state[index] -= 2*M_PI;
                  else if (diff < -2 * M_PI)
                     output_state[index] += 2 * M_PI;
                  else if (-2 * M_PI <= diff && diff < -M_PI)
                     output_state[index] += 2 * M_PI;
                  else if (-M_PI <= diff && diff < 0)
                     output_state[index] += M_PI;
               }
            }
         }*/
      }
   }
}


// the input states are quaternion-base, return the output states in Euler
// when converting velocities and accelerations, take the velocity term from the second
//  half of the "state" vector, NOT the first part of the state derivative vector
// ie. use &y[sdm->nq], not dy
void convert_state_from_quaternion(double time, double q[], double u[], double udot[], double q_out[], double u_out[], double udot_out[], double previous_state[])
{
   int i, j, first_index, state_index;
   double theta1, theta2, theta3, theta1dot, theta2dot, theta3dot, theta1dotdot, theta2dotdot, theta3dotdot;
   double omega1, omega2, omega3, alpha1, alpha2, alpha3;
   double c1, c2, c3, s1, s2, s3;
   dpEulerAngleOrder order;

   // copy the original input values to the output array (so all non-quaternion elements will be there)
   for (i = 0; i < sdm->nu; i++)
   {
      if (q != NULL && q_out != NULL)
         q_out[i] = q[i];
      if (u != NULL && u_out != NULL)
         u_out[i] = u[i];
      if (udot != NULL && udot_out != NULL)
         udot_out[i] = udot[i];
   }

   // positions
   // input Q array is in quaternions, convert it to Euler angles
   convert_quaternions_to_euler(time, 0, q, previous_state, q_out);

   // velocities and accelerations
   if (u == NULL || u_out == NULL)
      return;

   for (i = 0; i < sdm->num_joints; i++)
   {
      if (sdm->joint[i].is_quaternion)
      {
         first_index = sdm->joint[i].first_q;
         if (sdm->joint[i].jnt_type == dpFree)
            first_index += 3;
         state_index = sdm->q[first_index].state_index;

         // if the axis is "mirrored", eg -X, -Y or -Z, flip its angle for the thetas
         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               q_out[state_index+j] = -q_out[state_index+j];
         }

            // get thetas
         theta1 = q_out[state_index];
         theta2 = q_out[state_index+1];
         theta3 = q_out[state_index+2];

         omega1 = u[state_index];
         omega2 = u[state_index+1];
         omega3 = u[state_index+2];

         // restore the output value if it was flipped for further calculations
         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               q_out[state_index+j] = -q_out[state_index+j];
         }

         // can't use theta and theta dot from prescribe gencoords since hasn't been called yet
         c1 = cos(theta1);
         c2 = cos(theta2);
         c3 = cos(theta3);
         s1 = sin(theta1);
         s2 = sin(theta2);
         s3 = sin(theta3);

         order = sdm->joint[i].euler_rots;

         if (order == dp123)
         {
            theta1dot = (omega1 * c3 - omega2 * s3) / c2;
            theta2dot = omega1 * s3 + omega2 * c3;
            theta3dot = (-1 * omega1 * c3 + omega2 * s3) * s2 / c2 + omega3;
         }
         else if (order == dp231)
         {
            theta1dot = (omega2 * c3 - omega3 * s3) / c2;
            theta2dot = omega2 * s3 + omega3 * c3;
            theta3dot = (-1 * omega2 * c3 + omega3 * s3) * s2 / c2 + omega1;
         }
         else if (order == dp312)
         {
            theta1dot = (omega3 * c3 - omega1 * s3) / c2;
            theta2dot = omega3 * s3 + omega1 * c3;
            theta3dot = (-1 * omega3 * c3 + omega1 * s3) * s2 / c2 + omega2;
         }
         else if (order == dp132)
         {
            theta1dot = (omega1 * c3 + omega3 * s3) / c2;
            theta2dot = -1 * omega1 * s3 + omega3 * c3;
            theta3dot = (omega1 * c3 + omega3 * s3) * s2 / c2 + omega2;
         }
         else if (order == dp213)
         {
            theta1dot = (omega2 * c3 + omega1 * s3) / c2;
            theta2dot = -1 * omega2 * s3 + omega1 * c3;
            theta3dot = (omega2 * c3 + omega1 * s3) * s2 / c2 + omega3;
         }
         else if (order == dp321)
         {
            theta1dot = (omega3 * c3 + omega2 * s3) / c2;
            theta2dot = -1 * omega3 * s3 + omega2 * c3;
            theta3dot = (omega3 * c3 + omega2 * s3) * s2 / c2 + omega1;
         }

         u_out[state_index+0] = theta1dot;
         u_out[state_index+1] = theta2dot;
         u_out[state_index+2] = theta3dot;

         // if the axis is "mirrored", eg -X, -Y or -Z, flip its angle
         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               u_out[state_index+j] = -u_out[state_index+j];
         }

         if (udot != NULL && udot_out != NULL)
         {
            alpha1 = udot[state_index];
            alpha2 = udot[state_index+1];
            alpha3 = udot[state_index+2];

            if (order == dp123)
            {
               theta1dotdot = (alpha1 * c3 - alpha2 * s3 - theta3dot * (omega1 * s3 + omega2 * c3)) / c2
                  + theta2dot * (omega1 * c3 - omega2 * s3) * s2 / (c2 * c2);
               theta2dotdot = alpha1 * s3 + alpha2 * c3 + theta3dot * (omega1 * c3 - omega2 * s3);
               theta3dotdot = (-1 * alpha1 * c3 + alpha2 * s3 + theta3dot * (omega1 * s3 + omega2 * c3)) * s2 / c2
                  + theta2dot * (-1 * omega1 * c3 + omega2 * s3) / (c2 * c2) + alpha3;
            }
            else if (order == dp231)
            {
               theta1dotdot = (alpha2 * c3 - alpha3 * s3 - theta3dot * (omega2 * s3 + omega3 * c3)) / c2
                  + theta2dot * (omega2 * c3 - omega3 * s3) * s2 / (c2 * c2);
               theta2dotdot = alpha2 * s3 + alpha3 * c3 + theta3dot * (omega2 * c3 - omega3 * s3);
               theta3dotdot = (-1 * alpha2 * c3 + alpha3 * s3 + theta3dot * (omega2 * s3 + omega3 * c3)) * s2 / c2
                  + theta2dot * (-1 * omega2 * c3 + omega3 * s3) / (c2 * c2) + alpha1;
            }
            else if (order == dp312)
            {
               theta1dotdot = (alpha3 * c3 - alpha1 * s3 - theta3dot * (omega3 * s3 + omega1 * c3)) / c2
                  + theta2dot * (omega3 * c3 - omega1 * s3) * s2 / (c2 * c2);
               theta2dotdot = alpha3 * s3 + alpha1 * c3 + theta3dot * (omega3 * c3 - omega1 * s3);
               theta3dotdot = (-1 * alpha3 * c3 + alpha1 * s3 + theta3dot * (omega3 * s3 + omega1 * c3)) * s2 / c2
                  + theta2dot * (-1 * omega3 * c3 + omega1 * s3) / (c2 * c2) + alpha2;
            }
            else if (order == dp132)
            {
               theta1dotdot = (alpha1 * c3 + alpha3 * s3 + theta3dot * (-1 * omega1 * s3 + omega3 * c3))/c2 
                  + theta2dot * (omega1 * c3 + omega3 * s3) * s2 / (c2 * c2);
               theta2dotdot = -1 * alpha1 * s3 + alpha3 * c3 - theta3dot * (omega1 * c3 + omega3 * s3);
               theta3dotdot = (alpha1 * c3 + alpha3 * s3 + theta3dot * (-1 * omega1 * s3 + omega3 * c3)) * s2 / c2
                  + (omega1 * c3 + omega3 * s3) * theta2dot / (c2 * c2) + alpha2;
            }
            else if (order == dp213)
            {
               theta1dotdot = (alpha2 * c3 + alpha1 * s3 + theta3dot * (-1 * omega2 * s3 + omega1 * c3))/c2 
                  + theta2dot * (omega2 * c3 + omega1 * s3) * s2 / (c2 * c2);
               theta2dotdot = -1 * alpha2 * s3 + alpha1 * c3 - theta3dot * (omega2 * c3 + omega1 * s3);
               theta3dotdot = (alpha2 * c3 + alpha1 * s3 + theta3dot * (-1 * omega2 * s3 + omega1 * c3)) * s2 / c2
                  + (omega2 * c3 + omega1 * s3) * theta2dot / (c2 * c2) + alpha3;
            }
            else if (order == dp321)
            {
               theta1dotdot = (alpha3 * c3 + alpha2 * s3 + theta3dot * (-1 * omega3 * s3 + omega2 * c3))/c2 
                  + theta2dot * (omega3 * c3 + omega2 * s3) * s2 / (c2 * c2);
               theta2dotdot = -1 * alpha3 * s3 + alpha2 * c3 - theta3dot * (omega3 * c3 + omega2 * s3);
               theta3dotdot = (alpha3 * c3 + alpha2 * s3 + theta3dot * (-1 * omega3 * s3 + omega2 * c3)) * s2 / c2
                  + (omega3 * c3 + omega2 * s3) * theta2dot / (c2 * c2) + alpha1;
            }

            udot_out[state_index+0] = theta1dotdot;
            udot_out[state_index+1] = theta2dotdot;
            udot_out[state_index+2] = theta3dotdot;

            // if the axis is "mirrored", eg -X, -Y or -Z, flip its angle
            for (j = 0; j < 3; j++)
            {
               if (sdm->q[first_index+j].axis < 0)
                  udot_out[state_index+j] = -udot_out[state_index+j];
            }
         }
      }
   }
}

// Convert the state vector from Euler to quaternion
ReturnCode convert_state_to_quaternion(double time, double q[], double u[], double udot[], double q_out[], double u_out[], double udot_out[])
{
   int i, j, first_index, state_index, axis1, axis2, axis3;
   double theta1, theta2, theta3, theta1dot, theta2dot, theta3dot, theta1dotdot, theta2dotdot, theta3dotdot;
   double omega1, omega2, omega3, alpha1, alpha2, alpha3, R[3][3];
   double c1, c2, c3, s1, s2, s3, A, B, C, D, E, F, Adot, Bdot, Cdot, Ddot, Edot, Fdot;
   dpEulerAngleOrder order;
   Quat quat;

   // copy the original input values to the output array (so all non-quaternion elements will be there)
   for (i = 0; i < sdm->nu; i++)
   {
      if (q != NULL && q_out != NULL)
         q_out[i] = q[i];
      if (u != NULL && u_out != NULL)
         u_out[i] = u[i];
      if (udot != NULL && udot_out != NULL)
         udot_out[i] = udot[i];
   }

   // positions
   // input Q array is in Euler angles, convert it to quaternions
   for (i = 0; i < sdm->num_joints; i++)
   {
      if (!sdm->joint[i].is_quaternion)
         continue;

      first_index = sdm->joint[i].first_q;
      if (sdm->joint[i].jnt_type == dpFree)
         first_index += 3;

      state_index = sdm->q[first_index].state_index;

      order = sdm->joint[i].euler_rots;

      axis1 = abs(sdm->q[first_index].axis);
      axis2 = abs(sdm->q[first_index+1].axis);
      axis3 = abs(sdm->q[first_index+2].axis);

      if (axis1 < 1 || axis1 > 3)
      {
         printf("Set initial quaternion conditions: illegal axis 1 in joint %d (q = %s)\n", i, sdm->q[first_index].name);
         return code_bad;
      }
      if ((axis2 < 1 || axis2 > 3))
      {
         printf("Set initial quaternion conditions: illegal axis 2 in joint %d (q = %s)\n", i, sdm->q[first_index+1].name);
         return code_bad;
      }
      if (axis3 < 1 || axis3 > 3)
      {
         printf("Set initial quaternion conditions: illegal 3 axis in joint %d (q = %s)\n", i, sdm->q[first_index+2].name);
         return code_bad;
      }

      theta1 = q[state_index];
      theta2 = q[state_index+1];
      theta3 = q[state_index+2];

      // if any of the axes is "mirrored" (ie -X, -Y or -Z) flip its angle
      if (sdm->q[first_index].axis < 0)
         theta1 *= -1;
      if (sdm->q[first_index+1].axis < 0)
         theta2 *= -1;
      if (sdm->q[first_index+2].axis < 0)
         theta3 *= -1;

      set_rotation_matrix_from_euler_angles(theta1, axis1, theta2, axis2, theta3, axis3, R, order);       

      convert_rotation_to_quaternion(R, quat);

      // to match sdfast representation,          
      q_out[state_index] = quat[1];
      q_out[state_index+1] = quat[2];
      q_out[state_index+2] = quat[3];
      q_out[sdm->joint[i].quat] = quat[0];

      //velocity and acceleration
      if (u != NULL && u_out != NULL)
      {
         // do thetas need to be flipped for mirrored joints??
         c1 = cos(theta1);
         c2 = cos(theta2);
         c3 = cos(theta3);
         s1 = sin(theta1);
         s2 = sin(theta2);
         s3 = sin(theta3);

         theta1dot = u[state_index];
         theta2dot = u[state_index+1];
         theta3dot = u[state_index+2];

         A = theta1dot * c2 * c3;
         B = theta2dot * s3;
         C = -1 * theta1dot * c2 * s3;
         D = theta2dot * c3;
         E = theta1dot * s2;
         F = theta3dot;

         if (order == dp123)
         {
            omega1 = A + B;
            omega2 = C + D;
            omega3 = E + F;
         }
         else if (order == dp132) 
         {
            omega1 = A - B;
            omega2 = -1*E + F;
            omega3 = -1*C + D;
         }
         else if (order == dp213) 
         {
            omega1 = -1 *C + D;
            omega2 = A - B;
            omega3 = -1 * E + F;
         }
         else if (order == dp231) 
         {
            omega1 = E + F;
            omega2 = A + B;
            omega3 = C + D;
         }
         else if (order == dp312)
         {
            omega1 = C + D;
            omega2 = E + F;
            omega3 = A + B;
         }
         else if (order == dp321) 
         {
            omega1 = -1 * E + F;
            omega2 = -1 * C + D;
            omega3 = A - B;
         }

         u_out[state_index+0] = omega1;
         u_out[state_index+1] = omega2;
         u_out[state_index+2] = omega3;

         // if the axis is "mirrored", eg -X, -Y or -Z, flip its angle
         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               u_out[state_index+j] = -u_out[state_index+j];
         }

         if (udot != NULL && udot_out != NULL)
         {
            theta1dotdot = udot[state_index+0];
            theta2dotdot = udot[state_index+1];
            theta3dotdot = udot[state_index+2];
            Adot = theta1dotdot * c2 * c3 - theta1dot * theta2dot * s2 * c3
               - theta1dot * theta3dot * c2 * s3;
            Bdot = theta2dotdot * s3 + theta2dot * theta3dot * c3;
            Cdot = -1 * theta1dotdot * c2 * s3 + theta1dot * theta2dot * s2 * s3
               - theta1dot * theta3dot * c2 * c3;
            Ddot = theta2dotdot * c3 - theta2dot * theta3dot * s3;
            Edot = theta1dotdot * s2 + theta1dot * theta2dot * c2;
            Fdot = theta3dotdot;

            if (order == dp123)
            {
               alpha1 = Adot + Bdot;
               alpha2 = Cdot + Ddot;
               alpha3 = Edot + Fdot;
            }
            else if (order == dp132) 
            {
               alpha1 = Adot - Bdot;
               alpha2 = -1 * Edot + Fdot;
               alpha3 = -1 * Cdot + Ddot;
            }
            else if (order == dp213) 
            {
               alpha1 = -1 * Cdot + Ddot;
               alpha2 = Adot - Bdot;
               alpha3 = -1 * Edot + Fdot;
            }
            else if (order == dp231) 
            {
               alpha1 = Edot + Fdot;
               alpha2 = Adot + Bdot;
               alpha3 = Cdot + Ddot;
            }
            else if (order == dp312)
            {
               alpha1 = Cdot + Ddot;
               alpha2 = Edot + Fdot;
               alpha3 = Adot + Bdot;
            }
            else if (order == dp321) 
            {
               alpha1 = -1 * Edot + Fdot;
               alpha2 = -1 * Cdot + Ddot;
               alpha3 = Adot - Bdot;
            }

            udot_out[state_index+0] = alpha1;
            udot_out[state_index+1] = alpha2;
            udot_out[state_index+2] = alpha3;

            // if the axis is "mirrored", eg -X, -Y or -Z, flip its output
            for (j = 0; j < 3; j++)
            {
               if (sdm->q[first_index+j].axis < 0)
                  udot_out[state_index+j] = -udot_out[state_index+j];
            }
         }
      }
   }
   return code_fine;
}

/* Convert array of moments from quaternion-space to Euler-space */
void convert_moment_array_from_quaternion(double state_euler[], double input_moments[], double output_moments[])
{
   int i, j, first_index, state_index;
   double theta1, theta2, theta3, c1, c2, c3, s1, s2, s3;
   double moment1, moment2, moment3, newmoments[3];
   dpEulerAngleOrder order;

   copy_vector(input_moments, output_moments, sdm->nu);

   for (i = 0; i < sdm->num_joints; i++)
   {
      if (sdm->joint[i].is_quaternion)
      {
         first_index = sdm->joint[i].first_q;
         if (sdm->joint[i].jnt_type == dpFree)
            first_index += 3;
         state_index = sdm->q[first_index].state_index;

         theta1 = state_euler[state_index+0];
         theta2 = state_euler[state_index+1];
         theta3 = state_euler[state_index+2];

         // flip the angles of mirrored joints (with negative axes)
         if (sdm->q[first_index].axis < 0)
            theta1 *= -1;
         if (sdm->q[first_index+1].axis < 0)
            theta2 *= -1;
         if (sdm->q[first_index+2].axis < 0)
            theta3 *= -1;

         c1 = cos(theta1);
         c2 = cos(theta2);
         c3 = cos(theta3);
         s1 = sin(theta1);
         s2 = sin(theta2);
         s3 = sin(theta3);

         moment1 = input_moments[first_index];
         moment2 = input_moments[first_index+1];
         moment3 = input_moments[first_index+2];

         order = sdm->joint[i].euler_rots;

         if (order == dp123)
         {
            newmoments[0] = c2 * c3 * moment1 - c2 * s3 * moment2 + s2 * moment3;
            newmoments[1] = s3 * moment1 + c3 * moment2;
            newmoments[2] = moment3;
         }
         else if (order == dp231)
         {
            newmoments[0] = c2 * c3 * moment2 - c2 * s3 * moment3 + s2 * moment1;
            newmoments[1] = s3 * moment2 + c3 * moment3;
            newmoments[2] = moment1;
         }
         else if (order == dp312)
         {
            newmoments[0] = c2 * c3 * moment3 - c2 * s3 * moment1 + s2 * moment2;
            newmoments[1] = s3 * moment3 + c3 * moment1;
            newmoments[2] = moment2;
         }
         else if (order == dp132)
         {
            newmoments[0] = c2 * c3 * moment1 + c2 * s3 * moment3 - s2 * moment2;
            newmoments[1] = -1 * s3 * moment1 + c3 * moment3;
            newmoments[2] = moment2;
         }
         else if (order == dp213)
         {
            newmoments[0] = c2 * c3 * moment2 + c2 * s3 * moment1 - s2 * moment3;
            newmoments[1] = -1 * s3 * moment2 + c3 * moment1;
            newmoments[2] = moment3;
         }
         else if (order == dp321)
         {
            newmoments[0] = c2 * c3 * moment3 + c2 * s3 * moment2 - s2 * moment1;
            newmoments[1] = -1 * s3 * moment3 + c3 * moment2;
            newmoments[2] = moment1;
         }
         output_moments[first_index] = newmoments[0];
         output_moments[first_index+1] = newmoments[1];
         output_moments[first_index+2] = newmoments[2];

         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               output_moments[first_index+j] = -output_moments[first_index+j];
         }
      }
   }
}

/* convert whole array of moments to quaternions */
void convert_moment_array_to_quaternion(double state_quat[], double input_moments[], double output_moments[])
{
   int i, j, first_index, state_index;
   double theta1, theta2, theta3, c1, c2, c3, s1, s2, s3;
   double moment1, moment2, moment3, newmoments[3], *state_euler = NULL;
   dpEulerAngleOrder order;

   // Convert the state vector.
   state_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   convert_state_from_quaternion(0, state_quat, NULL, NULL, state_euler, NULL, NULL, NULL);

   // Copy the input moments to the output moments (to take care of the ones
   // that do not need conversion).
   copy_vector(input_moments, output_moments, sdm->nu);

   for (i = 0; i < sdm->num_joints; i++)
   {
      if (sdm->joint[i].is_quaternion && state_euler != NULL)
      {
         first_index = sdm->joint[i].first_q;
         if (sdm->joint[i].jnt_type == dpFree)
            first_index += 3;
         state_index = sdm->q[first_index].state_index;

         theta1 = state_euler[state_index+0];
         theta2 = state_euler[state_index+1];
         theta3 = state_euler[state_index+2];

         c1 = cos(theta1);
         c2 = cos(theta2);
         c3 = cos(theta3);
         s1 = sin(theta1);
         s2 = sin(theta2);
         s3 = sin(theta3);

         moment1 = input_moments[state_index];
         moment2 = input_moments[state_index+1];
         moment3 = input_moments[state_index+2];
         order = sdm->joint[i].euler_rots;

         if (order == dp123)
         {
            newmoments[0] = c3 / c2 * moment1 + s3 * moment2 - s2 * c3 / c2 * moment3;
            newmoments[1] =  -1 * s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
            newmoments[2] = moment3;
         }
         else if (order == dp231)
         {
            newmoments[0] = moment3;
            newmoments[1] = c3 / c2 * moment1 + s3 * moment2 - s2 * c3 / c2 * moment3;
            newmoments[2] =  -1 * s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
         }
         else if (order == dp312)
         {
            newmoments[0] =  -1 * s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
            newmoments[1] = moment3;
            newmoments[2] = c3 / c2 * moment1 + s3 * moment2 - s2 * c3 / c2 * moment3;
         }
         else if (order == dp132)
         {
            newmoments[0] = c3 / c2 * moment1 - s3 * moment2 + s2 * c3 / c2 * moment3;
            newmoments[1] = moment3;
            newmoments[2] = s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
         }
         else if (order == dp213)
         {
            newmoments[0] = s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
            newmoments[1] = c3 / c2 * moment1 - s3 * moment2 + s2 * c3 / c2 * moment3;
            newmoments[2] = moment3;
         }
         else if (order == dp321)
         {
            newmoments[0] = moment3;
            newmoments[1] = s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
            newmoments[2] = c3 / c2 * moment1 - s3 * moment2 + s2 * c3 / c2 * moment3;
         }
         for (j = 0; j < 3; j++)
         {
            int axis;
            axis = abs(sdm->q[first_index+j].axis) - 1;

            output_moments[state_index+j] = newmoments[axis];
         }

         for (j = 0; j < 3; j++)
         {
            if (sdm->q[first_index+j].axis < 0)
               output_moments[state_index+j] = -output_moments[state_index+j];
         }
      }
   }
   FREE_IFNOTNULL(state_euler);
}

// convert a single moment from Euler-space to quaternion
void convert_moment_to_quaternion(double theta1, double theta2, double theta3, double input_moments[], double output_moments[], dpEulerAngleOrder order)
{
   int j;
   double c1, c2, c3, s1, s2, s3, moment1, moment2, moment3, newmoments[3];
   
   c1 = cos(theta1);
   c2 = cos(theta2);
   c3 = cos(theta3);
   s1 = sin(theta1);
   s2 = sin(theta2);
   s3 = sin(theta3);

   moment1 = input_moments[0];
   moment2 = input_moments[1];
   moment3 = input_moments[2];

   if (EQUAL_WITHIN_TOLERANCE(c2, 0.0, 1e-5))
      printf("DIVIDE BY ZERO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");

   if (order == dp123)
   {
      newmoments[0] = c3 / c2 * moment1 + s3 * moment2 - s2 * c3 / c2 * moment3;
      newmoments[1] =  -1 * s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
      newmoments[2] = moment3;
   }
   else if (order == dp231)
   {
      newmoments[0] = moment3;
      newmoments[1] = c3 / c2 * moment1 + s3 * moment2 - s2 * c3 / c2 * moment3;
      newmoments[2] =  -1 * s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
   }
   else if (order == dp312)
   {
      newmoments[0] =  -1 * s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
      newmoments[1] = moment3;
      newmoments[2] = c3 / c2 * moment1 + s3 * moment2 - s2 * c3 / c2 * moment3;
   }
   else if (order == dp132)
   {
      newmoments[0] = c3 / c2 * moment1 - s3 * moment2 + s2 * c3 / c2 * moment3;
      newmoments[1] = moment3;
      newmoments[2] = s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
   }
   else if (order == dp213)
   {
      newmoments[0] = s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
      newmoments[1] = c3 / c2 * moment1 - s3 * moment2 + s2 * c3 / c2 * moment3;
      newmoments[2] = moment3;
   }
   else if (order == dp321)
   {
      newmoments[0] = moment3;
      newmoments[1] = s3 / c2 * moment1 + c3 * moment2 + s2 * s3 / c2 * moment3;
      newmoments[2] = c3 / c2 * moment1 - s3 * moment2 + s2 * c3 / c2 * moment3;
   }

   for (j = 0; j < 3; j++)
   {
      output_moments[j] = newmoments[j];
   }
}


// Convert the state vector from Euler to quaternion
ReturnCode convert_angles_to_quaternions(double q[], double u[], double udot[], double q_out[], double u_out[], double udot_out[], dpEulerAngleOrder order, int axes[])
{
   int i, j, first_index, state_index, axis1, axis2, axis3;
   double theta1, theta2, theta3, theta1dot, theta2dot, theta3dot, theta1dotdot, theta2dotdot, theta3dotdot;
   double omega1, omega2, omega3, alpha1, alpha2, alpha3, R[3][3];
   double c1, c2, c3, s1, s2, s3, A, B, C, D, E, F, Adot, Bdot, Cdot, Ddot, Edot, Fdot;
   Quat quat;

   // copy the original input values to the output array (so all non-quaternion elements will be there)
   for (i = 0; i < 3; i++)
   {
      if (q != NULL && q_out != NULL)
         q_out[i] = q[i];
      if (u != NULL && u_out != NULL)
         u_out[i] = u[i];
      if (udot != NULL && udot_out != NULL)
         udot_out[i] = udot[i];
   }

   // positions
   axis1 = abs(axes[0]);
   axis2 = abs(axes[1]);
   axis3 = abs(axes[2]);

   if (axis1 < 1 || axis1 > 3)
   {
//      printf("Set initial quaternion conditions: illegal axis 1 in joint %d (q = %s)\n", i, sdm->q[first_index].name);
      return code_bad;
   }
   if ((axis2 < 1 || axis2 > 3))
   {
//      printf("Set initial quaternion conditions: illegal axis 2 in joint %d (q = %s)\n", i, sdm->q[first_index+1].name);
      return code_bad;
   }
   if (axis3 < 1 || axis3 > 3)
   {
//      printf("Set initial quaternion conditions: illegal 3 axis in joint %d (q = %s)\n", i, sdm->q[first_index+2].name);
      return code_bad;
   }

   theta1 = q[0];
   theta2 = q[1];
   theta3 = q[2];

   // if any of the axes is "mirrored" (ie -X, -Y or -Z) flip its angle
   if (axes[0] < 0)
      theta1 *= -1;
   if (axes[1] < 0)
      theta2 *= -1;
   if (axes[2] < 0)
      theta3 *= -1;

   set_rotation_matrix_from_euler_angles(theta1, axis1, theta2, axis2, theta3, axis3, R, order);       

   convert_rotation_to_quaternion(R, quat);

   // to match sdfast representation,          
   q_out[0] = quat[1];
   q_out[1] = quat[2];
   q_out[2] = quat[3];
   q_out[3] = quat[0];

   // if any of the axes is "mirrored" (ie -X, -Y or -Z) undo the mirroring for velocity and acceleration calculations
   if (axes[0] < 0)
      theta1 *= -1;
   if (axes[1] < 0)
      theta2 *= -1;
   if (axes[2] < 0)
      theta3 *= -1;


   //velocity and acceleration
   if (u != NULL && u_out != NULL)
   {
      c1 = cos(theta1);
      c2 = cos(theta2);
      c3 = cos(theta3);
      s1 = sin(theta1);
      s2 = sin(theta2);
      s3 = sin(theta3);

      theta1dot = u[0];
      theta2dot = u[1];
      theta3dot = u[2];

      A = theta1dot * c2 * c3;
      B = theta2dot * s3;
      C = -1 * theta1dot * c2 * s3;
      D = theta2dot * c3;
      E = theta1dot * s2;
      F = theta3dot;

      if (order == dp123)
      {
         omega1 = A + B;
         omega2 = C + D;
         omega3 = E + F;
      }
      else if (order == dp132) 
      {
         omega1 = A - B;
         omega2 = -1*E + F;
         omega3 = -1*C + D;
      }
      else if (order == dp213) 
      {
         omega1 = -1 *C + D;
         omega2 = A - B;
         omega3 = -1 * E + F;
      }
      else if (order == dp231) 
      {
         omega1 = E + F;
         omega2 = A + B;
         omega3 = C + D;
      }
      else if (order == dp312)
      {
         omega1 = C + D;
         omega2 = E + F;
         omega3 = A + B;
      }
      else if (order == dp321) 
      {
         omega1 = -1 * E + F;
         omega2 = -1 * C + D;
         omega3 = A - B;
      }

      u_out[0] = omega1;
      u_out[1] = omega2;
      u_out[2] = omega3;

      // if the axis is "mirrored", eg -X, -Y or -Z, flip its angle
      for (j = 0; j < 3; j++)
      {
         if (axes[j] < 0)
            u_out[abs(axes[j])-1] = -u_out[abs(axes[j])-1];
      }

      if (udot != NULL && udot_out != NULL)
      {
         theta1dotdot = udot[0];
         theta2dotdot = udot[1];
         theta3dotdot = udot[2];
         Adot = theta1dotdot * c2 * c3 - theta1dot * theta2dot * s2 * c3
            - theta1dot * theta3dot * c2 * s3;
         Bdot = theta2dotdot * s3 + theta2dot * theta3dot * c3;
         Cdot = -1 * theta1dotdot * c2 * s3 + theta1dot * theta2dot * s2 * s3
            - theta1dot * theta3dot * c2 * c3;
         Ddot = theta2dotdot * c3 - theta2dot * theta3dot * s3;
         Edot = theta1dotdot * s2 + theta1dot * theta2dot * c2;
         Fdot = theta3dotdot;

         if (order == dp123)
         {
            alpha1 = Adot + Bdot;
            alpha2 = Cdot + Ddot;
            alpha3 = Edot + Fdot;
         }
         else if (order == dp132) 
         {
            alpha1 = Adot - Bdot;
            alpha2 = -1 * Edot + Fdot;
            alpha3 = -1 * Cdot + Ddot;
         }
         else if (order == dp213) 
         {
            alpha1 = -1 * Cdot + Ddot;
            alpha2 = Adot - Bdot;
            alpha3 = -1 * Edot + Fdot;
         }
         else if (order == dp231) 
         {
            alpha1 = Edot + Fdot;
            alpha2 = Adot + Bdot;
            alpha3 = Cdot + Ddot;
         }
         else if (order == dp312)
         {
            alpha1 = Cdot + Ddot;
            alpha2 = Edot + Fdot;
            alpha3 = Adot + Bdot;
         }
         else if (order == dp321) 
         {
            alpha1 = -1 * Edot + Fdot;
            alpha2 = -1 * Cdot + Ddot;
            alpha3 = Adot - Bdot;
         }

         udot_out[0] = alpha1;
         udot_out[1] = alpha2;
         udot_out[2] = alpha3;

         // if the axis is "mirrored", eg -X, -Y or -Z, flip its output
         for (j = 0; j < 3; j++)
         {
            if (axes[j] < 0)
               udot_out[abs(axes[j])-1] = -udot_out[abs(axes[j])-1];
         }
      }
   }
   return code_fine;
}


/* copy_nndouble: this utility function copies a double, but only if the
 * source is not NULL.
 */
ReturnCode copy_nndouble(double* from, double** to)
{
   if (from == NULL)
      *to = NULL;
   else
   {
      if ((*to = (double*)simm_malloc(sizeof(double))) == NULL)
         return code_bad;
      **to = *from;
   }

   return code_fine;
}


/* copy_nndouble: this utility function copies a double, but only if the
 * source is not equal to a default value.
 */
ReturnCode copy_nddouble(double* from, double** to, double* deffrom, double* defto)
{
   if (from == deffrom)
      *to = defto;
   else
   {
      if (from == NULL)
         *to = NULL;
      else
      {
         if ((*to = (double*)simm_malloc(sizeof(double))) == NULL)
            return code_bad;
         **to = *from;
      }
   }

   return code_fine;
}


/* copy_nndouble: this utility function copies an integer, but only if the
 * source is not NULL.
 */
ReturnCode copy_nnint(int* from, int** to)
{
   if (from == NULL)
      *to = NULL;
   else
   {
      if ((*to = (int*)simm_malloc(sizeof(int))) == NULL)
         return code_bad;
      **to = *from;
   }

   return code_fine;
}


/* copyMuscle: this function makes a copy of a muscle. */
void copyMuscle(dpMuscleStruct* from, dpMuscleStruct* defFrom,
                dpMuscleStruct* to, dpMuscleStruct* defTo,
                dpModelStruct* ms, dpModelStruct* dp)
{
   int i;

   // Start by zero-ing out the entire muscle structure.
   memset(to, 0, sizeof(dpMuscleStruct));

   if (from->name == defFrom->name)
      to->name = defTo->name;
   else
      mstrcpy(&to->name, from->name);

   to->display = from->display;
   to->selected = from->selected;

   copy_nddouble(from->max_isometric_force, &to->max_isometric_force,
                 defFrom->max_isometric_force, defTo->max_isometric_force);

   copy_nddouble(from->pennation_angle, &to->pennation_angle,
                 defFrom->pennation_angle, defTo->pennation_angle);

   copy_nddouble(from->optimal_fiber_length, &to->optimal_fiber_length,
                 defFrom->optimal_fiber_length, defTo->optimal_fiber_length);

   copy_nddouble(from->resting_tendon_length, &to->resting_tendon_length,
                 defFrom->resting_tendon_length, defTo->resting_tendon_length);

   copy_nddouble(from->max_contraction_vel, &to->max_contraction_vel,
                 defFrom->max_contraction_vel, defTo->max_contraction_vel);

   to->num_dynamic_params = from->num_dynamic_params;

   if (to->num_dynamic_params > 0)
   {
      // A muscle's param name array is always the same as its default muscle's.
      to->dynamic_param_names = defTo->dynamic_param_names;
      to->dynamic_params = (double**)simm_malloc(to->num_dynamic_params * sizeof(double*));

      for (i = 0; i < to->num_dynamic_params; i++)
         copy_nddouble(from->dynamic_params[i], &to->dynamic_params[i],
                       defFrom->dynamic_params[i], defTo->dynamic_params[i]);
   }
   else
   {
      to->dynamic_param_names = NULL;
      to->dynamic_params = NULL;
   }

   if (from->active_force_len_func == defFrom->active_force_len_func)
      to->active_force_len_func = defTo->active_force_len_func;
   else
   {
      if (from->active_force_len_func == NULL)
         to->active_force_len_func = NULL;
      else
      {
         to->active_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         *to->active_force_len_func = *from->active_force_len_func;
      }
   }

   if (from->passive_force_len_func == defFrom->passive_force_len_func)
      to->passive_force_len_func = defTo->passive_force_len_func;
   else
   {
      if (from->passive_force_len_func == NULL)
         to->passive_force_len_func = NULL;
      else
      {
         to->passive_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         *to->passive_force_len_func = *from->passive_force_len_func;
      }
   }

   if (from->tendon_force_len_func == defFrom->tendon_force_len_func)
      to->tendon_force_len_func = defTo->tendon_force_len_func;
   else
   {
      if (from->tendon_force_len_func == NULL)
         to->tendon_force_len_func = NULL;
      else
      {
         to->tendon_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         *to->tendon_force_len_func = *from->tendon_force_len_func;
      }
   }

   if (from->force_vel_func == defFrom->force_vel_func)
      to->force_vel_func = defTo->force_vel_func;
   else
   {
      if (from->force_vel_func == NULL)
         to->force_vel_func = NULL;
      else
      {
         to->force_vel_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         *to->force_vel_func = *from->force_vel_func;
      }
   }

   if (from->excitation_abscissa == defFrom->excitation_abscissa)
      to->excitation_abscissa = defTo->excitation_abscissa;
   else
   {
      to->excitation_abscissa = (void**)simm_malloc(sizeof(void*));
      *to->excitation_abscissa = *from->excitation_abscissa;
   }

   if (from->excitation_func == defFrom->excitation_func)
      to->excitation_func = defTo->excitation_func;
   else
   {
      if (from->excitation_func == NULL)
         to->excitation_func = NULL;
      else
      {
         to->excitation_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         *to->excitation_func = *from->excitation_func;
      }
   }

   if (from->muscle_model_index == defFrom->muscle_model_index)
      to->muscle_model_index = defTo->muscle_model_index;
   else
   {
      to->muscle_model_index = (int*)simm_malloc(sizeof(int));
      *to->muscle_model_index = *from->muscle_model_index;
   }

   if (from->muscle_model_opt_index == defFrom->muscle_model_opt_index)
      to->muscle_model_opt_index = defTo->muscle_model_opt_index;
   else
   {
      to->muscle_model_opt_index = (int*)simm_malloc(sizeof(int));
      *to->muscle_model_opt_index = *from->muscle_model_opt_index;
   }

   to->numWrapStructs = from->numWrapStructs;
   if (from->wrapStruct == defFrom->wrapStruct)
      to->wrapStruct = defTo->wrapStruct;
   else
   {
      to->wrapStruct = (dpMuscleWrapStruct**)simm_malloc(to->numWrapStructs * sizeof(dpMuscleWrapStruct*));
      for (i = 0; i < to->numWrapStructs; i++)
      {
         to->wrapStruct[i] = (dpMuscleWrapStruct*)simm_calloc(1, sizeof(dpMuscleWrapStruct));
         to->wrapStruct[i]->wrap_algorithm = from->wrapStruct[i]->wrap_algorithm;
         to->wrapStruct[i]->startPoint = from->wrapStruct[i]->startPoint;
         to->wrapStruct[i]->endPoint = from->wrapStruct[i]->endPoint;
         to->wrapStruct[i]->wrap_object = from->wrapStruct[i]->wrap_object;
      }
   }

   if (from->path != NULL)
   {
      to->path = (dpMusclePathStruct *)simm_malloc(sizeof(dpMusclePathStruct));

      to->path->num_orig_points = from->path->num_orig_points;

      to->path->mp_orig = (dpMusclePoint*)simm_malloc(to->path->num_orig_points * sizeof(dpMusclePoint));
      memcpy(to->path->mp_orig, from->path->mp_orig, to->path->num_orig_points * sizeof(dpMusclePoint));
      for (i = 0; i < to->path->num_orig_points; i++)
      {
         if (to->path->mp_orig[i].num_wrap_pts > 0 && from->path->mp_orig[i].wrap_pts)
         {
            to->path->mp_orig[i].wrap_pts = (double*)simm_malloc(to->path->mp_orig[i].num_wrap_pts * 3 * sizeof(double));
            memcpy(to->path->mp_orig[i].wrap_pts, from->path->mp_orig[i].wrap_pts, to->path->mp_orig[i].num_wrap_pts * 3 * sizeof(double));
         }
         else
         {
            to->path->mp_orig[i].wrap_pts = NULL;
         }

         to->path->mp_orig[i].isVia = from->path->mp_orig[i].isVia;
         if (to->path->mp_orig[i].isVia)
         {
            to->path->mp_orig[i].viaRange.start = from->path->mp_orig[i].viaRange.start;
            to->path->mp_orig[i].viaRange.end = from->path->mp_orig[i].viaRange.end;
            to->path->mp_orig[i].viaRange.gencoord = from->path->mp_orig[i].viaRange.gencoord;
         }
      }

      to->path->mp_array_size = to->path->num_orig_points + (to->numWrapStructs * 2);
      to->path->mp = (dpMusclePoint**)simm_malloc(sizeof(dpMusclePoint*) * to->path->mp_array_size);

      /* When wrapping is calculated by the simulation, the mp[] array will be filled in
      * and num_points set appropriately.
      */
      to->path->num_points = 0;
   }
   else
   {
      to->path = NULL;
   }
}


/* Given a gencoord name, return the index of the sdm->q list element with the same name. */
int get_q_index_by_name(const char name[])
{
   int i;

   for (i=0; i<sdm->nu; i++)
      if (STRINGS_ARE_EQUAL(name, sdm->q[i].name))
         return i;

   return -1;
}

/* Given a segment name, return the index of the sdm->body_segment list element with the same name. */
int get_segment_index_by_name(const char name[])
{
   int i;

   for (i=0; i<sdm->num_body_segments; i++)
      if (STRINGS_ARE_EQUAL(name, sdm->body_segment[i].name))
         return i;

   return -1;
}

/* Given a joint name, return the index of the sdm->joint list element with the same name. */
int get_joint_index_by_name(const char name[])
{
   int i;

   for (i=0; i<sdm->num_joints; i++)
      if (STRINGS_ARE_EQUAL(name, sdm->joint[i].name))
         return i;

   return -1;
}

/* Given a muscle name, return the index of the sdm->muscles list element with the same name. */
int get_muscle_index_by_name(const char name[])
{
   int i;

   for (i=0; i<sdm->num_muscles; i++)
      if (STRINGS_ARE_EQUAL(name, sdm->muscles[i].name))
         return i;

   return -1;
}

/* Clear the given list and free its arrays, then allocate space for a new list of the given size. */
void allocate_name_list(dpNameList* list, int size)
{
   clear_name_list(list);
   FREE_IFNOTNULL((void*)list->name);
   FREE_IFNOTNULL((void*)list->index);

   list->list_size = size;
   if (size > 0)
   {
      list->name = (const char**)simm_calloc(list->list_size, sizeof(const char*));
      list->index = (int*)simm_calloc(list->list_size, sizeof(int));
   }
}

/* Remove all elements from a list. */
void clear_name_list(dpNameList* list)
{
   int i;

   for (i=0; i<list->num_items; i++)
      FREE_IFNOTNULL((void*)list->name[i]);

   list->num_items = 0;
}

/* Given a name, if it is already in the list, return its index. If it is not already in the list,
 * add it and return its index.
 */
int add_name_to_list(const char* name, dpNameList* list)
{
   int i;

   // First make sure the object is not already in the list.
   for (i=0; i<list->num_items; i++)
   {
      if (list->name[i] == name)
         return i;
   }

   // Now make sure the list is big enough to hold the object.
   if (list->list_size == 0)
   {
      list->list_size = 10;
      list->name = (const char**)simm_calloc(list->list_size, sizeof(const char*));
      list->index = (int*)simm_calloc(list->list_size, sizeof(int));
   }
   else if (list->num_items == list->list_size)
   {
      ReturnCode rc;
      list->list_size += 10;
      list->name = (const char**)simm_realloc((void*)list->name, list->list_size * sizeof(const char*), &rc);
      list->index = (int*)simm_realloc((void*)list->index, list->list_size * sizeof(int), &rc);
   }

   // Now add the object.
   mstrcpy((char**)&list->name[list->num_items], name);
   list->num_items++;

   return list->num_items - 1;
}

/* Remove the specified name from the given list of names.
 */
void remove_name_from_list(const char* name, dpNameList* list)
{
   int i, j;

   for (i=0; i<list->num_items; i++)
   {
      if (list->name[i] == name)
      {
         free((void*)list->name[i]);
         for (j=i; j<list->num_items-1; j++)
         {
            list->name[j] = list->name[j+1];
            list->index[j] = list->index[j+1];
         }
         list->num_items--;
         return;
      }
   }
}

/* If the given name is an element of the given list of names, return its index.
 * Otherwise, return -1.
 */
int find_name_in_list(const char* name, dpNameList* list)
{
   int i;

   for (i=0; i<list->num_items; i++)
   {
      if (STRINGS_ARE_EQUAL(list->name[i], name))
         return i;
   }

   return -1;
}

/* copy a list of names from one dpNameList structure to another */
void copy_name_list(dpNameList* from, dpNameList* to)
{
   int i;

   memset(to, 0, sizeof(dpNameList));

   if (from->state && from->num_items > 0)
   {
      to->state = 1;
      to->list_size = to->num_items = from->num_items;
      to->name = (const char**)simm_calloc(to->list_size, sizeof(const char*));
      to->index = (int*)simm_calloc(to->list_size, sizeof(int));
      for (i=0; i<to->num_items; i++)
      {
         mstrcpy((char**)&to->name[i], from->name[i]);
         to->index[i] = from->index[i];
      }
   }
}
