/*******************************************************************************

   READMUSCLES.C

   Author: Peter Loan

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains routines that read-in muscle
      descriptions from an input file.

   Routines:
      read_muscle_file        : main routine to read a muscle file
      read_muscle             : reads a muscle definition from a file
      read_muscle_parameter   : reads a [double] muscle parameter from a file
      read_muscle_function    : reads a [muscle] spline function from a file
      read_excitation         : reads a muscle excitation pattern from a file
      init_muscle             : initializes a muscle structure
      read_dynamic_parameter_names : reads dynamic (muscle model) parameters 

*******************************************************************************/

#include "universal.h"


/*************** DEFINES (for this file only) *********************************/
#define MUSCLE_ARRAY_INCREMENT 40
#define MP_ARRAY_INCREMENT 10


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/
static dpFunctionType excitationType = dpStepFunction;


/*************** GLOBAL VARIABLES for a few files only) ****************/



/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern dpModelStruct* model;
extern dpSimulationParameters params;
extern char buffer[];


/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static ReturnCode read_muscle(FILE* fp, dpModelStruct* model, dpMuscleStruct* muscle);
static ReturnCode read_muscle_parameter(FILE* fp, dpMuscleStruct* muscle, dpMuscleStruct* default_muscle,
                                        double** param, double* default_param, char param_name[],
                                        double min_value, double max_value);
static ReturnCode read_muscle_function(FILE* fp, dpModelStruct* model, dpMuscleStruct* muscle,
                                       dpFunction*** musc_func, dpFunction** default_mfunc,
                                       char ending[], char description[]);
static ReturnCode read_excitation(FILE* fp, dpModelStruct* model, dpMuscleStruct* muscle);
static ReturnCode init_musclepoint(dpMusclePoint *mp);
static ReturnCode init_musclepath(dpMusclePathStruct *path);
static ReturnCode init_muscle(dpModelStruct* model, dpMuscleStruct* muscle);
static void postprocess_muscle(dpMuscleStruct* muscle);
static ReturnCode read_dynamic_parameter_names(FILE* fp, dpMuscleStruct* muscle);
static ReturnCode init_dynamic_param_array(dpMuscleStruct* muscle, dpMuscleStruct* default_muscle);
static void check_dynamic_param_array(FILE* fp, dpMuscleStruct* muscle, char text_string[]);
static const char* get_wrap_algorithm_name (int i);
static const char* get_function_type_name(dpFunctionType funcType);
static const char* get_function_tag(dpFunctionType funcType, int beginEnd);
static void finish_muscle_definition(FILE* fp);



/* READ_MUSCLE_FILE: this is the main routine for reading a muscle input
 * file. It runs cpp on the muscle file, then reads it in. When it finds
 * the string "beginmuscle" it calls read_muscle() to read in a muscle
 * definition.
 */
ReturnCode read_muscle_file(dpModelStruct* model, char filename[], dpIncludeMuscles include)
{
   FILE *fp;
   int i, count, muscle_array_size;
   ReturnCode rc;
   dpBoolean default_muscle_defined = dpNo;
   char tmp_name[CHARBUFFER];

   if (include == dpNoMuscles)
      return code_fine;

   get_temp_file_name(NULL, tmp_name);
   if ((fp = preprocess_file(filename, tmp_name)) == NULL)
   {
      sim_message(none, "Unable to open muscle input file %s.", filename);
      return code_bad;
   }

   model->num_muscles = 0;
   muscle_array_size = MUSCLE_ARRAY_INCREMENT;
   model->muscles = (dpMuscleStruct*)simm_malloc(muscle_array_size*sizeof(dpMuscleStruct));
   memset(&model->default_muscle, 0, sizeof(dpMuscleStruct));

   while (1)
   {
      if (read_string(fp, buffer) == EOF)
         break;

      if (buffer[0] == '#')
      {
         read_nonempty_line(fp, buffer);
         continue;
      }

      if (STRINGS_ARE_EQUAL(buffer, "beginmuscle"))
      {
         read_string(fp, buffer);
         if (STRINGS_ARE_EQUAL(buffer,"defaultmuscle"))
         {
            if (default_muscle_defined == dpYes)
            {
               sim_message(none, "Default muscle already defined. Ignoring redefinition.");
               finish_muscle_definition(fp);
            }
            else if (model->num_muscles > 0)
            {
               sim_message(none, "Error reading default muscle:");
               sim_message(none, "    You must define the default muscle before any other muscles.");
               return code_bad;
            }
            else
            {
               default_muscle_defined = dpYes;
               model->default_muscle.display = dpYes;
               model->default_muscle.excitation_abscissa = (void**)simm_malloc(sizeof(void*));//(int*)simm_malloc(sizeof(int));
               *(model->default_muscle.excitation_abscissa) = TIME;
               model->default_muscle.index = DEFAULT_MUSCLE_INDEX;
               if (model->default_muscle.num_dynamic_params > 0)
                  model->default_muscle.dynamic_params = (double**)simm_calloc(model->default_muscle.num_dynamic_params, sizeof(double*));
               mstrcpy(&model->default_muscle.name, "defaultmuscle");
               rc = read_muscle(fp, model, &model->default_muscle);
               if (rc == code_bad)
                  sim_message(exit_program, "Could not read defaultmuscle.");
            }
         }
         else
         {
            /* Check for duplicate muscle names */
            for (i=0, count=0; i<model->num_muscles; i++)
            {
               if (STRINGS_ARE_EQUAL(buffer, model->muscles[i].name))
                  count++;
            }
            if (count > 0)
               sim_message(exit_program, "Warning: %d muscles with name %s", count+1, buffer);

            if (model->num_muscles == muscle_array_size)
            {
               muscle_array_size += MUSCLE_ARRAY_INCREMENT;
               model->muscles = (dpMuscleStruct*)simm_realloc(model->muscles,
                  muscle_array_size*sizeof(dpMuscleStruct), &rc);
               if (rc == code_bad)
                  sim_message(exit_program, "Could not read in all of the muscles.");
            }
            rc = init_muscle(model, &model->muscles[model->num_muscles]);
            if (rc == code_bad)
               return code_bad;
            mstrcpy(&model->muscles[model->num_muscles].name,buffer);
            rc = read_muscle(fp, model, &model->muscles[model->num_muscles]);
            if (rc == code_fine)
            {
               // check if muscle should be added to muscle list or not
               // in this step, include all muscles for the "with EMG" option, then remove muscles from the 
               // list if there is no EMG data in the input motion
               if ((include == dpVisibleMuscles && model->muscles[model->num_muscles].display == dpYes)
                  || (include == dpMusclesWithExcitation && model->muscles[model->num_muscles].excitation_func)
                  || (include == dpAllMuscles) || (include == dpMusclesWithEMG))
               {
                  postprocess_muscle(&model->muscles[model->num_muscles]);
                  model->muscles[model->num_muscles].index = model->num_muscles++;
               }
            }
            else
               return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer,"begindynamicparameters"))
      {
         if (default_muscle_defined == dpYes || model->num_muscles > 0)
         {
            sim_message(none, "Error reading dynamic parameter names:");
            sim_message(none, "    You must define the names before the default muscle or any other muscles.");
            return code_bad;
         }
         else
         {
            if (read_dynamic_parameter_names(fp, &model->default_muscle) == code_bad)
               return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, get_function_tag(dpFunctionTypeUndefined, 0))) // old-style function definitions
      {
         if (read_function(fp, model, dpNaturalCubicSpline, get_function_tag(dpFunctionTypeUndefined, 1)) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, get_function_tag(dpStepFunction, 0)))
      {
         if (read_function(fp, model, dpStepFunction, get_function_tag(dpStepFunction, 1)) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, get_function_tag(dpLinearFunction, 0)))
      {
         if (read_function(fp, model, dpLinearFunction, get_function_tag(dpLinearFunction, 1)) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, get_function_tag(dpNaturalCubicSpline, 0)))
      {
         if (read_function(fp, model, dpNaturalCubicSpline, get_function_tag(dpNaturalCubicSpline, 1)) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, get_function_tag(dpGCVSpline, 0)))
      {
         if (read_function(fp, model, dpGCVSpline, get_function_tag(dpGCVSpline, 1)) == code_bad)
            return code_bad;
      }
      else
      {
         sim_message(recover, "Unrecognized string \"%s\" found in muscle file.", buffer);
      }
   }

   fclose(fp);
   (void)unlink(tmp_name);

   sim_message(none, "Read muscle file %s.", filename);

   return code_fine;
}



/* READ_MUSCLE: this routine reads a muscle definition from a file. It is
 * basically a big if-else block with one section for each muscle parameter.
 */
static ReturnCode read_muscle(FILE* fp, dpModelStruct* model, dpMuscleStruct* muscle)
{
   int rc, function_num;
   double dummy;

   while (1)
   {
      if (read_string(fp, buffer) == EOF)
         break;

      if (STRINGS_ARE_EQUAL(buffer, "beginpoints"))
      {
         // specifying points in the default muscle is not supported
         if (muscle == &model->default_muscle)
         {
            while (1)
            {
               read_string(fp, buffer);
               if (STRINGS_ARE_EQUAL(buffer, "endpoints"))
                  break;
            }
         }
         else
         {
            if (read_muscle_attachment_points(fp, model, muscle) == code_bad)
            {
               sim_message(abort_action, "Error reading coordinates for muscle %s.", muscle->name);
               return code_bad;
            }
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "begingroups"))
      {
         if (read_muscle_groups(fp) == code_bad)
         {
            sim_message(none, "EOF found while reading groups for muscle %s.",  muscle->name);
            sim_message(none, "Note: groups are ignored by the Pipeline.");
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer,"begintendonforcelengthcurve"))
      {
         if (read_muscle_function(fp, model, muscle, &muscle->tendon_force_len_func,
            model->default_muscle.tendon_force_len_func, "endtendonforcelengthcurve",
            "tendon_force_length_curve") == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "tendon_force_length_curve"))
      {
         read_string(fp, buffer);
         if (sscanf(buffer, "f%d", &function_num) != 1)
         {
            if (muscle != &model->default_muscle && &model->default_muscle != NULL)
            {
               muscle->active_force_len_func = model->default_muscle.tendon_force_len_func;
               sim_message(none, "Error reading tendon_force_length_curve for muscle %s.", muscle->name);
               sim_message(none, "Using function from default muscle.");
               return code_fine;
            }
            else
            {
               sim_message(none, "Error reading tendon_force_length_curve for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         muscle->tendon_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         if (muscle->tendon_force_len_func == NULL)
            return code_bad;
         *muscle->tendon_force_len_func = model->function[enter_function(model, function_num, dpYes)];
      }
      else if (STRINGS_ARE_EQUAL(buffer,"beginactiveforcelengthcurve"))
      {
         if (read_muscle_function(fp, model, muscle, &muscle->active_force_len_func,
            model->default_muscle.active_force_len_func,"endactiveforcelengthcurve",
            "active_force_length_curve") == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "active_force_length_curve"))
      {
         read_string(fp, buffer);
         if (sscanf(buffer, "f%d", &function_num) != 1)
         {
            if (muscle != &model->default_muscle && &model->default_muscle != NULL)
            {
               muscle->active_force_len_func = model->default_muscle.active_force_len_func;
               sim_message(none, "Error reading active_force_length_curve for muscle %s.", muscle->name);
               sim_message(none, "Using function from default muscle.");
               return code_fine;
            }
            else
            {
               sim_message(none, "Error reading active_force_length_curve for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         muscle->active_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         if (muscle->active_force_len_func == NULL)
            return code_bad;
         *muscle->active_force_len_func = model->function[enter_function(model, function_num, dpYes)];
      }
      else if (STRINGS_ARE_EQUAL(buffer,"beginpassiveforcelengthcurve"))
      {
         if (read_muscle_function(fp, model, muscle, &muscle->passive_force_len_func,
            model->default_muscle.passive_force_len_func,"endpassiveforcelengthcurve",
            "passive_force_length_curve") == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "passive_force_length_curve"))
      {
         read_string(fp, buffer);
         if (sscanf(buffer, "f%d", &function_num) != 1)
         {
            if (muscle != &model->default_muscle && &model->default_muscle != NULL)
            {
               muscle->active_force_len_func = model->default_muscle.active_force_len_func;
               sim_message(none, "Error reading passive_force_length_curve for muscle %s.", muscle->name);
               sim_message(none, "Using function from default muscle.");
               return code_fine;
            }
            else
            {
               sim_message(none, "Error reading passive_force_length_curve for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         muscle->passive_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         if (muscle->passive_force_len_func == NULL)
            return code_bad;
         *muscle->passive_force_len_func = model->function[enter_function(model, function_num, dpYes)];
      }
      else if (STRINGS_ARE_EQUAL(buffer,"beginforcevelocitycurve"))
      {
         if (read_muscle_function(fp, model, muscle, &muscle->force_vel_func,
            model->default_muscle.force_vel_func,"endforcevelocitycurve",
            "force_velocity_curve") == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "force_velocity_curve"))
      {
         read_string(fp, buffer);
         if (sscanf(buffer, "f%d", &function_num) != 1)
         {
            if (muscle != &model->default_muscle && &model->default_muscle != NULL)
            {
               muscle->force_vel_func = model->default_muscle.force_vel_func;
               sim_message(none, "Error reading force_velocity_curve for muscle %s.", muscle->name);
               sim_message(none, "Using function from default muscle.");
               return code_fine;
            }
            else
            {
               sim_message(none, "Error reading force_velocity_curve for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         muscle->force_vel_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         if (muscle->force_vel_func == NULL)
            return code_bad;
         *muscle->force_vel_func = model->function[enter_function(model, function_num, dpYes)];
      }
      else if (STRINGS_ARE_EQUAL(buffer, "beginexcitation")) // deprecated
      {
         if (read_excitation(fp, model, muscle) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "excitation"))
      {
         int function_num;
         char name[CHARBUFFER];

         read_line(fp, buffer);
         if (sscanf(buffer, "f%d(%s)", &function_num, name) != 2)
         {
            if (muscle != &model->default_muscle)
            {
               muscle->excitation_func = model->default_muscle.excitation_func;
               sim_message(none, "Error reading excitation for muscle %s.", muscle->name);
               sim_message(none, "Using function from default muscle.");
               return code_fine;
            }
            else
            {
               sim_message(none, "Error reading excitation for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         muscle->excitation_abscissa = (void**)simm_malloc(sizeof(void*));//(int*)simm_malloc(sizeof(int));
         if (muscle->excitation_abscissa == NULL)
            return code_bad;
         strip_brackets_from_string(name);
         if (STRINGS_ARE_EQUAL(name, "time"))
            *muscle->excitation_abscissa = TIME;
         else
            *muscle->excitation_abscissa = enter_gencoord(model, name);
         if (*muscle->excitation_abscissa == NULL)
         {
            sim_message(none, "Error reading abscissa of excitation function for muscle %s", muscle->name);
            sim_message(none, "%s is neither \"time\" nor the name of a generalized coordinate", buffer);
            finish_muscle_definition(fp);
            return code_bad;
         }
         muscle->excitation_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
         if (muscle->excitation_func == NULL)
            return code_bad;
         *muscle->excitation_func = model->function[enter_function(model, function_num, dpYes)];
      }
      else if (STRINGS_ARE_EQUAL(buffer, "excitation_format")) // deprecated
      {
         fscanf(fp, "%s", buffer);
         // If the excitation format is defined after the excitation, then you can store
         // the format directly in the excitation function. If it is defined before, store
         // it in a static global for use later.
         if (STRINGS_ARE_EQUAL(buffer, "spline_fit") || STRINGS_ARE_EQUAL(buffer, get_function_type_name(dpNaturalCubicSpline)))
         {
            if (muscle->excitation_func)
            {
               if ((*muscle->excitation_func)->type != dpNaturalCubicSpline)
               {
                  (*muscle->excitation_func)->type = dpNaturalCubicSpline;
                  calc_function_coefficients(*muscle->excitation_func);
               }
            }
            else
            {
               excitationType = dpNaturalCubicSpline;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer, get_function_type_name(dpGCVSpline)))
         {
            if (muscle->excitation_func)
            {
               if ((*muscle->excitation_func)->type != dpGCVSpline)
               {
                  (*muscle->excitation_func)->type = dpGCVSpline;
                  calc_function_coefficients(*muscle->excitation_func);
               }
            }
            else
            {
               excitationType = dpGCVSpline;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer, get_function_type_name(dpStepFunction)))
         {
            if (muscle->excitation_func)
            {
               if ((*muscle->excitation_func)->type != dpStepFunction)
               {
                  (*muscle->excitation_func)->type = dpStepFunction;
                  calc_function_coefficients(*muscle->excitation_func);
               }
            }
            else
            {
               excitationType = dpStepFunction;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer, get_function_type_name(dpLinearFunction)))
         {
            if (muscle->excitation_func)
            {
               if ((*muscle->excitation_func)->type != dpLinearFunction)
               {
                  (*muscle->excitation_func)->type = dpLinearFunction;
                  calc_function_coefficients(*muscle->excitation_func);
               }
            }
            else
            {
               excitationType = dpLinearFunction;
            }
         }
         else
         {
            sim_message(none, "Error reading excitation_format for muscle %s.", muscle->name);
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "max_force"))
      {
         if (read_muscle_parameter(fp, muscle, &model->default_muscle, &muscle->max_isometric_force,
            model->default_muscle.max_isometric_force, "max_force",
            0.0, DONT_CHECK_DOUBLE) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "tendon_slack_length"))
      {
         if (read_muscle_parameter(fp, muscle, &model->default_muscle, &muscle->resting_tendon_length,
            model->default_muscle.resting_tendon_length, "tendon_slack_length",
            0.0, DONT_CHECK_DOUBLE) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "optimal_fiber_length"))
      {
         if (read_muscle_parameter(fp, muscle, &model->default_muscle, &muscle->optimal_fiber_length,
            model->default_muscle.optimal_fiber_length, "optimal_fiber_length",
            0.0, DONT_CHECK_DOUBLE) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "pennation_angle"))
      {
         if (read_muscle_parameter(fp, muscle, &model->default_muscle, &muscle->pennation_angle,
            model->default_muscle.pennation_angle, "pennation_angle",
            0.0, 90.0) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "min_thickness"))
      {
         if (fscanf(fp, "%lg", &dummy) != 1)
         {
            sim_message(none, "Error reading min_thickness for muscle %s.", muscle->name);
            sim_message(none, "Note: This parameter is ignored by the Pipeline.");
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "max_thickness"))
      {
         if (fscanf(fp, "%lg", &dummy) != 1)
         {
            sim_message(none, "Error reading max_thickness for muscle %s.", muscle->name);
            sim_message(none, "Note: This parameter is ignored by the Pipeline.");
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "activation"))
      {
         if (fscanf(fp, "%lg", &dummy) != 1)
         {
            sim_message(none, "Error reading activation for muscle %s.", muscle->name);
            sim_message(none, "Note: This parameter is ignored by the Pipeline.");
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "max_contraction_velocity"))
      {
         if (read_muscle_parameter(fp, muscle, &model->default_muscle, &muscle->max_contraction_vel,
            model->default_muscle.max_contraction_vel,"max_contraction_velocity",
            0.0,DONT_CHECK_DOUBLE) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "muscle_model"))
      {
         muscle->muscle_model_index = (int*)simm_malloc(sizeof(int));
         rc = fscanf(fp, "%d", muscle->muscle_model_index);
         if (rc != 1 || *(muscle->muscle_model_index) < 1 || *(muscle->muscle_model_index) > 20)
         {
            if (muscle != &model->default_muscle && model->default_muscle.muscle_model_index != NULL)
            {
               *(muscle->muscle_model_index) = *(model->default_muscle.muscle_model_index);
               sim_message(recover, "Error reading muscle_model for muscle %s.", muscle->name);
               sim_message(none, "Using value from default muscle.");
            }
            else
            {
               sim_message(abort_action, "Error reading muscle_model for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         else
         {
            (*muscle->muscle_model_index)--;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "muscle_model_opt"))
      {
         muscle->muscle_model_opt_index = (int*)simm_malloc(sizeof(int));
         rc = fscanf(fp, "%d", muscle->muscle_model_opt_index);
         if (rc != 1 || *(muscle->muscle_model_opt_index) < 1 || *(muscle->muscle_model_opt_index) > 20)
         {
            if (muscle != &model->default_muscle && model->default_muscle.muscle_model_opt_index != NULL)
            {
               *(muscle->muscle_model_opt_index) = *(model->default_muscle.muscle_model_opt_index);
               sim_message(recover, "Error reading muscle_model_opt for muscle %s.", muscle->name);
               sim_message(none, "Using value from default muscle.");
            }
            else
            {
               sim_message(abort_action, "Error reading muscle_model_opt for muscle %s.", muscle->name);
               return code_bad;
            }
         }
         else
         {
            (*muscle->muscle_model_opt_index)--;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "visible"))
      {
         if (fscanf(fp, "%s", buffer) != 1)
         {
            sim_message(none, "Error reading \"visible\" parameter for muscle %s.", muscle->name);
            //sim_message(none, "Note: This parameter is ignored by the Pipeline.");
            return code_bad;
         }
         if (STRINGS_ARE_EQUAL(buffer, "true") || STRINGS_ARE_EQUAL(buffer, "yes"))
         {
            muscle->display = dpYes;
         }
         else if (STRINGS_ARE_EQUAL(buffer, "false") || STRINGS_ARE_EQUAL(buffer, "no"))
         {
            muscle->display = dpNo;
         }
         else
         {
            sim_message(none, "Error reading \"visible\" parameter for muscle %s.", muscle->name);
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "output")) // deprecated
      {
         fscanf(fp, "*s", buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "wrapobjects"))
      {
         sim_message(none, "Error reading wrap specification for muscle: %s", muscle->name);
         sim_message(none, "  (1) The \"wrapobjects\" keyword has been changed to \"wrapobject\".");
         sim_message(none, "  (2) Space characters are no longer allowed in wrap object names.");
         sim_message(none, "  (3) You may now specify the ellipsoid wrapping method following the wrap");
         sim_message(none, "      object\'s name with the keywords: \"hybrid\", \"midpoint\", or \"axial\".");
         sim_message(none, "      The wrapping method defaults to hybrid if none is specified.");

         read_line(fp, buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "wrapobject"))
      {
         char* p;
         int   i;
         dpMuscleWrapStruct* wrap = NULL;
         ReturnCode rc;
         dpBoolean algorithmSpecified = dpNo;

         read_line(fp, buffer);

         p = strtok(buffer, " \t");

         if (p)
         {
            wrap = (dpMuscleWrapStruct*)simm_malloc(sizeof(dpMuscleWrapStruct));

            if (wrap)
            {
               /* initialize the muscle wrap association */
               wrap->wrap_object = get_wrap_object(p);
               wrap->startPoint = -1;
               wrap->endPoint = -1;

               if (wrap->wrap_object == NULL)
               {
                  sim_message(none, "Unknown wrap object (%s) in muscle %s. Object will be ignored.", p, muscle->name);
                  FREE_IFNOTNULL(wrap);
               }
               else
               {
                  wrap->wrap_algorithm = wrap->wrap_object->wrap_algorithm;

                  for (i = 0; i < 2; i++)
                  {
                     wrap->mp_wrap[i].is_auto_wrap_point = dpYes;
                     wrap->mp_wrap[i].wrap_distance = 0.0;
                     wrap->mp_wrap[i].num_wrap_pts = 0;
                     wrap->mp_wrap[i].wrap_pts = NULL;
                     wrap->mp_wrap[i].isVia = dpNo;
                  }

                  for (i = 0; i < 3; i++)
                  {
                     wrap->c[i] = MINMDOUBLE;
                     wrap->r1[i] = MINMDOUBLE;
                     wrap->r2[i] = MINMDOUBLE;
                  }
               }
            }
         }

         /* if a wrap algorithm was specified, override the default algorithm */
         p = strtok(NULL, " \t");

         if (p && wrap)
         {
            for (i = 0; i < WE_NUM_WRAP_ALGORITHMS; i++)
            {
               if (STRINGS_ARE_EQUAL(p, get_wrap_algorithm_name(i)))
               {
                  wrap->wrap_algorithm = i;
                  algorithmSpecified = dpYes;
                  break;
               }
            }
         }

         /* if a range of muscle points was specified, override the defaults */
         if (algorithmSpecified == dpYes)
            p = strtok(NULL, " \t");

         if (p && wrap)
         {
            if (STRINGS_ARE_EQUAL(p, "range"))
            {
               p = strtok(NULL, " \t");
               if (p)
               {
                  wrap->startPoint = atoi(p);
                  if (wrap->startPoint == 0)
                  {
                     sim_message(recover, "Error reading wrap starting range in muscle %s.", muscle->name);
                  }
               }
               else
               {
                  sim_message(recover, "Error reading wrap starting range in muscle %s.", muscle->name);
               }
               p = strtok(NULL, " \t");
               if (p)
               {
                  wrap->endPoint = atoi(p);
                  if (wrap->endPoint == 0)
                  {
                     sim_message(recover, "Error reading wrap ending range in muscle %s.", muscle->name);
                  }
               }
               else
               {
                  sim_message(recover, "Error reading wrap ending range in muscle %s.", muscle->name);
               }
               if (wrap->startPoint < 1)
                  wrap->startPoint = -1;
               if (wrap->endPoint < 1)
                  wrap->endPoint = -1;
               if (wrap->startPoint > wrap->endPoint && wrap->endPoint != -1)
               {
                  wrap->startPoint = wrap->endPoint - 1;
                  sim_message(recover, "Changing wrap range to (%d, %d) for muscle %s.", wrap->startPoint, wrap->endPoint, muscle->name);
               }
            }
            else
            {
               sim_message(recover, "Unrecognized string \"%s\" found while reading muscle %s.", p, muscle->name);
            }
         }

         /* Now add the wrap structure to the muscle */
         if (wrap)
         {
            if (muscle->numWrapStructs == 0)
               muscle->wrapStruct = (dpMuscleWrapStruct**)simm_malloc(sizeof(dpMuscleWrapStruct*));
            else
               muscle->wrapStruct = (dpMuscleWrapStruct**)simm_realloc(muscle->wrapStruct,
               (muscle->numWrapStructs+1) * sizeof(dpMuscleWrapStruct*), &rc);
            muscle->wrapStruct[muscle->numWrapStructs++] = wrap;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "min_material"))
      {
         fscanf(fp, "%*s");
      }
      else if (STRINGS_ARE_EQUAL(buffer, "max_material"))
      {
         fscanf(fp, "%*s");
      }
      else if (STRINGS_ARE_EQUAL(buffer, "endmuscle"))
      {
         return code_fine;
      }
      else
      {
         check_dynamic_param_array(fp, muscle, buffer);
      }
   }

   return code_fine;
}



/* READ_MUSCLE_PARAMETER: this routine reads a single muscle parameter from
 * a file. The parameter must be stored as a pointer to a double in the
 * muscle structure. If there is an error when reading the parameter, the
 * pointer will be set to the corresponding parameter in the default muscle.
 */

static ReturnCode read_muscle_parameter(FILE* fp, dpMuscleStruct* muscle, dpMuscleStruct* default_muscle,
                                        double** param, double* default_param, char param_name[],
                                        double min_value, double max_value)
{
   *param = (double *)simm_malloc(sizeof(double));
   if (*param == NULL)
      return code_bad;

   if (fscanf(fp, "%lf", *param) != 1)
   {
      if (muscle != default_muscle && default_param != NULL)
      {
         free(*param);
         *param = default_param;
         sim_message(recover, "Error reading %s for muscle %s.", param_name, muscle->name);
         sim_message(none, "Using value from default muscle.");
         return code_fine;
      }
      else
      {
         free(*param);
         sim_message(abort_action, "Error reading %s for muscle %s.", param_name, muscle->name);
         return code_bad;
      }
   }
   else
   {
      if (min_value != DONT_CHECK_DOUBLE)
      {
         if (**param < min_value)
         {
            sim_message(none, "Error reading %s for muscle %s:", param_name, muscle->name);
            sim_message(abort_action, "   value of %lf is below min allowable value (%lf).", **param, min_value);
            free(*param);
            return code_bad;
         }
      }

      if (max_value != DONT_CHECK_DOUBLE)
      {
         if (**param > max_value)
         {
            sim_message(none, "Error reading %s for muscle %s:", param_name, muscle->name);
            sim_message(abort_action, "   value of %lf is above max allowable value (%lf).", **param, max_value);
            free(*param);
            return code_bad;
         }
      }
   }

   return code_fine;
}


/* READ_MUSCLE_FUNCTION: this routine reads a function for a muscle
 * (e.g. active force-length curve). If there is an error when reading from
 * the file, it sets the function equal to the corresponding function in the
 * default muscle.
 */
static ReturnCode read_muscle_function(FILE* fp, dpModelStruct* model, dpMuscleStruct* muscle,
                                       dpFunction*** mfunc, dpFunction** default_mfunc,
                                       char ending[], char description[])
{
   dpFunction newFunc;

   *mfunc = (dpFunction**)simm_malloc(sizeof(dpFunction*));
   if (*mfunc == NULL)
      return code_bad;

   malloc_function(&newFunc, FUNCTION_ARRAY_INCREMENT);

   if (read_double_array(fp, ending, description, &newFunc) == code_bad)
   {
      free_function(&newFunc);
      free(*mfunc);

      if (muscle != &model->default_muscle && default_mfunc != NULL)
      {
         *mfunc = default_mfunc;
         sim_message(recover, "Error reading %s for muscle %s.", description, muscle->name);
         sim_message(none, "Using function from default muscle.");
         return code_fine;
      }
      else
      {
         sim_message(abort_action, "Error reading %s for muscle %s.", description, muscle->name);
         return code_bad;
      }
   }

   newFunc.type = dpNaturalCubicSpline;
   newFunc.cutoff_frequency = -1.0;

   // DKB Oct 27, 2009 - added the defined status else coefficients do not get calculated
   // and simulations give garbage results
   newFunc.status = dpFunctionUserDefined;
   calc_function_coefficients(&newFunc);

   *(*mfunc) = load_simm_function(model, &newFunc);

   return code_fine;
}



static ReturnCode read_excitation(FILE* fp, dpModelStruct* model, dpMuscleStruct* muscle)
{
   int abscissa = -2;
   dpFunction newFunc;

   /* First read the name of the abscissa of the function. It should be either
    * the word "time" or the name of a generalized coordinate, of which this
    * excitation pattern will be a function.
    */
   if (fscanf(fp, "%s", buffer) != 1)
   {
      sim_message(abort_action, "Error reading name of abscissa in excitation function for muscle %s.", muscle->name);
      return code_bad;
   }

  // FREE_IFNOTNULL(muscle->excitation_abscissa);
   // commented out DKB Oct 7, 2009 - causing crash because free corrupts memory where
   // excitation func for muscle0 is stored (inverse.jnt model)
   muscle->excitation_abscissa = (void**)simm_malloc(sizeof(void*));
   if (muscle->excitation_abscissa == NULL)
      return code_bad;

   if (STRINGS_ARE_EQUAL(buffer, "time"))
   {
      *muscle->excitation_abscissa = TIME;
   }
   else
   {
      *muscle->excitation_abscissa = enter_gencoord(model, buffer);
      if (*muscle->excitation_abscissa == NULL) 
      {
         sim_message(none, "Error reading abscissa of excitation function for muscle %s.", muscle->name);
         sim_message(abort_action, "%s is neither \"time\" nor the name of a generalized coordinate.", buffer);
         return code_bad;
      }
   }

   muscle->excitation_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
   if (muscle->excitation_func == NULL)
      return code_bad;

   malloc_function(&newFunc, FUNCTION_ARRAY_INCREMENT);
   newFunc.type = excitationType;
   newFunc.cutoff_frequency = -1.0;

   if (read_double_array(fp, "endexcitation", "excitation-sequence", &newFunc) == code_bad)
   {
      free_function(&newFunc);

      if (muscle != &model->default_muscle && model->default_muscle.excitation_func != NULL)
      {
         muscle->excitation_func = model->default_muscle.excitation_func;
         sim_message(recover, "Error reading excitation sequence for muscle %s.", muscle->name);
         sim_message(none, "Using same sequence as for default muscle.");
         return code_fine;
      }
      else
      {
         sim_message(abort_action, "Error reading excitation sequence for muscle %s.", muscle->name);
         return code_bad;
      }
   }

   // DKB Oct 27, 2009 - added the defined status else coefficients do not get calculated
   // and simulations give garbage results
   newFunc.status = dpFunctionUserDefined;
   calc_function_coefficients(&newFunc);

   *muscle->excitation_func = load_simm_function(model, &newFunc);

   return code_fine;
}


/* READ_MUSCLE_GROUPS: This routine reads names of muscle groups from a file
 * until the string "endgroups" is found. Muscle groups are used in SIMM,
 * but have no use in an SD/FAST simulation. So that a SIMM muscle file can
 * be read in by the GMC without changes, however, this routine needs to exist.
 * It just reads strings from the file until "endgroups" is found.
 */
ReturnCode read_muscle_groups(FILE* fp)
{
   while (1)
   {
      if (fscanf(fp, "%s", buffer) != 1)
         return code_bad;

      if (STRINGS_ARE_EQUAL(buffer, "endgroups"))
         return code_fine;
   }
}


/* READ_MUSCLE_ATTACHMENT_POINTS: this routine reads muscle attachment points
 * from a file. It keeps reading from the file until the string "endpoints"
 * is encountered. It initially mallocs space for some points, and if it
 * fills up that space and finds more points, it reallocs the muscle point
 * array so that it can add the additional points.
 */

ReturnCode read_muscle_attachment_points(FILE* fp, dpModelStruct* model, dpMuscleStruct *muscle)
{
   int i, numpoints, function_num ;
   double range1, range2;
   dpMusclePoint* mp;
   char gencoord_name[CHARBUFFER], segment_name[CHARBUFFER], line2[CHARBUFFER], *line, com[CHARBUFFER];
   ReturnCode rc = code_fine;

   // initialize the muscle path
   muscle->path = (dpMusclePathStruct *)simm_malloc(sizeof(dpMusclePathStruct));
   if (init_musclepath(muscle->path) == code_bad)
      return code_bad;

   numpoints = 0;

   while (1)
   {
      if (fscanf(fp,"%s",buffer) != 1)
      {
         sim_message(none,"EOF while reading muscle attachment points");
         return code_bad;
      }

      if (STRINGS_ARE_EQUAL(buffer,"endpoints"))
      {
         muscle->path->num_orig_points = numpoints;
         return code_fine;
      }

      /* Check to see if you need to increase the size of the muscle-point
       * array before reading any more points.
       */
      if (numpoints >= muscle->path->mp_orig_array_size)
      {
         muscle->path->mp_orig_array_size += MP_ARRAY_INCREMENT;
         muscle->path->mp_orig = (dpMusclePoint *)simm_realloc(muscle->path->mp_orig, muscle->path->mp_orig_array_size * sizeof(dpMusclePoint), &rc);
         if (rc == code_bad)
         {
            muscle->path->mp_orig_array_size -= MP_ARRAY_INCREMENT;
            return rc;
         }
      }

      mp = &muscle->path->mp_orig[numpoints];

      // initialize the point
      init_musclepoint(mp);

      if (sscanf(buffer,"%lg", &mp->point[XX]) != 1)
      {
         if (sscanf(buffer,"f%d%s", &function_num, gencoord_name) != 2)
            return code_bad;
         strip_brackets_from_string(gencoord_name);
         mp->function[XX] = model->function[enter_function(model, function_num, dpYes)];
         mp->gencoord[XX] = enter_gencoord(model, gencoord_name);
         mp->isMovingPoint = dpYes;
         if (mp->gencoord[XX] == NULL) 
         {
            sim_message(none, "Gencoord %s referenced in muscle file but not defined in model.", gencoord_name);
            return code_bad;
         }
         if (mp->function[XX] == NULL)
         {
            sim_message(none, "Function %d referenced in muscle file but not defined in muscle file.", function_num);
            return code_bad;
         }
      }

      (void)read_nonempty_line(fp, line2);

      /* Read the y coordinate or function */
      line = parse_string(line2, type_string, (void*)buffer);
      if (sscanf(buffer, "%lg", &mp->point[YY]) != 1)
      {
         if (sscanf(buffer, "f%d%s", &function_num, gencoord_name) != 2)
            return code_bad;
         strip_brackets_from_string(gencoord_name);
         mp->function[YY] = model->function[enter_function(model, function_num, dpYes)];
         mp->gencoord[YY] = enter_gencoord(model, gencoord_name);
         mp->isMovingPoint = dpYes;
         if (mp->gencoord[YY] == NULL)  
         {
            sim_message(none, "Gencoord %s referenced in muscle file but not defined in model.", gencoord_name);
            return code_bad;
         }
         if (mp->function[YY] == NULL)
         {
            sim_message(none, "Function %d referenced in muscle file but not defined in muscle file.", function_num);
            return code_bad;
         }
      }

      /* Read the z coordinate or function */
      line = parse_string(line, type_string, (void*)buffer);
      if (sscanf(buffer, "%lg", &mp->point[ZZ]) != 1)
      {
         if (sscanf(buffer, "f%d%s", &function_num, gencoord_name) != 2)
            return code_bad;
         strip_brackets_from_string(gencoord_name);
         mp->function[ZZ] = model->function[enter_function(model, function_num, dpYes)];
         mp->gencoord[ZZ] = enter_gencoord(model, gencoord_name);
         mp->isMovingPoint = dpYes;
         if (mp->gencoord[ZZ] == NULL) 
         {
            sim_message(none, "Gencoord %s referenced in muscle file but not defined in model.", gencoord_name);
            return code_bad;
         }
         if (mp->function[ZZ] == NULL)
         {
            sim_message(none, "Function %d referenced in muscle file but not defined in muscle file.", function_num);
            return code_bad;
         }
      }

      /* read the keyword "segment" */
      line = parse_string(line, type_string, (void*)buffer);

      /* read the segment name */
      line = parse_string(line, type_string, (void*)segment_name);
      if (segment_name == NULL)
      {
         sim_message(none, "Error reading segment name in muscle %s", muscle->name);
         return code_bad;
      }

      mp->segment = enter_segment(model, segment_name);
      if (mp->segment == UNDEFINED_SEGMENT)
      {
         sim_message(none, "Segment %s referenced in muscle file but not defined in model.", segment_name);
         return code_bad;
      }

      /* In SD/FAST, the origin of a body's reference frame is located at the
       * center of mass. So subtract the coordinates of the center of mass
       * (w.r.t the SIMM reference frame) to get the coords of the muscle
       * point w.r.t the center of mass. The mass center coords for segment i
       * are located in the body segment array at element i+1 (because ground
       * is -1, but is the 0th element in the array).
       */
      for (i=0; i<3; i++)
         mp->point[i] -= model->body_segment[mp->segment+1].mass_center[i];

      /* read a string from the leftover part of the line. For most points,
       * the leftover will be NULL. For wrapping points, the first string
       * in the leftover should be "ranges".
       */
      line = parse_string(line, type_string, (void*)com);

      if (STRINGS_ARE_EQUAL(com, "range"))
      {
         line = parse_string(line, type_string, (void*)gencoord_name);
         if (gencoord_name == NULL)
         {
            return code_bad;
         }

         if (get_xypair_from_string(line, &range1, &range2) == code_bad)
            return code_bad;
         else
         {
            mp->viaRange.start = MIN(range1, range2);
            mp->viaRange.end = MAX(range1, range2);
         }
         mp->viaRange.gencoord = enter_gencoord(model, gencoord_name);
         if (mp->viaRange.gencoord == NULL) 
         {
            sim_message(none, "Gencoord %s referenced in muscle file but not defined in joints file.", gencoord_name);
            return code_bad;
         }
         mp->isVia = dpYes;
      }
      // support old files
      if (STRINGS_ARE_EQUAL(com, "ranges"))
      {
         line = parse_string(line, type_int, (void*)&i);
         if (i <= 0)
         {
            sim_message(none, "The number of ranges must be greater than 0.");
            return code_bad;
         }
         line = parse_string(line,type_string,(void*)gencoord_name);
         if (gencoord_name == NULL)
            return code_bad;

         if (get_xypair_from_string(line,&range1, &range2) == code_bad)
            return code_bad;
         else
         {
            mp->viaRange.start = MIN(range1, range2);
            mp->viaRange.end = MAX(range1, range2);
         }

         mp->viaRange.gencoord = enter_gencoord(model,gencoord_name);
         if (mp->viaRange.gencoord == NULL) 
            return code_bad;
         mp->isVia = dpYes;
      }
      numpoints++;
   }
   return code_fine;
}


static ReturnCode init_musclepoint(dpMusclePoint *mp)
{
   int i;

   mp->segment = UNDEFINED_SEGMENT;
   mp->selected = dpNo;
   mp->isMovingPoint = dpNo;
   mp->isVia = dpNo;
   mp->is_auto_wrap_point = dpNo;
   mp->viaRange.gencoord = NULL;
   mp->viaRange.start = MINMDOUBLE;
   mp->viaRange.end = MAXMDOUBLE;
   mp->num_wrap_pts = 0;
   mp->wrap_pts = NULL;
   mp->wrap_distance = 0;
   for (i = 0; i < 3; i++)
   {
      mp->point[i] = ERROR_DOUBLE;
      mp->function[i] = NULL;
      mp->gencoord[i] = NULL; 
   }
   return code_fine;
}

/* INIT_MUSCLE: this routine initializes the musclepath structure that holds the 
 * muscle points.
 */

static ReturnCode init_musclepath(dpMusclePathStruct *path)
{
   path->num_orig_points = 0;
   path->mp_orig_array_size = MP_ARRAY_INCREMENT;
   path->mp_orig = (dpMusclePoint *)simm_malloc(path->mp_orig_array_size * sizeof(dpMusclePoint));
   if (path->mp_orig == NULL)
      return code_bad;

   path->num_points = 0;
   path->mp_array_size = 0;
   path->mp = NULL;

   
   return code_fine;
}

/* INIT_MUSCLE: this routine initializes much of the muscle structure. Before
 * a given muscle is read-in from an input file, each of the pointers to its
 * muscle-specific parameters is set to point into the default-muscle structure.
 * Thus if a certain parameter is not specified for a muscle (e.g. force-length
 * curve), the default-muscle's parameter is used. This cuts down significantly
 * on the amount of information that must be supplied in a muscle file since
 * most muscles share the same activeflcurve, tendonflcurve, passiveflcurve,
 * and fvcurve, as well as some tendon characteristics.
 */
static ReturnCode init_muscle(dpModelStruct* model, dpMuscleStruct* muscle)
{
   int i;

   memset(muscle, 0, sizeof(dpMuscleStruct));

   muscle->name = model->default_muscle.name;
   muscle->max_isometric_force = model->default_muscle.max_isometric_force;
   muscle->pennation_angle = model->default_muscle.pennation_angle;
   muscle->max_contraction_vel = model->default_muscle.max_contraction_vel;
   muscle->optimal_fiber_length = model->default_muscle.optimal_fiber_length;
   muscle->resting_tendon_length = model->default_muscle.resting_tendon_length;
   muscle->excitation_func = model->default_muscle.excitation_func;
   muscle->excitation_abscissa = model->default_muscle.excitation_abscissa;
   muscle->tendon_force_len_func = model->default_muscle.tendon_force_len_func;
   muscle->active_force_len_func = model->default_muscle.active_force_len_func;
   muscle->passive_force_len_func = model->default_muscle.passive_force_len_func;
   muscle->force_vel_func = model->default_muscle.force_vel_func;
   muscle->muscle_model_index = model->default_muscle.muscle_model_index;
   muscle->muscle_model_opt_index = model->default_muscle.muscle_model_opt_index;

   if (init_dynamic_param_array(muscle, &model->default_muscle) == code_bad)
      return code_bad;

   // Always make room for the moment arms because you don't know yet if the
   // user has requested them.
   muscle->nummomentarms = model->nu;
   muscle->momentarms = (double*)simm_malloc(muscle->nummomentarms * sizeof(double));
   if (muscle->momentarms == NULL)
      return code_bad;

   return code_fine;
}


static void postprocess_muscle(dpMuscleStruct* muscle)
{
   muscle->path->mp_array_size = muscle->path->num_orig_points + (muscle->numWrapStructs * 2);
   muscle->path->mp = (dpMusclePoint**)simm_malloc(sizeof(dpMusclePoint*) * muscle->path->mp_array_size);

   /* When wrapping is calculated, the mp[] array will be filled in
    * and num_points set appropriately.
    */
   muscle->path->num_points = 0;
}


static ReturnCode read_dynamic_parameter_names(FILE* fp, dpMuscleStruct* muscle)
{
   int array_size = ARRAY_INCREMENT;
   ReturnCode rc = code_fine;

   muscle->num_dynamic_params = 0;
   muscle->dynamic_param_names = (char**)simm_malloc(array_size * sizeof(char*));

   if (muscle->dynamic_param_names == NULL)
      return code_bad;

   while (1)
   {
      if (fscanf(fp, "%s", buffer) != 1)
      {
         sim_message(none, "EOF while reading dynamic parameters");
         return code_bad;
      }

      if (STRINGS_ARE_EQUAL(buffer, "enddynamicparameters"))
         break;

      // Check to see if you need to increase the size of the array before
      // reading any more parameter names.
      if (muscle->num_dynamic_params >= array_size)
      {
         array_size += ARRAY_INCREMENT;
         muscle->dynamic_param_names = (char**)simm_realloc(muscle->dynamic_param_names,
            array_size*sizeof(char*), &rc);

         if (rc == code_bad)
            return code_bad;
      }

      mstrcpy(&muscle->dynamic_param_names[muscle->num_dynamic_params], buffer);

      muscle->num_dynamic_params++;
   }

   muscle->dynamic_param_names = (char**)simm_realloc(muscle->dynamic_param_names,
      muscle->num_dynamic_params*sizeof(char*), &rc);

   return rc;
}


static ReturnCode init_dynamic_param_array(dpMuscleStruct* muscle, dpMuscleStruct* default_muscle)
{
   int i;

   muscle->num_dynamic_params = default_muscle->num_dynamic_params;
   muscle->dynamic_param_names = default_muscle->dynamic_param_names;
   muscle->dynamic_params = (double**)simm_calloc(muscle->num_dynamic_params, sizeof(double*));

   if (muscle->dynamic_params == NULL)
      return code_bad;

   for (i=0; i<muscle->num_dynamic_params; i++)
      muscle->dynamic_params[i] = default_muscle->dynamic_params[i];

   return code_fine;
}


static void check_dynamic_param_array(FILE* fp, dpMuscleStruct* muscle, char text_string[])
{
   int i;

   for (i=0; i<muscle->num_dynamic_params; i++)
   {
      if (STRINGS_ARE_EQUAL(text_string,muscle->dynamic_param_names[i]))
      {
         muscle->dynamic_params[i] = (double*)simm_malloc(sizeof(double));
         fscanf(fp,"%lg", muscle->dynamic_params[i]);
         return;
      }
   }

   sim_message(recover, "Unrecognized string \"%s\" found while reading muscle %s.", text_string, muscle->name);
}


static const char* get_wrap_algorithm_name (int i)
{
   static char* names[WE_NUM_WRAP_ALGORITHMS] = { "hybrid", "midpoint", "axial" };
   
   if (i >= 0 && i < WE_NUM_WRAP_ALGORITHMS)
      return names[i];
   
   return NULL;
}

ReturnCode init_functions(dpModelStruct* model)
{
   int i;

   model->num_functions = 0;
   model->function = (dpFunction **)simm_calloc(model->function_array_size, sizeof(dpFunction*));
   for (i = 0; i < model->function_array_size; i++)
   {
      //model->function[i] = (dpFunction *)simm_malloc(sizeof(dpFunction));
      model->function[i] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         //memset(&model->function[i], 0, sizeof(dpFunction));
      model->function[i]->status = dpFunctionUndefined;
      model->function[i]->usernum = UNDEFINED_USERNUM;
      model->function[i]->type = dpNaturalCubicSpline;
      model->function[i]->used = dpNo;

/*      model->function[i]->source = dpMuscleFile;
      model->function[i]->numpoints = 0;
      model->function[i]->coefficient_array_size = 0;
      model->function[i]->cutoff_frequency = 0;
      model->function[i]->x = NULL;
      model->function[i]->y = NULL;
      model->function[i]->b = NULL;
      model->function[i]->c = NULL;
      model->function[i]->d = NULL;*/

   }

   return code_fine;
}


/* READ_FUNCTION: this routine reads a function for a muscle
 * (e.g. active force-length curve). If there is an error when reading from
 * the file, it sets the function equal to the corresponding function in the
 * default muscle.
 */
ReturnCode read_function(FILE* fp, dpModelStruct* model, dpFunctionType funcType, const char endTag[])
{
   dpFunction func;
   char fname[CHARBUFFER];

   func.type = funcType;
   func.cutoff_frequency = 0.0;
   func.source = dpMuscleFile;

   // Malloc space for the points and coefficients of this function.
   (void)malloc_function(&func, FUNCTION_ARRAY_INCREMENT);

   if (read_string(fp, buffer) == EOF)
   {
      sim_message(none, "Unexpected EOF found reading a function");
      return code_bad;
   }

   // Read the function name, which is usually a number (e.g. "f19").
   if (sscanf(buffer, "f%d", &func.usernum) != 1)
   {
      sim_message(none,"Error reading a function name");
      return code_bad;
   }

   if (func.usernum < 0)
   {
      sim_message(none,"Function numbers must be non-negative (%d not allowed)", func.usernum);
      return code_bad;
   }

   (void)sprintf(fname, "f%d", func.usernum);

   // Get the array of x-y pairs which comprises the function.
   if (read_double_array(fp, endTag, fname, &func) == code_bad)
   {
      free(func.x);
      free(func.y);
      free(func.b);
      free(func.c);
      free(func.d);
      return code_bad;
   }

   // If you made it to here, a valid function with two or more points has been
   // read-in. Calculate the coefficients, and load it into the model's function list.
   // DKB Oct 27, 2009 - added the defined status else coefficients do not get calculated
   // and simulations give garbage results
   func.status = dpFunctionUserDefined;
   calc_function_coefficients(&func);

   if (load_user_function(model, &func) == NULL)
      return code_bad;

   return code_fine;
}


static const char* get_function_type_name(dpFunctionType funcType)
{
   switch (funcType)
   {
      case dpStepFunction:
         return "step_function";
      case dpLinearFunction:
         return "linear_function";
      case dpNaturalCubicSpline:
         return "natural_cubic";
      case dpGCVSpline:
         return "gcv_spline";
      case dpFunctionTypeUndefined:
      default:
         return "function_type_undefined"; // should probably be an error
   }
}

static const char* get_function_tag(dpFunctionType funcType, int beginEnd)
{
   // beginEnd = 0 for "begin, beginEnd = 1 for "end"
   switch (funcType)
   {
      case dpStepFunction:
         return (beginEnd) ? "endstepfunction" : "beginstepfunction";
      case dpLinearFunction:
         return (beginEnd) ? "endlinearfunction" : "beginlinearfunction";
      case dpNaturalCubicSpline:
         return (beginEnd) ? "endnaturalcubicspline" : "beginnaturalcubicspline";
      case dpGCVSpline:
         return (beginEnd) ? "endgcvspline" : "begingcvspline";
      case dpFunctionTypeUndefined:
      default:
         return (beginEnd) ? "endfunction" : "beginfunction"; // old-style function definitions
   }
}

/* This function scans forward to the end of a muscle definition.
 * It should be called only after encountering an error in the definition,
 * as it skips over everything until "endmuscle" is found.
 */
static void finish_muscle_definition(FILE* fp)
{
   while (1)
   {
      fscanf(fp, "%s", buffer);
      if (STRINGS_ARE_EQUAL(buffer, "endmuscle"))
         break;
   }
}
