/*******************************************************************************

   READTOOLS.C

   Author: Peter Loan
           Krystyne Blaikie

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains tools for reading ascii text from
      input files, and for extracting numbers and strings from the text.

   Routines:
      get_muscle_file_name      : reads the muscle file name from the parameters file
      read_parameters_file      : reads the parameters file for the stand-alone app
      read_dllparameters_file   : reads the parameters file for the DLL app
      preprocess_file           : runs acpp on a file, then opens it
      read_string               : reads a string from a file
      read_line                 : reads a possibly empty line from a file
      read_nonempty_line        : reads a non-empty line of chars from a file
      read_til_tokens           : reads from a file until certain tokens are found
      clean_string              : removes leading and trailing white space
      read_muscle_groups        : reads a set of muscle group names from a file
      read_muscle_attachment_points : reads muscle attachment points from a file
      read_double_array         : reads an array of x-y pairs from a file
      string_contains_closing_paren : checks to see if string has closing paren
      get_xy_pair_from_string   : reads x-y pair from a file, e.g. (0.00,0.00)
      parse_string              : extracts variable w/specified type from string
      enter_segment             : enters a segment name into a model's database
      enter_gencoord            : enters a gencoord name into a model's database
      enter_function            : enters a function's number into a model's database
      load_input_data           : does set up for reading input data (motion or kinetics files)
      read_kinetics_file        : reads kinetics data from a .ktx file
      read_motion_file          : reads kinetics data from a .mot file
      verify_kinetics_data      : checks kinetics data from .ktx file for completeness
      verify_motion_data        : checks kinetics data from .mot file for completeness
      name_is_gencoord          : checks if a string is a gencoord or gc deriv
      name_is_body_segment      : checks if a string is a body segment name
      name_is_muscle            : checks if a string is a muscle name
      name_is_forceplate        : checks if a string is a forceplate name
      string_has_suffix         : compares one string to the end of another
      simm_malloc               : wrapper routine for malloc
      simm_calloc               : wrapper routine for calloc
      simm_realloc              : wrapper routine for realloc
      simm_fopen                : wrapper routine for fopen
      convert_string            : converts special chars in a string to '_'
      is_absolute_path          : determines whether filename is an absolute path
      read_spring               : reads a spring element from parameters file
      read_spring_floor         : reads a spring floor element from parameters file
      append_if_necessary       : appends a DIR_SEP_CHAR to a file name if necessary
      identify_elements_to_transfer : identify columns from motion data to transfer to output data
      is_preference_on          : determine how the preference has been set
      get_function_suffix       : return a string containing the function's type 
      get_suffix                : get the suffix in the given string
      get_temp_file_name        : get a temporary file name

*******************************************************************************/

#include "universal.h"


/*************** DEFINES (for this file only) *********************************/
typedef enum {POSX = 0, POSY, POSZ, VECX, VECY, VECZ, FREETX, FREETY, FREETZ} FTcomponent;


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/
static char* velocity_suffix = "_vel";
static char* acceleration_suffix = "_acc";
static char* force_suffix = "_force";
static char* torque_suffix = "_torque";
static char* moment_suffix = "_moment";
static char* linear_text = "_lin";
static char* natural_cubic_text = "_spl";
static char* gcv_text = "_gcv";
static char* px_suffix = "_force_px";
static char* py_suffix = "_force_py";
static char* pz_suffix = "_force_pz";
static char* vx_suffix = "_force_vx";
static char* vy_suffix = "_force_vy";
static char* vz_suffix = "_force_vz";
static char* ax_suffix = "_torque_x";
static char* ay_suffix = "_torque_y";
static char* az_suffix = "_torque_z";


/**************** GLOBAL VARIABLES (used in only a few files) *****************/
char buffer[CHARBUFFER];
char buffer2[CHARBUFFER];
char *bone_path = NULL;


/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern dpModelStruct* sdm;                      // defined in main.c or dllmain.c
extern dpBoolean verbose;                       // defined in output.c
extern dpBoolean isStandAlone;                  // defined in main.c or dllmain.c
extern dpSimulationParameters params;           // defined in main.c or dllmain.c

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static void get_axis_and_sign_from_quadrant(char quadrant[], int* axis, int* sign);
static dpWrapObjectType get_wrap_type_index_from_name(char type_name[]);
static int get_wrap_algorithm_index_from_name(char type_name[]);
static ReturnCode verify_kinetics_data(dpModelStruct* sdm, MotionData* data);
static void map_motion_force(dpModelStruct* sdm, MotionData* data, int segNum, FTcomponent comp,
                             dpFunctionType functionType, double cutoffFrequency, int column);
static void map_motion_torque(dpModelStruct* sdm, MotionData* data, int segNum, FTcomponent comp,
                              dpFunctionType functionType, double cutoffFrequency, int column);
static ReturnCode verify_motion_data(dpModelStruct* sdm, MotionData* data);
static int name_is_gencoord(char name[], dpModelStruct* sdm, char suffix[], dpFunctionType* functionType, double* cutoffFrequency);
static int name_is_body_segment(char name[], dpModelStruct* sdm, char suffix[], dpFunctionType* functionType, double* cutoffFrequency);
static int name_is_forceplate(char name[], dpModelStruct *sdm, MotionData *motion, char suffix[], char **component, dpFunctionType* functionType, double *cutoffFrequency);
static dpBoolean string_has_suffix(char str1[], char suffix[]);
static dpBoolean string_contains_closing_paren(char str_buffer[]);
static int find_highest_usernum(dpModelStruct* model);
static ReturnCode read_spring_floor(FILE* fp);
static ReturnCode read_spring_surface(FILE* fp);
static ReturnCode read_spring(FILE* fp);
static ReturnCode read_force_matte(FILE* fp);
static void add_foot_print(MotionData* motion, const char name[], double scale, double translation[], double euler_rotation[]);
static void add_force_plate(MotionData* motion, const char name[], double scale[], double translation[], double euler_rotation[]);
static void update_column_names(MotionData *motion);

/* GET_MUSCLE_FILE_NAME: read the muscle file name from the parameters file and return it. */
ReturnCode get_muscle_file_name(char params_file[], char** muscle_file)
{
   char* filename;
   FILE* fp;   

   if ((fp = fopen(params_file, "r")) == NULL)
      return code_bad;

   while (1)
   {
      if (fscanf(fp, "%s", buffer) != 1)
         break;

      if (buffer[0] == '#')
      {
         read_line(fp, buffer);
         continue;
      }

      if (STRINGS_ARE_EQUAL(buffer, "muscle_file") && muscle_file)
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading muscle_file name in %s.", params_file);
         else
            mstrcpy(muscle_file, filename);
      }
   }

   fclose(fp);

   return code_fine;
}

/* READ_PARAMETERS_FILE: open the parameters file and read and store all the information about
 * input and output files, input and output options.
 */
ReturnCode read_parameters_file(char params_file[], dpModelStruct* sdm, char **muscle_file, char **input_data_file, char** output_motion_file, char** output_kinetics_file)
{
   int i, j, k, index;
   double mass, inertia[3][3];
   char* filename;
   FILE* fp;   

   if ((fp = fopen(params_file,"r")) == NULL)
      return code_bad;

#if CONTACT_DETECTION
   init_object_info();
#endif

   while (1)
   {
      if (fscanf(fp, "%s", buffer) != 1)
         break;

      if (buffer[0] == '#')
      {
         read_line(fp, buffer);
         continue;
      }

      if (STRINGS_ARE_EQUAL(buffer,"verbose"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            verbose = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            verbose = dpNo;
         else
            sim_message(exit_program, "Error reading verbose mode (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"kinetics_file") && input_data_file)
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading kinetics_file name in %s.", params_file);
         else
            mstrcpy(input_data_file,filename);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_motion_file") && output_motion_file)
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading output_motion_file name in %s.", params_file);
         else
            mstrcpy(output_motion_file,filename);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_kinetics_file") && output_kinetics_file)
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading output_kinetics_file name in %s.", params_file);
         else
            mstrcpy(output_kinetics_file,filename);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "muscle_file"))
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading muscle_file name in %s.", params_file);
         else
            mstrcpy(muscle_file,filename);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "bone_path"))
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading bone_path name in %s.", params_file);
         else
            mstrcpy(&bone_path,filename);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"input_motion_file") && input_data_file)
      {
         read_line(fp, buffer);
         filename = clean_string(buffer);
         if (filename == NULL || strlen(filename) < 1)
            sim_message(exit_program, "Error reading input_motion_file name in %s.", params_file);
         else
            mstrcpy(input_data_file,filename);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"kinematic_smoothing_cutoff"))
      {
         double temp;
         if (fscanf(fp,"%lf", &temp) != 1)
            sim_message(exit_program, "Error reading kinematic_smoothing_cutoff from file %s.", params_file);
         params.inputOptions.kinematic_cutoff = (int)temp;
         params.inputOptions.kinematic_cutoff_defined = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"force_smoothing_cutoff"))
      {
         double temp;
         if (fscanf(fp,"%lf", &temp) != 1)
            sim_message(exit_program, "Error reading force_smoothing_cutoff from file %s.", params_file);
         params.inputOptions.force_cutoff = (int)temp;
         params.inputOptions.force_cutoff_defined = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"emg_smoothing_cutoff") ||
         STRINGS_ARE_EQUAL(buffer,"EMG_smoothing_cutoff"))
      {
         double temp;
         if (fscanf(fp,"%lf", &temp) != 1)
            sim_message(exit_program, "Error reading emg_smoothing_cutoff from file %s.", params_file);
         params.inputOptions.emg_cutoff = (int)temp;
         params.inputOptions.emg_cutoff_defined = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"include_muscles"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"all"))
            params.inputOptions.muscles = dpAllMuscles;
         else if (STRINGS_ARE_EQUAL(buffer,"none"))
            params.inputOptions.muscles = dpNoMuscles;
         else if (STRINGS_ARE_EQUAL(buffer,"visible"))
            params.inputOptions.muscles = dpVisibleMuscles;
         else if (STRINGS_ARE_EQUAL(buffer,"withEMG"))
            params.inputOptions.muscles = dpMusclesWithEMG;
         else if (STRINGS_ARE_EQUAL(buffer,"withExcitation"))
            params.inputOptions.muscles = dpMusclesWithExcitation;
         else
            sim_message(exit_program, "Error reading include muscles (all/none/visible/withEMG/withExcitation) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "include_kinematics"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            params.inputOptions.include_kinematics = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            params.inputOptions.include_kinematics = dpNo;
         else
            sim_message(exit_program, "Error reading include_kinematics (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "include_forces"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            params.inputOptions.include_forces = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            params.inputOptions.include_forces = dpNo;
         else
            sim_message(exit_program, "Error reading include_forces (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "include_emg"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            params.inputOptions.include_emg = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            params.inputOptions.include_emg = dpNo;
         else
            sim_message(exit_program, "Error reading include_emg (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"baumgarte_stabilization"))
      {
         double temp;
         if (fscanf(fp,"%lf", &temp) != 1)
            sim_message(exit_program, "Error reading baumgarte_stabilization from file %s.", params_file);
         params.baumgarte_stabilization = temp;
      }
#ifndef INVERSE
      else if (STRINGS_ARE_EQUAL(buffer,"spring"))
      {
         read_spring(fp);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"spring_floor"))
      {
         if (read_spring_floor(fp) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"spring_surface"))
      {
         if (read_spring_surface(fp) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"force_matte"))
      {
         if (read_force_matte(fp) == code_bad)
            return code_bad;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"start_time"))
      {
         if (fscanf(fp,"%lf", &params.startTime) != 1)
            sim_message(exit_program, "Error reading start_time from file %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"end_time"))
      {
         if (fscanf(fp,"%lf", &params.endTime) != 1)
            sim_message(exit_program, "Error reading end_time from file %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"step_size"))
      {
         if (fscanf(fp,"%lf", &params.stepSize) != 1)
            sim_message(exit_program, "Error reading step_size from file %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_gencoord_values"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.gencoordValues, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.gencoordValues);
                  params.outputOptions.gencoordValues.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.gencoordValues);
         else
            sim_message(exit_program, "Error reading output_gencoord_values (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_gencoord_velocities"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.gencoordVelocities, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.gencoordVelocities);
                  params.outputOptions.gencoordVelocities.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.gencoordVelocities);
         else
            sim_message(exit_program, "Error reading output_gencoord_velocities (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_gencoord_accelerations"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.gencoordAccelerations, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.gencoordAccelerations);
                  params.outputOptions.gencoordAccelerations.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.gencoordAccelerations);
         else
            sim_message(exit_program, "Error reading output_gencoord_accelerations (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_activations"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleActivations.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleActivations.state = 1;
            allocate_name_list(&params.outputOptions.muscleActivations, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleActivations);
                  params.outputOptions.muscleActivations.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.muscleActivations);
         else
            sim_message(exit_program, "Error reading output_muscle_activations (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_lengths"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleLengths.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleLengths.state = 1;
            allocate_name_list(&params.outputOptions.muscleLengths, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleLengths);
                  params.outputOptions.muscleLengths.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.muscleLengths);
         else
            sim_message(exit_program, "Error reading output_muscle_lengths (true/yes or false/no) in %s.", params_file);
      }
     else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_fiber_lengths"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleFiberLengths.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleFiberLengths.state = 1;
            allocate_name_list(&params.outputOptions.muscleFiberLengths, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleFiberLengths);
                  params.outputOptions.muscleFiberLengths.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.muscleFiberLengths);
         else
            sim_message(exit_program, "Error reading output_muscle_fiber_lengths (true/yes or false/no) in %s.", params_file);
      }
     else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_fiber_velocities"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleFiberVelocities.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleFiberVelocities.state = 1;
            allocate_name_list(&params.outputOptions.muscleFiberVelocities, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleFiberVelocities);
                  params.outputOptions.muscleFiberVelocities.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.muscleFiberVelocities);
         else
            sim_message(exit_program, "Error reading output_muscle_fiber_velocities (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_forces"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleForces.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleForces.state = 1;
            allocate_name_list(&params.outputOptions.muscleForces, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleForces);
                  params.outputOptions.muscleForces.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.muscleForces);
         else
            sim_message(exit_program, "Error reading output_muscle_forces (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_muscles"))
      {
         sim_message(none, "Parameter output_muscles is no longer supported.");
         sim_message(none, "Please use one or more of the following instead:");
         sim_message(none, "   output_muscle_activations");
         sim_message(none, "   output_muscle_lengths");
         sim_message(none, "   output_muscle_forces");
         sim_message(none, "   output_muscle_moment_arms");
         sim_message(none, "   output_muscle_moments");
         sim_message(exit_program, "   output_total_muscle_moments");
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_contact_forces"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            params.outputOptions.contactForces = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            params.outputOptions.contactForces = dpNo;
         else
            sim_message(exit_program, "Error reading output_contact_forces (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_gencoord_moments"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.traditionalInverse = dpYes;
            allocate_name_list(&params.outputOptions.gencoordMoments, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.gencoordMoments);
                  params.outputOptions.gencoordMoments.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
         {
            // can't set it to dpNo because output_corrected_gencoord_moments may have already set it to yes
            //params.outputOptions.gencoordMoments = dpNo;
         }
         else
            sim_message(exit_program, "Error reading output_gencoord_moments (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_corrected_gencoord_moments"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.traditionalInverse = dpNo;
            allocate_name_list(&params.outputOptions.gencoordMoments, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.gencoordMoments);
                  params.outputOptions.gencoordMoments.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
         {
            // can't set it to dpNo because output_gencoord_moments may have already set it to yes
            //params.outputOptions.gencoordMoments = dpNo;
         }
         else
            sim_message(exit_program, "Error reading output_corrected_gencoord_moments (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_gencoord_powers"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.gencoordPowers, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.gencoordPowers);
                  params.outputOptions.gencoordPowers.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.gencoordPowers);
         else
            sim_message(exit_program, "Error reading output_gencoord_powers (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_optimized_muscle_activations"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.optimizedActivations_Muscles.state = 0;
         params.outputOptions.optimizedActivations_Gencoords.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.optimizedActivations_Muscles.state = 1;
            allocate_name_list(&params.outputOptions.optimizedActivations_Muscles, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.optimizedActivations_Muscles);
                  params.outputOptions.optimizedActivations_Muscles.index[index] = i;
            }
            params.outputOptions.optimizedActivations_Gencoords.state = 1;
            allocate_name_list(&params.outputOptions.optimizedActivations_Gencoords, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.optimizedActivations_Gencoords);
                  params.outputOptions.optimizedActivations_Gencoords.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
         {
            clear_name_list(&params.outputOptions.optimizedActivations_Muscles);
            clear_name_list(&params.outputOptions.optimizedActivations_Gencoords);
         }
         else
            sim_message(exit_program, "Error reading output_optimized_muscle_activations (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_moment_arms"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleMomentArms_Gencoords.state = 0;
         params.outputOptions.muscleMomentArms_Muscles.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleMomentArms_Gencoords.state = 1;
            params.outputOptions.muscleMomentArms_Muscles.state = 1;
            allocate_name_list(&params.outputOptions.muscleMomentArms_Muscles, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleMomentArms_Muscles);
                  params.outputOptions.muscleMomentArms_Muscles.index[index] = i;
            }
            allocate_name_list(&params.outputOptions.muscleMomentArms_Gencoords, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.muscleMomentArms_Gencoords);
                  params.outputOptions.muscleMomentArms_Gencoords.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
         {
            clear_name_list(&params.outputOptions.muscleMomentArms_Muscles);
            clear_name_list(&params.outputOptions.muscleMomentArms_Gencoords);
         }
         else
            sim_message(exit_program, "Error reading output_muscle_moment_arms (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_muscle_moments"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.muscleMoments_Gencoords.state = 0;
         params.outputOptions.muscleMoments_Muscles.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.muscleMoments_Gencoords.state = 1;
            params.outputOptions.muscleMoments_Muscles.state = 1;
            allocate_name_list(&params.outputOptions.muscleMoments_Muscles, sdm->num_muscles);
            for (i=0; i<sdm->num_muscles; i++)
            {
                  index = add_name_to_list(sdm->muscles[i].name, &params.outputOptions.muscleMoments_Muscles);
                  params.outputOptions.muscleMoments_Muscles.index[index] = i;
            }
            allocate_name_list(&params.outputOptions.muscleMoments_Gencoords, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.muscleMoments_Gencoords);
                  params.outputOptions.muscleMoments_Gencoords.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
         {
            clear_name_list(&params.outputOptions.muscleMoments_Muscles);
            clear_name_list(&params.outputOptions.muscleMoments_Gencoords);
         }
         else
            sim_message(exit_program, "Error reading output_muscle_moments (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_total_muscle_moments"))
      {
         fscanf(fp,"%s", buffer);
         params.outputOptions.totalMuscleMoments.state = 0;
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            params.outputOptions.totalMuscleMoments.state = 1;
            allocate_name_list(&params.outputOptions.totalMuscleMoments, sdm->nu);
            for (i=0; i<sdm->nu; i++)
            {
               if (sdm->q[i].type == dpUnconstrainedQ || sdm->q[i].type == dpPrescribedQ)
               {
                  index = add_name_to_list(sdm->q[i].name, &params.outputOptions.totalMuscleMoments);
                  params.outputOptions.totalMuscleMoments.index[index] = i;
               }
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.totalMuscleMoments);
         else
            sim_message(exit_program, "Error reading output_total_muscle_moments (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_joint_reaction_forces"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.jointReactionForces, sdm->num_joints);
            for (i=0; i<sdm->num_joints; i++)
            {
               index = add_name_to_list(sdm->joint[i].name, &params.outputOptions.jointReactionForces);
               params.outputOptions.jointReactionForces.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.jointReactionForces);
         else
            sim_message(exit_program, "Error reading output_joint_reaction_forces (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_joint_reaction_torques"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.jointReactionTorques, sdm->num_joints);
            for (i=0; i<sdm->num_joints; i++)
            {
               index = add_name_to_list(sdm->joint[i].name, &params.outputOptions.jointReactionTorques);
               params.outputOptions.jointReactionTorques.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.jointReactionTorques);
         else
            sim_message(exit_program, "Error reading output_joint_reaction_torques (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_mass_center_positions"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.massCenterPositions, sdm->num_body_segments);
            for (i=0; i<sdm->num_body_segments; i++)
            {
               // indices stored are the indices into the body_segment list, not the SD/FAST indices
               index = add_name_to_list(sdm->body_segment[i].name, &params.outputOptions.massCenterPositions);
               params.outputOptions.massCenterPositions.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.massCenterPositions);
         else
            sim_message(exit_program, "Error reading output_mass_center_positions (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_mass_center_velocities"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
         {
            allocate_name_list(&params.outputOptions.massCenterVelocities, sdm->num_body_segments);
            for (i=0; i<sdm->num_body_segments; i++)
            {
               // indices stored are the indices into the body_segment list, not the SD/FAST indices
               index = add_name_to_list(sdm->body_segment[i].name, &params.outputOptions.massCenterVelocities);
               params.outputOptions.massCenterVelocities.index[index] = i;
            }
         }
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            clear_name_list(&params.outputOptions.massCenterVelocities);
         else
            sim_message(exit_program, "Error reading output_mass_center_velocities (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"output_system_energy"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            params.outputOptions.systemEnergy = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            params.outputOptions.systemEnergy = dpNo;
         else
            sim_message(exit_program, "Error reading output_system_energy (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "output_spring_forces"))
      {
         fscanf(fp,"%s", buffer);
         if (STRINGS_ARE_EQUAL(buffer,"true") || STRINGS_ARE_EQUAL(buffer,"yes"))
            params.outputOptions.springForces = dpYes;
         else if (STRINGS_ARE_EQUAL(buffer,"false") || STRINGS_ARE_EQUAL(buffer,"no"))
            params.outputOptions.springForces = dpNo;
         else
            sim_message(exit_program, "Error reading output_spring_forces (true/yes or false/no) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "num_output_spring_forces"))
      {
         if (fscanf(fp,"%d", &params.outputOptions.numSpringForces) != 1)
            sim_message(exit_program, "Error reading num_output_spring_forces (integer) in %s.", params_file);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"object"))
      {
#if CONTACT_DETECTION
         if (read_object(fp) == code_bad)
            sim_message(exit_program, "Error reading contact object in %s.", params_file);
#else
         sim_message(none, "Error reading \"object\" in %s.", params_file);
         sim_message(none, "   You must set CONTACT_DETECTION=1 in your project settings");
         sim_message(exit_program, "   in order to use contact detection.");
#endif
      }
      else if (STRINGS_ARE_EQUAL(buffer,"begin_group"))
      {
#if CONTACT_DETECTION
         if (read_object_group(fp) == code_bad)
            sim_message(exit_program, "Error reading contact group in %s.", params_file);
#else
         sim_message(none, "Error reading \"begin_group\" in %s.", params_file);
         sim_message(none, "   You must set CONTACT_DETECTION=1 in your project settings");
         sim_message(exit_program, "   in order to use contact detection.");
#endif
      }
      else if (STRINGS_ARE_EQUAL(buffer,"contact_pair"))
      {
#if CONTACT_DETECTION
         if (read_contact_pair(fp) == code_bad)
            sim_message(exit_program, "Error reading contact pair in %s.", params_file);
#else
         sim_message(none, "Error reading \"contact_pair\" in %s.", params_file);
         sim_message(none, "   You must set CONTACT_DETECTION=1 in your project settings");
         sim_message(exit_program, "   in order to use contact detection.");
#endif
      }
#endif
      else
      {
         // see if the name is a segment
         for (i=0; i<sdm->num_body_segments; i++)
         {
            if (STRINGS_ARE_EQUAL(buffer,sdm->body_segment[i].name))
            {
               fscanf(fp,"%s", buffer2);
               if (STRINGS_ARE_EQUAL(buffer2,"mass"))
               {
                  if (fscanf(fp,"%lf", &mass) != 1)
                  {
                     sim_message(recover, "Error reading mass for %s in %s.", buffer, params_file);
                  }
                  else
                  {
                     sdmass(i-1,mass);
                     if (check_for_sderror("READ_PARAMETERS_FILE") == 19)
                     {
                        sim_message(recover, "READ_PARAMETERS_FILE: trying to set a non-? parameter.");
                        sim_message(none, "Mass of %s will remain at %lf.", buffer, sdm->body_segment[i].mass);
                     }
                     else
                        sdm->body_segment[i].mass = mass;
                  }
                  break;
               }
               else if (STRINGS_ARE_EQUAL(buffer2,"inertia"))
               {
                  if (fscanf(fp,"%lf %lf %lf %lf %lf %lf %lf %lf %lf", &inertia[0][0],
                     &inertia[0][1], &inertia[0][2], &inertia[1][0], &inertia[1][1],
                     &inertia[1][2], &inertia[2][0], &inertia[2][1], &inertia[2][2]) != 9)
                  {
                     sim_message(recover, "Error reading 3x3 inertia tensor for %s in %s.", buffer, params_file);
                  }
                  else
                  {
                     sdiner(i-1,inertia);
                     if (check_for_sderror("READ_PARAMETERS_FILE") == 19)
                     {
                        sim_message(recover, "READ_PARAMETERS_FILE: trying to set a non-? parameter.");
                        sim_message(none, "Inertia of %s will remain at:", buffer);
                        print_3x3matrix(sdm->body_segment[i].inertia,NULL,stderr);
                     }
                     else
                     {
                        for (j=0; j<3; j++)
                           for (k=0; k<3; k++)
                              sdm->body_segment[i].inertia[j][k] = inertia[j][k];
                     }
                  }
                  break;
               }
               else
               {
                  sim_message(recover, "Unknown string (%s) found after segment name %s in %s.",
                     buffer, buffer2, params_file);
               }
            }
         }
         // if it's not a segment, see if it is a Q
         if (i == sdm->num_body_segments)
         {
            for (i=0; i<sdm->nu; i++)
            {
               if (STRINGS_ARE_EQUAL(buffer,sdm->q[i].name))
               {
                  if (fscanf(fp,"%lf %lf", &sdm->q[i].initial_value, &sdm->q[i].initial_velocity) != 2)
                     sim_message(recover, "Error reading initial value and velocity for %s in %s.", buffer, params_file);
                     break;
               }
            }

            // If it's not a Q, see if it's a muscle name.
            if (i == sdm->nu)
            {
               for (i = 0; i < sdm->num_muscles; i++)
               {
                  if (STRINGS_ARE_EQUAL(buffer, sdm->muscles[i].name))
                  {
                     char* buffer_ptr;
                     double value = 0.0;
                     ReturnCode rc = code_fine;
                     dpMuscleStruct* muscle = &sdm->muscles[i];

                     read_line(fp, buffer);

                     // Read as many doubles as you can from the line, storing them in
                     // the muscle's stateParams array. You don't know how many there should
                     // be because this depends on the particular muscle model.
                     
                     for (buffer_ptr = buffer; NOT_EQUAL_WITHIN_ERROR(value, ERROR_DOUBLE); )
                     {
                        buffer_ptr = parse_string(buffer_ptr, type_double, (void*)&value);
                        if (NOT_EQUAL_WITHIN_ERROR(value, ERROR_DOUBLE))
                        {
                           if (muscle->numStateParams == 0)
                           {
                              muscle->stateParams = (double*)simm_malloc(sizeof(double));
                           }
                           else
                           {
                              muscle->stateParams = (double*)simm_realloc(muscle->stateParams, (muscle->numStateParams + 1) * sizeof(double), &rc);
                              if (rc == code_bad)
                                 sim_message(exit_program, "Ran out of memory while reading state vector for %s in %s.", buffer, params_file);
                           }
                           muscle->stateParams[muscle->numStateParams++] = value;
                        }
                     }
                     break;
                  }
               }
               if (i == sdm->num_muscles)
               {
                  for (i = 0; i < sdm->num_wrap_objects; i++)
                  {
                     if (STRINGS_ARE_EQUAL(buffer, sdm->wrap_object[i].name))
                     {
                        char type[50], algorithm[50], quadrant[50];
                        int active, segment, wrap_axis, wrap_sign, wrap_algorithm = WE_HYBRID_ALGORITHM;
                        dpWrapObjectType wrap_type;
                        double dimension[3], translation[3], rotation[3];
                        dpWrapObject* wrap = &sdm->wrap_object[i];

                        if (fscanf(fp, "%s %d %s %s %lg %lg %lg %s %lg %lg %lg %lg %lg %lg",
                           type, &active, algorithm, buffer2, &dimension[0], &dimension[1],
                           &dimension[2], quadrant, &translation[0], &translation[1], &translation[2],
                           &rotation[0], &rotation[1], &rotation[2]) != 14)
                        {
                           sim_message(recover, "Error reading parameters for wrap object %s in %s.", buffer, params_file);
                           break;
                        }
                        if ((wrap_type = get_wrap_type_index_from_name(type)) == dpWrapNone)
                        {
                           sim_message(none, "Error reading type of wrap object %s in %s.", buffer, params_file);
                           sim_message(recover, "   Must be one of \"sphere\", \"ellipsoid\", \"cylinder\", \"torus\".");
                           break;
                        }
                        if (wrap_type == dpWrapEllipsoid)
                        {
                           if ((wrap_algorithm = get_wrap_algorithm_index_from_name(algorithm)) == -1)
                           {
                              sim_message(recover, "Error reading algorithm type of ellipsoid wrap object %s in %s.", buffer, params_file);
                              sim_message(recover, "   Must be one of \"hybrid\", \"midpoint\", \"axial\".");
                              break;
                           }
                        }
                        segment = enter_segment(sdm, buffer2);
                        if (segment == UNDEFINED_SEGMENT)
                        {
                           sim_message(recover, "Error reading body segment of wrap object %s in %s.", buffer, params_file);
                           break;
                        }
                        get_axis_and_sign_from_quadrant(quadrant, &wrap_axis, &wrap_sign);
                        if (wrap_axis == -1)
                        {
                              sim_message(recover, "Error reading quadrant of wrap object %s in %s.", buffer, params_file);
                              sim_message(recover, "   Must be one of \"x\", \"-x\", \"y\", \"-y\", \"z\", \"-z\", \"all\".");
                              break;
                        }
                        wrap->wrap_type = wrap_type;
                        wrap->active = (active == 1) ? dpYes : dpNo;
                        wrap->wrap_algorithm = wrap_algorithm;
                        wrap->segment = segment;
                        memcpy(wrap->radius, dimension, 3*sizeof(double));
                        if (wrap->wrap_type == dpWrapCylinder)
                           wrap->height = dimension[1];
                        wrap->wrap_axis = wrap_axis;
                        wrap->wrap_sign = wrap_sign;
                        translation[XX] -= sdm->body_segment[segment+1].mass_center[XX];
                        translation[YY] -= sdm->body_segment[segment+1].mass_center[YY];
                        translation[ZZ] -= sdm->body_segment[segment+1].mass_center[ZZ];
                        MAKE_IDENTITY_MATRIX(wrap->from_local_xform);
                        x_rotate_matrix_bodyfixed(wrap->from_local_xform, rotation[XX]*DTOR);
                        y_rotate_matrix_bodyfixed(wrap->from_local_xform, rotation[YY]*DTOR);
                        z_rotate_matrix_bodyfixed(wrap->from_local_xform, rotation[ZZ]*DTOR);
                        wrap->from_local_xform[3][XX] = translation[XX];
                        wrap->from_local_xform[3][YY] = translation[YY];
                        wrap->from_local_xform[3][ZZ] = translation[ZZ];
                        invert_4x4transform(wrap->from_local_xform, wrap->to_local_xform);
                        break;
                     }
                  }
               }
               if (i == sdm->num_wrap_objects)
                  sim_message(exit_program, "Unknown string (%s) found in %s.", buffer, params_file);
            }
         }
      }
   }

   fclose(fp);

#if CONTACT_DETECTION
   if (check_object_info() == code_bad)
      return code_bad;
#endif

   return code_fine;
}

/* GET_AXIS_AND_SIGN_FROM_QUADRANT: return integer values for axis and sign from input string */
static void get_axis_and_sign_from_quadrant(char quadrant[], int* axis, int* sign)
{
   *axis = *sign = -1;

   _strupr(quadrant);
   if (STRINGS_ARE_EQUAL(quadrant, "ALL"))
   {
      *axis = 0;
      *sign = 0;
      return;
   }
   else if (STRINGS_ARE_EQUAL(quadrant, "X") || STRINGS_ARE_EQUAL(quadrant, "+X"))
   {
      *axis = 0;
      *sign = 1;
      return;
   }
   else if (STRINGS_ARE_EQUAL(quadrant, "-X"))
   {
      *axis = 0;
      *sign = -1;
      return;
   }
   else if (STRINGS_ARE_EQUAL(quadrant, "Y") || STRINGS_ARE_EQUAL(quadrant, "+Y"))
   {
      *axis = 1;
      *sign = 1;
      return;
   }
   else if (STRINGS_ARE_EQUAL(quadrant, "-Y"))
   {
      *axis = 1;
      *sign = -1;
      return;
   }
   else if (STRINGS_ARE_EQUAL(quadrant, "Z") || STRINGS_ARE_EQUAL(quadrant, "+Z"))
   {
      *axis = 2;
      *sign = 1;
      return;
   }
   else if (STRINGS_ARE_EQUAL(quadrant, "-Z"))
   {
      *axis = 2;
      *sign = -1;
      return;
   }
}

/* GET_WRAP_TYPE_INDEX_FROM_NAME: return the dpWrapObjectType corresponding to the input string name. */
static dpWrapObjectType get_wrap_type_index_from_name(char type_name[])
{
   if (STRINGS_ARE_EQUAL(type_name, "sphere"))
      return dpWrapSphere;
   else if (STRINGS_ARE_EQUAL(type_name, "ellipsoid"))
      return dpWrapEllipsoid;
   else if (STRINGS_ARE_EQUAL(type_name, "cylinder"))
      return dpWrapCylinder;
   else if (STRINGS_ARE_EQUAL(type_name, "torus"))
      return dpWrapTorus;

   return dpWrapNone;
}

/* GET_WRAP_ALGORITHM_INDEX_FROM_NAME: return the wrap algorithm index corresponding to the input string name. */
static int get_wrap_algorithm_index_from_name(char type_name[])
{
   if (STRINGS_ARE_EQUAL(type_name, "hybrid"))
      return WE_HYBRID_ALGORITHM;
   else if (STRINGS_ARE_EQUAL(type_name, "midpoint"))
      return WE_MIDPOINT_ALGORITHM;
   else if (STRINGS_ARE_EQUAL(type_name, "axial"))
      return WE_AXIAL_ALGORITHM;

   return -1;
}

/* READ_DLL_PARAMETERS_FILE: open and read the parameters file for a simulation run from SIMM, using the DLL.
 * Read initial setup conditions, and muscle states.
 */
ReturnCode read_dllparameters_file(char params_file[], dpModelStruct* sdm)
{
   int i, j, k;
   double mass, inertia[3][3];
   char* filename;
   FILE* fp;   

   if ((fp = fopen(params_file,"r")) == NULL)
      return code_bad;

   while (1)
   {
      if (fscanf(fp, "%s", buffer) != 1)
         break;

      if (buffer[0] == '#')
      {
         read_line(fp, buffer);
         continue;
      }

      // See if the string is the name of a Q.
      for (i = 0; i < sdm->nu; i++)
      {
         if (STRINGS_ARE_EQUAL(buffer, sdm->q[i].name))
         {
            if (fscanf(fp, "%lf %lf", &sdm->q[i].initial_value, &sdm->q[i].initial_velocity) != 2)
               sim_message(recover, "Error reading initial value and velocity for %s in %s.", buffer, params_file);
            break;
         }
      }

      // If it's not a Q, see if it's a muscle name.
      if (i == sdm->nu)
      {
         for (i = 0; i < sdm->num_muscles; i++)
         {
            if (STRINGS_ARE_EQUAL(buffer, sdm->muscles[i].name))
            {
               char* buffer_ptr;
               double value = 0.0;
               ReturnCode rc = code_fine;
               dpMuscleStruct* muscle = &sdm->muscles[i];

               read_line(fp, buffer);

               // Read as many doubles as you can from the line, storing them in
               // the muscle's stateParams array. You don't know how many there should
               // be because this depends on the particular muscle model.
               
               for (buffer_ptr = buffer; NOT_EQUAL_WITHIN_ERROR(value, ERROR_DOUBLE); )
               {
                  buffer_ptr = parse_string(buffer_ptr, type_double, (void*)&value);
                  if (NOT_EQUAL_WITHIN_ERROR(value, ERROR_DOUBLE))
                  {
                     if (muscle->numStateParams == 0)
                     {
                        muscle->stateParams = (double*)simm_malloc(sizeof(double));
                     }
                     else
                     {
                        muscle->stateParams = (double*)simm_realloc(muscle->stateParams, (muscle->numStateParams + 1) * sizeof(double), &rc);
                        if (rc == code_bad)
                           sim_message(exit_program, "Ran out of memory while reading state vector for %s in %s.", buffer, params_file);
                     }
                     muscle->stateParams[muscle->numStateParams++] = value;
                  }
               }
               break;
            }
         }
         if (i == sdm->num_muscles)
            sim_message(exit_program, "Unknown string (%s) found in %s.", buffer, params_file);
      }
   }

   fclose(fp);

   return code_fine;
}


#if CONTACT_DETECTION
/* Read contact objects from a parameters file. This function
 * is called by the DLL version of the simulation because the
 * contact objects are not included in dpModelStruct.
 */
ReturnCode read_contact_parameters_file(char params_file[])
{
   FILE* fp;   

   // If the model does not contain any contact objects,
   // SIMM won't write a file, so exit quietly if the
   // file is not found.
   
   if ((fp = fopen(params_file,"r")) == NULL)
      return code_fine;

   init_object_info();

   while (1)
   {
      if (fscanf(fp, "%s", buffer) != 1)
         break;

      if (buffer[0] == '#')
      {
         read_line(fp, buffer);
         continue;
      }

      if (STRINGS_ARE_EQUAL(buffer,"object"))
      {
         if (read_object(fp) == code_bad)
         {
            fclose(fp);
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer,"begin_group"))
      {
         if (read_object_group(fp) == code_bad)
         {
            fclose(fp);
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer,"contact_pair"))
      {
         if (read_contact_pair(fp) == code_bad)
         {
            fclose(fp);
            return code_bad;
         }
      }
      else
      {
         sim_message(exit_program, "Unknown string (%s) found in %s.", buffer, params_file);
         return code_bad;
      }
   }

   fclose(fp);

   if (check_object_info() == code_bad)
      return code_bad;

   return code_fine;
}
#endif


/* PREPROCESS_FILE: this routine runs the a C preprocessor on the specified
 * input file, putting the output in the specified output file. It then
 * opens the processed file and returns a pointer to a FILE structure for it.
 * NOTE: the original (Unix) version of preprocess_file() for the pipeline
 * called the ANSI system() routine to execute the C preprocessor.
 * Unfortunately, not all Win32 compilers support system(), nor do they all
 * provide command-line compilers.  So we avoid the problem by call a
 * C preprocessor linked directly into our executable (acpp.lib).
 */

static void acpp(char in_file[], const char out_file[], char option_string[])

{
   int argc = 0;
   char** argv;
   char* remaining_options = option_string;
   char option[CHARBUFFER];

   // room for 50 arguments should be enough
   argv = (char**) malloc(50 * sizeof(char*));
   mstrcpy(&argv[argc++], "acpp");

   while (1)
   {
      remaining_options = parse_string(remaining_options, type_string, option);

      if (option == NULL || STRING_IS_NULL(option))
         break;

      mstrcpy(&argv[argc++], option);
   }

   mstrcpy(&argv[argc++], (char*) in_file);
   mstrcpy(&argv[argc++], (char*) out_file);

   remove(out_file);

   if (acpp_main(argc,argv) != 0)
      fprintf(stderr, "Error running acpp on input file: %s.", in_file);

   while (argc > 0)
      free(argv[--argc]);

   free(argv);
}

/* PREPROCESS_FILE: preprocess the input file. */
FILE* preprocess_file(char infile[], char outfile[])
{
   FILE* fp = fopen(infile,"r");

   if (fp == NULL)
      return (NULL);

   fclose(fp);

   acpp(infile, outfile, "-P");

#ifndef WIN32
   chmod(outfile,438);
#endif

   return fopen(outfile, "r");
}



/* READ_STRING: this routine reads a character string from a file. It skips over
 * any white space at the beginning, and returns EOF if it hit the end of
 * the file before completing the string.
 */
int read_string(FILE* fp, char str_buffer[])
{
   int c;

   // Scan thru white space until you find the first character in the string.
   while (1)
   {
      c = getc(fp);
      if (c == EOF)
         return EOF;
      *str_buffer = c;
      if (CHAR_IS_NOT_WHITE_SPACE(*str_buffer))
         break;
   }
   str_buffer++;

   // Now read characters until you find white space or EOF.
   while (1)
   {
      c = getc(fp);
      if (c == EOF)
         return EOF;
      *str_buffer = c;
      if (CHAR_IS_WHITE_SPACE(*str_buffer))
         break;
      str_buffer++;
   }

   // Null-terminate the string
   *str_buffer = STRING_TERMINATOR;

   // You found a valid string without hitting EOF, so return a value that you
   // know will never equal EOF, no matter how EOF is defined.
   return EOF + 1;
}



/* READ_LINE: reads a line (possibly empty) from a file */
int read_line(FILE* fp, char str_buffer[])
{
   int c;

   // Read characters until you hit a carriage return or EOF.
   while (1)
   {
      c = getc(fp);
      if (c == EOF)
         return EOF;
      *str_buffer = c;
      if (*str_buffer == CARRIAGE_RETURN)
         break;
      str_buffer++;
   }

   // Null-terminate the string.
   *str_buffer = STRING_TERMINATOR;

   // You found a valid line without hitting EOF, so return a value that you
   // know will never equal EOF, no matter how EOF is defined.
   return EOF + 1;
}



/* READ_NONEMPTY_LINE: Reads the first non-empty line from a file. */
int read_nonempty_line(FILE* fp, char str_buffer[])
{
   int c;

   // Scan thru the white space until you find the first character
   while (1)
   {
      c = getc(fp);
      if (c == EOF)
         return EOF;
      *str_buffer = c;
      if (*str_buffer != SPACE && *str_buffer != TAB)
         break;
   }
   str_buffer++;

   // Now read characters until you find a carriage return or EOF
   while (1)
   {
      c = getc(fp);
      if (c == EOF)
         return EOF;
      *str_buffer = c;
      if (*str_buffer == CARRIAGE_RETURN)
         break;
      str_buffer++;
   }

   *str_buffer = STRING_TERMINATOR;

   // You found a valid line without hitting EOF, so return a value that you
   // know will never equal EOF, no matter how EOF is defined.
   return EOF + 1;
}


/* READ_TIL_TOKENS: Read characters from 'file' into 'buf' until one of the
 * characters specified in 'delimiters' is encountered.  Put the delimiting
 * character back into 'file'.
 */
int read_til_tokens(FILE* file, char buf[], const char delimiters[])
{
   char* p = buf;

   while (1)
   {
      char c = fgetc(file);

      if (feof(file) || ferror(file))
         break;
      else
      {
         if (strchr(delimiters, c))
         {
            ungetc(c, file);
            break;
         }
         else
            *p++ = c;
      }
   }
   *p = '\0';

   // remove trailing whitespace if necessary
   while (p > buf && isspace(p[-1]))
       *(--p) = '\0';

   // Return 1 if the string is not empty.
   if (p > buf)
      return 1;
   else
      return 0;
}


/* CLEAN_STRING: clears leading and trailing white space from a string */
char* clean_string(char buf[])
{
   char* ptr;

   if (strlen(buf) < 1)
      return NULL;

   // Remove trailing white space
   ptr = &buf[strlen(buf)-1];
   while (CHAR_IS_WHITE_SPACE(*ptr))
      ptr--;
   *(ptr+1) = STRING_TERMINATOR;

   // Remove leading white space
   ptr = buf;
   while (CHAR_IS_WHITE_SPACE(*ptr))
      ptr++;

   return ptr;
}


/* READ_DOUBLE_ARRAY: this routine reads an array of pairs-of-doubles
 * (e.g. "(2.30, -0.05)"), as in the definition of a function. It keeps
 * reading until it encounters the ending string (e.g. "endfunction")
 * which is passed in. This routine is not trivial because it allows
 * spaces to be placed liberally in the string that it extracts the
 * doubles from (e.g. "( 2.30 , -0.05 )").
 */
ReturnCode read_double_array(FILE* fp, const char ending[], char name[], dpFunction* func)
{
   int new_size;

   func->numpoints = 0;

   while (1)
   {
      if (read_string(fp,buffer) == EOF)
      {
         sim_message(abort_action, "Unexpected EOF reading function %s.", name);
         return code_bad;
      }

      if (STRINGS_ARE_EQUAL(buffer,ending))
         break;

      // If the string just read is not the ending string (e.g. "endfunction"),
      // then it must be part or all of an x-y pair "(x,y)". It cannot be more
      // than one x-y pair because there must be a space between each x-y pair.
      // Since there can be spaces within an x-y pair, you now want to continue
      // reading strings from the file until you encounter a closing parenthesis.
      

      // Before parsing this next string, first make sure that there is enough
      // space in the function structure for the next x-y pair.
      if (func->numpoints == func->coefficient_array_size)
      {
         new_size = func->coefficient_array_size + FUNCTION_ARRAY_INCREMENT;
         if (realloc_function(func,new_size) == code_bad)
            return code_bad;
      }

      while (1)
      {
         if (string_contains_closing_paren(buffer) == dpYes)
            break;

         if (read_string(fp,&buffer[strlen(buffer)]) == EOF)
         {
            sim_message(abort_action, "Unexpected EOF reading function %s.", name);
            return code_bad;
         }
      }

      if (get_xypair_from_string(buffer,&func->x[func->numpoints],
         &func->y[func->numpoints]) == code_bad)
      {
         sim_message(abort_action, "Error reading x-y pair in function %s.", name);
         return code_bad;
      }
      
      (func->numpoints)++;
   }

   if (func->numpoints < 2)
   {
      sim_message(abort_action, "Function %s contains fewer than 2 points.", name);
      return code_bad;
   }

   return code_fine;
}



/* STRING_CONTAINS_CLOSING_PAREN: this routine scans a string to see if
 * it contains a closing parenthesis.
 */

static dpBoolean string_contains_closing_paren(char str_buffer[])
{

   while (*str_buffer != STRING_TERMINATOR)
   {
      if (*(str_buffer++) == ')')
         return dpYes;
   }

   return dpNo;

}

/* STRIP_BRACKETS_FROM_STRING: remove the () from the given gencoord name */
void strip_brackets_from_string(char name[])
{
   int i, j;
   char buffer[CHARBUFFER];

   // strip the () from the gencoord name
   strcpy(buffer, name);
   i = 0;
   j = 0;
   while (1)
   {
      if (buffer[i] == '(')
      { 
         i++;
         continue;
      }
      if (buffer[i] == ')')
      {
         name[j] = '\0';
         break;
      }
      name[j++] = buffer[i++];
   }
}

/* GET_XYPAIR_FROM_STRING: this routine parses a string to extract a pair
 * of doubles from it. The string should be of the form: "(x, y)".
 */

ReturnCode get_xypair_from_string(char str_buffer[], double* x, double* y)
{

   char junk;
   char* buffer_ptr;

   *x = *y = ERROR_DOUBLE;

   buffer_ptr = parse_string(str_buffer,type_char,(void*)&junk); // open paren
   buffer_ptr = parse_string(buffer_ptr,type_double,(void*)x);   // x coord
   buffer_ptr = parse_string(buffer_ptr,type_char,(void*)&junk); // comma
   buffer_ptr = parse_string(buffer_ptr,type_double,(void*)y);   // y coord
   buffer_ptr = parse_string(buffer_ptr,type_char,(void*)&junk); // close paren*/

   if (*x == ERROR_DOUBLE || *y == ERROR_DOUBLE)
      return code_bad;

   return code_fine;

}



/* PARSE_STRING: this routine scans a string and extracts a variable from
 * it. The type of variable that it extracts is specified in one of the
 * arguments. It returns the unused portion of the string so that more
 * variables can be extracted from it.
 */

char* parse_string(char str_buffer[], VariableType var_type, void* dest_var)
{
   char tmp_buffer[CHARBUFFER], *buffer_ptr;

   if (str_buffer == NULL)
      return NULL;

   buffer_ptr = tmp_buffer;

   while (CHAR_IS_WHITE_SPACE(*str_buffer))
      str_buffer++;

   if (var_type == type_char)
   {
      *((char*)dest_var) = *str_buffer;
      return str_buffer + 1;
   }

   if (var_type == type_string)
   {
      if (STRING_IS_NULL(str_buffer))
         *((char*)dest_var) = STRING_TERMINATOR;
      else
         sscanf(str_buffer,"%s", (char*)dest_var);
      return str_buffer + strlen((char*)dest_var);
   }

   if (var_type == type_double)
   {
      *((double*)dest_var) = ERROR_DOUBLE;
      if (STRING_IS_NOT_NULL(str_buffer))
      {
         while (*str_buffer == '-' || *str_buffer == '.' || (*str_buffer >= '0' && *str_buffer <= '9'))
            *(buffer_ptr++) = *(str_buffer++);
         *buffer_ptr = STRING_TERMINATOR;
         if (sscanf(tmp_buffer,"%lf",(double*)dest_var) != 1)
            *((double*)dest_var) = ERROR_DOUBLE;
      }
      return str_buffer;
   }

   if (var_type == type_int)
   {
      *((int*)dest_var) = ERROR_INT;
      if (STRING_IS_NOT_NULL(str_buffer))
      {
         while (*str_buffer == '-' || (*str_buffer >= '0' && *str_buffer <= '9'))
            *(buffer_ptr++) = *(str_buffer++);
         *buffer_ptr = STRING_TERMINATOR;
         sscanf(tmp_buffer,"%d",(int*)dest_var);
      }
      return str_buffer;
   }

   sim_message(none, "Unknown variable type (%d) passed to parse_string().", (int)var_type);

   return str_buffer;
}



/* ENTER_SEGMENT: this routine takes the name of a body segment and enters it
 * into the model's array of segment names. It returns the index of the array
 * entry where the name is stored (minus 1 since SD/FAST body numbers start at -1).
 * It is used when reading muscle attachment points. When a segment name is read
 * from a muscle file, you want to find out which SD/FAST body segment number the
 * name corresponds to. If the name is not in the SDModel structure, it is an error
 * since all of the body segments have already been defined. Before checking the
 * name, remove any special characters from it since this has been done to the
 * array of valid segment names (because they have to compile as #defines).
 */
int enter_segment(dpModelStruct* sdm, char segment_name[])
{
   int i;

   // Because body segment numbers in SD/FAST start at -1 (ground),
   // this routine returns segment array indices minus one, so that
   // it also numbers segments starting at -1.
   convert_string(segment_name);

   for (i=0; i<sdm->num_body_segments; i++)
      if (STRINGS_ARE_EQUAL(segment_name,sdm->body_segment[i].name))
         return i-1;

   return UNDEFINED_SEGMENT;
}



/* ENTER_GENCOORD: take the name of a gencoord find the corresponding gencoord in the list of gencoords.
 */
dpQStruct* enter_gencoord(dpModelStruct* sdm, char gencoord_name[])
{
   int i;

   convert_string(gencoord_name);

   for (i=0; i<sdm->nu; i++)
      if (STRINGS_ARE_EQUAL(gencoord_name,sdm->q[i].name))
         return &sdm->q[i];

   return NULL;
}

/* ENTER_FUNCTION: whenever a function number is read from a file, the
 * user-specified number is compared to the existing array of functions to
 * see if that one has already been defined. If it has, the routine just
 * returns the internal number of that function. If it has not yet been
 * defined, it adds the number to a new element in the function list.
 */
int enter_function(dpModelStruct* model, int usernum, dpBoolean permission_to_add)
{
   int i, new_func, old_count;
   ReturnCode rc = code_fine;

   if (usernum != UNDEFINED_USERFUNCNUM)
   {
      for (i=0; i<model->function_array_size; i++)
      {
         if (model->function[i]->usernum == usernum)
         {
            if (model->function[i]->status == dpFunctionSimmDefined)
            {
               // If the usernum is already being used by a SIMM-defined function,
               // change the usernum of the SIMM-defined function.
               if (permission_to_add == dpYes)
               {
                  model->function[i]->usernum = find_highest_usernum(model) + 1;
                  break;
               }
               else // You can't match a user-defined function with a SIMM-defined one.
               {
                  return INVALID_FUNCTION;
               }
            }
            else
            {
               return i;//&model->function[i];
            }
         }
      }
   }

   if (permission_to_add == dpNo)
      return INVALID_FUNCTION;

   for (i=0; i<model->function_array_size; i++)
      if (model->function[i]->used == dpNo)
         break;
   new_func = i;

   if (new_func == model->function_array_size)
   {
      old_count = model->function_array_size;
      model->function_array_size += ARRAY_INCREMENT;
      model->function = (dpFunction**)simm_realloc(model->function, model->function_array_size*sizeof(dpFunction*), &rc);

      if (rc == code_bad)
      {
         model->function_array_size = old_count;
         return INVALID_FUNCTION;
      }

      for (i=old_count; i<model->function_array_size; i++)
      {
         model->function[i] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         model->function[i]->used = dpNo;
         model->function[i]->status = dpFunctionUndefined;
         model->function[i]->usernum = UNDEFINED_USERNUM;
         model->function[i]->coefficient_array_size = 0;
         model->function[i]->x = NULL;
         model->function[i]->y = NULL;
         model->function[i]->b = NULL;
         model->function[i]->c = NULL;
         model->function[i]->d = NULL;
         model->function[i]->numpoints = 0;
         model->function[i]->type = dpNaturalCubicSpline;
         model->function[i]->cutoff_frequency = 0;
      }
   }

   model->function[new_func]->used = dpYes;
   model->function[new_func]->status = dpFunctionUndefined;
   model->function[new_func]->usernum = usernum;

   return new_func;//&model->function[new_func];
}


/* LOAD_USER_FUNCTION: This function passes ownership of the x, y, b, c, d arrays in 'func' to the model. */
dpFunction* load_user_function(dpModelStruct* model, dpFunction* func)
{
   int funcIndex = enter_function(model, func->usernum, dpYes);

   if (funcIndex == INVALID_FUNCTION)
      return NULL;

   if (model->function[funcIndex]->status != dpFunctionUndefined)
      sim_message(none, "Warning: redefinition of function f%d. Replacing earlier definition.", func->usernum);

   model->function[funcIndex]->type = func->type;
   model->function[funcIndex]->numpoints = func->numpoints;
   model->function[funcIndex]->coefficient_array_size = func->coefficient_array_size;
   model->function[funcIndex]->cutoff_frequency = func->cutoff_frequency;
   model->function[funcIndex]->x = func->x;
   model->function[funcIndex]->y = func->y;
   model->function[funcIndex]->b = func->b;
   model->function[funcIndex]->c = func->c;
   model->function[funcIndex]->d = func->d;
   model->function[funcIndex]->status = dpFunctionUserDefined;
   model->function[funcIndex]->usernum = func->usernum;
   model->function[funcIndex]->used = dpYes;
   model->function[funcIndex]->source = func->source;

   return model->function[funcIndex];
}


/* LOAD_SIMM_FUNCTION: This function passes ownership of the x, y, b, c, d arrays in 'func' to the model. */
dpFunction* load_simm_function(dpModelStruct* model, dpFunction* func)
{
   int i, funcIndex;

   // Find an unused slot in the function array.
   for (i=0; i<model->function_array_size; i++)
      if (model->function[i]->used == dpNo)
         break;
   funcIndex = i;

   if (funcIndex == model->function_array_size)
   {
      ReturnCode rc = code_fine;
      int i, old_count = model->function_array_size;
      model->function_array_size += ARRAY_INCREMENT;
      model->function = (dpFunction**)simm_realloc(model->function, model->function_array_size*sizeof(dpFunction*), &rc);

      if (rc == code_bad)
      {
         model->function_array_size = old_count;
         return NULL;
      }

      for (i=old_count; i<model->function_array_size; i++)
      {
         model->function[i] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         //memset(&model->function[i], 0, sizeof(dpFunction));
         model->function[i]->status = dpFunctionUndefined;
         model->function[i]->usernum = UNDEFINED_USERNUM;
         model->function[i]->type = dpNaturalCubicSpline;
         model->function[i]->used = dpNo;

      }
   }

   model->function[funcIndex]->usernum = find_highest_usernum(model) + 1;
   model->function[funcIndex]->source = dpMuscleFile;
   model->function[funcIndex]->type = func->type;
   model->function[funcIndex]->numpoints = func->numpoints;
   model->function[funcIndex]->coefficient_array_size = func->coefficient_array_size;
   model->function[funcIndex]->cutoff_frequency = func->cutoff_frequency;
   model->function[funcIndex]->x = func->x;
   model->function[funcIndex]->y = func->y;
   model->function[funcIndex]->b = func->b;
   model->function[funcIndex]->c = func->c;
   model->function[funcIndex]->d = func->d;
   model->function[funcIndex]->status = dpFunctionSimmDefined;
   model->function[funcIndex]->used = dpYes;

   return model->function[funcIndex];
}


/* FIND_HIGHEST_USERNUM: return the highest user function number */
static int find_highest_usernum(dpModelStruct* model)
{
   int i, highest_user_num = -1;

   for (i=0; i<model->function_array_size; i++)
   {
      if (model->function[i]->used == dpYes && model->function[i]->usernum > highest_user_num)
         highest_user_num = model->function[i]->usernum;
   }

   return highest_user_num;
}


/* LOAD_INPUT_DATA: This routine reads a file of input data either a motion file or a kinetics file.
 * The file can contain positions, velocities, and accelerations of Qs (gencoords), and
 * external forces and torques applied to the body segments. It can be used in a forward or inverse simulation.
 */
ReturnCode load_input_data(dpModelStruct* sdm, MotionData* data, char filename[])
{
   const char *suffix = get_suffix(filename);

   data->num_frames = data->current_frame = 0;
   data->num_elements = data->num_otherdata = 0;
   data->num_forces = data->num_torques = data->num_musc_excitations = 0;
   data->time_column = -1;
   data->x_column = -1;
   data->range_start = data->range_end = data->step_size = 0.0;
   data->units = NULL;
   data->q_data = NULL;
   data->event_color[0] = 1.0;
   data->event_color[1] = 0.0;
   data->event_color[2] = 1.0;
   data->enforce_constraints = dpYes;
   data->enforce_loops = dpYes;
   data->show_markers = dpYes;
   data->show_marker_trails = dpNo;
   data->show_forces = dpYes;
   data->show_force_trails = dpNo;
   data->show_foot_prints = dpYes;
   data->show_force_plates = dpYes;
   data->show_gencoord_moments = dpNo;
   data->show_other_objects = dpNo;
   data->wrap = dpNo;
   data->num_foot_prints = 0;
   data->foot_print = NULL;
   data->num_force_plates = 0;
   data->force_plate = NULL;

   if (suffix == NULL || STRINGS_ARE_EQUAL(suffix, "mot"))
   {
      return read_motion_file(sdm, data, filename);
   }
   else if (STRINGS_ARE_EQUAL(suffix, "ktx"))
   {
      mstrcpy(&data->name, "input_data");
      return read_kinetics_file(sdm, data, filename);
   }
   else
   {
      sim_message(none, "Unrecognized suffix on input file %s (must be .ktx or .mot).", filename);
      return code_bad;
   }
}

/* READ_KINETICS_FILE: read the information stored in the given kinetics file.
 * This function is used to read the data from old style ".ktx" files that only stored input kinetics data (forces and torques).
 * It has now been replaced by read_motion_file, while reads file containing all input data (kinematic and kinetic and other).
 */
ReturnCode read_kinetics_file(dpModelStruct* sdm, MotionData* data, char filename[])
{
   int i, j, num_read, num_columns;
   dpFunctionType functionType;
   double cutoffFrequency;
   dpBoolean read_frames = dpNo, read_elements = dpNo, read_otherdata = dpNo;
   char tmp_name[CHARBUFFER];

   get_temp_file_name(NULL, tmp_name);
   if ((data->fp = preprocess_file(filename, tmp_name)) == NULL)
   {
      sim_message(none, "Unable to open kinetics file %s.", filename);
      return code_bad;
   }

   while (1)
   {
      if (read_string(data->fp, buffer) == EOF)
      {
         sim_message(abort_action, "Premature EOF found in kinetics file.");
         return code_bad;
      }

      if (buffer[0] == '#')
      {
         read_nonempty_line(data->fp,buffer);
         continue;
      }

      if (STRINGS_ARE_EQUAL(buffer,"frames") || STRINGS_ARE_EQUAL(buffer,"numframes"))
      {
         if (fscanf(data->fp,"%d", &data->num_frames) != 1)
         {
            sim_message(abort_action, "Error reading number of frames from kinetics file.");
            return code_bad;
         }
         else
            read_frames = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"elements") || STRINGS_ARE_EQUAL(buffer,"numelements"))
      {
         if (fscanf(data->fp,"%d", &data->num_elements) != 1)
         {
            sim_message(abort_action, "Error reading number of elements from kinetics file.");
            return code_bad;
         }
         else
            read_elements = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer,"otherdata"))
      {
         if (fscanf(data->fp,"%d", &data->num_otherdata) != 1)
         {
            sim_message(abort_action, "Error reading number of otherdata from kinetics file.");
            return code_bad;
         }
         else
            read_otherdata = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "enforce_constraints"))
      {
         if (read_string(data->fp, buffer) == EOF)
         {
            sim_message(abort_action, "Error reading enforce_constraints from kinetics file (must be: yes/no/on/off/true/false).");
            return code_bad;
         }
         data->enforce_constraints = is_preference_on(buffer);
      }
      else
      {
         sim_message(recover, "Unrecognized string \"%s\" found in kinetics file.", buffer);
      }
      
      if (read_frames == dpYes && read_elements == dpYes && read_otherdata == dpYes)
         break;
   }

   data->elementnames = (char**)simm_calloc(data->num_elements, sizeof(char*));
   if (data->elementnames == NULL)
      return code_bad;

   for (num_read=0, i=0; i<data->num_elements; i++)
   {
      num_read += fscanf(data->fp,"%s", buffer);
      mstrcpy(&data->elementnames[i],buffer);
   }

   if (num_read < data->num_elements)
   {
      sim_message(abort_action, "Error reading names of data elements. Only %d of %d read.",
         num_read, data->num_elements);
      return code_bad;
   }

   for (num_columns = 0, i=0; i<data->num_elements; i++)
   {
      if (name_is_body_segment(data->elementnames[i], sdm, force_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         num_columns += 6;
      else if (name_is_body_segment(data->elementnames[i], sdm, torque_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         num_columns += 3;
      else
         num_columns++;
   }

   data->motiondata = (double**)simm_calloc(num_columns, sizeof(double*));
   if (data->motiondata == NULL)
   {
      return code_bad;
   }

   for (i=0; i<num_columns; i++)
   {
      data->motiondata[i] = (double*)simm_calloc(data->num_frames, sizeof(double));
      if (data->motiondata[i] == NULL)
      {
         return code_bad;
      }
   }

   for (i=0; i<data->num_frames; i++)
   {
      num_read = 0;
      for (j=0; j<num_columns; j++)
         num_read += fscanf(data->fp,"%lg", &data->motiondata[j][i]);
      if (num_read < num_columns)
      {
         data->num_frames = i;
         sim_message(recover, "Error reading kinetics file %s:\n  Only %d frames found.", filename, data->num_frames);
         break;
      }
   }

   (void)fclose(data->fp);
   (void)unlink(tmp_name);

   sim_message(none, "Read %d frames of data from kinetics file %s.", data->num_frames, filename);

   return verify_kinetics_data(sdm, data);
}

/* READ_MOTION_FILE: read and store the information in a given motion file.  
 */
ReturnCode read_motion_file(dpModelStruct* sdm, MotionData* data, char filename[])
{
   int i, j, num_read;
   dpFunctionType functionType;
   double cutoffFrequency;
   ReturnCode rc;
   char tmp_name[CHARBUFFER];

   get_temp_file_name(NULL, tmp_name);
   if ((data->fp = preprocess_file(filename, tmp_name)) == NULL)
   {
      sim_message(none, "Unable to open motion file %s.", filename);
      return code_bad;
   }

   while (1)
   {
      if (read_string(data->fp, buffer) == EOF)
      {
         sim_message(abort_action, "Premature EOF found in motion file.");
         return code_bad;
      }

      if (buffer[0] == '#')
      {
         read_nonempty_line(data->fp, buffer);
         continue;
      }

      if (STRINGS_ARE_EQUAL(buffer, "endheader"))
         break;

      else if (STRINGS_ARE_EQUAL(buffer, "name"))
      {
         read_line(data->fp, buffer);
         mstrcpy(&data->name, buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "wrap"))
      {
         data->wrap = dpYes;
      }
      else if (STRINGS_ARE_EQUAL(buffer, "enforce_constraints"))
      {
         if (read_string(data->fp, buffer) == EOF)
         {
            sim_message(abort_action, "Error reading enforce_constraints from motion file (must be: yes/no/on/off/true/false).");
            return code_bad;
         }
         data->enforce_constraints = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "datacolumns"))
      {
         if (fscanf(data->fp,"%d", &data->num_elements) != 1)
         {
            sim_message(abort_action, "Error reading number of data columns from motion file.");
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer,"keys"))
      {
         read_line(data->fp, buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer,"datarows"))
      {
         if (fscanf(data->fp, "%d", &data->num_frames) != 1)
         {
            sim_message(abort_action, "Error reading number of frames from motion file.");
            return code_bad;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "range"))
      {
         fscanf(data->fp, "%*lg %*lg");
      }
      else if (STRINGS_ARE_EQUAL(buffer, "units"))
      {
         read_line(data->fp, buffer);
         mstrcpy(&data->units, buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "cursor"))
      {
         fscanf(data->fp, "%*f %*f %*f");
      }
      else if (STRINGS_ARE_EQUAL(buffer, "event") || STRINGS_ARE_EQUAL(buffer, "force_event") || STRINGS_ARE_EQUAL(buffer, "marker_event"))
      {
         char name[CHARBUFFER];

         if (data->event)
            data->event = (dpMotionEvent*)simm_realloc(data->event, (data->num_events + 1) * sizeof(dpMotionEvent), &rc);
         else
            data->event = (dpMotionEvent*)simm_malloc((data->num_events + 1) * sizeof(dpMotionEvent));

         if (!data->event)
         {
            sim_message(recover, "Unable to malloc space to hold motion event.");
         }
         else if (fscanf(data->fp, "%lg", &data->event[data->num_events].xCoord) != 1)
         {
            sim_message(recover, "Error reading event time in motion %s.", data->name);
         }
         else if (read_nonempty_line(data->fp, name) == EOF)
         {
            sim_message(recover, "Error reading event name in motion %s.", data->name);
         }
         else
         {
            mstrcpy(&data->event[data->num_events].name, name);
            if (STRINGS_ARE_EQUAL(name, "rhs"))
            {
               data->event[data->num_events].type = dpHeelStrikeEvent;
               data->event[data->num_events].segment = dpRightFootEvent;
            }
            else if (STRINGS_ARE_EQUAL(name, "lhs"))
            {
               data->event[data->num_events].type = dpHeelStrikeEvent;
               data->event[data->num_events].segment = dpLeftFootEvent;
            }
            else if (STRINGS_ARE_EQUAL(name, "rto"))
            {
               data->event[data->num_events].type = dpToeOffEvent;
               data->event[data->num_events].segment = dpRightFootEvent;
            }
            else if (STRINGS_ARE_EQUAL(name, "lto"))
            {
               data->event[data->num_events].type = dpToeOffEvent;
               data->event[data->num_events].segment = dpLeftFootEvent;
            }
            else
            {
               data->event[data->num_events].type = dpUnknownEventType;
               data->event[data->num_events].segment = dpUnknownSegmentEvent;
            }

            if (STRINGS_ARE_EQUAL(buffer, "force_event"))
               data->event[data->num_events].source = dpForceEvent;
            else if (STRINGS_ARE_EQUAL(buffer, "marker_event"))
               data->event[data->num_events].source = dpMarkerEvent;
            else
               data->event[data->num_events].source = dpUserEvent;
            data->event[data->num_events].displayFlag = dpYes;

            data->num_events++;
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "event_color"))
      {
         if (fscanf(data->fp, "%f %f %f", &data->event_color[0],
             &data->event_color[1], &data->event_color[2]) != 3)
         {
            sim_message(none, "Error reading event_color for motion %s.", data->name);
         }
         else
         {
            // Support for 'old style' colors (range: 0 to 255). If any of the
            // color components is greater than 1.0, assume that it's an old
            // style color definition.
            
            if (data->event_color[0] > 1.0 || data->event_color[1] > 1.0 || data->event_color[2] > 1.0)
            {
               data->event_color[0] /= 255.0;
               data->event_color[1] /= 255.0;
               data->event_color[2] /= 255.0;
            }
            for (i = 0; i < 3; i++)
            {
               if (data->event_color[i] < 0.0)
                  data->event_color[i] = 0.0;
            }
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "calc_derivatives"))
      {
         // Whether old-style (time_step) or new-style (yes/no), this option is not used by dynamic simulations.
         fscanf(data->fp, "%*s");
      }
      else if (STRINGS_ARE_EQUAL(buffer, "x_column"))
      {
         fscanf(data->fp, "%d", &data->x_column);
         data->x_column--;    // make it zero-based
      }
      else if (STRINGS_ARE_EQUAL(buffer, "time_column"))
      {
         fscanf(data->fp, "%d", &data->time_column);
         data->time_column--;    // make it zero-based
      }
      else if (STRINGS_ARE_EQUAL(buffer, "accept_realtime_motion"))
      {
         // not used by dynamic simulations
      }
      else if (STRINGS_ARE_EQUAL(buffer, "sliding_time_scale"))
      {
         // not used by dynamic simulations
      }
      // The following header parameters are not used by dynamic simulations,
      // but they are stored in MotionData so they can be written to the
      // output motion.
      else if (STRINGS_ARE_EQUAL(buffer, "enforce_loops"))
      {
         read_string(data->fp, buffer);
         data->enforce_loops = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_markers"))
      {
         read_string(data->fp, buffer);
         data->show_markers = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_marker_trails"))
      {
         read_string(data->fp, buffer);
         data->show_marker_trails = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_forces"))
      {
         read_string(data->fp, buffer);
         data->show_forces = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_force_trails"))
      {
         read_string(data->fp, buffer);
         data->show_force_trails = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_foot_prints"))
      {
         read_string(data->fp, buffer);
         data->show_foot_prints = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_force_plates"))
      {
         read_string(data->fp, buffer);
         data->show_force_plates = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_gencoord_moments"))
      {
         read_string(data->fp, buffer);
         data->show_gencoord_moments = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "show_other_objects"))
      {
         read_string(data->fp, buffer);
         data->show_other_objects = is_preference_on(buffer);
      }
      else if (STRINGS_ARE_EQUAL(buffer, "rightFootPrint") || STRINGS_ARE_EQUAL(buffer, "leftFootPrint"))
      {
         double scale, translation[3], euler_rotation[3];
         if (fscanf(data->fp, "%lg %lg %lg %lg %lg %lg %lg", &scale, &translation[0], &translation[1], &translation[2],
            &euler_rotation[0], &euler_rotation[1], &euler_rotation[2]) != 7)
         {
            sim_message(none, "Error reading \"%s\" in motion file:", buffer);
            sim_message(none, "You must specify: scale tx ty tz rx ry rz.");
         }
         else
         {
            add_foot_print(data, buffer, scale, translation, euler_rotation);
         }
      }
      else if (STRINGS_ARE_EQUAL(buffer, "forceplate"))
      {
         double scale[3], translation[3], euler_rotation[3];
         if (fscanf(data->fp, "%lg %lg %lg %lg %lg %lg %lg %lg %lg", &scale[0], &scale[1], &scale[2],
            &translation[0], &translation[1], &translation[2], &euler_rotation[0], &euler_rotation[1], &euler_rotation[2]) != 9)
         {
            sim_message(none, "Error reading \"%s\" in motion file:", buffer);
            sim_message(recover, "You must specify: sx sy sz tx ty tz rx ry rz.");
         }
         else
         {
            add_force_plate(data, buffer, scale, translation, euler_rotation);
         }
      }
      else
      {
         sim_message(recover, "Unrecognized string \"%s\" found in motion file.", buffer);
      }
   }

   data->elementnames = (char**)simm_calloc(data->num_elements, sizeof(char*));
   if (data->elementnames == NULL)
   {
      sim_message(exit_program, "Unable to allocate enough memory to read motion file.");
   }

   for (num_read=0, i=0; i<data->num_elements; i++)
   {
      num_read += fscanf(data->fp,"%s", buffer);
      mstrcpy(&data->elementnames[i],buffer);
   }

   if (num_read < data->num_elements)
   {
      sim_message(recover, "Error reading names of data columns. Only %d of %d read.", num_read, data->num_elements);
      return code_bad;
   }

   data->motiondata = (double**)simm_calloc(data->num_elements, sizeof(double*));
   if (data->motiondata == NULL)
   {
      sim_message(exit_program, "Unable to allocate enough memory to read motion file.");
   }

   for (i=0; i<data->num_elements; i++)
   {
      data->motiondata[i] = (double*)simm_calloc(data->num_frames, sizeof(double));
      if (data->motiondata[i] == NULL)
      {
         sim_message(exit_program, "Unable to allocate enough memory to read motion file.");
      }
   }

   for (i=0; i<data->num_frames; i++)
   {
      num_read = 0;
      for (j=0; j<data->num_elements; j++)
         num_read += fscanf(data->fp,"%lg", &data->motiondata[j][i]);
      if (num_read < data->num_elements)
      {
         data->num_frames = i;
         sim_message(recover, "Error reading motion file %s:\n  Only %d frames found.", filename, data->num_frames);
         break;
      }
   }

   (void)fclose(data->fp);
   (void)unlink(tmp_name);

   sim_message(none, "Read %d frames of data from motion file %s.", data->num_frames, filename);

   return verify_motion_data(sdm, data);
}


/* VERIFY_KINETICS_DATA: this routine checks the validity of the kinetics data.
 * It checks the column names in the file to make sure they match the model's
 * gencoords (for positions, velocities, and accelerations), and body segments
 * (for forces and torques). Then it stores the data appropriately in the
 * MotionData structure.
 */
static ReturnCode verify_kinetics_data(dpModelStruct* sdm, MotionData* data)
{
   int i, gc, seg, column, num_otherdata;
   dpFunctionType functionType;
   double cutoffFrequency;
   ForceStruct* fs;
   TorqueStruct* ts;

   // The size of these arrays is fixed at the number of gencoords in the model.
   data->q_data = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));
   data->u_data = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));
   data->udot_data = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));
   data->q_moments = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));

   // These arrays are initially allocated to their maximum possible size, and then
   // they are re-alloced later.
   
   data->forces = (ForceStruct**)simm_calloc(data->num_elements, sizeof(ForceStruct*));
   data->torques = (TorqueStruct**)simm_calloc(data->num_elements, sizeof(TorqueStruct*));

   for (column = 0, num_otherdata = 0, i = 0; i<data->num_elements; i++)
   {
      functionType = dpStepFunction;

      if ((gc = name_is_gencoord(data->elementnames[i], sdm, NULL, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->q_data[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         init_function(data->q_data[gc], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         column++;
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, velocity_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->u_data[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         init_function(data->u_data[gc], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         column++;
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, acceleration_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->udot_data[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         init_function(data->udot_data[gc], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         column++;
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, torque_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->q_moments[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         init_function(data->q_moments[gc], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         column++;
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, moment_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->q_moments[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         init_function(data->q_moments[gc], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         column++;
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, force_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         data->forces[data->num_forces] = (ForceStruct*)simm_calloc(1, sizeof(ForceStruct));
         fs = data->forces[data->num_forces++];
         fs->segment = seg;
         init_function(&fs->appl_point[XX], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         init_function(&fs->appl_point[YY], data->motiondata[column+1], data->num_frames, functionType, cutoffFrequency);
         init_function(&fs->appl_point[ZZ], data->motiondata[column+2], data->num_frames, functionType, cutoffFrequency);
         init_function(&fs->force_vec[XX], data->motiondata[column+3], data->num_frames, functionType, cutoffFrequency);
         init_function(&fs->force_vec[YY], data->motiondata[column+4], data->num_frames, functionType, cutoffFrequency);
         init_function(&fs->force_vec[ZZ], data->motiondata[column+5], data->num_frames, functionType, cutoffFrequency);
         column += 6;
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, torque_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         data->torques[data->num_torques] = (TorqueStruct*)simm_calloc(1, sizeof(TorqueStruct));
         ts = data->torques[data->num_torques++];
         ts->segment = seg;
         init_function(&ts->torque_vec[XX], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         init_function(&ts->torque_vec[YY], data->motiondata[column+1], data->num_frames, functionType, cutoffFrequency);
         init_function(&ts->torque_vec[ZZ], data->motiondata[column+2], data->num_frames, functionType, cutoffFrequency);
         column += 3;
      }
      else if (STRINGS_ARE_EQUAL(data->elementnames[i], "time"))
      {
         data->time_column = i;
         column++;
      }
      else
      {
         num_otherdata++;
         column++;
      }
   }

   // Realloc these arrays, now that their actual size is known.
   data->forces = (ForceStruct**)realloc(data->forces, data->num_forces * sizeof(ForceStruct*));
   data->torques = (TorqueStruct**)realloc(data->torques, data->num_torques * sizeof(TorqueStruct*));

   if (setup_functions(sdm, data) == code_bad)
      return code_bad;

   if (num_otherdata > 0)
   {
      sim_message(none, "  %d column%sof data in kinetics file %s classified as 'otherdata'", num_otherdata,
         num_otherdata > 1 ? "s " : " ", num_otherdata > 1 ? "were" : "was");
      sim_message(none, "  and will be ignored.");
   }

   sim_message(none, "Verified data in kinetics file.");

   return code_fine;
}

/* MAP_MOTION_FORCE: use the data in the input motion file to initialize data in the applied force structure.
 */
static void map_motion_force(dpModelStruct* sdm, MotionData* data, int segNum, FTcomponent comp,
                             dpFunctionType functionType, double cutoffFrequency, int column)
{
   int i;
   ForceStruct *fs;

   // Scan the force structs, looking for one which matches the segment, but which does not
   // have the 'comp' component defined. If you find one, use the column of data to initialize
   // that component. If you don't, make a new force struct and define that component.
   for (i=0; i<data->num_forces; i++)
   {
      if (comp <= POSZ && data->forces[i]->segment == segNum &&
          data->forces[i]->appl_point[comp].status == dpFunctionUndefined)
      {
         init_function(&data->forces[i]->appl_point[comp], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         return;
      }
      else if (comp >= VECX && comp <= VECZ && data->forces[i]->segment == segNum &&
               data->forces[i]->force_vec[comp - 3].status == dpFunctionUndefined)
      {
         init_function(&data->forces[i]->force_vec[comp - 3], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         return;
      }
      else if (comp >= FREETX && data->forces[i]->segment == segNum &&
               data->forces[i]->free_torque[comp - 6].status == dpFunctionUndefined)
      {
         init_function(&data->forces[i]->free_torque[comp - 6], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
         return;
      }
   }

   // You didn't find a slot for this component, so start a new ForceStruct.
   data->forces[data->num_forces] = (ForceStruct*)simm_calloc(1, sizeof(ForceStruct));
   fs = data->forces[data->num_forces++];
   fs->segment = segNum;
   for (i=0; i<3; i++)
   {
      fs->appl_point[i].status = dpFunctionUndefined;
      fs->force_vec[i].status = dpFunctionUndefined;
      fs->free_torque[i].status = dpFunctionUndefined;
   }

   if (comp <= POSZ)
      init_function(&fs->appl_point[comp], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
   else if (comp <= VECZ)
      init_function(&fs->force_vec[comp - 3], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
   else
      init_function(&fs->free_torque[comp - 6], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
}

/* MAP_MOTION_TORQUE: use the input motion data to initialize the data in the applied torque structure.
 */
static void map_motion_torque(dpModelStruct* sdm, MotionData* data, int segNum, FTcomponent comp,
                              dpFunctionType functionType, double cutoffFrequency, int column)
{
   int i;
   TorqueStruct *ts;

   // Scan the torque structs, looking for one which matches the segment, but which does not
   // have the 'comp' component defined. If you find one, use the column of data to initialize
   // that component. If you don't, make a new torque struct and define that component.
   for (i = 0; i < data->num_torques; i++)
   {
      if (data->torques[i]->segment == segNum && data->torques[i]->torque_vec[comp].status == dpFunctionUndefined)
      {
         init_function(&data->torques[i]->torque_vec[comp], data->motiondata[column], data->num_frames,
                          functionType, cutoffFrequency);
         return;
      }
   }

   // You didn't find a slot for this component, so start a new TorqueStruct.
   data->torques[data->num_torques] = (TorqueStruct*)simm_calloc(1, sizeof(TorqueStruct));
   ts = data->torques[data->num_torques++];
   ts->segment = segNum;
   for (i = 0; i < 3; i++)
      ts->torque_vec[i].status = dpFunctionUndefined;
   init_function(&ts->torque_vec[comp], data->motiondata[column], data->num_frames, functionType, cutoffFrequency);
}

/* MAP_MUSCLE_EXCITATION: store the information about muscle excitation from an input motion file into the model's muscles.
 */
static map_muscle_excitation(dpModelStruct* sdm, MotionData* data, int muscNum, dpFunctionType functionType, double cutoffFrequency, int column)
{
   dpFunction* sf;
   dpMuscleStruct* muscle = &sdm->muscles[muscNum];

   // Put this data in the next unused musc_excitation slot.
   sf = data->musc_excitations[data->num_musc_excitations] = (dpFunction*)simm_malloc(sizeof(dpFunction));
   data->num_musc_excitations++;

   init_function(sf, data->motiondata[column], data->num_frames, functionType, cutoffFrequency);

   // Now point the muscle's excitation function to this one.
   muscle->excitation_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
   *muscle->excitation_func = sf;

   // If the muscle's excitation abscissa is NULL or equal to the default muscle's, set it to TIME.
   // Otherwise, leave it alone (this allows the column of excitation data to be interpreted as a
   // function of a gencoord, if the excitation_abscissa is so specified in the muscle file).
   if (muscle->excitation_abscissa == NULL || muscle->excitation_abscissa == sdm->default_muscle.excitation_abscissa)
   {
      muscle->excitation_abscissa = (void**)simm_malloc(sizeof(void*));
      *muscle->excitation_abscissa = TIME;
   }
}


/* UPDATE_COLUMN_NAMES: If smoothing data has been specified, add the suffix to the column names so that
 * the data will get smoothed.
 * Gencoord data is smoothed using the kinematic cutoff frequency.
 * External forces and gencoord torques are smoothed using the force cutoff frequency.
 * Muscle activations/excitations are smoothed using the emg cutoff frequency.
 */
static void update_column_names(MotionData *motion)
{
   int i, j, motionObject;
   char suffix[CHARBUFFER], new_name[CHARBUFFER], *component = NULL;
   dpFunctionType type;
   double cutoff;

   //what if you don't want any smoothing???
   for (i = 0; i < motion->num_elements; i++)
   {
      // gencoords
      if (name_is_gencoord(motion->elementnames[i], sdm, NULL, &type, &cutoff) >= 0)
      {
         if (params.inputOptions.kinematic_cutoff_defined && params.inputOptions.kinematic_cutoff > 0)
            sprintf(new_name, "%s_gcv_%d", motion->elementnames[i], params.inputOptions.kinematic_cutoff);
         else
            sprintf(new_name, "%s_gcv", motion->elementnames[i]);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }

      // gencoord velocities
      else if (name_is_gencoord(motion->elementnames[i], sdm, "_vel", &type, &cutoff) >= 0)
      {
         if (params.inputOptions.kinematic_cutoff_defined && params.inputOptions.kinematic_cutoff > 0)
            sprintf(new_name, "%s_gcv_%d", motion->elementnames[i], params.inputOptions.kinematic_cutoff);
         else
            sprintf(new_name, "%s_gcv", motion->elementnames[i]);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }

      // gencoord accelerations
      else if (name_is_gencoord(motion->elementnames[i], sdm, "_acc", &type, &cutoff) >= 0)
      {
         if (params.inputOptions.kinematic_cutoff_defined && params.inputOptions.kinematic_cutoff > 0)
            sprintf(new_name, "%s_gcv_%d", motion->elementnames[i], params.inputOptions.kinematic_cutoff);
         else
            sprintf(new_name, "%s_gcv", motion->elementnames[i]);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }

      // gencoord moments
      else if (name_is_gencoord(motion->elementnames[i], sdm, "_torque", &type, &cutoff) >= 0 || name_is_gencoord(motion->elementnames[i], sdm, "_moment", &type, &cutoff) >= 0)
      {
         if (params.inputOptions.force_cutoff_defined && params.inputOptions.force_cutoff > 0)
            sprintf(new_name, "%s_gcv_%d", motion->elementnames[i], params.inputOptions.force_cutoff);
         else
            sprintf(new_name, "%s_gcv", motion->elementnames[i]);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }

      // external forces, but not forceplates
      else if (name_is_body_segment(motion->elementnames[i], sdm, "_force", &type, &cutoff) >= 0)  // check what suffix should be
      {
		  {
		     if (params.inputOptions.force_cutoff_defined && params.inputOptions.force_cutoff > 0)
			    sprintf(new_name, "%s_gcv_%d", motion->elementnames[i], params.inputOptions.force_cutoff);
			 else
			    sprintf(new_name, "%s_gcv", motion->elementnames[i]);
			 FREE_IFNOTNULL(motion->elementnames[i]);
			 mstrcpy(&motion->elementnames[i], new_name);
		  }
      }

      // forceplate torques - check this before forceplate forces (which don't have the "force" suffix
      else if (name_is_forceplate(motion->elementnames[i], sdm, motion, "_torque", &component, NULL, NULL) >= 0)
      {
         if (params.inputOptions.force_cutoff_defined && params.inputOptions.force_cutoff > 0)
            sprintf(new_name, "ground_torque%s_gcv_%d", component, params.inputOptions.force_cutoff);
         else
            sprintf(new_name, "ground_torque%s_gcv", component);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }

      // forceplate forces
      else if (name_is_forceplate(motion->elementnames[i], sdm, motion, NULL, &component, NULL, NULL) >= 0)
      {
         if (params.inputOptions.force_cutoff_defined && params.inputOptions.force_cutoff > 0)
            sprintf(new_name, "ground_force%s_gcv_%d", component, params.inputOptions.force_cutoff);
         else
            sprintf(new_name, "ground_force%s_gcv", component);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }

      // muscle activations
      else if (name_is_muscle(motion->elementnames[i], sdm, NULL, &type, &cutoff) >= 0)
      {
         if (params.inputOptions.emg_cutoff_defined && params.inputOptions.emg_cutoff > 0)
            sprintf(new_name, "%s_gcv_%d", motion->elementnames[i], params.inputOptions.emg_cutoff);
         else
            sprintf(new_name, "%s_gcv", motion->elementnames[i]);
         FREE_IFNOTNULL(motion->elementnames[i]);
         mstrcpy(&motion->elementnames[i], new_name);
      }
   }
}



/* VERIFY_MOTION_DATA: this routine checks the validity of the motion data.
 * It checks the column names in the file to make sure they match the model's
 * gencoords (for positions, velocities, and accelerations), body segments (for
 * forces and torques), and muscles (for excitations). Then it stores the data
 * appropriately in the MotionData structure.
 */
static ReturnCode verify_motion_data(dpModelStruct* sdm, MotionData* data)
{
   int i, gc, seg, musc, num_otherdata;
   dpFunctionType functionType;
   double cutoffFrequency;
   ForceStruct* fs;
   TorqueStruct* ts;

   // The size of these arrays is fixed at the number of gencoords in the model.
   data->q_data = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));
   data->u_data = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));
   data->udot_data = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));
   data->q_moments = (dpFunction**)simm_calloc(sdm->nq, sizeof(dpFunction*));

   // this is done in write_dp_input_motion so needs to be done in standalone so that results match
   if (isStandAlone)
      update_column_names(data);

   // These arrays are initially allocated to their maximum possible size, and then
   // they are re-alloced later.
   
   data->musc_excitations = (dpFunction**)simm_calloc(sdm->num_muscles, sizeof(dpFunction*));
   data->forces = (ForceStruct**)simm_calloc(data->num_elements, sizeof(ForceStruct*));
   data->torques = (TorqueStruct**)simm_calloc(data->num_elements, sizeof(TorqueStruct*));

   for (num_otherdata = 0, i = 0; i<data->num_elements; i++)
   {
      functionType = dpStepFunction;

      if ((gc = name_is_gencoord(data->elementnames[i], sdm, NULL, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->q_data[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         if (params.inputOptions.kinematic_cutoff_defined)
         {
            functionType = dpGCVSpline;
            cutoffFrequency = params.inputOptions.kinematic_cutoff;
         }
         init_function(data->q_data[gc], data->motiondata[i], data->num_frames, functionType, cutoffFrequency);
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, velocity_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->u_data[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         if (params.inputOptions.kinematic_cutoff_defined)
         {
            functionType = dpGCVSpline;
            cutoffFrequency = params.inputOptions.kinematic_cutoff;
         }
         init_function(data->u_data[gc], data->motiondata[i], data->num_frames, functionType, cutoffFrequency);
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, acceleration_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->udot_data[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         if (params.inputOptions.kinematic_cutoff_defined)
         {
            functionType = dpGCVSpline;
            cutoffFrequency = params.inputOptions.kinematic_cutoff;
         }
         init_function(data->udot_data[gc], data->motiondata[i], data->num_frames, functionType, cutoffFrequency);
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, torque_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->q_moments[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         if (params.inputOptions.force_cutoff_defined)
         {
            functionType = dpGCVSpline;
            cutoffFrequency = params.inputOptions.force_cutoff;
         }
         init_function(data->q_moments[gc], data->motiondata[i], data->num_frames, functionType, cutoffFrequency);
      }
      else if ((gc = name_is_gencoord(data->elementnames[i], sdm, moment_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_GENCOORD)
      {
         data->q_moments[gc] = (dpFunction*)simm_calloc(1, sizeof(dpFunction));
         if (params.inputOptions.force_cutoff_defined)
         {
            functionType = dpGCVSpline;
            cutoffFrequency = params.inputOptions.force_cutoff;
         }
         init_function(data->q_moments[gc], data->motiondata[i], data->num_frames, functionType, cutoffFrequency);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, px_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         map_motion_force(sdm, data, seg, POSX, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, py_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         map_motion_force(sdm, data, seg, POSY, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, pz_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         map_motion_force(sdm, data, seg, POSZ, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, vx_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         map_motion_force(sdm, data, seg, VECX, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, vy_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         map_motion_force(sdm, data, seg, VECY, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, vz_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         map_motion_force(sdm, data, seg, VECZ, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, ax_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         // If the torque is applied to ground, assume that it is a free torque and is part of a ground-reaction force.
         if (seg == GROUND)
            map_motion_force(sdm, data, seg, FREETX, functionType, cutoffFrequency, i);
         else
            map_motion_torque(sdm, data, seg, POSX, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, ay_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         // If the torque is applied to ground, assume that it is a free torque and is part of a ground-reaction force.
         if (seg == GROUND)
            map_motion_force(sdm, data, seg, FREETY, functionType, cutoffFrequency, i);
         else
            map_motion_torque(sdm, data, seg, POSY, functionType, cutoffFrequency, i);
      }
      else if ((seg = name_is_body_segment(data->elementnames[i], sdm, az_suffix, &functionType, &cutoffFrequency)) != UNDEFINED_SEGMENT)
      {
         // If the torque is applied to ground, assume that it is a free torque and is part of a ground-reaction force.                   
         if (seg == GROUND)
            map_motion_force(sdm, data, seg, FREETZ, functionType, cutoffFrequency, i);
         else
            map_motion_torque(sdm, data, seg, POSZ, functionType, cutoffFrequency, i);
      }
      else if ((musc = name_is_muscle(data->elementnames[i], sdm, NULL, &functionType, &cutoffFrequency)) != -1)
      {
         map_muscle_excitation(sdm, data, musc, functionType, cutoffFrequency, i);
      }
      else if (STRINGS_ARE_EQUAL(data->elementnames[i], "time"))
      {
         data->time_column = i;
      }
      else
      {
         num_otherdata++;
      }
   }

   // Realloc these arrays, now that their actual size is known.
   data->forces = (ForceStruct**)realloc(data->forces, data->num_forces * sizeof(ForceStruct*));
   data->torques = (TorqueStruct**)realloc(data->torques, data->num_torques * sizeof(TorqueStruct*));
   data->musc_excitations = (dpFunction**)realloc(data->musc_excitations,
                                                      data->num_musc_excitations * sizeof(dpFunction*));

   if (setup_functions(sdm, data) == code_bad)
      return code_bad;

   if (num_otherdata > 0)
   {
      sim_message(none, "  %d column%sof data in motion file %s classified as 'otherdata'", num_otherdata,
         num_otherdata > 1 ? "s " : " ", num_otherdata > 1 ? "were" : "was");
      sim_message(none, "  and will be ignored.");
   }

   sim_message(none, "Verified data in motion file.");

   return code_fine;
}


/* NAME_IS_GENCOORD: this routine checks to see if a string is the name
 * of one of the model's gencoords, with an optional suffix and also
 * possibly the extra suffix "_lin", "_spl", or "_gcv" (possibly followed a cutoff
 * frequency). It returns the index of the gencoord if there is a match.
 */
static int name_is_gencoord(char name[], dpModelStruct* sdm, char suffix[], dpFunctionType* functionType, double* cutoffFrequency)
{
   int i, len, lenQ, Qnum, index, maxLen;
   char* ptr;

   len = strlen(name);

   // First check to see if the string begins with the name of a Q.
   // To handle models which have overlapping gencoord names, like
   // "q1" and "q10", check for a match with all gencoords, and take
   // the longest match.
   for (i = 0, index = -1, maxLen = -1; i < sdm->nu; i++)
   {
      lenQ = strlen(sdm->q[i].name);
      if (len >= lenQ)
      {
         if (!strncmp(name, sdm->q[i].name, lenQ))
         {
            if (lenQ > maxLen)
            {
               index = i;
               maxLen = lenQ;
            }
         }
      }
   }

   if (index == -1)
      return UNDEFINED_GENCOORD;

   // You've found a matching Q name, so move ptr past the Q name and
   // get ready to check the suffixes.
   Qnum = index;
   ptr = &name[maxLen];
   len -= maxLen;

   // If a suffix was passed in, check to see if the name ends in that suffix.
   // If it does, remove the suffix and continue on. If it does not, return -1
   // because the suffix is not optional.
   if (suffix)
   {
      int suffix_len = strlen(suffix);

      if (len >= suffix_len)
      {
         if (!strncmp(ptr, suffix, suffix_len))
         {
            ptr += suffix_len;
            len -= suffix_len;
         }
         else
         {
            return UNDEFINED_GENCOORD;
         }
      }
      else
      {
         return UNDEFINED_GENCOORD;
      }
   }

   // If functionType and cutoffFrequency are not NULL, check to see if the name ends in
   // one of the function type labels (followed by an optional cutoff frequency). If it
   // does, set *functionType to the appropriate type. If no function label is found, set
   // the type to dpNaturalCubicSpline.
   if (functionType && cutoffFrequency)
   {
      int matched_spl = 0, matched_lin = 0;
      int lin_len = strlen(linear_text);
      int spl_len = strlen(natural_cubic_text);
      int gcv_len = strlen(gcv_text);

      *functionType = dpNaturalCubicSpline;
      *cutoffFrequency = -1.0; // by default there is no smoothing

      if (len >= lin_len)
      {
         if (!strncmp(ptr, linear_text, lin_len))
         {
            *functionType = dpLinearFunction;
            ptr += lin_len;
            len -= lin_len;
            matched_lin = 1;
         }
      }

      if (!matched_lin && len >= spl_len)
      {
         if (!strncmp(ptr, natural_cubic_text, spl_len))
         {
            *functionType = dpNaturalCubicSpline;
            ptr += spl_len;
            len -= spl_len;
            matched_spl = 1;
         }
      }

      if (!matched_lin && !matched_spl && len >= gcv_len)
      {
         if (!strncmp(ptr, gcv_text, gcv_len))
         {
            ptr += gcv_len;
            len -= gcv_len;
            *functionType = dpGCVSpline;
            if (len > 0)
            {
               char* intPtr = buffer;

               // Move over the underscore and look for an integer.
               if (*(ptr++) != '_')
               {
                  return UNDEFINED_GENCOORD;
               }
               else
               {
                  len--; // for the underscore character
                  while (*ptr >= '0' && *ptr <= '9')
                  {
                     *(intPtr++) = *(ptr++);
                     len--;
                  }
                  *intPtr = STRING_TERMINATOR;
                  *cutoffFrequency = atof(buffer);
               }
            }
         }
      }
   }

   // If there are extra characters after the suffixes, return an error.
   if (len > 0)
      return UNDEFINED_GENCOORD;

   return Qnum;
}


/* NAME_IS_BODY_SEGMENT: this routine checks to see if a string is the name
 * of one of the model's body segments, with an optional suffix and also possibly
 * the extra suffix "_spl" or "_gcv" (possibly followed by a cutoff frequency).
 * Since the C version of SD/FAST uses -1 as the number of the ground segment,
 * UNDEFINED_SEGMENT means there was no match.
 */
static int name_is_body_segment(char name[], dpModelStruct* sdm, char suffix[], dpFunctionType* functionType, double* cutoffFrequency)
{
   int i, len, lenS, Snum, index, maxLen;
   char* ptr;

   len = strlen(name);

   // First check to see if the string begins with the name of a body segment.
   // To handle models which have overlapping segment names, like
   // "body1" and "body10", check for a match with all segments, and take
   // the longest match.
   for (i = 0, index = -1, maxLen = -1; i < sdm->num_body_segments; i++)
   {
      lenS = strlen(sdm->body_segment[i].name);
      if (len >= lenS)
      {
         if (!strncmp(name, sdm->body_segment[i].name, lenS))
         {
            if (lenS > maxLen)
            {
               index = i;
               maxLen = lenS;
            }
         }
      }
   }

   if (index == -1)
      return UNDEFINED_SEGMENT;

   // You've found a matching segment name, so move ptr past the segment name and
   // get ready to check the suffixes.
   Snum = index - 1;
   ptr = &name[maxLen];
   len -= maxLen;

   // If a suffix was passed in, check to see if the name ends in that suffix.
   // If it does, remove the suffix and continue on. If it does not, return UNDEFINED_SEGMENT
   // because the suffix is not optional.
   if (suffix)
   {
      int suffix_len = strlen(suffix);

      if (len >= suffix_len)
      {
         if (!strncmp(ptr, suffix, suffix_len))
         {
            ptr += suffix_len;
            len -= suffix_len;
         }
         else
         {
            return UNDEFINED_SEGMENT;
         }
      }
      else
      {
         return UNDEFINED_SEGMENT;
      }
   }


   // If functionType and cutoffFrequency are not NULL, check to see if the name ends in
   // one of the function type labels (followed by an optional cutoff frequency). If it
   // does, set *functionType to the appropriate type. If no function label is found, set
   // the type to dpNaturalCubicSpline.
   if (functionType && cutoffFrequency)
   {
      int matched_spl = 0, matched_lin = 0;
      int lin_len = strlen(linear_text);
      int spl_len = strlen(natural_cubic_text);
      int gcv_len = strlen(gcv_text);

      *functionType = dpNaturalCubicSpline;
      *cutoffFrequency = -1.0; // by default there is no smoothing

      if (len >= lin_len)
      {
         if (!strncmp(ptr, linear_text, lin_len))
         {
            *functionType = dpLinearFunction;
            ptr += lin_len;
            len -= lin_len;
            matched_lin = 1;
         }
      }

      if (!matched_lin && len >= spl_len)
      {
         if (!strncmp(ptr, natural_cubic_text, spl_len))
         {
            *functionType = dpNaturalCubicSpline;
            ptr += spl_len;
            len -= spl_len;
            matched_spl = 1;
         }
      }

      if (!matched_lin && !matched_spl && len >= gcv_len)
      {
         if (!strncmp(ptr, gcv_text, gcv_len))
         {
            ptr += gcv_len;
            len -= gcv_len;
            *functionType = dpGCVSpline;
            if (len > 0)
            {
               char* intPtr = buffer;

               // Move over the underscore and look for an integer.
               if (*(ptr++) != '_')
               {
                  return UNDEFINED_SEGMENT;
               }
               else
               {
                  len--; // for the underscore character
                  while (*ptr >= '0' && *ptr <= '9')
                  {
                     *(intPtr++) = *(ptr++);
                     len--;
                  }
                  *intPtr = STRING_TERMINATOR;
                  *cutoffFrequency = atof(buffer);
               }
            }
         }
      }
   }

   // If there are extra characters after the suffixes, return an error.
   if (len > 0)
      return UNDEFINED_SEGMENT;

   return Snum;
}

/* NAME_IS_FORCEPLATE: this routine checks to see if a string is the name of a forceplate defined in the given motion file.
 * If the name is a forceplate, return its number, and if they exist, any component string (eg. _px, _py...), the type of function,
 * and the cutoff.
 */
static int name_is_forceplate(char name[], dpModelStruct* sdm, MotionData *motion, char suffix[], char **component, dpFunctionType* functionType, double* cutoffFrequency)
{
   int i, len, lenS, Snum, index, maxLen, plateNum, Flen = 0;
   char* ptr, buffer[CHARBUFFER];

   len = strlen(name);

   // If the column name starts with "forceplate" and a number, then it is forceplate data.
   if (sscanf(name, "forceplate%d", &plateNum) != 1)
      return -1;

   sprintf(buffer, "forceplate%d", plateNum);
   Flen = strlen(buffer);

   // Move ptr past the forceplate name and number.
   ptr = &name[Flen];
   len -= Flen;

   // If a suffix was passed in, check to see if the name ends in that suffix.
   // If it does, remove the suffix and continue on. If it does not, return UNDEFINED_SEGMENT
   // because the suffix is not optional.
   if (suffix)
   {
      int suffix_len = strlen(suffix);

      if (len >= suffix_len)
      {
         if (!strncmp(ptr, suffix, suffix_len))
         {
            ptr += suffix_len;
            len -= suffix_len;
         }
         else
         {
            return UNDEFINED_SEGMENT;
         }
      }
      else
      {
         return UNDEFINED_SEGMENT;
      }
      // if the suffix was found, return the matching component string
      // if no matching component was found there's something wrong so return undefined segment
      if (!strncmp(ptr, "_x", 3))
         mstrcpy(component, "_x");
      else if (!strncmp(ptr, "_y", 3))
         mstrcpy(component, "_y");
      else if (!strncmp(ptr, "_z", 3))
         mstrcpy(component, "_z");
      else
         return UNDEFINED_SEGMENT;
   }
   // if there is no suffix, if the remaining name matches _px, _py, _pz, _vx, _vy, or _vz return that component string
   // if no matching component was found there's something wrong so return undefined segment
   else
   {
      if (!strncmp(ptr, "_px", 3))
         mstrcpy(component, "_px");
      else if (!strncmp(ptr, "_py", 3))
         mstrcpy(component, "_py");
      else if (!strncmp(ptr, "_pz", 3))
         mstrcpy(component, "_pz");
      else if (!strncmp(ptr, "_vx", 3))
         mstrcpy(component, "_vx");
      else if (!strncmp(ptr, "_vy", 3))
         mstrcpy(component, "_vy");
      else if (!strncmp(ptr, "_vz", 3))
         mstrcpy(component, "_vz");
      else 
         return UNDEFINED_SEGMENT;
   }

   ptr += 3;
   len -= 3;

   // If functionType and cutoffFrequency are not NULL, check to see if the name ends in
   // natural_cubic_text or gcv_text (followed by an optional cutoff frequency). If it
   // does, set *functionType to the appropriate type. If no function label is found, set
   // the type to step_func.
   if (functionType && cutoffFrequency)
   {
      int matched_spl = 0;
      int spl_len = strlen(natural_cubic_text);
      int gcv_len = strlen(gcv_text);

      *functionType = dpStepFunction;
      *cutoffFrequency = -1.0; // by default there is no smoothing

      if (len >= spl_len)
      {
         if (!strncmp(ptr, natural_cubic_text, spl_len))
         {
            *functionType = dpNaturalCubicSpline;
            ptr += spl_len;
            len -= spl_len;
            matched_spl = 1;
         }
      }

      if (!matched_spl && len >= gcv_len)
      {
         if (!strncmp(ptr, gcv_text, gcv_len))
         {
            ptr += gcv_len;
            len -= gcv_len;
            *functionType = dpGCVSpline;
            if (len > 0)
            {
               char* intPtr = buffer;

               // Move over the underscore and look for an integer.
               if (*(ptr++) != '_')
               {
                  return -1;
               }
               else
               {
                  len--; // for the underscore character
                  while (*ptr >= '0' && *ptr <= '9')
                  {
                     *(intPtr++) = *(ptr++);
                     len--;
                  }
                  *intPtr = STRING_TERMINATOR;
                  *cutoffFrequency = atof(buffer);
               }
            }
         }
      }
   }

   return plateNum;
}

/* NAME_IS_MUSCLE: this routine checks to see if a string is the name
 * of one of the model's muscles, with an optional suffix and also
 * possibly the extra suffix "_spl" or "_gcv" (possibly followed a cutoff
 * frequency). It returns the index of the muscle if there is a match.
 */
int name_is_muscle(char name[], dpModelStruct* sdm, char suffix[], dpFunctionType* functionType, double* cutoffFrequency)
{
   int i, len, lenM, Mnum, index, maxLen;
   char* ptr;

   len = strlen(name);

   // First check to see if the string begins with the name of a muscle.
   // To handle models which have overlapping muscle names, like
   // "trap1" and "trap10", check for a match with all muscles, and take
   // the longest match.
   for (i = 0, index = -1, maxLen = -1; i < sdm->num_muscles; i++)
   {
      lenM = strlen(sdm->muscles[i].name);
      if (len >= lenM)
      {
         if (!strncmp(name, sdm->muscles[i].name, lenM))
         {
            if (lenM > maxLen)
            {
               index = i;
               maxLen = lenM;
            }
         }
      }
   }

   if (index == -1)
      return -1;

   // You've found a matching muscle name, so move ptr past the name and
   // get ready to check the suffixes.
   Mnum = index;
   ptr = &name[maxLen];
   len -= maxLen;

   // If a suffix was passed in, check to see if the name ends in that suffix.
   // If it does, remove the suffix and continue on. If it does not, return -1
   // because the suffix is not optional.
   if (suffix)
   {
      int suffix_len = strlen(suffix);

      if (len >= suffix_len)
      {
         if (!strncmp(ptr, suffix, suffix_len))
         {
            ptr += suffix_len;
            len -= suffix_len;
         }
         else
         {
            return -1;
         }
      }
      else
      {
         return -1;
      }
   }

   // If functionType and cutoffFrequency are not NULL, check to see if the name ends in
   // one of the function type labels (followed by an optional cutoff frequency). If it
   // does, set *functionType to the appropriate type. If no function label is found, set
   // the type to dpStepFunction.
   if (functionType && cutoffFrequency)
   {
      int matched_spl = 0, matched_lin = 0;
      int lin_len = strlen(linear_text);
      int spl_len = strlen(natural_cubic_text);
      int gcv_len = strlen(gcv_text);

      *functionType = dpStepFunction;
      *cutoffFrequency = -1.0; // by default there is no smoothing

      if (len >= lin_len)
      {
         if (!strncmp(ptr, linear_text, lin_len))
         {
            *functionType = dpLinearFunction;
            ptr += lin_len;
            len -= lin_len;
            matched_lin = 1;
         }
      }

      if (!matched_lin && len >= spl_len)
      {
         if (!strncmp(ptr, natural_cubic_text, spl_len))
         {
            *functionType = dpNaturalCubicSpline;
            ptr += spl_len;
            len -= spl_len;
            matched_spl = 1;
         }
      }

      if (!matched_lin && !matched_spl && len >= gcv_len)
      {
         if (!strncmp(ptr, gcv_text, gcv_len))
         {
            ptr += gcv_len;
            len -= gcv_len;
            *functionType = dpGCVSpline;
            if (len > 0)
            {
               char* intPtr = buffer;

               // Move over the underscore and look for an integer.
               if (*(ptr++) != '_')
               {
                  return -1;
               }
               else
               {
                  len--; // for the underscore character
                  while (*ptr >= '0' && *ptr <= '9')
                  {
                     *(intPtr++) = *(ptr++);
                     len--;
                  }
                  *intPtr = STRING_TERMINATOR;
                  *cutoffFrequency = atof(buffer);
               }
            }
         }
      }
   }

   // If there are extra characters after the suffixes, return an error.
   if (len > 0)
      return -1;

   return Mnum;
}


/* STRING_HAS_SUFFIX: this function checks if str1 ends in suffix. */

static dpBoolean string_has_suffix(char str1[], char suffix[])
{

   int len, suffix_len;

   len = strlen(str1);
   suffix_len = strlen(suffix);

   if (len >= suffix_len)
   {
      if (strncmp(&str1[len - suffix_len], suffix, suffix_len) == 0)
         return dpYes;
   }

   return dpNo;
}



/* SIMM_MALLOC: this is a wrapper routine for malloc. It checks the size
 * you want to malloc, then calls malloc. If there is an error, it prints
 * an error message then returns NULL.
 */

void* simm_malloc(unsigned mem_size)
{
   void* ptr;

   // To make sure you don't try to malloc 0 bytes (which is a problem
   // with some compilers because malloc returns NULL when you try it).
   // This is not the most elegant solution, but it works in the case
   // in which simm_malloc() is accidentally called with size 0.
   
   if (mem_size <= 0)
      mem_size = sizeof(int);

   ptr = malloc(mem_size);

   if (ptr == NULL)
      sim_message(exit_program, "Ran out of memory. Unable to malloc %d bytes.", (int)mem_size);

   return ptr;
}



/* SIMM_CALLOC: this is a wrapper routine for calloc. It checks the size
 * you want to calloc, then calls calloc. If there is an error, it prints
 * an error message then returns NULL.
 */

void* simm_calloc(unsigned num_elements, unsigned elem_size)
{
   void* ptr;

   // To make sure you don't try to malloc 0 bytes (which is a problem
   // with some compilers because malloc returns NULL when you try it).
   // This is not the most elegant solution, but it works in the case
   // in which simm_calloc() is accidentally called with size or num_elements 0.
   
   if (num_elements*elem_size <= 0)
   {
      num_elements = 1;
      elem_size = sizeof(int);
   }

   ptr = calloc(num_elements,elem_size);

   if (ptr == NULL)
      sim_message(exit_program, "Ran out of memory. Unable to calloc %d bytes.", (int)(num_elements*elem_size));

   return ptr;
}



/* SIMM_REALLOC: this is a wrapper routine for realloc. It checks the size
 * you want to realloc to, then calls realloc. If there is an error, it prints
 * an error message then returns NULL.
 */

void* simm_realloc(void* ptr, unsigned mem_size, ReturnCode* rc)
{
   void* new_ptr;

   // To make sure you don't try to malloc 0 bytes (which is a problem
   // with some compilers because malloc returns NULL when you try it).
   // This is not the most elegant solution, but it works in the case
   // in which simm_calloc() is accidentally called with size 0.
   
   if (mem_size <= 0)
      mem_size = sizeof(int);

   new_ptr = realloc(ptr,mem_size);

   if (new_ptr == NULL)
   {
      sim_message(none, "Ran out of memory. Unable to realloc %d bytes.", (int)mem_size);
      *rc = code_bad;
      return ptr;
   }

   *rc = code_fine;
   return new_ptr;
}



/* CONVERT_STRING: this routine scans a string and converts all special
 * characters into underscores. A special character is any character
 * other than a letter or number. The resulting string is one token,
 * and can therefore be used as a variable name in SIMM-written C code.
 */

void convert_string(char str[])
{

   int i, len;

   len = strlen(str);

   for (i=0; i<len; i++)
   {
      if (str[i] >= 97 && str[i] <= 122)  // lowercase letters
         continue;
      if (str[i] >= 65 && str[i] <= 90)   // uppercase letters
         continue;
      if (str[i] >= 48 && str[i] <= 57)   // numbers
         continue;
      str[i] = '_';
   }

}

/* SIMM_FOPEN: open a file. */
FILE* simm_fopen(const char* name, const char* mode)
{
#ifdef WIN32
   if (mode && mode[0] == 'w')
      remove(name);
#endif
   
   return fopen(name, mode);
}


#ifdef WIN32
   #include "windows.h"
#endif

/* GET_TEMP_FILE_NAME: get a temporary file name. */
void get_temp_file_name(const char* fileName, char tempFileName[])
{
   static char sTempFilePath[256];
   static int sTempFilePathInited = 0;

   if (!sTempFilePathInited)
   {
      GetTempPath(sizeof(sTempFilePath), sTempFilePath);
      sTempFilePathInited = 1;
   }

   strcpy(tempFileName, sTempFilePath);

   if (fileName)
   {
      strcat(tempFileName, fileName);
   }
   else
   {
      static int counter = 0;
      int len = strlen(tempFileName);
      char* p = &tempFileName[len];

      sprintf(p, "%d-%d", (int)GetCurrentThreadId(), counter++);
   }
}

/* IS_ABSOLUTE_PATH: determine if the given path is absolute. */
dpBoolean is_absolute_path(char *path)
{
   if ((path[1] == ':') && (path[2] == DIR_SEP_CHAR))
      return dpYes;
   return dpNo;
}

/* READ_SPRING: Read a spring element from the parameters file and store it in an array in the model struct. */
static ReturnCode read_spring(FILE* fp)
{
   int i, segnum, index = 0;
   int num_params = 7;
   double pt[3]; 
   double friction, param_a, param_b, param_c, param_d, param_e, param_f;
   ReturnCode rc = code_fine;

   // Read segment name, get its SD/FAST index, and check that it is valid.
   if (fscanf(fp, "%s", buffer) != 1)
   {
      sim_message(exit_program, "Error reading segment name for spring %d.", sdm->num_springs + 1);
      return code_bad;
   }
   segnum = enter_segment(sdm,buffer);
   if (segnum  == UNDEFINED_SEGMENT)
      sim_message(exit_program, "Unknown segment name (%s) in spring definition.", buffer);

   // Read the coordinates of the spring attachement point, defined from the segment origin.
   if (fscanf(fp, "%lf %lf %lf", &pt[0], &pt[1], &pt[2]) != 3)
   {
      sim_message(exit_program, "Error reading spring coordinates for spring %d.", sdm->num_springs + 1);
      return code_bad;
   }

   // Read the name of the object it is contacting (a floor or surface).
   if (fscanf(fp, "%s ", buffer) != 1)
   {
      sim_message(exit_program, "Error reading spring floor/surface name for spring %d.", sdm->num_springs + 1);
      return code_bad;
   }

   // Read the spring parameters - friction and 6 other parameters.
   if (fscanf(fp, "%lf %lf %lf %lf %lf %lf %lf", &friction, &param_a, &param_b, &param_c, &param_d, &param_e, &param_f) != num_params)
   {
      sim_message(exit_program, "Error reading parameters for spring %d.", sdm->num_springs + 1);
      return code_bad;
   }

   // If you make it to here, then the spring was successfully read and the segment name matched one in the model, 
   // so add the spring to the list.  Reallocate space if the array is already full.
   if (sdm->spring_array_size == 0)
      sdm->spring_array_size += ARRAY_INCREMENT;
   if (sdm->num_springs == 0)
   {
      sdm->spring = (dpSpringStruct*)simm_malloc(sdm->spring_array_size * sizeof(dpSpringStruct));
   }
   else if (sdm->num_springs == sdm->spring_array_size)
   {
      sdm->spring_array_size += ARRAY_INCREMENT;
      sdm->spring = (dpSpringStruct *)simm_realloc(sdm->spring, (sdm->spring_array_size)*sizeof(dpSpringStruct), &rc);
      if (rc == code_bad)
         return code_bad;
   }

   // The spring coordinates read in are specified with respect to the segment's SIMM origin.  The coordinates are
   // stored with respect to the segment's body mass center.
   index = sdm->num_springs;
   sdvsub(pt, sdm->body_segment[segnum + 1].mass_center, sdm->spring[index].point);
   sdm->spring[index].sd_segment = segnum;
   sdm->spring[index].friction = friction;
   sdm->spring[index].param_a = param_a;
   sdm->spring[index].param_b = param_b;
   sdm->spring[index].param_c = param_c;
   sdm->spring[index].param_d = param_d;
   sdm->spring[index].param_e = param_e;
   sdm->spring[index].param_f = param_f;
   sdm->spring[index].force[0] = 0.0;
   sdm->spring[index].force[1] = 0.0;
   sdm->spring[index].force[2] = 0.0;

   // Find the index of the associated contact surface or floor.
   mstrcpy(&sdm->spring[index].contact_name, buffer);
   for (i = 0; i < sdm->num_spring_floors; i++)
   {
      if (STRINGS_ARE_EQUAL(sdm->spring[index].contact_name, sdm->spring_floor[i].name))
      {
         sdm->spring[index].contact_index = i;
         sdm->spring[index].is_floor_contact = dpYes;
      }
   }
   for (i = 0; i < sdm->num_spring_surfaces; i++)
   {
      if (STRINGS_ARE_EQUAL(sdm->spring[index].contact_name, sdm->spring_surface[i].name))
      {
         sdm->spring[index].contact_index = i;
         sdm->spring[index].is_floor_contact = dpNo;
      }
   }

   sdm->num_springs++;
   return code_fine;

}

/* READ_SPRING_FLOOR: Read a spring floor element from the parameters file and store it in the array in the model struct */
static ReturnCode read_spring_floor(FILE* fp)
{
   int segnum, count;
   ReturnCode rc = code_fine;
   dpPolyhedronStruct ph;
   char floor_name[CHARBUFFER], filename[CHARBUFFER], seg_name[CHARBUFFER], buf2[CHARBUFFER], buf[2][CHARBUFFER];

   // Read the floor name from the file.
   if (fscanf(fp,"%s", floor_name) != 1)
   {
      sim_message(exit_program, "Error reading spring_floor floor name definition.");
      return code_bad;
   }

   // Read file name (with spaces allowed).
   if (_read_til_tokens(fp, filename, "\t\r\n") == 0)
   {
      sim_message(exit_program, "Error reading filename for spring floor %s", floor_name);
      return code_bad;
   }
   _strip_outer_whitespace(filename);

   // Read the rest of the line.  It should contain the segment name and any other keywords like visible/invisible/active/inactive
   // which will be processed after the spring floor has been created.
   read_line(fp, buf2);

   // Read the segment name from the string.
   count = sscanf(buf2, "%s %s %s", seg_name, buf[0], buf[1]);
   if (count < 1)
   {
      sim_message(exit_program, "Could not read segment name in spring floor %s", floor_name);
      return code_bad;
   }

   // Check that the floor is attached to a valid segment.
   strcpy(buffer, seg_name);
   segnum = enter_segment(sdm, buffer);
   if (segnum == UNDEFINED_SEGMENT)
   {
      sim_message(exit_program, "Unknown segment name (%s) in spring_floor definition.", seg_name);
      return code_bad;
   }

   // Read the bone file.
   ph.body_segment = segnum;
   if (read_polyhedron(&ph, filename) == code_bad)
   {
      sim_message(exit_program, "Could not create spring_floor %s.", floor_name);
      return code_bad;
   }
   mstrcpy(&ph.name, floor_name);
   
   // If you make it to here, then the polyhedron was successfully loaded,
   // and the segment name matched one in the model, so add the floor to
   // the list, and store the polyhedron.  Reallocate space for the array
   // if it is already full.
   if (sdm->spring_floor_array_size == 0)
      sdm->spring_floor_array_size += ARRAY_INCREMENT;

   if (sdm->num_spring_floors == 0)
   {
      sdm->spring_floor = (dpSpringFloor*)simm_malloc(sdm->spring_floor_array_size * sizeof(dpSpringFloor));
   }
   else if (sdm->num_spring_floors == sdm->spring_floor_array_size)
   {
      sdm->spring_floor_array_size += ARRAY_INCREMENT;
      sdm->spring_floor = (dpSpringFloor*)simm_realloc(sdm->spring_floor, (sdm->spring_floor_array_size) * sizeof(dpSpringFloor), &rc);
      if (rc == code_bad)
      {
         sim_message(abort_action, "Ran out of memory. Unable to create spring floor.");
         return code_bad;
      }
   }

   // Now that you have allocated space for the floor, fill in the info.
   sdm->spring_floor[sdm->num_spring_floors].sd_segment = segnum;
   mstrcpy(&sdm->spring_floor[sdm->num_spring_floors].name, floor_name);
   sdm->spring_floor[sdm->num_spring_floors].sd_segment = segnum;
   sdm->spring_floor[sdm->num_spring_floors].ph = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
   copy_polyhedron(&ph, sdm->spring_floor[sdm->num_spring_floors].ph);

   sdm->spring_floor[sdm->num_spring_floors].active = dpYes;
   sdm->spring_floor[sdm->num_spring_floors].visible = dpYes;

   sdm->spring_floor[sdm->num_spring_floors].plane.a = sdm->spring_floor[sdm->num_spring_floors].ph->polygon[0].normal[0];
   sdm->spring_floor[sdm->num_spring_floors].plane.b = sdm->spring_floor[sdm->num_spring_floors].ph->polygon[0].normal[1];
   sdm->spring_floor[sdm->num_spring_floors].plane.c = sdm->spring_floor[sdm->num_spring_floors].ph->polygon[0].normal[2];
   sdm->spring_floor[sdm->num_spring_floors].plane.d = sdm->spring_floor[sdm->num_spring_floors].ph->polygon[0].d;

   // Process any optional parameters at the end of the line, like visible/invisible/active/inactive.
   count--;    // segment already read
   for (count--; count >= 0; count--)
   {
      if (STRINGS_ARE_EQUAL(buf[count], "visible"))
         sdm->spring_floor[sdm->num_spring_floors].visible = dpYes;
      else if (STRINGS_ARE_EQUAL(buf[count], "invisible"))
         sdm->spring_floor[sdm->num_spring_floors].visible = dpNo;
      else if (STRINGS_ARE_EQUAL(buf[count], "active"))
         sdm->spring_floor[sdm->num_spring_floors].active = dpYes;
      else if (STRINGS_ARE_EQUAL(buf[count], "inactive"))
         sdm->spring_floor[sdm->num_spring_floors].active = dpNo;
   }

   sdm->num_spring_floors++;

   return code_fine;
}

/* READ_SPRING_SURFACE: Read a spring surface element from the parameters file and store it in the array in the model struct. */
static ReturnCode read_spring_surface(FILE* fp)
{
   int segnum, count;
   ReturnCode rc = code_fine;
   dpPolyhedronStruct ph;
   char name[CHARBUFFER], filename[CHARBUFFER], seg_name[CHARBUFFER], buf2[CHARBUFFER], buf[2][CHARBUFFER];
   double scale[3];
   dpBoolean scaled = dpNo;

   // Read the floor name from the file.
   if (fscanf(fp,"%s", name) != 1)
   {
      sim_message(exit_program, "Error reading spring_surface name definition.");
      return code_bad;
   }

   // Read file name (with spaces allowed).
   if (_read_til_tokens(fp, filename, "\t\r\n") == 0)
   {
      sim_message(exit_program, "Error reading filename for spring surface %s", name);
      return code_bad;
   }
   _strip_outer_whitespace(filename);

   // Read the rest of the line.  It should contain the segment name and any other keywords like visible/invisible/active/inactive
   // which will be processed after the spring floor has been created.
   read_line(fp, buf2);

   // Read the segment name from the string, also read any scale factors and optional parameters
   count = sscanf(buf2, "%s %lf %lf %lf %s %s", seg_name, &scale[0], &scale[1], &scale[2], buf[0], buf[1]);
   if (count < 1 || count > 6)
   {
      sim_message(exit_program, "Could not read segment name in spring surface %s", name);
      return code_bad;
   }

   // Check that the floor is attached to a valid segment.
   strcpy(buffer, seg_name);
   segnum = enter_segment(sdm, buffer);
   if (segnum == UNDEFINED_SEGMENT)
   {
      sim_message(exit_program, "Unknown segment name (%s) in spring_surface definition.", seg_name);
      return code_bad;
   }

   // Read the bone file.
   ph.body_segment = segnum;
   if (read_polyhedron(&ph, filename) == code_bad)
   {
      sim_message(exit_program, "Could not create spring_surface %s.", name);
      return code_bad;
   }

   mstrcpy(&ph.name, name);
   
   // If you make it to here, then the polyhedron was successfully loaded,
   // and the segment name matched one in the model, so add the floor to
   // the list, and store the polyhedron.  Reallocate space for the array
   // if it is already full.
   if (sdm->spring_surface_array_size == 0)
      sdm->spring_surface_array_size += ARRAY_INCREMENT;

   if (sdm->num_spring_surfaces == 0)
   {
      sdm->spring_surface = (dpSpringSurface*)simm_malloc(sdm->spring_surface_array_size * sizeof(dpSpringSurface));
   }
   else if (sdm->num_spring_surfaces == sdm->spring_surface_array_size)
   {
      sdm->spring_surface_array_size += ARRAY_INCREMENT;
      sdm->spring_surface = (dpSpringSurface*)simm_realloc(sdm->spring_surface, (sdm->spring_surface_array_size) * sizeof(dpSpringSurface), &rc);
      if (rc == code_bad)
      {
         sim_message(abort_action, "Ran out of memory. Unable to create spring surface.");
         return code_bad;
      }
   }

   // Now that you have allocated space for the surface, fill in the info.
   mstrcpy(&sdm->spring_surface[sdm->num_spring_surfaces].name, name);
   sdm->spring_surface[sdm->num_spring_surfaces].sd_segment = segnum;
   sdm->spring_surface[sdm->num_spring_surfaces].ph = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
   copy_polyhedron(&ph, sdm->spring_surface[sdm->num_spring_surfaces].ph);

   sdm->spring_surface[sdm->num_spring_surfaces].active = dpYes;
   sdm->spring_surface[sdm->num_spring_surfaces].visible = dpYes;

   // Process any optional parameters at the end of the line, like visible/invisible/active/inactive.
   count--;    // segment already read
   if (count >= 3)
   {
      scaled = dpYes;
      count -= 3;
   }
   for (count--; count >= 0; count--)
   {
      if (STRINGS_ARE_EQUAL(buf[count], "visible"))
         sdm->spring_surface[sdm->num_spring_surfaces].visible = dpYes;
      else if (STRINGS_ARE_EQUAL(buf[count], "invisible"))
         sdm->spring_surface[sdm->num_spring_surfaces].visible = dpNo;
      else if (STRINGS_ARE_EQUAL(buf[count], "active"))
         sdm->spring_surface[sdm->num_spring_surfaces].active = dpYes;
      else if (STRINGS_ARE_EQUAL(buf[count], "inactive"))
         sdm->spring_surface[sdm->num_spring_surfaces].active = dpNo;
   }

   if (scaled)
      post_process_polyhedron(sdm->spring_surface[sdm->num_spring_surfaces].ph, scale);

   sdm->num_spring_surfaces++;

   return code_fine;
}

/* READ_FORCE_MATTE: read a force matte element from the parameters file and store it in
 * the array in the model struct */
static ReturnCode read_force_matte(FILE* fp)
{
   int segnum, count;
   ReturnCode rc = code_fine;
   dpPolyhedronStruct ph;
   char *clean_name, matte_name[CHARBUFFER], filename[CHARBUFFER], seg_name[CHARBUFFER];

   // Read the three elements of a force matte definition. There must be
   // a tab after the filename because the name can contain spaces.
   count = fscanf(fp, "%s", matte_name);
   count += read_til_tokens(fp, filename, "\t\r\n");
   clean_name = clean_string(filename);
   read_line(fp, buffer);
   count += sscanf(buffer, "%s", seg_name);

   if (count != 3)
   {
      sim_message(exit_program, "Error reading force_matte definition (make sure a tab follows the filename).");
      return code_bad;
   }

   // Check that the matte is attached to a valid segment.
   strcpy(buffer, seg_name);
   segnum = enter_segment(sdm, buffer);
   if (segnum == UNDEFINED_SEGMENT)
   {
      sim_message(exit_program, "Unknown segment name (%s) in force_matte definition.", seg_name);
      return code_bad;
   }

   // Read the bone file.
   ph.body_segment = segnum;
   if (read_polyhedron(&ph, clean_name) == code_bad)
   {
      sim_message(exit_program, "Could not create force_matte %s.", matte_name);
      return code_bad;
   }

   mstrcpy(&ph.name, matte_name);
   
   // If you make it to here, then the polyhedron was successfully loaded,
   // and the segment name matched one in the model, so add the matte to
   // the list, and store the polyhedron.  Reallocate space for the array
   // if it is already full.
   if (sdm->force_matte_array_size == 0)
      sdm->force_matte_array_size += ARRAY_INCREMENT;

   if (sdm->num_force_mattes == 0)
   {
      sdm->force_matte = (dpForceMatte*)simm_malloc(sdm->force_matte_array_size * sizeof(dpForceMatte));
   }
   else if (sdm->num_force_mattes == sdm->force_matte_array_size)
   {
      sdm->force_matte_array_size += ARRAY_INCREMENT;
      sdm->force_matte = (dpForceMatte*)simm_realloc(sdm->force_matte, (sdm->force_matte_array_size) * sizeof(dpForceMatte), &rc);
      if (rc == code_bad)
      {
         sim_message(abort_action, "Ran out of memory. Unable to create force matte.");
         return code_bad;
      }
   }

   // Now that you have allocated space for the matte, fill in the info.
   mstrcpy(&sdm->force_matte[sdm->num_force_mattes].name, matte_name);
   sdm->force_matte[sdm->num_force_mattes].segment = segnum;
   sdm->force_matte[sdm->num_force_mattes].ph = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
   copy_polyhedron(&ph, sdm->force_matte[sdm->num_force_mattes].ph);

   sdm->num_force_mattes++;

   return code_fine;
}

/* APPEND_IF_NECESSARY: append the character to the string. */
void append_if_necessary (char str[], char c)
{
   int n = strlen(str);
   
   if (str[n-1] != c)
   {
      str[n] = c;
      str[n+1] = '\0';
   }
}

/* GET_SUFFIX: strip the suffix from the given string. */
const char* get_suffix(const char str[])
{
   int cp = 0;

   cp = strlen(str) - 1;

   while (cp >= 0 && str[cp] != '.')
      cp--;

   if (cp == 0)
      return (NULL);

   return &str[cp+1];
}

/* GET_FUNCTION_SUFFIX: return the character string corresponding to the given function type. */
const char* get_function_suffix(dpFunction* func)
{
   switch (func->type)
   {
      case dpLinearFunction:
         return linear_text;
      case dpNaturalCubicSpline:
         return natural_cubic_text;
      case dpGCVSpline:
         return gcv_text;
      case dpFunctionTypeUndefined:
      case dpStepFunction:
      default:
         return "";
   }
}

/* ADD_FOOT_PRINT: add a footprint to the list of footprints in the motion data and set its parameters to the given data.
 */
static void add_foot_print(MotionData* motion, const char name[], double scale, double translation[], double euler_rotation[])
{
   int i, index;
   ReturnCode rc;

   index = motion->num_foot_prints++;

   if (motion->foot_print == NULL)
      motion->foot_print = (dpFootPrint**)simm_malloc(motion->num_foot_prints * sizeof(dpFootPrint*));
   else
      motion->foot_print = (dpFootPrint**)simm_realloc(motion->foot_print, motion->num_foot_prints * sizeof(dpFootPrint*), &rc);

   if (motion->foot_print == NULL)
   {
      motion->num_foot_prints--;
      return;
   }

   motion->foot_print[index] = (dpFootPrint*)simm_calloc(1, sizeof(dpFootPrint));

   if (motion->foot_print[index] == NULL)
   {
      motion->num_foot_prints--;
      return;
   }

   mstrcpy(&motion->foot_print[index]->name, name);

   motion->foot_print[index]->scale = scale;

   for (i=0; i<3; i++)
   {
      motion->foot_print[index]->translation[i] = translation[i];
      motion->foot_print[index]->euler_rotation[i] = euler_rotation[i];
   }
}

/* ADD_FORCE_PLATE: add a forceplate to the list of forceplates in the motion data and set its parameters to the given data.
 */
static void add_force_plate(MotionData* motion, const char name[], double scale[], double translation[], double euler_rotation[])
{
   int i, index;
   ReturnCode rc;

   index = motion->num_force_plates++;

   if (motion->force_plate == NULL)
      motion->force_plate = (dpForcePlate**)simm_malloc(motion->num_force_plates * sizeof(dpForcePlate*));
   else
      motion->force_plate = (dpForcePlate**)simm_realloc(motion->force_plate, motion->num_force_plates * sizeof(dpForcePlate*), &rc);

   if (motion->force_plate == NULL)
   {
      motion->num_force_plates--;
      return;
   }

   motion->force_plate[index] = (dpForcePlate*)simm_calloc(1, sizeof(dpForcePlate));

   if (motion->force_plate[index] == NULL)
   {
      motion->num_force_plates--;
      return;
   }

   mstrcpy(&motion->force_plate[index]->name, name);

   for (i=0; i<3; i++)
   {
      motion->force_plate[index]->scale[i] = scale[i];
      motion->force_plate[index]->translation[i] = translation[i];
      motion->force_plate[index]->euler_rotation[i] = euler_rotation[i];
   }
}

/* IS_PREFERENCE_ON: return the boolean value of the input string. */
dpBoolean is_preference_on(const char value[])
{
   if (value)
   {
      char buf[CHARBUFFER];
      strcpy_s(buf, CHARBUFFER, value);
      _strupr(buf);

      if (STRINGS_ARE_EQUAL(buf, "Y") ||
         STRINGS_ARE_EQUAL(buf, "YES") ||
         STRINGS_ARE_EQUAL(buf, "TRUE") ||
         STRINGS_ARE_EQUAL(buf, "T") ||
         STRINGS_ARE_EQUAL(buf, "ON") ||
         STRINGS_ARE_EQUAL(buf, "1"))
      {
         return dpYes;
      }
   }

   return dpNo;
}


/* IDENTIFY_ELEMENTS_TO_TRANSFER: Identify which columns in the input motion should be transfered to the output motion. The ones that
 * should be transferred are the ones that are not used by the simulation.
 */
int identify_elements_to_transfer(MotionData* data, dpBoolean transfer[])
{
   int i, count = 0;
   double cutoffFrequency;
   dpFunctionType functionType;

   for (i=0; i<data->num_elements; i++)
   {
      if (name_is_gencoord(data->elementnames[i], sdm, NULL, &functionType, &cutoffFrequency) != UNDEFINED_GENCOORD)
         continue;
      if (name_is_gencoord(data->elementnames[i], sdm, velocity_suffix, &functionType, &cutoffFrequency) != UNDEFINED_GENCOORD)
         continue;
      if (name_is_gencoord(data->elementnames[i], sdm, acceleration_suffix, &functionType, &cutoffFrequency) != UNDEFINED_GENCOORD)
         continue;
      if (name_is_gencoord(data->elementnames[i], sdm, torque_suffix, &functionType, &cutoffFrequency) != UNDEFINED_GENCOORD)
         continue;
      if (name_is_gencoord(data->elementnames[i], sdm, moment_suffix, &functionType, &cutoffFrequency) != UNDEFINED_GENCOORD)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, px_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, py_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, pz_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, vx_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, vy_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, vz_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, ax_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, ay_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_body_segment(data->elementnames[i], sdm, az_suffix, &functionType, &cutoffFrequency) != UNDEFINED_SEGMENT)
         continue;
      if (name_is_muscle(data->elementnames[i], sdm, NULL, &functionType, &cutoffFrequency) != -1)
         continue;
      if (STRINGS_ARE_EQUAL(data->elementnames[i], "time"))
         continue;

      transfer[i] = dpYes;
      count++;
   }

   return count;
}


/* -------------------------------------------------------------------------
   _read_til_tokens - Read characters from 'file' into 'buf' until one of the
     characters specified in 'delimiters' is encountered.  Put the delimiting
     character back into 'file'.
---------------------------------------------------------------------------- */
int _read_til_tokens(FILE* file, char* buf, const char* delimiters)
{
   char* p = buf;

   while (1)
   {
      char c = fgetc(file);

      if (feof(file) || ferror(file))
         break;
      else
      {
         if (strchr(delimiters, c))
         {
            ungetc(c, file);
            break;
         }
         else
            *p++ = c;
      }
   }
   *p = '\0';

   /* remove trailing whitespace if necessary */
   while (p > buf && isspace(p[-1]))
       *(--p) = '\0';

   /* Return 1 if the string is not empty. */
   if (p > buf)
      return 1;
   else
      return 0;
}

/* -------------------------------------------------------------------------
   _strip_outer_whitespace - remove exterior (but not interior) whitespace from a string
---------------------------------------------------------------------------- */
void _strip_outer_whitespace(char* str)
{
   // Remove trailing whitespace.
   char* p = str + strlen(str) - 1;

   for ( ; p >= str && isspace(*p); p--)
      *p = '\0';

   // Remove leading whitespace.
   for (p = str; *p; p++)
      if (*p <= 0 || ! isspace(*p))
         break;

   if (*p && p != str)
      memmove(str, p, strlen(p) + 1);
}

