/*******************************************************************************

   SDUFUNCS.C

   Authors: Peter Loan
            Krystyne Blaikie

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains the main routine for a dynamics
      simulation, and several other utility functions that are
      independent of the musculoskeletal model in the simulation.
      This file is generated by SIMM when you choose the
      "project files" option in the "Save Dynamics" dialog box.
      You will probably want to generate it once and then customize
      it to your specific simulation.

   Routines:
      calc_derivatives        :  compute system derivatives
      sduforce                :  apply all applicable forces to the body segments
      sdumotion               :  apply prescribed motions to fixed and prescribed gencoords
      sduperr                 :  calculate position errors for constrained gencoords
      sduverr                 :  calculate velocity errors for constrained gencoords
      sduaerr                 :  calculate acceleration errors for constrained gencoords
      sduconsfrc              :  calculate and apply constraint forces for constrained gencoords
      init_motion             :  set on/off state of prescribed motion for all Qs based on their types
      set_initial_conditions  :  set initial conditions for all Qs in the model
      assemble_model          :  assemble the model, and make sure the gencoord velocities are properly set
      set_up_input_data       :  load kinetics/motion data for the system


*******************************************************************************/
#include "universal.h"

/*************** DEFINES (for this file only) *********************************/

/* Tolerances and Integration parameters. */
#define ASSEMBLY_TOL 1e-7

/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/


/**************** GLOBAL VARIABLES (used in only a few files) *****************/


/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern SystemInfo si;
extern MotionData* input_data;
extern dpSimulationParameters params;
extern dpModelStruct* sdm;
extern char buffer[];

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/
static ReturnCode set_initial_conditions_quaternions(double time, double y[], double dy[], MotionData *input_motion);


/* This routine (called by integrate()) computes system derivatives, given the 
 * time and state of the system.  It returns a user flag if impacts are found.
 */
void calc_derivatives(double t, double y[], double dy[], double param[], int *status)
{
   int i, err;
#if CONTACT_DETECTION
   int calc_impact = NO;
#endif

   *status = NO_FLAG;

   // normalize the state vector in case quaternions have become un-normalized
   sdnormst(y, y);

   sdstate(t, y, &y[sdm->nq]);
   err = check_for_sderror("CALC_DERIVATIVES");
   if (err)
      goto allDone;

   sdumotion(t, y, &y[sdm->nq]);

   calc_muscle_derivatives(t, y, dy, param, &err);

   sduforce(t, y, &y[sdm->nq]);

   sdderiv(dy, &dy[sdm->nq]);

   // normalize the state vector in case quaternions have become un-normalized
   sdnormst(y, y);

#if CONTACT_DETECTION
   /* determine contact points */
   determine_contacts(y, dy);
   get_contact_info();
   get_bilat_contact_info();

   if (sdm->num_contacts == 0)
      goto allDone;

   // Check for impacts. If any are found, set the flag and return.
   if (param[0] < 0.0)
   {
      for (i = 0; i < sdm->num_contacts; i++)
      {
         if ((sdm->contacts[i].contact_type == dpImpact))
         {
            *status = IMPACT_FLAG;
            goto allDone;
         }
      }
   }

   // if no impacts were found, find resting contact forces, apply them,
   // and calculate accelerations
   sdstate(t, y, &y[sdm->nq]);
   sdumotion(t, y, &y[sdm->nq]);
   sduforce(t, y, &y[sdm->nq]);

   compute_contact_forces(calc_impact);
   apply_forces_at_contacts(sdm->num_contacts, sdm->contacts);

   sdderiv(dy, &dy[sdm->nq]);
#endif

   err = check_for_sderror("CALC_DERIVATIVES");

allDone:
    // We're done with calculating derivatives.  If an error occurred, we'll
    // assume it means the integrator picked a bad time/state for this guess,
    // and we'll reject the guess.  This will make the integrator cut its
    // step size, hopefully producing a better guess next time.
    //
    // If the error is "singular mass matrix" we'll ignore it, however,
    // since in this case that's just an artifact of our use of massless
    // frames when in certain configurations and does not actually cause
    // problems.
    //
    // If we reject this guess, we'll clear out dy[] which may contain
    // garbage (e.g. NaNs) resulting from the bad state.
   if (err == 43 || err == 44)  // 43: tree gimbal joint in gimbal lock, 44: loop gimbal joint in gimbal lock
   {
      sim_message(none, "Cannot continue: A joint is in gimbal lock.");
      *status = BAD_STATE;
   }
   else if (err == 47)  // 47: singular mass matrix
   {
      sim_message(none, "Singular mass matrix\n");
      *status = BAD_STATE;
   }
   else if (err != 0)
   {
      sim_message(none, "BAD STATE");
      *status = BAD_STATE;
   } 
}


/* SDUFORCE: this function is used to apply forces to the body segments.
 * The first section of code determines the current frame for the external
 * force data, if any exist. You may want to change this code, depending
 * on what type of external force[s] you want to apply, and how many
 * frames of data there are. Sduforce() then calls the functions to
 * apply the muscles forces, the restraint moments, and the external
 * forces, respectively.  It also calls the functions to apply spring
 * forces for spring-based contact.
 */
int sduforce(double t, double q[], double u[])
{
   if (input_data)
      apply_external_forces(t, q, input_data);

   apply_muscle_forces(sdm);

   apply_gencoord_restraint_moments(sdm, q);

   apply_spring_forces(sdm);

   check_for_sderror("SDUFORCE");

   return 1;
}


/* Apply prescribed motions to fixed and prescribed gencoords */
int sdumotion(double t, double q[], double u[])
{
   set_fixed_gencoords(sdm);

   if (sdm->newInverseSimulation == dpNo)
      set_prescribed_gencoords(t, q, u, sdm, input_data);

   return 1;
}


/* calculate position errors for constrained gencoords */
void sduperr(double t, double q[], double errors[])
{
   int i, j, q_ind, constraint_num;
   double q_ind_value, q_dep_value, q_value, u_ind_value = 0.0, a_ind_value = 0.0;
   dpConstraintObject *co;
   double *q_euler = NULL;

   q_euler = (double *)simm_malloc(sdm->nu * sizeof(double));

   // convert the state positions to Euler angles
   convert_state_from_quaternion(t, q, NULL, NULL, q_euler, NULL, NULL, NULL);

   // constrained gencoords
   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         q_ind = sdm->q[i].q_ind;
         q_ind_value = q_euler[q_ind] * sdm->q[q_ind].conversion;
         q_dep_value = q_euler[i];
         q_value = interpolate_function(q_ind_value, sdm->q[i].constraint_function, zeroth, u_ind_value, a_ind_value) / sdm->q[i].conversion;
         errors[sdm->q[i].constraint_num] = q_dep_value - q_value;
      }
   }

   // constraint objects
   if (sdm->enforce_constraints == dpYes)
   {
      for (i = 0; i < sdm->num_constraint_objects; i++)
      {
         co = &sdm->constraint_object[i];
         if (co->active == dpNo)
            continue;
         for (j = 0; j < co->numPoints; j++)
         {
            constraint_num = co->points[j].constraint_num;
            errors[constraint_num] = calculate_constraint_position_error(co, PERR, j);
         }
      }
   }

   FREE_IFNOTNULL(q_euler);
}


/* calculate velocity errors for constrained gencoords */
void sduverr(double t, double q[], double u[], double errors[])
{
   int i, j, q_ind, constraint_num;
   double q_ind_value, u_ind_value, u_dep_value, a_ind_value = 0.0, v_value;
   dpConstraintObject *co;
   double *q_euler = NULL, *u_euler = NULL;

   q_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   u_euler = (double *)simm_malloc(sdm->nu * sizeof(double));

   // convert the state positions to Euler angles
   convert_state_from_quaternion(t, q, u, NULL, q_euler, u_euler, NULL, NULL);

   // constrained gencoords
   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         q_ind = sdm->q[i].q_ind;
         q_ind_value = q_euler[q_ind] * sdm->q[q_ind].conversion;
         u_ind_value = u_euler[q_ind] * sdm->q[q_ind].conversion;
         u_dep_value = u_euler[i];
         v_value = interpolate_function(q_ind_value, sdm->q[i].constraint_function, first, u_ind_value, a_ind_value)/sdm->q[i].conversion;
         errors[sdm->q[i].constraint_num] = u_dep_value - v_value;
      }
   }

   // constraint objects
   if (sdm->enforce_constraints == 1)
   {
      for (i = 0; i < sdm->num_constraint_objects; i++)
      {
         co = &sdm->constraint_object[i];
         if (co->active == dpNo)
            continue;

         for (j = 0; j < co->numPoints; j++)
         {
            constraint_num = co->points[j].constraint_num;
            errors[constraint_num] = calculate_constraint_velocity_error(co, j);
         }
      }
   }

   FREE_IFNOTNULL(q_euler);
   FREE_IFNOTNULL(u_euler);
}


/* calculate acceleration errors for constrained gencoords */
void sduaerr(double t, double q[], double u[], double udot[], double errors[])
{
   int i, j, q_ind, constraint_num;
   double q_ind_value, u_ind_value, a_ind_value, a_dep_value, a_value;
   dpConstraintObject *co;
   double *q_euler = NULL, *u_euler = NULL, *udot_euler = NULL;

   q_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   u_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   udot_euler = (double *)simm_malloc(sdm->nu * sizeof(double));

   // convert the state positions to Euler angles
   convert_state_from_quaternion(t, q, u, udot, q_euler, u_euler, udot_euler, NULL);

   // constrained gencoords 
   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         q_ind = sdm->q[i].q_ind;
         q_ind_value = q_euler[q_ind] * sdm->q[q_ind].conversion;
         u_ind_value = u_euler[q_ind] * sdm->q[q_ind].conversion;
         a_ind_value = udot_euler[q_ind] * sdm->q[q_ind].conversion;
         a_dep_value = udot_euler[i];
         a_value = interpolate_function(q_ind_value, sdm->q[i].constraint_function, second, u_ind_value, a_ind_value)/sdm->q[i].conversion;
         errors[sdm->q[i].constraint_num] = a_dep_value - a_value;
      }
   }

   // constraint objects 
   if (sdm->enforce_constraints == dpYes)
   {
      for (i = 0; i < sdm->num_constraint_objects; i++)
      {
         co = &sdm->constraint_object[i];
         if (co->active == dpNo)
            continue;

         for (j = 0; j < co->numPoints; j++)
         {
            constraint_num = co->points[j].constraint_num;
            errors[constraint_num] = calculate_constraint_acceleration_error(co, j);
         }
      }
   }

   FREE_IFNOTNULL(q_euler);
   FREE_IFNOTNULL(u_euler);
   FREE_IFNOTNULL(udot_euler);
}


/* calculate and apply constraint forces for constrained gencoords */
void sduconsfrc(double t, double q[], double u[], double mults[])
{
   int i, j, q_ind, constraint_num;
   double q_ind_value, u_ind_value, a_ind_value = 0.0, int_spline;
   dpConstraintObject* co;
   double* q_euler = (double *)simm_malloc(sdm->nu * sizeof(double));
   double* moment_euler = (double *)simm_calloc(sdm->nu, sizeof(double));
   double* moment_quat = (double *)simm_calloc(sdm->nu, sizeof(double));

   // Convert the states from quaternions to Euler angles.
   convert_state_from_quaternion(t, q, NULL, NULL, q_euler, NULL, NULL, NULL);

   // Determine and store the constraint moments in Euler coordinates.
   for (i=0; i<sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         q_ind = sdm->q[i].q_ind;
         q_ind_value = q_euler[q_ind] * sdm->q[q_ind].conversion;
         u_ind_value = sdm->q[q_ind].conversion;

         moment_euler[sdm->q[i].state_index] += mults[sdm->q[i].constraint_num];

         int_spline = interpolate_function(q_ind_value, sdm->q[i].constraint_function, first, u_ind_value, a_ind_value) / sdm->q[i].conversion;
         moment_euler[sdm->q[q_ind].state_index] += -mults[sdm->q[i].constraint_num] * int_spline;
      }
   }

   // Convert the moments to quaternion coordinates.
   convert_moment_array_to_quaternion(q, moment_euler, moment_quat);

   // Apply the moments.
   for (i=0; i<sdm->nu; i++)
   {
      int axis = sdm->q[i].axis;
      if (sdm->q[i].is_part_of_quaternion)
      {
         axis = abs(axis) - 1;
         if (sdm->joint[sdm->q[i].joint].jnt_type == dpFree)
            axis += 3;
      }
      sdhinget(sdm->q[i].joint, axis, moment_quat[i]);
   }

   // constraint objects
   for (i=0; i<sdm->num_constraint_objects; i++)
   {
      co = &sdm->constraint_object[i];
      if (co->active == dpYes)
      {
         for (j=0; j<co->numPoints; j++)
         {
            constraint_num = co->points[j].constraint_num;
            apply_constraint_forces(co, mults[constraint_num], j);
         }
      }
   }

   FREE_IFNOTNULL(q_euler);
   FREE_IFNOTNULL(moment_euler);
   FREE_IFNOTNULL(moment_quat);

   check_for_sderror("SDUCONSFRC");
}


/* Set on/off state of prescribed motion for all Qs based on their types */
void init_motion(void)
{
   set_prescribed_motion(sdm, dpUnconstrainedQ, 0);
   set_prescribed_motion(sdm, dpConstrainedQ, 0);
   set_prescribed_motion(sdm, dpPrescribedQ, 1);
   set_prescribed_motion(sdm, dpFixedQ, 1);

   check_for_sderror("INIT_MOTION");
}


/* Set initial conditions for all Qs in the model. The initial values
 * and velocities of all Qs, regardless of type, are set here. Values
 * for unconstrained Qs are taken from the 'initial_value' field in the
 * sdm struct.
 */
ReturnCode set_initial_conditions(double *t, double y[], double dy[], MotionData *input_motion)
{
   int i;
   double q_ind_value, *q_euler;

   q_euler = (double *)simm_malloc(sdm->nu * sizeof(double));

   *t = params.startTime;

   for (i = 0; i < sdm->neq; i++)
      y[i] = dy[i] = 0.0;

   // set the initial (Euler) positions
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type != dpConstrainedQ)
      {
         y[i] = sdm->q[i].initial_value / sdm->q[i].conversion;
      }
   }

   // set the initial (Euler) velocities
   for (i = 0; i < sdm->nu; i++) 
   {
      if (sdm->q[i].type != dpFixedQ)
      {
         y[i + sdm->nq] = dy[i] = sdm->q[i].initial_velocity / sdm->q[i].conversion;
      }
      else
      {
         if (sdm->q[i].initial_velocity != 0.0)
            sim_message(none, "Warning: initial velocity for %s (FIXED) was not 0.0. Setting to zero...", sdm->q[i].name);
         y[i + sdm->nq] = dy[i] = 0.0;
      }
   }

   // if there are any quaternions, their initial values need to be set properly since
   // their initial values are given in Euler angles and need to be converted.
   // NOTE: since quaternion positions are not prescribed (using sdprespos) if input motion data
   // is being used, the quaternion positions must be set using these values.  (Unlike Euler
   // joints whose positions will be calculated from input data when sdprespos is called, in sdumotion).
   if (set_initial_conditions_quaternions(*t, y, dy, input_motion) != code_fine)
      return code_bad;

   // convert the state positions to Euler angles
   convert_state_from_quaternion(*t, y, NULL, NULL, q_euler, NULL, NULL, NULL);
// up to here have set all Qs except constrained ones.  Y is in quaternions, q_euler is in euler angles
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == dpConstrainedQ)
      {
         if (sdm->q[i].constraint_function == NULL)
            sim_message(exit_program, "Constrained gencoord %s does not have a function defined.", sdm->q[i].name);

         if (sdm->q[i].constraint_num == -1)
            sim_message(exit_program, "Constrained gencoord %s does not have an associated constraint number.", sdm->q[i].name);

         if (sdm->q[i].q_ind == -1)
            sim_message(exit_program, "Constrained gencoord %s does not have an independent q.", sdm->q[i].name);

         q_ind_value = q_euler[sdm->q[i].q_ind] * sdm->q[sdm->q[i].q_ind].conversion;
         q_euler[sdm->q[i].state_index] = interpolate_function(q_ind_value, sdm->q[i].constraint_function, zeroth, 0.0, 0.0) / sdm->q[i].conversion;
      }
   }

   convert_state_to_quaternion(*t, q_euler, NULL, NULL, y, NULL, NULL);

   FREE_IFNOTNULL(q_euler);

   return code_fine;
}


/* Set the initial state of system (from euler angles specified in input file or params.txt/dllparams.txt) 
 * in quaternion frame.  
 * NOTE: the quaternion representation used here is NOT the same as the SD/FAST representation
 * here quaternion is  q0 = cos(theta/2) q1 = lambda1*sin(theta/2) q2 = lambda2*sin(theta/2) q3 = lambda3*sin(theta/2)
 * SDFAST quaternion is q0 = lambda1*sin(theta/2) q1 = lambda2*sin(theta/2) q2 = lambda3*sin(theta/2) q4 = cos(theta/2) 
 */
static ReturnCode set_initial_conditions_quaternions(double time, double y[], double dy[], MotionData *input_motion)
{
   int i, j, first_index, state_index, axes[3];
   double q_euler[3], u_euler[3], q_quat[4], u_quat[3];
   dpBoolean validPrescribedJoint = dpNo;

   for (i = 0; i < sdm->num_joints; i++)
   {
      if (sdm->joint[i].is_quaternion)
      {
         first_index = sdm->joint[i].first_q;
         if (sdm->joint[i].jnt_type == dpFree)
            first_index += 3;

         state_index = sdm->q[first_index].state_index;

         validPrescribedJoint = dpYes;

         // set the initial position and velocity
         for (j = 0; j < 3; j++)
         {
            axes[j] = sdm->q[first_index+j].axis;
            q_euler[j] = sdm->q[first_index+j].initial_value / sdm->q[first_index+j].conversion;
            u_euler[j] = sdm->q[first_index+j].initial_velocity / sdm->q[first_index+j].conversion;

            if (sdm->q[first_index+j].type != dpPrescribedQ || input_motion == NULL || input_motion->q_data[state_index+j] == NULL)
               validPrescribedJoint = dpNo;
         }

         // if an input motion file is specified (for inverse dynamics), the initial position of quaternion 
         // joints needs to be specified from the input data since it does not get calculated and prescribed
         // in sdumotion (because positions are not prescribed for quaternions).
         // All three degrees of freedom must be of dpPrescribed type and have at position data specified
         if (validPrescribedJoint)
         {
            for (j = 0; j < 3; j++)
               q_euler[j] = interpolate_function(time, input_motion->q_data[state_index+j], zeroth, 0.0, 0.0) / sdm->q[first_index+j].conversion;

         }
         convert_angles_to_quaternions(q_euler, u_euler, NULL, q_quat, u_quat, NULL, sdm->joint[i].euler_rots, axes);

         // copy the quaternion state back into the state vector
         for (j = 0; j < 3; j++)
         {
            y[state_index+j] = q_quat[j];
            y[sdm->nq + state_index+j] = u_quat[j];
         }
         y[sdm->joint[i].quat] = q_quat[3];
      }
   }
   
   // normalize the state
   sdnormst(y, y);

   return code_fine;
}



/* Assemble the model, and make sure the gencoord velocities are properly set.
 */
ReturnCode assemble_model(double t, double y[])
{
   int i, fcnt, err, *lock;

   lock = (int*)simm_malloc(sdm->nq * sizeof(int));

   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == dpFixedQ)
         lock[i] = 1;
      else
         lock[i] = 0;
   }
   for (i = sdm->nu; i < sdm->nq; i++)
      lock[i] = 0;

   // assemble model
   sdassemble(t, y, lock, ASSEMBLY_TOL, 500, &fcnt, &err);
   if (err)
   {
      sim_message(none, "Assembly failed, err = %d.", err);
      //sim_message(none, "Model set to closest solution:");
      //sim_message(none, "Closest solution:");
      //for (i = 0; i < sdm->nu; i++)
      //   sim_message(none, "%s = %lf (%d)", sdm->q[i].name, y[i], lock[i]);
      sim_message(exit_program, "");
      return code_bad;
   }

   /* You may want to call verify_assembly() if you know for sure that
    * none of the gencoords should be outside their range of motion.
   if (verify_assembly(sdm,y) == code_bad)
      sim_message(exit_program, "Verify assembly failed.");
    */

   sdinitvel(t, y, lock, ASSEMBLY_TOL, 500, &fcnt, &err);
   if (err)
   {
      sim_message(none, "Velocity analysis failed.");
      sim_message(exit_program, "Check that prescribed gencoord velocities are not being set to new values.");
      return code_bad;
   }

   free(lock);

   return code_fine;
}


void set_up_input_data(char filename[], MotionData** data)
{
   int i, index_pres = -1;

   if ((*data = (MotionData*)simm_calloc(1, sizeof(MotionData))) == NULL)
   {
      sim_message(exit_program, "Unable to allocate enough memory to run.");
      return;
   }
   else if (load_input_data(sdm, *data, filename) == code_bad)
   {
      // You could try to recover from an error reading the kinetics/motion
      // file, but it's probably serious enough to exit and make the user
      // fix the problem.
      free_motion_data(*data);
      FREE_IFNOTNULL(*data);
      sim_message(exit_program, "Encountered fatal error while reading %s.", filename);
      return;
   }

   // Copy the enforce_constraints parameter from the motion into the model structure.
   // This parameter is dpYes by default, but can be set to dpNo in the motion if the motion
   // contains data for all of the gencoords affected by the constraint objects.
   sdm->enforce_constraints = (*data)->enforce_constraints;

   // If there is position, velocity, and acceleration data for a Q in the kinetics file, make it prescribed.
   if ((*data)->q_data)
   {
      int position_present, velocity_present, acceleration_present;

      for (i = 0; i < sdm->nu; i++)
      {
         // Position data must be explicitly specified (though it can be splined or not).
         // Velocity can either be specified explicitly, or calculated from position
         // (if position data is splined). Acceleration can be specified explicitly or
         // calculated from position or velocity.
         position_present = (int)((*data)->q_data[i]);
         velocity_present = (*data)->u_data[i] || 
                              ((*data)->q_data[i] && (*data)->q_data[i]->type != dpStepFunction);
         acceleration_present = (*data)->udot_data[i] ||
                                ((*data)->q_data[i] && (*data)->q_data[i]->type != dpStepFunction) ||
                                ((*data)->u_data[i] && (*data)->u_data[i]->type != dpStepFunction);

         if (sdm->q[i].type == dpFixedQ)
         {
            if (position_present || velocity_present || acceleration_present)
            {
               sim_message(recover, "Warning: Pos, vel, and/or acc data for FIXED Q %s was specified in %s.", sdm->q[i].name, filename);
               sim_message(none, "Data will be ignored.");
            }
         }
         else if (sdm->q[i].type == dpConstrainedQ)
         {
            if (position_present || velocity_present || acceleration_present)
            {
               sim_message(recover, "Warning: Pos, vel, and/or acc data for CONSTRAINED Q %s was specified in %s.", sdm->q[i].name, filename);
               sim_message(none, "Data will be ignored.");
            }
         }
         else
         {
            if (position_present && velocity_present && acceleration_present)
            {
               sdm->q[i].type = dpPrescribedQ;
               index_pres = i;
            }
            else if (position_present || velocity_present || acceleration_present)
            {
               if (!position_present)
               {
                  sim_message(recover, "Warning: Position data for %s was not specified in %s.", sdm->q[i].name, filename);
                  sim_message(none, "Gencoord will not be prescribed.");
               }
               else if (!velocity_present)
               {
                  sim_message(recover, "Warning: Velocity data for %s was not specified in %s.", sdm->q[i].name, filename);
                  sim_message(none, "Gencoord will not be prescribed.");
               }
               else if (!acceleration_present)
               {
                  sim_message(recover, "Warning: Acceleration data for %s was not specified in %s.", sdm->q[i].name, filename);
                  sim_message(none, "Gencoord will not be prescribed.");
               }
            }
         }
      }
   }

   // If all unconstrained gencoords are prescribed with data in the
   // kinetics file, then this is really more of an inverse dynamics
   // simulation than a forward one. So override the default values
   // of params.start_time, params.end_time, and params.step_size to
   // match the data in the kinetics file.
   for (i = 0; i < sdm->nu; i++)
   {
      if (sdm->q[i].type == dpUnconstrainedQ)
         break;
   }

   if (i == sdm->nu)
   {
#if 0
      // No gencoords are left as unconstrained. Use index_pres, which is the
      // index of one of the Qs that is prescribed in the file, to get the
      // time of the first and last frames of data in the file. Divide this
      // time range by the number of frames to get the step size (reporting
      // interval) for the integrator. If the startTime and endTime are already
      // equal to the times of the first and last frames, don't change stepSize
      // because the user may have changed it to re-sample the data.
      if (NOT_EQUAL_WITHIN_ERROR(params.startTime, (*data)->q_data[index_pres]->x[0]) ||
          NOT_EQUAL_WITHIN_ERROR(params.endTime, (*data)->q_data[index_pres]->x[(*data)->q_data[index_pres]->numpoints - 1]))
      {
         params.startTime = (*data)->q_data[index_pres]->x[0];
         params.endTime = (*data)->q_data[index_pres]->x[(*data)->q_data[index_pres]->numpoints - 1];
         params.stepSize = (params.endTime - params.startTime) / ((*data)->q_data[index_pres]->numpoints - 1);
      }
      sim_message(none, "All unconstrained Qs are prescribed with data from %s.", filename);
      sim_message(none, "   assuming inverse dynamics analysis, changing:");
      sim_message(none, "      start time to %lf", params.startTime);
      sim_message(none, "      end time to   %lf", params.endTime);
      sim_message(none, "      step size to  %lf", params.stepSize);
#endif
      // If the "new" style of joint moment calculation was selected,
      // then set the newInverseSimulation flag and make room for the moment curves.
      if (params.outputOptions.traditionalInverse == dpNo)
      {
         sdm->newInverseSimulation = dpYes;
         add_moments_to_motion_data(sdm, *data);
      }
   }
   else
   {
      sdm->newInverseSimulation = dpNo;
   }
}
