/*******************************************************************************

   SPRING_BASED.C

   Author: Krystyne Blaikie

   Copyright (c) 2000-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains the functions to calculate and apply
      spring forces for spring-based contact.

   Routines:
      apply_spring_forces:    apply forces at contact springs
      calculate_spring_force: calculate the force in a spring
      enter_spring_floor:     enters a spring floor in the master list
      show_spring_info:       display information about spring elements to screen
      get_spring_contact_segment_sd_index:   get the index of the segment the contact floor or surface is attached to

*******************************************************************************/
#include "universal.h"

/*************** DEFINES (for this file only) *********************************/


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/


/**************** GLOBAL VARIABLES (used in only a few files) *****************/


/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern dpModelStruct* sdm;

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/


/*************************** FUNCTION PROTOTYPES ******************************/
static int enter_spring_floor(char name[]);
static int enter_spring_surface(char name[]);
static dpBoolean calculate_spring_contact_force(dpSpringStruct spring, int contact_segment, double normal[], double springPt_in_contactFrame[], double projPt_in_contactFrame[], double force[]);

/* GET_SPRING_CONTACT_SEGMENT_SD_INDEX: Return the index of the segment to which the floor or surface contact element is attached. 
 */
int get_spring_contact_segment_sd_index(int springIndex)
{
   int index = -1;

   if (sdm->spring[springIndex].contact_index < 0)
      return index;

   if (sdm->spring[springIndex].is_floor_contact)
      index = sdm->spring_floor[sdm->spring[springIndex].contact_index].sd_segment;
   else
      index = sdm->spring_surface[sdm->spring[springIndex].contact_index].sd_segment;

   return index;
}

/* APPLY_SPRING_FORCES: Calculate and apply forces from spring elements for spring-based contact detection.
 * If the spring points have penetrated below the plane/polygons a force will be generated to prevent the penetration.
 * Forces cannot pull the objects together.  For spring floors, the first polygon in the polyhedron is assumed to
 * be an infinite plane.
 */
void apply_spring_forces(dpModelStruct *sdm)
{
   int i, contact_segment, closestPoly, polygon;
   double dist, smallestDist, normal[3], force[3];
   double force_in_spring_frame[3], force_on_contact_segment[3], spring_pt_in_contact_frame[3], proj_pt_in_contact_frame[3];
   dpSpringStruct *spring = NULL;
   dpBoolean applyForce;

   for (i = 0; i < sdm->num_springs; i++)
   {
      spring = &sdm->spring[i];
      contact_segment = get_spring_contact_segment_sd_index(i);

      if (contact_segment < -1)
         continue;

      applyForce = dpNo;
      clear_vector(force_in_spring_frame, 3);
      closestPoly = -1;
      smallestDist = 99999.9;

      if (spring->is_floor_contact)
      {
         dpSpringFloor floor = sdm->spring_floor[spring->contact_index];
         if (floor.active == dpNo)
            continue;

         // Convert spring point into the plane frame.
         convert_point2(spring->sd_segment, spring->point, contact_segment, spring_pt_in_contact_frame);

         // Project point onto plane.
         dist = project_point_onto_plane(spring_pt_in_contact_frame, &floor.plane, proj_pt_in_contact_frame);

         // Store the plane's normal vector.
         normal[0] = floor.plane.a;
         normal[1] = floor.plane.b;
         normal[2] = floor.plane.c;

         // Calculate the spring force in the floor frame.
         applyForce = calculate_spring_contact_force(*spring, contact_segment, normal, spring_pt_in_contact_frame, proj_pt_in_contact_frame, force);
      }
      else
      {
         dpSpringSurface surface = sdm->spring_surface[spring->contact_index];  
         if (surface.active == dpNo)
            continue;

         // Convert spring point into contact surface frame.
         convert_point2(spring->sd_segment, spring->point, contact_segment, spring_pt_in_contact_frame);

         // For each polygon in the surface, project the point onto the polygon and determine if the point
         // intersects the polygon (if the projected point is within the polygon).  Determine which polygon 
         //the point is closest to (above or below the surface). 
         for (polygon = 0; polygon < surface.ph->num_polygons; polygon++)
         {
            dpPlaneStruct plane;
            dpBoolean intersects;

            plane.a = surface.ph->polygon[polygon].normal[0];
            plane.b = surface.ph->polygon[polygon].normal[1];
            plane.c = surface.ph->polygon[polygon].normal[2];
            plane.d = surface.ph->polygon[polygon].d;

            dist = project_point_onto_plane(spring_pt_in_contact_frame, &plane, proj_pt_in_contact_frame);
            intersects = point_in_polygon3D(proj_pt_in_contact_frame, surface.ph, polygon);

            if ((fabs(dist) < smallestDist) && intersects)
            {
               closestPoly = polygon;
               smallestDist = fabs(dist);
            }
         }

         // When the closest polygon has been found, if the use the one force from the closest polygon
         if (closestPoly >= 0)
         {
            dpPlaneStruct plane;
            dpBoolean intersects;

            plane.a = surface.ph->polygon[closestPoly].normal[0];
            plane.b = surface.ph->polygon[closestPoly].normal[1];
            plane.c = surface.ph->polygon[closestPoly].normal[2];
            plane.d = surface.ph->polygon[closestPoly].d;
            normal[0] = surface.ph->polygon[closestPoly].normal[0];
            normal[1] = surface.ph->polygon[closestPoly].normal[1];
            normal[2] = surface.ph->polygon[closestPoly].normal[2];

            dist = project_point_onto_plane(spring_pt_in_contact_frame, &plane, proj_pt_in_contact_frame);
            intersects = point_in_polygon3D(proj_pt_in_contact_frame, surface.ph, closestPoly);

            if (intersects && dist < 0)
               applyForce = calculate_spring_contact_force(*spring, contact_segment, normal, spring_pt_in_contact_frame, proj_pt_in_contact_frame, force);
         }
      }

      if (applyForce)
      {
         // Apply forces to both segments and copy spring force into array.
         // The force on the floor is in the direction opposite to the floor normal. Transform the force from the floor frame
         // to the segment frame (if the point penetrated the plane/polygon).
         sdvmul(-1.0, force, force_on_contact_segment);
         sdpointf(contact_segment, proj_pt_in_contact_frame, force_on_contact_segment);

         sdtrans(contact_segment, force, spring->sd_segment, force_in_spring_frame);
         sdpointf(spring->sd_segment, spring->point, force_in_spring_frame);
      }

      sdvcopy(force_in_spring_frame, spring->force);
   }
}

/* Calculate the spring force based on the contact model presented in
 * Neptune, R.R., Wright, I.C., Bogert, A.J. van den (2000).  A method
 * for numerical simulation of single limb ground contact events: application
 * to heel-toe running.  Journal of Computer Methods in Biomechanics and
 * Biomechanical Engineering 3(4): 321-334.
 */
static dpBoolean calculate_spring_contact_force(dpSpringStruct spring, int contact_segment, double normal[], double springPt_in_contactFrame[], double projPt_in_contactFrame[], double force_in_contactFrame[])
{
   int i;
   int velocity_dir, penetration;
   double coef_a, coef_b, coef_c, coef_d, coef_e, coef_f;
   double depth, p2p1[3];
   double vel1[3], vel2[3], rel_vel[3];
   double normal_vel[3], tang_vel[3], norm_vel_mag;
   double normal_force_mag, tang_vel_mag, friction_factor1, friction_factor2;
   double normal_force[3], tangential_force[3];
   double viscous_damping_coef = 3000.0;

   coef_a = spring.param_a;   // Coefficient for velocity independent term.
   coef_b = spring.param_b;   // Power on displacement in velocity independent term.
   coef_c = spring.param_c;   // Coefficient for velocity dependent term.
   coef_d = spring.param_d;   // Power on displacement in velocity dependent term.
   coef_e = spring.param_e;   // Power on velocity in velocity dependent term.
   coef_f = spring.param_f;   // Area normalization term (approximate area associated with ground contact point divided by 32.2 cm^2.

   // Find the vector from pt2 to pt1 along the polygon normal.  Pt1 is on the spring,
   // pt2 is on the floor, both are specified in the surface segment frame.  Find the length of the vector (depth
   // of penetration and its direction (-1: point is below the polygon, 1: point is above the polygon).
   sdvsub(springPt_in_contactFrame, projPt_in_contactFrame, p2p1);
   penetration = (int)SIGN(DOT_VECTORS(p2p1, normal));
   depth = VECTOR_MAGNITUDE(p2p1);

   // If the point is above the polygon, no force is applied.
   if (penetration == 1)
      return dpNo;
   
   // Find the relative velocity of pt1 wrt pt2 and convert to plane frame.
   sdvel(spring.sd_segment, spring.point, vel1);
   sdvel(contact_segment, projPt_in_contactFrame, vel2);
   sdvsub(vel1, vel2, rel_vel);
   sdtrans(GROUND, rel_vel, contact_segment, rel_vel);

   // Find normal and tangential relative velocities
   sdvmul(DOT_VECTORS(rel_vel, normal), normal, normal_vel);
   sdvsub(rel_vel, normal_vel, tang_vel);
   norm_vel_mag = VECTOR_MAGNITUDE(normal_vel);
   tang_vel_mag = VECTOR_MAGNITUDE(tang_vel);

   // If the point is not moving toward the polygon, no force is applied.
   //if (EQUAL_WITHIN_TOLERANCE(norm_vel_mag, 0.0, 0.005))
   //   return dpNo;

   // Find the direction of the normal relative velocity
   // -1 : point is moving into the plane, 1 : point is moving away from the plane
   velocity_dir = (int)SIGN(DOT_VECTORS(rel_vel, normal));

   force_in_contactFrame[0] = force_in_contactFrame[1] = force_in_contactFrame[2] = 0.0;

   // Calculate the spring force if the spring point has penetrated the plane.
   normal_force_mag = coef_f * (coef_a * pow(depth, coef_b) + velocity_dir * coef_c * pow(depth, coef_d) * pow(norm_vel_mag, coef_e));

   // If the force is not pulling on the objects, calculate friction force.
   if (normal_force_mag > 0.0)
   {
      sdvmul(normal_force_mag, normal, normal_force);
      friction_factor1 = viscous_damping_coef * tang_vel_mag * coef_f;
      friction_factor2 = spring.friction * normal_force_mag;
      if (friction_factor1 > friction_factor2)
         friction_factor1 = friction_factor2;
      if (tang_vel_mag < 0.0000001)
         friction_factor1 = 0.0;
      else 
         friction_factor1 = -1.0 * friction_factor1 / tang_vel_mag;

      sdvmul(friction_factor1, tang_vel, tangential_force);
      sdvadd(tangential_force, normal_force, force_in_contactFrame);
   }

   return dpYes;
}

/* ENTER_SPRING_FLOOR: Given a spring_floor name, return the index of the floor in the array,
 * or -1 if it is not in the array.
 */
static int enter_spring_floor(char name[])
{
   int i;

   for (i = 0; i < sdm->num_spring_floors; i++)
      if (STRINGS_ARE_EQUAL(name, sdm->spring_floor[i].name))
         return i;

   // If the name is not found in the list, return -1.
   return -1;
}

/* ENTER_SPRING_SURFACE: Given a spring_surface name, return the index of the surface in the array,
 * or -1 if it is not in the array.
 */
static int enter_spring_surface(char name[])
{
   int i;

   for (i = 0; i < sdm->num_spring_surfaces; i++)
      if (STRINGS_ARE_EQUAL(name, sdm->spring_surface[i].name))
         return i;

   // If the name is not found in the list, return -1.
   return -1;
}

/* Display information about spring components. */
void show_spring_info(void)
{
   int i, spring_sd_seg, contact_num, contact_seg;

   for (i = 0; i < sdm->num_spring_floors; i++)
   {
      sim_message(none, "floor %s:", sdm->spring_floor[i].name);
      sim_message(none, "\ton segment %d (%s)", sdm->spring_floor[i].sd_segment,
         sdm->body_segment[sdm->spring_floor[i].sd_segment + 1].name);
      sim_message(none, "\tPlane a: %f  b: %f  c: %f  d: %f", sdm->spring_floor[i].plane.a,
         sdm->spring_floor[i].plane.b, sdm->spring_floor[i].plane.c, 
         sdm->spring_floor[i].plane.d);
   }
   for (i = 0; i < sdm->num_spring_surfaces; i++)
   {
      sim_message(none, "surface %s:", sdm->spring_surface[i].name);
      sim_message(none, "\ton segment %d (%s)", sdm->spring_surface[i].sd_segment,
         sdm->body_segment[sdm->spring_floor[i].sd_segment + 1].name);
      sim_message(none, "\thas %d polygons", sdm->spring_surface[i].ph->num_polygons);
   }
   for (i = 0; i < sdm->num_springs; i++)
   {
      spring_sd_seg = sdm->spring[i].sd_segment;
      contact_num = sdm->spring[i].contact_index;
      contact_seg = get_spring_contact_segment_sd_index(i);
      sim_message(none, "spring%d:", i);
      sim_message(none, "\tattached to segment %d (%s)", spring_sd_seg, sdm->body_segment[spring_sd_seg + 1].name);
      if (contact_num < 0)
         sim_message(none, "\t  and floor %s (not defined)", sdm->spring[i].contact_name);
      else
      {
         sim_message(none, "\t  and floor %d (%s) on segment %d (%s)", contact_num, sdm->spring[i].contact_name, contact_seg, sdm->body_segment[contact_seg + 1].name);
      }
      sim_message(none, "\tspring point: %f %f %f", sdm->spring[i].point[0], sdm->spring[i].point[1], sdm->spring[i].point[2]);
      sim_message(none, "\tparams:");
      sim_message(none, "\tfriction = %f", sdm->spring[i].friction);
      sim_message(none, "\ta = %f", sdm->spring[i].param_a);
      sim_message(none, "\tb = %f", sdm->spring[i].param_b);
      sim_message(none, "\tc = %f", sdm->spring[i].param_c);
      sim_message(none, "\td = %f", sdm->spring[i].param_d);
      sim_message(none, "\te = %f", sdm->spring[i].param_e);
      sim_message(none, "\tf = %f", sdm->spring[i].param_f);
   }
}

